#----------------
import os
import sys
import math
import c4d
sys.path.insert( 0, os.path.dirname( __file__ ) )
import Ornatrix.Ornatrix as Ox
import TestUtilities
sys.path.pop( 0 )
#----------------
from unittest import TestCase

#
# TestCase
class Test_ClumpModifier( TestCase ):

	def setUp( self ):
		self.doc = c4d.documents.GetActiveDocument()
		Ox.DeleteAll( self.doc )
		# Create base object
		self.hair = TestUtilities.AddGuidesToNewPlane( planeSize=400, planeSegmentCount=20, guideCount=20, pointsPerStrandCount=2)
		self.gfm = self.hair.GetDown()
		# Add clump
		self.clump = c4d.BaseObject( Ox.res.ID_OX_CLUMP_MOD )
		self.clump[Ox.res.clump_ClumpCountAttribute] = 20
		self.clump[Ox.res.clump_FlyawayFractionAttribute] = 0
		self.clump.InsertUnderLast( self.hair )
		# Build
		Ox.ComputeStack( self.doc )
		# Generate clumps
		TestUtilities.SendDescriptionCommand (self.clump, Ox.res.clump_Button_CreateClumps )
		# Build
		Ox.ComputeStack( self.doc )
		self.verticesBeforeClump = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.gfm )
		self.verticesAfterClump = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )

	def tearDown( self ):
		#Ox.DeleteAll( self.doc )
		self.doc = None

	# Test if Clump modifier changed original guides
	def test_Clump_StrandsChanged( self ):
		TestUtilities.CheckPointsNotAllNearEqualByStrandId( self, self.verticesBeforeClump, self.verticesAfterClump )

	# Test if Clump modifier changed original guides
	def test_Clump_Undo( self, undoCount=1 ):
		# Undo second clump
		if undoCount == 2:
			self.doc.DoUndo()
			Ox.ComputeStack( self.doc )
			self.clump = self.doc.SearchObject(self.clump.GetName())
			verticesAfterUndo = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
			TestUtilities.CheckPointsAllNearEqualByStrandId( self, self.verticesAfterClump, verticesAfterUndo )
		# Undo first clump
		self.doc.DoUndo()
		Ox.ComputeStack( self.doc )
		self.clump = self.doc.SearchObject(self.clump.GetName())
		verticesAfterUndo = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
		TestUtilities.CheckPointsAllNearEqualByStrandId( self, self.verticesBeforeClump, verticesAfterUndo )

	# Test if Clump modifier changed original guides
	def test_Clump_HairSelection( self ):
		# Strand edit mode
		self.doc.SetActiveObject( self.clump, c4d.SELECTION_NEW )
		self.clump[Ox.res.clump_ClumpEditMode] = Ox.res.clump_ClumpEditMode_ClumpHairs
		Ox.ComputeStack( self.doc )
		# Select half clumped strands
		strandIds = TestUtilities.GetStrandIds( self.clump )
		selectedStrandIds = strandIds[0:int(len(strandIds)/2)]
		TestUtilities.SelectStrandsByStrandIds( self.doc, self.clump, selectedStrandIds, Ox.res.clump_Command_SelectClumpsHairsByStrandId )
		Ox.ComputeStack( self.doc )
		# Clump selected strands
		TestUtilities.SendDescriptionCommand (self.clump, Ox.res.clump_Button_CreateClumps )
		Ox.ComputeStack( self.doc )
		# Test
		verticesAfterSecondClump = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
		nearEqualPercentage = TestUtilities.GetNearEqualPointsPercentage( self, self.verticesAfterClump, verticesAfterSecondClump )
		TestUtilities.CheckPercentageNearEqual( self, nearEqualPercentage, 0.5, 0.01, 'Equal strands' )

	def test_Clump_HairSelection_Undo( self ):
		self.test_Clump_HairSelection()
		self.test_Clump_Undo( 2 )

	# Test if Clump modifier changed original guides
	def test_ClumpClumpSelection( self ):
		# Strand edit mode
		self.doc.SetActiveObject( self.clump, c4d.SELECTION_NEW )
		self.clump[Ox.res.clump_ClumpEditMode] = Ox.res.clump_ClumpEditMode_Clumps
		Ox.ComputeStack( self.doc )
		# Select half clumped strands
		TestUtilities.SelectStrandsByRange( self.doc, self.clump, 0, 10, 20, Ox.res.clump_Command_SelectClumpsByClumpIndex )
		Ox.ComputeStack( self.doc )
		# Clump selected strands
		TestUtilities.SendDescriptionCommand (self.clump, Ox.res.clump_Button_CreateClumps )
		Ox.ComputeStack( self.doc )
		# Test
		verticesAfterSecondClump = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
		nearEqualPercentage = TestUtilities.GetNearEqualPointsPercentage( self, self.verticesAfterClump, verticesAfterSecondClump )
		TestUtilities.CheckPercentageNearEqual( self, nearEqualPercentage, 0.5, 0.1, 'Equal strands' )

	def test_ClumpClumpSelection_Undo( self ):
		self.test_ClumpClumpSelection()
		self.test_Clump_Undo( 2 )

	def test_VertexColorChannels( self ):
		plane = self.gfm[Ox.res.gfm_InputMeshAttribute]
		numSegments = plane[c4d.PRIM_PLANE_SUBW]
		planeWidth = plane[c4d.PRIM_PLANE_WIDTH]
		segmentWidth = planeWidth / numSegments
		planeMinBound = -planeWidth / 2.0
		blackMaxBound = planeMinBound + segmentWidth

		# Make the plane modifiable
		res = c4d.utils.SendModelingCommand(
				command = c4d.MCOMMAND_MAKEEDITABLE,
				list = [plane],
				mode = c4d.MODELINGCOMMANDMODE_ALL,
				doc = self.doc )

		if res is False or res is True:
			raise TypeError( "Test_ClumpModifier: Error when making plane editable." )
		elif isinstance(res, list):
			plane = res[0]
			self.doc.InsertObject( plane )
		c4d.EventAdd()

		self.gfm[Ox.res.gfm_RootGenerationMethodAttribute] = 6 # Face center roots
		self.gfm[Ox.res.gfm_GuidePointCountAttribute] = 5 # 5 points per strand

		# Color the first column of faces black and get vertices in that area
		TestUtilities.CreateVertexColors( self.doc, plane, 4 + 2 * ( numSegments - 1 ) )
		rootIdsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.gfm, blackMaxBound, 2 )
		verticesBeforeClumpInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.gfm, rootIdsInBlackArea )
		clumpedVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.clump, rootIdsInBlackArea )

		# Test amount channel
		self.clump[Ox.res.clump_AmountChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		Ox.ComputeStack( self.doc )
		newVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.clump, rootIdsInBlackArea )

		TestUtilities.CheckPointsNotAllNearEqual( self, clumpedVerticesInBlackArea, newVerticesInBlackArea )
		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeClumpInBlackArea, newVerticesInBlackArea )
		self.clump[Ox.res.clump_AmountChannelAttribute] = 0

		# Test twist channel
		self.clump[Ox.res.clump_TwistAttribute] = 0
		Ox.ComputeStack( self.doc )
		verticesWithoutTwistInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.clump, rootIdsInBlackArea )

		self.clump[Ox.res.clump_TwistChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		self.clump[Ox.res.clump_TwistAttribute] = math.pi / 3.0
		Ox.ComputeStack( self.doc )
		newVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.clump, rootIdsInBlackArea )

		TestUtilities.CheckPointsAllNearEqual( self, verticesWithoutTwistInBlackArea, newVerticesInBlackArea )
		self.clump[Ox.res.clump_TwistChannelAttribute] = 0

	def test_ClumpClumpWithPatternHandleChangesToClumpParams( self ):
		verticesBeeforePattern = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )

		# create pattern
		braidHair = c4d.BaseObject( Ox.res.ID_OX_HAIR_OBJECT )
		braid = c4d.BaseObject( Ox.res.ID_OX_BRAID_GUIDES_MOD )
		self.doc.InsertObject( braidHair  )
		self.doc.InsertObject( braid, braidHair )

		# add pattern
		braids = c4d.InExcludeData()
		braids.InsertObject( braidHair, 1 )
		self.clump[Ox.res.clump_PatternsAttribute] = braids
		Ox.ComputeStack( self.doc )
		verticesAfterPattern = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
		TestUtilities.CheckPointsNotAllNearEqualByStrandId( self, verticesBeeforePattern , verticesAfterPattern )

		# change a parameter
		# had to manually add this undo section, otherwise it would undo inserting the braids as pattern
		self.doc.StartUndo()
		self.doc.AddUndo(c4d.UNDOTYPE_CHANGE, self.clump)
		self.clump[Ox.res.clump_ClumpPatternRotationAttribute] = 0.5
		self.doc.EndUndo()
		Ox.ComputeStack( self.doc )
		verticesAfterChange = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
		TestUtilities.CheckPointsNotAllNearEqualByStrandId( self, verticesAfterPattern, verticesAfterChange )

		# undo change
		self.doc.DoUndo()
		Ox.ComputeStack( self.doc )
		self.clump = self.doc.SearchObject(self.clump.GetName())
		verticesAfterUndo = TestUtilities.GetVerticesInObjectCoordinatesByStrandId( self.clump )
		# compare to before change.
		TestUtilities.CheckPointsAllNearEqualByStrandId( self, verticesAfterPattern, verticesAfterUndo )


#----------------------------------
# main
testObject = None
if __name__ == '__main__':
	Ox.runTest( doc, Test_ClumpModifier )