#----------------
import os
import sys
import math
import c4d
sys.path.insert( 0, os.path.dirname( __file__ ) )
import Ornatrix.Ornatrix as Ox
import TestUtilities
sys.path.pop( 0 )
#----------------
from unittest import TestCase

#
# TestCase
class Test_MultiplierModifier( TestCase ):

	def setUp( self ):
		self.doc = c4d.documents.GetActiveDocument()
		Ox.DeleteAll( self.doc )
		# Create base object
		self.hair = TestUtilities.AddGuidesToNewPlane( planeSize=100, planeSegmentCount=3, guideCount=20, pointsPerStrandCount=5, rootGenerationMethod=6 )
		self.gfm = self.hair.GetDown()
		# Add multiplier
		self.multiplier = c4d.BaseObject( Ox.res.ID_OX_MULTIPLIER_MOD )
		self.multiplier.InsertUnderLast( self.hair )
		# Build
		Ox.ComputeStack( self.doc )

	def tearDown( self ):
		self.doc = None

	def test_VertexColorChannels( self ):
		plane = self.gfm[Ox.res.gfm_InputMeshAttribute]
		numSegments = plane[c4d.PRIM_PLANE_SUBW]
		planeWidth = plane[c4d.PRIM_PLANE_WIDTH]
		segmentWidth = planeWidth / numSegments
		planeMinBound = -planeWidth / 2.0
		blackMaxBound = planeMinBound + segmentWidth

		# Make the plane modifiable
		res = c4d.utils.SendModelingCommand(
				command = c4d.MCOMMAND_MAKEEDITABLE,
				list = [plane],
				mode = c4d.MODELINGCOMMANDMODE_ALL,
				doc = self.doc )

		if res is False or res is True:
			raise TypeError( "Test_MultiplierModifier: Error when making plane editable." )
		elif isinstance(res, list):
			plane = res[0]
			self.doc.InsertObject( plane )
		c4d.EventAdd()

		# Color the first column of faces black and get vertices in that area
		TestUtilities.CreateVertexColors( self.doc, plane, 4 + 2 * ( numSegments - 1 ) )
		rootIdsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.gfm, blackMaxBound, 2 )
		verticesBeforeMultiplierInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.gfm, rootIdsInBlackArea )

		# Test probability channel
		self.multiplier[Ox.res.mlt_ProbabilityChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		Ox.ComputeStack( self.doc )
		newRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		newVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, newRootsInBlackArea )
		self.assertEqual( len( rootIdsInBlackArea ), len( newRootsInBlackArea ) )
		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeMultiplierInBlackArea, newVerticesInBlackArea )
		self.multiplier[Ox.res.mlt_ProbabilityChannelAttribute] = 0

		# Test copies channel
		self.multiplier[Ox.res.mlt_CopiesChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		Ox.ComputeStack( self.doc )
		newRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		self.assertEqual( len( newRootsInBlackArea ), 0 )
		self.multiplier[Ox.res.mlt_CopiesChannelAttribute] = 0

		# Test spread channel
		self.multiplier[Ox.res.mlt_SpreadTipAttribute] = 0
		Ox.ComputeStack( self.doc )
		oldRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		oldVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, oldRootsInBlackArea )
		self.multiplier[Ox.res.mlt_SpreadTipAttribute] = 10
		self.multiplier[Ox.res.mlt_SpreadChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		Ox.ComputeStack( self.doc )
		newRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		newVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, newRootsInBlackArea )
		self.assertEqual( len( oldRootsInBlackArea ), len( newRootsInBlackArea ) )
		TestUtilities.CheckPointsAllNearEqual( self, oldVerticesInBlackArea, newVerticesInBlackArea )
		self.multiplier[Ox.res.mlt_SpreadChannelAttribute] = 0

		# Test fluff channel
		self.multiplier[Ox.res.mlt_FluffTipAttribute] = 0
		Ox.ComputeStack( self.doc )
		oldRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		oldVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, oldRootsInBlackArea )
		self.multiplier[Ox.res.mlt_FluffTipAttribute] = 10
		self.multiplier[Ox.res.mlt_FluffChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		Ox.ComputeStack( self.doc )
		newRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		newVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, newRootsInBlackArea )
		self.assertEqual( len( oldRootsInBlackArea ), len( newRootsInBlackArea ) )
		TestUtilities.CheckPointsAllNearEqual( self, oldVerticesInBlackArea, newVerticesInBlackArea )
		self.multiplier[Ox.res.mlt_FluffChannelAttribute] = 0

		# Test shift channel
		self.multiplier[Ox.res.mlt_ShiftRootAttribute] = 0
		Ox.ComputeStack( self.doc )
		oldRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		oldVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, oldRootsInBlackArea )
		self.multiplier[Ox.res.mlt_ShiftRootAttribute] = 10
		self.multiplier[Ox.res.mlt_ShiftChannelAttribute] = TestUtilities.Ox_FirstVertexColorChannelIndex
		Ox.ComputeStack( self.doc )
		newRootsInBlackArea = TestUtilities.GetAndFilterRootsByMaxBound( self.multiplier, blackMaxBound, 2 )
		newVerticesInBlackArea = TestUtilities.GetVerticesAndFilterByIds( self.multiplier, newRootsInBlackArea )
		self.assertEqual( len( oldRootsInBlackArea ), len( newRootsInBlackArea ) )
		TestUtilities.CheckPointsAllNearEqual( self, oldVerticesInBlackArea, newVerticesInBlackArea )
		self.multiplier[Ox.res.mlt_ShiftChannelAttribute] = 0



#----------------------------------
# main
testObject = None
if __name__ == '__main__':
	Ox.runTest( doc, Test_MultiplierModifier )