import MayaTest
import TestUtilities
import pymel.core as pm
import pymel.core.system as pmsys
import pymel.core.datatypes as dt

class Test_MultiplierNode( MayaTest.OxTestCase ):

	def test_BasicMultiplier( self ):
		# Check that at least some vertices are modified
		TestUtilities.TestStrandOperatorChangesGuideCount( self, TestUtilities.MultiplierNodeName )

	def test_MultiplierRandomizationIsIndependentPerStrand( self ):
		guidesShape = TestUtilities.AddGuidesToNewPlane( guideCount = 2 )
		guidesFromMesh = TestUtilities.GetNodeByType( TestUtilities.GuidesFromMeshNodeName )
		guidesFromMesh.length.set( 5 )
		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( guidesShape, TestUtilities.MultiplierNodeName ) )

		multiplier.copies.set( 3 )
		verticesBefore = TestUtilities.GetStrandPoints( guidesShape, 3 )

		multiplier.copies.set( 2 )
		verticesAfter = TestUtilities.GetStrandPoints( guidesShape, 2 )

		self.assertEqual( verticesAfter, verticesBefore )

	def test_OutputSmoothness( self ):
		"""Tests for jagginess due to segment transformations (#3877)"""
		guidesShape = pm.PyNode( pm.mel.OxAddBraidGuides() )
		braidGuides = pm.PyNode( pm.mel.ls( type = TestUtilities.BraidGuidesNodeName )[0] )
		braidGuides.frequency.set( 1 )
		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( guidesShape, TestUtilities.MultiplierNodeName ) )
		strandCount = pm.mel.OxGetStrandCount( guidesShape )
		# Simply testing if any of the strands goes back in the z-direction is enough for this bug
		for strandIndex in range( strandCount ):
			vertices = TestUtilities.GetStrandPoints( guidesShape, strandIndex, True )
			oldVertex = vertices[0]
			for vertex in vertices:
				self.assertGreaterEqual( vertex[2], oldVertex[2] )
				oldVertex = vertex

	def test_ProbabilityVertexColorChannel( self ):
		plane = pm.polyPlane( sx = 3, sy = 3 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		blackFaces = meshShape.f[0:3]
		whiteFaces = meshShape.f[12:]
		TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces )
		blackTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, blackFaces )
		whiteTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, whiteFaces )

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		guidesFromMesh = pm.PyNode( pm.mel.ls( type = TestUtilities.GuidesFromMeshNodeName )[0] )
		hairShape = pm.PyNode( pm.mel.ls( type = TestUtilities.HairShapeName )[0] )
		guidesFromMesh.distribution.set( 6 ) # Face center distribution

		originalStrandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )

		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.MultiplierNodeName ) )
		multiplier.copies.set( 3 )
		multiplier.probability.set( 1.0 )

		oldStrandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		oldStrandIndicesInWhiteArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, whiteTriangleVectorsList )

		multiplier.probabilityChannel.set( 2001 )

		newStrandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		newStrandIndicesInWhiteArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, whiteTriangleVectorsList )

		self.assertEqual( len( oldStrandIndicesInWhiteArea ), len( newStrandIndicesInWhiteArea ) )
		self.assertNotEqual( len( oldStrandIndicesInBlackArea ), len( newStrandIndicesInBlackArea ) )
		self.assertEqual( len( originalStrandIndicesInBlackArea ), len( newStrandIndicesInBlackArea ) )

	def test_CopiesVertexColorChannel( self ):
		plane = pm.polyPlane( sx = 3, sy = 3 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		blackFaces = meshShape.f[0:3]
		whiteFaces = meshShape.f[12:]
		TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces )
		blackTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, blackFaces )
		whiteTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, whiteFaces )

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		guidesFromMesh = pm.PyNode( pm.mel.ls( type = TestUtilities.GuidesFromMeshNodeName )[0] )
		hairShape = pm.PyNode( pm.mel.ls( type = TestUtilities.HairShapeName )[0] )
		guidesFromMesh.distribution.set( 6 ) # Face center distribution

		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.MultiplierNodeName ) )
		multiplier.copies.set( 3 )
		multiplier.probability.set( 1.0 )

		oldStrandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		oldStrandIndicesInWhiteArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, whiteTriangleVectorsList )

		multiplier.copiesChannel.set( 2001 )

		newStrandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		newStrandIndicesInWhiteArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, whiteTriangleVectorsList )

		self.assertEqual( len( oldStrandIndicesInWhiteArea ), len( newStrandIndicesInWhiteArea ) )
		self.assertNotEqual( len( oldStrandIndicesInBlackArea ), len( newStrandIndicesInBlackArea ) )
		self.assertEqual( 0, len( newStrandIndicesInBlackArea ) )

	def test_SpreadVertexColorChannel( self ):
		plane = pm.polyPlane( sx = 3, sy = 3 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		blackFaces = meshShape.f[0:3]
		TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces )
		blackTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, blackFaces )

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		guidesFromMesh = pm.PyNode( pm.mel.ls( type = TestUtilities.GuidesFromMeshNodeName )[0] )
		hairShape = pm.PyNode( pm.mel.ls( type = TestUtilities.HairShapeName )[0] )
		guidesFromMesh.distribution.set( 6 ) # Face center distribution

		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.MultiplierNodeName ) )
		multiplier.copies.set( 3 )
		multiplier.probability.set( 1.0 )

		strandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		strandIds = pm.mel.OxGetStrandIds( hairShape )
		strandIdsInBlackArea = [strandIds[i] for i in strandIndicesInBlackArea]

		multiplier.spreadTip.set( 0.0 )
		originalVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )
		multiplier.spreadTip.set( 50.0 )
		spreadVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )
		multiplier.spreadChannel.set( 2001 )
		newVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )

		TestUtilities.CheckPointsNotAllNearEqual( self, originalVerticesInBlackArea, spreadVerticesInBlackArea )
		TestUtilities.CheckPointsAllNearEqual( self, originalVerticesInBlackArea, newVerticesInBlackArea )

	def test_FluffVertexColorChannel( self ):
		plane = pm.polyPlane( sx = 3, sy = 3 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		blackFaces = meshShape.f[0:3]
		TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces )
		blackTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, blackFaces )

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		guidesFromMesh = pm.PyNode( pm.mel.ls( type = TestUtilities.GuidesFromMeshNodeName )[0] )
		hairShape = pm.PyNode( pm.mel.ls( type = TestUtilities.HairShapeName )[0] )
		guidesFromMesh.distribution.set( 6 ) # Face center distribution

		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.MultiplierNodeName ) )
		multiplier.copies.set( 3 )
		multiplier.probability.set( 1.0 )

		strandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		strandIds = pm.mel.OxGetStrandIds( hairShape )
		strandIdsInBlackArea = [strandIds[i] for i in strandIndicesInBlackArea]

		multiplier.fluffTip.set( 0.0 )
		originalVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )
		multiplier.fluffTip.set( 50.0 )
		fluffedVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )
		multiplier.fluffChannel.set( 2001 )
		newVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )

		TestUtilities.CheckPointsNotAllNearEqual( self, originalVerticesInBlackArea, fluffedVerticesInBlackArea )
		TestUtilities.CheckPointsAllNearEqual( self, originalVerticesInBlackArea, newVerticesInBlackArea )

	def test_ShiftVertexColorChannel( self ):
		plane = pm.polyPlane( sx = 3, sy = 3 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		blackFaces = meshShape.f[0:3]
		TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces )
		blackTriangleVectorsList = TestUtilities.GetFaceVertexPositionsAsVectors( meshShape, blackFaces )

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		guidesFromMesh = pm.PyNode( pm.mel.ls( type = TestUtilities.GuidesFromMeshNodeName )[0] )
		hairShape = pm.PyNode( pm.mel.ls( type = TestUtilities.HairShapeName )[0] )
		guidesFromMesh.distribution.set( 6 ) # Face center distribution

		multiplier = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.MultiplierNodeName ) )
		multiplier.copies.set( 3 )
		multiplier.probability.set( 1.0 )

		strandIndicesInBlackArea = TestUtilities.GetStrandIndicesInTriangles( hairShape, blackTriangleVectorsList )
		strandIds = pm.mel.OxGetStrandIds( hairShape )
		strandIdsInBlackArea = [strandIds[i] for i in strandIndicesInBlackArea]

		multiplier.shiftTip.set( 0.0 )
		originalVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )
		multiplier.shiftTip.set( 50.0 )
		shiftedVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )
		multiplier.shiftChannel.set( 2001 )
		newVerticesInBlackArea = TestUtilities.GetVerticesSubsetPositionsInObjectCoordinates( hairShape, strandIdsInBlackArea )

		TestUtilities.CheckPointsNotAllNearEqual( self, originalVerticesInBlackArea, shiftedVerticesInBlackArea )
		TestUtilities.CheckPointsAllNearEqual( self, originalVerticesInBlackArea, newVerticesInBlackArea )

