import MayaTest
import TestUtilities
import pymel.core as pm
import pymel.core.datatypes as dt
import pymel.core.runtime as pr
import maya.OpenMaya as om
import maya.OpenMayaUI as omUI
import maya.cmds as cmds
import datetime
import os
import unittest
import random

def SetBrushTool():
	try:
		pm.setToolTo( "oxBrushTool1" )
	except:
		pm.mel.eval( "OxBrushTool oxBrushTool1" )
		pm.setToolTo( "oxBrushTool1" )

def SetSelectTool():
	pm.setToolTo( 'selectSuperContext' )

class Test_EditGuidesShape( MayaTest.OxTestCase ):

	def ExecuteEditGuidesOffsetCommand( self, editGuidesShape ):
		editGuidesCommandArguments = '-eg ';
		editGuidesCommandArguments += str( pm.mel.OxGetStrandCount( editGuidesShape ) )
		for strandIndex in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			pointCount = pm.mel.OxGetStrandPointCount( editGuidesShape, strandIndex )
			editGuidesCommandArguments += ' ' + str( strandIndex )
			editGuidesCommandArguments += ' ' + str( pointCount )
			for pointIndex in ( 0, pointCount ):
				editGuidesCommandArguments += ' 4 0 0'

		# Guides shape needs to be selected for the command to actually function
		# In some tests it was silently failing before
		pm.select( editGuidesShape )
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + ' ' + editGuidesCommandArguments + ';' )

	def test_EditGuidesCommand( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		meshShape = plane[0]

		# Add guides from mesh with vertex distribution
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		# Add guide shape to it
		self.assertEqual( 1, len( pm.ls( type = TestUtilities.HairShapeName ) ) )
		self.assertEqual( 4, len( meshShape.vtx ) )
		self.assertEqual( len( meshShape.vtx ), pm.mel.OxGetStrandCount(editGuidesShape) )

		# Record initial vertex positions
		tipPositions = []
		for i in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			tipPositions.append( pm.mel.OxGetStrandPoint( editGuidesShape, i, 1 ) )

		# Move vertices using the edit guides command
		self.ExecuteEditGuidesOffsetCommand( editGuidesShape )

		# Vertex positions after editing should all be offset by 4 units
		for i in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			beforeAfterDelta = dt.Vector( pm.mel.OxGetStrandPoint( editGuidesShape, i, 1 ) ).distanceTo( tipPositions[i] )
			self.assertLess( abs( beforeAfterDelta - 4 ), 0.0001, str( beforeAfterDelta ) )

		# Undo the change
		pm.undo()

		# Guide tips should be back to original positons
		for i in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			beforeAfterDelta = dt.Vector( pm.mel.OxGetStrandPoint( editGuidesShape, i, 1 ) ).distanceTo( tipPositions[i] )
			self.assertLess( beforeAfterDelta, 0.0001 )

		# Redo the change and test for offset again
		pm.redo()
		for i in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			beforeAfterDelta = dt.Vector( pm.mel.OxGetStrandPoint( editGuidesShape, i, 1 ) ).distanceTo( tipPositions[i] )
			self.assertLess( abs( beforeAfterDelta - 4 ), 0.0001, str( beforeAfterDelta ) )

	# Tests for a bug where assertions in GuidesDelta are thrown after edit guides command is executed post scene load
	def test_EditGuidesCommandAfterSceneReload( self ):
		# Add guides from mesh with vertex distribution
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( pm.polyPlane( sx = 1, sy = 1 )[0], rootGenerationMethod = 4 )

		# Query strand count to make sure EditGuidesShape evaluates prior to saving
		pm.mel.OxGetStrandCount(editGuidesShape)

		# Save and reload the scene
		filePath = pm.saveAs( self.addTempFile( 'temp.ma' ) )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]

		# Try to edit guides
		self.ExecuteEditGuidesOffsetCommand( editGuidesShape )

		# If no errors happened, we are happy

	def test_SaveLoadEditedGuidesToBinaryBuffer( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		meshShape = plane[0]

		# Add guides from mesh with vertex distribution
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		# Move vertices using the edit guides command
		self.ExecuteEditGuidesOffsetCommand( editGuidesShape )

		# Record vertex positions for later comparison
		tipPositions = []
		for i in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			tipPositions.append( pm.mel.OxGetStrandPoint( editGuidesShape, i, 1 ) )

		# Save and reload the scene using ASCII serialization
		filePath = pm.saveAs( self.addTempFile( 'temp.ma' ) )
		pm.newFile( force = 1 )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.HairShapeName )[0]
		self.assertEqual( 4, pm.mel.OxGetStrandCount( editGuidesShape ) )
		TestUtilities.CheckTipPositionsEqual( self, editGuidesShape, tipPositions )

		# Save and reload the scene using binary serialization
		filePath = pm.saveAs( self.addTempFile( 'temp.mb' ) )
		pm.newFile( force = 1 )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.HairShapeName )[0]
		self.assertEqual( 4, pm.mel.OxGetStrandCount( editGuidesShape ) )

		# Compare the tip positions of each guide, they should be the same as before
		TestUtilities.CheckTipPositionsEqual( self, editGuidesShape, tipPositions )

	def SetupBrushTest( self, editGuidesShape ):
		pm.select( editGuidesShape )

		# Move guides from default 0,0,0 position to make sure brushes properly interpret object transforms
		pm.move( 0, 0, 10 )

		viewCamera = pm.ls( 'perspShape' )[0]
		viewCamera.setFocalLength( 35.0 )
		viewCamera.setCenterOfInterest( 17.9338065896 )
		viewCameraTransform = pm.listRelatives(viewCamera, parent=True, fullPath=True)[0]
		viewCameraTransform.setMatrix( [0.671720589323, -4.33680868994e-19, -0.740804596287, 0.0, -0.00437469919309, 0.999982563358, -0.00396673499979, 0.0, 0.740791679142, 0.00590533484136, 0.671708876772, 0.0, 7.35972487679, 2.27535619882, 16.8408901451, 1.0], worldSpace = True )

		# Need to refresh screen so updated transform is used in calculations
		pm.refresh()

		# Set brushing mode and assign comb brush
		#pm.mel.OxSetEditGuidesMode( 3, 0 )
		#pm.mel.OxSetCurrentBrush( 1 )
		SetBrushTool()

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_CombBrush( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( guideLength = 5.0, pointsPerStrandCount = 10 )
		self.SetupBrushTest( editGuidesShape )

		viewportName = 'modelPanel4'
		viewportWidth = pm.control(viewportName, query=True, width=True)
		viewportHeight = pm.control(viewportName, query=True, height=True)

		# Set move brush
		try:
			pm.mel.eval( "OxCombBrushTool oxCombBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxCombBrushTool1" )
		pm.mel.eval( "OxCombBrushTool -e -s " + str( viewportHeight / 2 ) + " oxCombBrushTool1" )

		# Get the positions of the four tip vertices before moving them with a brush
		strandTipsBeforeBrushing = [pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 1, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 2, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 3, 1 )]

		# Begin brushing in the center of the view in upper corner and drag brush horizontally to the right
		pm.mel.OxDragBrush( viewportWidth / 2.0, viewportHeight / 2.0, viewportWidth * ( 2.0 / 3.0 ), viewportHeight / 2.0, 3 )

		strandTipsAfterBrushing = [pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 1, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 2, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 3, 1 )]

		# Make sure that the strand tips have indeed moved due to brushing
		epsilon = 0.01
		for i in range(0, 3):
			distanceDelta = dt.Vector( strandTipsBeforeBrushing[i] ).distanceTo( dt.Vector( strandTipsAfterBrushing[i] ) )
			self.assertGreater( distanceDelta, epsilon )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_DeleteBrush( self ):
		editGuidesShape = TestUtilities.AddEditGuidesAboveHairToNewPlane()
		self.SetupBrushTest( editGuidesShape )

		viewportName = 'modelPanel4'
		viewportWidth = pm.control(viewportName, query=True, width=True)
		viewportHeight = pm.control(viewportName, query=True, height=True)

		# Set delete brush
		try:
			pm.mel.eval( "OxDeleteBrushTool oxDeleteBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxDeleteBrushTool1" )
		pm.mel.eval( "OxDeleteBrushTool -e -s " + str( viewportHeight / 2 ) + " oxDeleteBrushTool1" )

		# Get the positions of the four tip vertices before moving them with a brush
		vertexCountBeforeDelete = pm.mel.OxGetVertexCount( editGuidesShape )

		# Begin brushing in the center of the view in upper corner and drag brush horizontally to the right
		pm.mel.OxDragBrush( viewportWidth / 2.0, viewportHeight / 2.0, viewportWidth * ( 2.0 / 3.0 ), viewportHeight / 2.0, 3 )

		vertexCountAfterDelete = pm.mel.OxGetVertexCount( editGuidesShape )

		# Make sure that the strand tips have indeed moved due to brushing
		self.assertNotEqual( vertexCountBeforeDelete, vertexCountAfterDelete )

	def test_ControlPointMoveUndo( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		tipPositionsBeforeMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		pm.select( editGuidesShape + ".ep[2:3]" )

		tipPositionsAfterMove = []

		with MayaTest.UndoChunk():
			# Move the two selected guide control points
			pm.move( 0, 0, 5, relative = True )
			pm.mel.OxEditGuides( cc = True )

			tipPositionsAfterMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

			for i in range( 2, len( tipPositionsBeforeMove ) ):
				self.assertGreaterEqual( dt.Vector( tipPositionsBeforeMove[i] ).distanceTo( tipPositionsAfterMove[i] ), 3,
				'Move: i: %(i)d, before: %(before)s, after: %(after)s' % { 'i': i, 'before': tipPositionsBeforeMove[i], 'after' : tipPositionsAfterMove[i] } )

		pm.undo()

		tipPositionsAfterUndo = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		# After undo the guide tips should return to their previous positions
		for i in range( 2, len( tipPositionsBeforeMove ) ):
			self.assertLessEqual( dt.Vector( tipPositionsBeforeMove[i] ).distanceTo( tipPositionsAfterUndo[i] ), 0.0001,
			'Undo: i: %(i)d, before: %(before)s, after: %(after)s' % { 'i': i, 'before': tipPositionsBeforeMove[i], 'after' : tipPositionsAfterUndo[i] } )

		# Repeating the move should put them into same positions as previous move
		# TODO: Selection is cleared after undo since 18011, so we need to select again until this is fixed
		pm.select( editGuidesShape + ".ep[2:3]" )
		pm.move( 0, 0, 5, relative = True )
		pm.mel.OxEditGuides( cc = True )
		tipPositionsAfterSecondMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		for i in range( 2, len( tipPositionsAfterMove ) ):
			self.assertLessEqual( dt.Vector( tipPositionsAfterMove[i] ).distanceTo( tipPositionsAfterSecondMove[i] ), 0.0001,
			'Repeat: i: %(i)d, before: %(before)s, after: %(after)s' % { 'i': i, 'before': tipPositionsAfterMove[i], 'after' : tipPositionsAfterSecondMove[i] } )

	def test_ChangingGuideCountDoesNotResetGuideShapeChanges( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		meshShape = plane[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		# Move vertices using the edit guides command
		self.ExecuteEditGuidesOffsetCommand( editGuidesShape )

		# Record vertex positions for later comparison
		tipPositions = []
		for i in range( 0, pm.mel.OxGetStrandCount( editGuidesShape ) ):
			tipPositions.append( pm.mel.OxGetStrandPoint( editGuidesShape, i, 1 ) )

		# Change the underlying root distribution method to something else, evaluate, and then change it back to per-vertex
		guidesFromMeshNode = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		guidesFromMeshNode.distribution.set( 0 )
		pm.mel.OxGetStrandCount( editGuidesShape )
		guidesFromMeshNode.distribution.set( 4 )

		# Shape changes should still remain
		TestUtilities.CheckTipPositionsEqual( self, editGuidesShape, tipPositions )

	def test_SelectStrands( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		meshShape = plane[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		pm.select( editGuidesShape + '.f[1:2]' )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 1, 1, 0] )

		pm.select( clear = True ) #deselects shape, but preserves last selected strands
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 1, 1, 0] ) #check last selected strands

		# check selected strands can be filtered with Polygon Face selection mask
		pm.select( editGuidesShape + '.f[1:2]' )
		selectionList = cmds.ls( selection=True )
		filteredList = cmds.filterExpand( selectionList, expand=False, selectionMask=34 )
		self.assertEqual( selectionList, filteredList )

		# check selected strand EPs can be filtered with Edit Points selection mask
		pm.select( editGuidesShape + '.ep[1:2]' )
		selectionList = cmds.ls( selection=True )
		filteredList = cmds.filterExpand( selectionList, expand=False, selectionMask=30 )
		self.assertEqual( selectionList, filteredList )

	def test_ApplyOperatorsOnSelection( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		meshShape = plane[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )
		pm.select( editGuidesShape + '.f[1:2]' )

		mfs = pm.PyNode( pm.mel.OxAddStrandOperator( '', TestUtilities.MeshFromStrandsNodeName ) )
		self.assertTrue( mfs.exists() )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_SelectStrandsGui( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		meshShape = plane[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		view = omUI.M3dView.active3dView()
		pm.select( editGuidesShape )
		pm.mel.OxSetEditGuidesMode( 2, 0 )
		pm.refresh()

		selList = om.MSelectionList()
		om.MGlobal.selectFromScreen( 0, 0, view.portWidth(), view.portHeight(), om.MGlobal.kReplaceList )

		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [1, 1, 1, 1] )

	def test_PushAwayFromMesh( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewSphere( radius = 5, pointsPerStrandCount = 10, segmentCount = 10 )
		pm.select( editGuidesShape + ".ep[0:91]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )
		tipPositionsBeforePush = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		tipPositionsAfterPush = []

		# Push away from mesh
		with MayaTest.UndoChunk():
			pm.mel.OxPushAwayFromMesh( editGuidesShape, 0.1 )
			tipPositionsAfterPush = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
			TestUtilities.CheckPointsNotAllNearEqual( self, tipPositionsBeforePush, tipPositionsAfterPush )

		pm.undo()
		tipPositionsAfterUndo = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsAllNearEqual( self, tipPositionsBeforePush, tipPositionsAfterUndo )

	def test_PushAwayFromMeshPreservesStrandLength( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewSphere( radius = 5, pointsPerStrandCount = 50, segmentCount = 10 )
		strandLengthsBeforePush = TestUtilities.GetStrandLengths( editGuidesShape )

		editGuidesShape.autoPushAwayFromSurface.set( True )
		pm.select( editGuidesShape + ".ep[0:91]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )

		strandLengthsAfterPush = TestUtilities.GetStrandLengths( editGuidesShape )
		for lengthBefore, lengthAfter in zip( strandLengthsBeforePush, strandLengthsAfterPush ):
			self.assertAlmostEqual( lengthBefore, lengthAfter, delta = 0.0001 )


	def test_ChangesArePreservedWhenToggledInStack( self ):
		# Add edit guides to a plane and move the tips
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()

		pm.mel.OxShowHairStackDialog( "-show" )
		pm.mel.OxUpdateHairStackDialog()
		guidesShape = pm.mel.OxGetHairStackDialogNodes()[0]

		pm.select( editGuidesShape + ".ep[0:3]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )
		tipPositionsAfterMove = TestUtilities.GetTipPositions( guidesShape, useObjectCoordinates = True )

		# Toggle the on state of the edit guides operator in the hair stack, the result should be changed
		pm.mel.OxEnableOperator( editGuidesShape, False )
		tipPositionsAfterDisable = TestUtilities.GetTipPositions( guidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsNotAllNearEqual( self, tipPositionsAfterMove, tipPositionsAfterDisable )

		# Toggle the edit guides operator back on, it should return the guides modification
		pm.mel.OxEnableOperator( editGuidesShape, True )
		tipPositionsAfterEnable = TestUtilities.GetTipPositions( guidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsAllNearEqual( self, tipPositionsAfterMove, tipPositionsAfterEnable )

	def test_PlantRoot( self ):
		# Create an edit guides on top of guides without any data
		editGuidesShape = TestUtilities.AddEditGuidesToNewSphere( radius = 5, guideCount = 1, rootGenerationMethod = 2 )

		self.assertEqual( 1, pm.mel.OxGetStrandCount( editGuidesShape ) )

		# select a strand, so we can copy points from it
		pm.select( editGuidesShape + ".f[0]", r = True );

		# Plant two roots
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + ' -cg 1 0 0 10 0.648876 0.259276 0.0918481 -cv 10 0 0 0 0 0 1.111111 0 0 2.222222 0 0 3.333333 0 0 4.444445 0 0 5.555555 0 0 6.666667 0 0 7.777778 0 0 8.888889 0 0 10 ;' )
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + ' -cg 1 0 0 10 0 0 -cv 10 0 0 0 0 0 1.111111 0 0 2.222222 0 0 3.333333 0 0 4.444445 0 0 5.555555 0 0 6.666667 0 0 7.777778 0 0 8.888889 0 0 10 ;' )

		# There should now be two roots
		self.assertEqual( 3, pm.mel.OxGetStrandCount( editGuidesShape ) )

		# Only the one previous root should be selected
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [1, 0, 0] )

		# Make sure the generated strand IDs are the same on all platforms
		strandIds = pm.mel.OxGetStrandIds( editGuidesShape )
		self.assertEqual( [240000, 1295363467, 1148853900], strandIds )

		tipPositionsBeforeSave = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		TestUtilities.SaveAndReloadScene( self, 'test_PlantRoot', False )

		editGuidesShape = pm.ls( type = TestUtilities.HairShapeName )[0]
		tipPositionsAfterLoad = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		self.assertGreaterEqual( 0.001, dt.Vector( tipPositionsBeforeSave[0] ).distanceTo( tipPositionsAfterLoad[0] ) )
		self.assertGreaterEqual( 0.001, dt.Vector( tipPositionsBeforeSave[1] ).distanceTo( tipPositionsAfterLoad[1] ) )

	def TestDeletePerRootChannel( self, editGuidesShape ):
		pm.select( editGuidesShape )

		# Initially we just have the selection channel
		self.assertEqual( 1, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )

		# Add a new channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -crc 1 Clustering;" )
		self.assertEqual( 2, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )

		# Delete the channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -drc 2 0 1;" )
		self.assertEqual( 1, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )

	def test_DeletePerRootChannel( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		self.TestDeletePerRootChannel( editGuidesShape )

	def test_DeletePerRootChannelOnHair( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()

		# Add hair before edit guides shape
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		hairFromGuides = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.HairFromGuidesNodeName ) )

		self.TestDeletePerRootChannel( editGuidesShape )

	def test_RenameChannel( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		pm.select( editGuidesShape )

		# Add a new channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -crc 1 Test1;" )
		self.assertEqual( "Test1", pm.mel.OxGetRootChannels( editGuidesShape )[1] )

		# Rename the channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + ' -rcn 1 1 "Test2";' )
		self.assertEqual( "Test2", pm.mel.OxGetRootChannels( editGuidesShape )[1] )

	# Tests for a bug where a crash would happen if previously edited per-vertex data would accept a new channel
	def test_AddPerVertexChannelAfterEditingPerVertexValues( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		pm.select( editGuidesShape )

		# Modify some values in selection per-vertex channel
		pm.mel.eval( "OxEditGuides -vcd 1 0 1 0 2 0.5 0.5;" )

		# Add a new per-vertex channel, this shouldn't cause a crash
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -cvc 1 Test;" )
		self.assertEqual( 1, len( pm.mel.OxGetVertexChannels( editGuidesShape ) ) )

	def TestPerVertexChannelValuesAreSavedWithScene( self, fileExtension ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		pm.select( editGuidesShape )
		pm.mel.eval( TestUtilities.EditGuidesCommandName + " -cvc 1 Test;" )
		pm.mel.eval( TestUtilities.EditGuidesCommandName + " -vcd 1 0 1 0 2 0.5 0.5;" )

		# Save and reload the scene, the per-vertex value changes we did above should still be there
		fileName = 'temp.' + fileExtension
		filePath = pm.saveAs( fileName )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.HairShapeName )[0]
		self.assertEqual( ["Test"], pm.mel.OxGetVertexChannels( editGuidesShape ) )
		self.assertSequenceAlmostEqual( [0.5, 0.5, 0, 0, 0, 0, 0, 0], pm.mel.OxGetVertexValuesForChannel( editGuidesShape, 0 ) )

	def test_PerVertexChannelValuesAreSavedWithAsciiScene( self ):
		self.TestPerVertexChannelValuesAreSavedWithScene( 'ma' )

	def test_PerVertexChannelValuesAreSavedWithBinaryScene( self ):
		self.TestPerVertexChannelValuesAreSavedWithScene( 'mb' )

	def test_DeletingPerVertexChannelAlsoDeletesItsModifiedValues( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		pm.select( editGuidesShape )

		# Add a new per-vertex channel and modify its values
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -cvc 1 Test;" )
		# The channel should have all zero values initially
		self.assertSequenceAlmostEqual( [0, 0, 0, 0, 0, 0, 0, 0], pm.mel.OxGetVertexValuesForChannel( editGuidesShape, 0 ) )
		pm.mel.eval( "OxEditGuides -vcd 1 1 1 0 2 0.2 0.1;" )
		self.assertSequenceAlmostEqual( [0, 0, 0.2, 0.1, 0, 0, 0, 0], pm.mel.OxGetVertexValuesForChannel( editGuidesShape, 0 ) )

		# delete the created per-vertex channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -dvc 1 1" )

		# Create per-vertex channel again and make sure that its values are all zero
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -cvc 1 Test;" )
		self.assertSequenceAlmostEqual( [0, 0, 0, 0, 0, 0, 0, 0], pm.mel.OxGetVertexValuesForChannel( editGuidesShape, 0 ) )

	def test_DeleteGuides( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()

		vertexCountBeforeDelete = pm.mel.OxGetVertexCount( editGuidesShape )
		pm.mel.OxEditGuides( 4, 0, 0, 0, 1, ds = True )

		vertexCountAfterDelete = pm.mel.OxGetVertexCount( editGuidesShape )
		self.assertEqual( vertexCountBeforeDelete - 2, vertexCountAfterDelete )

		pm.select( editGuidesShape )
		hairFromGuides = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.HairFromGuidesNodeName ) )

		pm.mel.OxDeleteStrandOperator( hairFromGuides )
		vertexCountAfterOperatorAddRemove = pm.mel.OxGetVertexCount( editGuidesShape )
		self.assertEqual( vertexCountBeforeDelete - 2, vertexCountAfterOperatorAddRemove )

	def test_DeleteSelectedGuides( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( rootGenerationMethod = 2, guideCount = 20 )

		# Remember how many guides we had initially
		initialGuideCount = pm.mel.OxGetStrandCount( editGuidesShape )

		# Select and delete first two guides
		pm.select( editGuidesShape + ".ep[10:12]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape )

		# There should be 3 fewer guides now that we deleted them
		self.assertEqual( initialGuideCount - 3, pm.mel.OxGetStrandCount( editGuidesShape ) )

	def test_DeleteGuidesInMultipleEditGuidesOperators( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		editGuidesShape2 = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.EditGuidesShapeName ) )
		editGuidesShape2.changeTracking.set( 1 )

		#pm.mel.OxEditGuides( 1, 0, dg = True )
		#pm.mel.eval( "OxEditGuides -dg 1 0;" )
		pm.select( editGuidesShape + ".f[0]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape )

		guideCountBeforeDelete = pm.mel.OxGetStrandCount( editGuidesShape2 )
		#pm.mel.OxEditGuides( 1, 1, dg = True )
		#pm.mel.eval( "OxEditGuides -dg 2 2 3;" )
		pm.select( editGuidesShape2 + ".f[1:2]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape2 )
		guideCountAfterDelete = pm.mel.OxGetStrandCount( editGuidesShape2 )

		self.assertEqual( guideCountBeforeDelete, guideCountAfterDelete + 2 )

	def test_MoveGuidesAfterDecreasingIncomingPointCount( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 5, rootGenerationMethod = 4 )

		# Evaluate the stack
		pm.mel.OxGetVertexCount( editGuidesShape )
		pm.mel.OxEditGuides( 1, 1, 1, 0, 5, 0.5, 0.5, 0.5, 0.5, 0.5, vcd = True )
		pm.select( editGuidesShape + ".ep[0:1]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )
		pm.mel.OxGetVertexCount( editGuidesShape )

		# Reduce point count to 2
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		guidesFromMesh.pointCount.set( 2 )
		guidesFromMesh.distribution.set( 0 )
		guidesFromMesh.c.set( 20 )
		pm.mel.OxGetVertexCount( editGuidesShape )
		pm.mel.OxEditGuides( 1, 10, 1, 0, 2, 1.0, 1.0, vcd = True )
		pm.select( editGuidesShape + ".ep[10:12]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )

		# No crash should occur at this point

	def test_IncreaseInputGuideCountWithSubsequentEditing( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, rootGenerationMethod = 0, guideCount = 4 )
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]

		# Evaluate the stack
		pm.mel.OxGetVertexCount( editGuidesShape )

		# Increase input guide count
		guidesFromMesh.c.set( 10 )
		strandCount = pm.mel.OxGetStrandCount( editGuidesShape )

		tipPositionsBeforeMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		# Move all guide tips
		pm.select( editGuidesShape + ".ep[0:" + str( strandCount - 1 ) + "]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )

		# Make sure that all guide tips indeed moved
		tipPositionsAfterMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		TestUtilities.CheckPointsAllNotNearEqual( self, tipPositionsBeforeMove, tipPositionsAfterMove )

	def test_ApplyAmount( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, rootGenerationMethod = 0, guideCount = 4 )
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		strandCount = pm.mel.OxGetStrandCount( editGuidesShape )

		# Move all guide tips
		tipPositionsBeforeMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		pm.select( editGuidesShape + ".ep[0:" + str( strandCount - 1 ) + "]" )
		pm.move( 0, 0, 20, relative = True )
		pm.mel.OxEditGuides( cc = True )

		# Make sure that all guide tips indeed moved
		tipPositionsAfterMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsAllNotNearEqual( self, tipPositionsBeforeMove, tipPositionsAfterMove )

		# Reduce apply amount, the tips should've been moved
		editGuidesShape.applyAmount.set( 0.5 )
		tipPositionsAfterApplyAmount = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsAllNotNearEqual( self, tipPositionsBeforeMove, tipPositionsAfterApplyAmount )
		TestUtilities.CheckPointsAllNotNearEqual( self, tipPositionsAfterMove, tipPositionsAfterApplyAmount )

		# Apply amount of 0 should be same as before moving
		editGuidesShape.applyAmount.set( 0 )
		tipPositionsAfterApplyAmount = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsAllNearEqual( self, tipPositionsBeforeMove, tipPositionsAfterApplyAmount )

		editGuidesShape.applyAmount.set( 1 )
		tipPositionsAfterApplyAmount = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		TestUtilities.CheckPointsAllNearEqual( self, tipPositionsAfterMove, tipPositionsAfterApplyAmount )

	def test_AdoptFromStackTop( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, rootGenerationMethod = 0, guideCount = 4 )
		originalVertices = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )
		curlNode = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.CurlNodeName ) )
		curlNode.magnitude.set( 100 )

		guidesShape = pm.ls( type = TestUtilities.HairShapeName )[0]
		verticesAfterCurl = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		TestUtilities.CheckPointsNotAllNearEqual( self, originalVertices, verticesAfterCurl )

		pm.select( editGuidesShape + ".ep[0:3]" )

		pm.mel.OxAdoptFromStackTop( editGuidesShape )

		verticesAfterAdoptFromStrandTop = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )

		# Vertices should have changed since we're applying curl on top of already curled guides
		TestUtilities.CheckPointsNotAllNearEqual( self, verticesAfterCurl, verticesAfterAdoptFromStrandTop )

		pm.mel.OxEnableOperator( curlNode, 0 )
		verticesAfterTurningOffCurl = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )

		TestUtilities.CheckPointsAllNearEqual( self, verticesAfterCurl, verticesAfterTurningOffCurl )

	def ImportMultipleCurvesExecuteAndCheck( self, line1Point1, line1Point2, line2Point1, line2Point2, initialVertexCount, plantedPointCount, curve1, curve2, editGuidesShape ):
		#add guides from curves
		pm.mel.eval( "OxImportMultipleCurves -p " + str( plantedPointCount ) + " -t 1 -c \"" + curve1+ " " + curve2 + "\" " + editGuidesShape + ";" )
		#assert there are now 6 strands
		self.assertEqual( pm.mel.OxGetStrandCount( editGuidesShape ), 6 )

		strandVertices = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		TestUtilities.CheckPointsAllNearEqual( self, line1Point1 , strandVertices[initialVertexCount + plantedPointCount + 1] )
		TestUtilities.CheckPointsAllNearEqual( self, line1Point2 , strandVertices[initialVertexCount + plantedPointCount + 3 ] )
		TestUtilities.CheckPointsAllNearEqual( self, line2Point1 , strandVertices[initialVertexCount + plantedPointCount * 2 + 1] )
		TestUtilities.CheckPointsAllNearEqual( self, line2Point2 , strandVertices[initialVertexCount + plantedPointCount * 2 + 3] )

	def test_ImportMultipleCurves( self ):
		#make stuff
		line1Point1 = [0,0,0]
		line1Point2 = [-1,10,5]
		line2Point1 = [5,0,0]
		line2Point2 = [1,10,5]

		curve1 = pm.curve( p=[line1Point1, line1Point2], k=[0,1], d=1 )
		curve2 = pm.curve( p=[line2Point1, line2Point2], k=[0,1], d=1 )

		plane = pm.polyPlane( sx = 1, sy = 1, w=25, h=25)
		pm.select( plane )
		meshShape = plane[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		guidesFromMeshNode = pm.mel.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromMeshNode, TestUtilities.RotateNodeName ) )

		#checking
		initialVertexCount = len( meshShape.vtx )
		plantedPointCount = 3

		#assert there are 4 strands
		self.assertEqual( pm.mel.OxGetStrandCount( editGuidesShape ), 4 )

		self.ImportMultipleCurvesExecuteAndCheck( line1Point1, line1Point2, line2Point1, line2Point2, initialVertexCount, plantedPointCount, curve1, curve2, editGuidesShape )

		# Simulate a view redraw by asking for bounding box. This is where hair will be put into object space coordinates.
		pm.exactWorldBoundingBox( editGuidesShape )

		# Undo the change
		pm.undo()

		pm.exactWorldBoundingBox( editGuidesShape )

		# Re-plant and test again
		self.ImportMultipleCurvesExecuteAndCheck( line1Point1, line1Point2, line2Point1, line2Point2, initialVertexCount, plantedPointCount, curve1, curve2, editGuidesShape )

	def test_ImportMultipleCurvesReplaceExistingGuides( self ):
		#make stuff
		line1Point1 = [5,0,5]
		line1Point2 = [-1,10,5]
		line2Point1 = [-4.5,0,5]
		line2Point2 = [1,10,5]

		curve1 = pm.curve( p=[line1Point1, line1Point2], k=[0,1], d=1 )
		curve2 = pm.curve( p=[line2Point1, line2Point2], k=[0,1], d=1 )

		plane = pm.polyPlane( sx = 1, sy = 1, w = 10, h = 10)
		pm.select( plane )
		meshShape = plane[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, rootGenerationMethod = 4 )

		#checking
		initialVertexCount = len( meshShape.vtx )
		plantedPointCount = 3

		#assert there are 4 guides
		self.assertEqual( pm.mel.OxGetStrandCount( editGuidesShape ), 4 )

		#add guides from curves
		pm.mel.eval( "OxImportMultipleCurves -p " + str(plantedPointCount)+" -t 1 -c \""+ curve1+ " "+ curve2 + "\" "+ editGuidesShape+";" )
		#after import we should still have 4 guides since newly added ones were merged in
		self.assertEqual( pm.mel.OxGetStrandCount( editGuidesShape ), 4 )

		existingGuidePointCount = 2
		strandVertices = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		TestUtilities.CheckPointsAllNearEqual( self, line1Point1 , strandVertices[2] )
		TestUtilities.CheckPointsAllNearEqual( self, line1Point2 , strandVertices[3] )

		# Only the direction vector of the second line should match because it's not aligned with the strand root
		TestUtilities.CheckPointsAllNearEqual( self, dt.Vector( line2Point2 ) - dt.Vector( line2Point1 ), dt.Vector( strandVertices[1] ) - dt.Vector( strandVertices[0] ) )

	def test_DeleteStrandsAndImportMultipleCurves( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewSphere( radius=5, rootGenerationMethod=0, guideCount=10, pointsPerStrandCount = 10, segmentCount = 10 )

		# assert there are 10 guides
		self.assertEqual( pm.mel.OxGetStrandCount( editGuidesShape ), 10 )

		# delete the guides
		pm.select( editGuidesShape + ".ep[0:9]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape )

		# add some curves
		curve1 = pm.curve( p=[(0, 0, 0), (3, 5, 6), (5, 6, 7), (9, 9, 9)] )
		curve2 = pm.curve( p=[(0, 0, 0), (5, 5, 6), (5, 6, 7), (15, 9, 9)] )
		curve3 = pm.curve( p=[(0, 0, 0), (5, 10, 6), (5, 6, 7), (0, 9, 9)] )

		selected = pm.ls(type='nurbsCurve')
		pm.select(selected, replace=True)

		#add guides from curves
		pm.mel.eval( "OxImportMultipleCurves -c \"" + curve1 + " " + curve2 + " " + curve3 + "\" "+ editGuidesShape+";" )
		#after import we should have 3 guides
		self.assertEqual( pm.mel.OxGetStrandCount( editGuidesShape ), 3 )

	def test_AddingChannelOnTopOfGuideGroupsDoesNotMutateGroups( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )

		# At this point all 4 guides should be in strand group 0
		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		self.assertEqual( [1, 1, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Add a new guide channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -crc 1 Test;" )

		# The four guides should still be in group 0
		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		self.assertEqual( [1, 1, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

	def test_AssignAndSetStrandGroupCommands( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )

		pm.select( editGuidesShape + ".ep[3]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "1" )

		# One guide in group 1
		pm.mel.OxSelectStrandGroup( editGuidesShape, "1" )
		self.assertEqual( [0, 0, 0, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Select again for consistency
		pm.mel.OxSelectStrandGroup( editGuidesShape, "1" )
		self.assertEqual( [0, 0, 0, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		self.assertEqual( [1, 1, 1, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		pm.select( editGuidesShape + ".ep[1:2]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "2" )

		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		self.assertEqual( [1, 0, 0, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		pm.mel.OxSelectStrandGroup( editGuidesShape, "1" )
		self.assertEqual( [0, 0, 0, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		pm.mel.OxSelectStrandGroup( editGuidesShape, "2" )
		self.assertEqual( [0, 1, 1, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Undo the change
		pm.undo()
		self.assertEqual( [0, 0, 0, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Test redo
		pm.redo()
		self.assertEqual( [0, 1, 1, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

	def test_AssignAndSelectUniqueStrandGroup( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )

		pm.select( editGuidesShape + ".ep[2:3]" )
		self.assertEqual( [0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Strand group argument of -1 will assign a unique strand group to the guides
		pm.mel.OxAssignStrandGroup( editGuidesShape, "-1" )

		# Two guides left in group 0
		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		self.assertEqual( [1, 1, 0, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Two guides were assigned (unique) group 1
		pm.mel.OxSelectStrandGroup( editGuidesShape, "1" )
		self.assertEqual( [0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

	# Tests for a bug where strands cannot be assigned another strand group when they were previously assigned a group
	def test_ReassignStrandGroup( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )
		pm.select( editGuidesShape + ".ep[2:3]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "1" )
		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		pm.mel.OxSelectStrandGroup( editGuidesShape, "1" )
		self.assertEqual( [0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Reassign strand group and select it
		pm.mel.OxAssignStrandGroup( editGuidesShape, "2" )
		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		pm.mel.OxSelectStrandGroup( editGuidesShape, "2" )
		self.assertEqual( [0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

	def test_SelectingGuidesInRootModeSetsGuideAndVertexSelectionChannels( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		pm.select( editGuidesShape + ".ep[2:3]" )
		self.assertEqual( [0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )
		self.assertEqual( 1, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )
		self.assertEqual( [0.0, 0.0, 1.0, 1.0], pm.mel.OxGetRootValuesForChannel( editGuidesShape, 0 ) )

	def test_RotateStrands( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )

		perStrandRotationAngle = pm.mel.OxGetPerStrandRotationAngle( editGuidesShape, 1 )
		self.assertIsNotNone( perStrandRotationAngle  )

		pm.select( editGuidesShape + '.f[1]' )
		pm.mel.eval( "OxEditRootsRotateStrand -r -ra 45;" )
		perStrandRotationAngleNew = pm.mel.OxGetPerStrandRotationAngle( editGuidesShape, 1 )
		self.assertIsNotNone(perStrandRotationAngleNew )

		degreesToRadians = 3.1415926535897932384626433832795 / 180.0

		self.assertAlmostEqual( perStrandRotationAngle + 45 * degreesToRadians, perStrandRotationAngleNew, places = 4 )

		#call command the second time
		perStrandRotationAngle = pm.mel.OxGetPerStrandRotationAngle( editGuidesShape, 1 )
		self.assertIsNotNone(perStrandRotationAngle )

		pm.select( editGuidesShape + '.f[1]' )
		pm.mel.eval( "OxEditRootsRotateStrand -r -ra 15;" )
		perStrandRotationAngleNew = pm.mel.OxGetPerStrandRotationAngle( editGuidesShape, 1 )
		self.assertIsNotNone(perStrandRotationAngleNew )

		self.assertAlmostEqual( perStrandRotationAngle + 15 * degreesToRadians, perStrandRotationAngleNew, places = 4 )

	def test_RotateStrandsDoesntChangePoints( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )

		# Modify the guides before EG so we can measure any potential changes to point rotations
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		frizz = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.FrizzNodeName ) )
		frizz.amount.set( 20 )

		verticesBeforeRotate = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# Rotate second guide
		pm.select( editGuidesShape + '.f[1]' )
		pm.mel.eval( "OxEditRootsRotateStrand -r -ra 45;" )

		verticesAfterRotate = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeRotate, verticesAfterRotate )

	def test_EditShapeAfterMovingGuide( self ):
		# Issue 3888: https://ephere.com:3000/issues/3888
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 4, guideCount = 1, rootGenerationMethod = 7, planeSegmentCount = 1 )
		editGuidesShape.preserveStrandLength.set( False )
		editGuidesShape.optimizeStrandGeometry.set( False )
		editGuidesShape.useStrandIk.set( False )
		editGuidesShape.controlPointInfluence.set( 0 )

		# Move the root into another mesh triangle
		pm.select( editGuidesShape + '.f[0]' )
		verticesBeforeRootMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )
		pm.move( -0.5, 0, 0, relative = True )
		pm.mel.OxEditGuides( cc = True )
		verticesBeforeMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )
		# Verify root move operation is successful
		for vertex in verticesBeforeRootMove:
			vertex[0] -= 0.5
		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeMove, verticesBeforeRootMove )

		# Move only the tip of the strand
		pm.select( editGuidesShape + '.ep[0]' )
		pm.move( 5, 0, 0, relative = True )
		pm.mel.OxEditGuides( cc = True )
		verticesAfterMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# This should be the only effect of the move
		verticesBeforeMove[3][0] += 5
		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeMove, verticesAfterMove )

	def test_MoveGuideAfterEditingShape( self ):
		# Issue 3888: https://ephere.com:3000/issues/3888
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 4, guideCount = 1, rootGenerationMethod = 7, planeSegmentCount = 1 )
		editGuidesShape.preserveStrandLength.set( False )
		editGuidesShape.optimizeStrandGeometry.set( False )
		editGuidesShape.useStrandIk.set( False )
		editGuidesShape.controlPointInfluence.set( 0 )

		# Move only the tip of the strand
		pm.select( editGuidesShape + '.ep[0]' )
		pm.move( 5, 0, 0, relative = True )
		pm.mel.OxEditGuides( cc = True )
		verticesBeforeRootMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# Move the root into another mesh triangle
		pm.select( editGuidesShape + '.f[0]' )
		pm.move( -0.5, 0, 0, relative = True )
		pm.mel.OxEditGuides( cc = True )
		verticesAfterRootMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		for vertex in verticesBeforeRootMove:
			vertex[0] -= 0.5
		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeRootMove, verticesAfterRootMove )

	def test_EnablingGuideGroupsAfterCreatingTheChannelShouldntAddAnotherChannel( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )

		# Initially we just have the selection channel
		self.assertEqual( 1, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )

		# Manually create a per-root "Groups" channel
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -crc 1 Groups;" )
		self.assertEqual( 2, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )

		# Enable guide groups
		editGuidesShape.useStrandGroups.set( True )

		# Number of root channels should remain two since Groups channel previous created needs to be reused
		self.assertEqual( 2, len( pm.mel.OxGetRootChannels( editGuidesShape ) ) )

	def test_AssigningStrandGroupInUnusedEditGuides( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape2 = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.EditGuidesShapeName ) )

		# We should have no problem assigning strand groups
		editGuidesShape2.useStrandGroups.set( True )

		pm.select( editGuidesShape2 + ".ep[1]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape2, "-1" )

		# One guide in group 1
		pm.mel.OxSelectStrandGroup( editGuidesShape2, "1" )
		self.assertEqual( [0, 1, 0, 0], pm.mel.OxGetSelectedStrands( editGuidesShape2 ) )

	# Tests that using "Use Strand Groups" option on EG preserves the incoming strand groups
	def test_InputGuidesStrandGroupsArePreserved( self ):
		# Create Edit Guides node and assign last two guides strand group of 1
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )
		pm.select( editGuidesShape + ".ep[2:3]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "1" )

		# Add another Edit Guides node on top with enabled strand groups
		editGuidesShape2 = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.EditGuidesShapeName ) )
		editGuidesShape2.useStrandGroups.set( True )
		pm.select( editGuidesShape2 + ".ep[0]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape2, "2" )

		# If the strand group from first Edit Guides shape was preserved we should be able to select the last two guides by group value of 1
		pm.mel.OxSelectStrandGroup( editGuidesShape2, "1" )
		self.assertEqual( [0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape2 ) )
		pm.mel.OxSelectStrandGroup( editGuidesShape2, "2" )
		self.assertEqual( [1, 0, 0, 0], pm.mel.OxGetSelectedStrands( editGuidesShape2 ) )

	def test_HidingGuidesPreservesStrandGroups( self ):
		# Set strand group of 1 to first and last guides
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )
		pm.select( editGuidesShape + ".ep[0]" )
		pm.select( editGuidesShape + ".ep[3]", add = True )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "1" )
		self.assertEqual( [1, 0, 0, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# Hide the selected guides, we should have an empty selection (hidden strands are deselected)
		pm.mel.OxSetSelectedGuidesVisibility( hide = True );
		self.assertEqual( [0, 0, 0, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

		# The first two strands should retain their strand group of 0
		pm.mel.OxSelectStrandGroup( editGuidesShape, "0" )
		self.assertEqual( [0, 1, 1, 0], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

	def test_HiddenGuidesPreservedWhenReloadingAsciiScene( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		pm.select( editGuidesShape + ".ep[0]" )
		pm.select( editGuidesShape + ".ep[3]", add = True )

		# Hide the selected guides
		pm.mel.OxSetSelectedGuidesVisibility( hide = True );
		self.assertEqual( [0, 1, 1, 0], pm.mel.OxGetVisibleStrands( editGuidesShape ) )

		# Save and reload the scene into an Ascii file, making sure that hidden guides survive the change
		filePath = pm.saveAs( self.addTempFile( 'temp.ma' ) )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		self.assertEqual( [0, 1, 1, 0], pm.mel.OxGetVisibleStrands( editGuidesShape ) )

	# Reproduces a crash when users turn off using guide groups after manually deleting the groups channel
	def test_DeleteGuideGroupChannelAndTurnOffGroupChannels( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2 )
		editGuidesShape.useStrandGroups.set( True )
		pm.mel.OxGetSelectedStrands( editGuidesShape )
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -drc 2 0 1;" )
		pm.mel.OxGetSelectedStrands( editGuidesShape )
		editGuidesShape.useStrandGroups.set( False )
		pm.mel.OxGetSelectedStrands( editGuidesShape )

	def TestMovingRootsDoesNotChangeStrandShape( self, editGuidesShape, moveTestVectors ):
		# Select and move the first root
		pm.select( editGuidesShape + ".f[0]" )
		strand1TangentBeforeMoving = dt.Vector( pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ) ) - dt.Vector( pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 ) )

		for moveTestVector in moveTestVectors:
			## The root should have been moved
			pm.move( moveTestVector[0], moveTestVector[1], moveTestVector[2], relative = True )
			pm.mel.OxEditGuides( cc = True )

			strand1TangentAfterMoving = dt.Vector( pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ) ) - dt.Vector( pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 ) )

			# The tangent direction of the strand should have not changed
			error = ( strand1TangentBeforeMoving - strand1TangentAfterMoving ).length()
			self.assertAlmostEqual( 0.0, error, places = 2, msg = "moveTestVector: " + str( moveTestVector ) + ", error: " +
						  str( error ) + ", Tangent before move: " + str( strand1TangentBeforeMoving ) + " != Tangent after move: " + str( strand1TangentAfterMoving ) )

	def test_MoveGuideToADifferentFace( self ):
		# Generate 4 guides with face-center distribution and edit them
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 4 )
		guidesFromSurfaceNode = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]

		# Change guide shapes by applying a surface comb node to them
		surfaceComb = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromSurfaceNode, TestUtilities.SurfaceCombNodeName ) )
		surfaceComb.algorithm.set( 0 )

		moveTestVectors = [[0.71, 0, -0.87], [-0.5, 0, -0.1], [0.8, 0, 0.79]]
		self.TestMovingRootsDoesNotChangeStrandShape( editGuidesShape, moveTestVectors )

	def test_MoveRotatedGuideToADifferentFace( self ):
		# Generate 4 guides with face-center distribution and edit them
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 4 )
		guidesFromSurfaceNode = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]

		# Change guide shapes by applying a surface comb node to them and rorate them
		surfaceCombNode = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromSurfaceNode, TestUtilities.SurfaceCombNodeName ) )
		surfaceCombNode.algorithm.set( 0 )
		rotateNode = pm.PyNode( pm.mel.OxAddStrandOperator( surfaceCombNode, TestUtilities.RotateNodeName ) )
		rotateNode.globalAngle.set( 2 )

		self.TestMovingRootsDoesNotChangeStrandShape( editGuidesShape, [[0.84, 0, -0.23]] )

	def test_MoveGuideAfterDeletingOtherGuides( self ):
		# Generate 4 guides with face-center distribution and edit them
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 6, guideCount = 10 )

		# Select and delete last two guides
		pm.select( editGuidesShape + ".f[1]" )
		pm.select( editGuidesShape + ".f[2]", add = True )
		pm.select( editGuidesShape + ".f[3]", add = True )
		pm.mel.eval( "OxEditGuides -dg 3 1 2 3;" )

		# Select and try to move the first root
		pm.select( editGuidesShape + ".f[0]" )

		root5PositionBeforeMoving = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 )

		# The root should have been moved
		pm.move( 0, 0, 0.2, relative = True )
		pm.mel.OxEditGuides( cc = True )
		root5PositionAfterMoving1 = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 )
		self.assertAlmostEqual( root5PositionBeforeMoving[2] + 0.2, root5PositionAfterMoving1[2] )

		# Moving the root another 0.2 units along z coordinate should produce incremented z value
		pm.move( 0, 0, -0.3, relative = True )
		pm.mel.OxEditGuides( cc = True )
		root5PositionAfterMoving2 = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 )
		self.assertAlmostEqual( root5PositionAfterMoving1[2] - 0.3, root5PositionAfterMoving2[2] )

	def test_MoveGuideAfterDeletingOtherGuidesAndShapeEditing( self ):
		# Generate 4 guides with face-center distribution and edit them
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 6, guideCount = 10 )

		#verticesBeforeDelete = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# Select and delete last two guides
		pm.select( editGuidesShape + ".f[1]" )
		pm.select( editGuidesShape + ".f[3]", add = True )
		#pm.select( editGuidesShape + ".f[3]", add = True )
		#pm.mel.eval( "OxEditGuides -dg 2 1 3;" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape )

		# Select and move the remaining two roots control points
		pm.select( editGuidesShape + ".ep[0:1]" )
		pm.move( 1, 0, 1, relative = True )
		pm.mel.OxEditGuides( cc = True )

		verticesAfterCpMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# Now move both roots
		pm.select( editGuidesShape + ".f[0:1]" )
		pm.move( 1, 0, 1, relative = True )
		pm.mel.OxEditGuides( cc = True )

		verticesAfterRootMove = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# The tangent vector of both guides should not have changed after moving the roots
		for i in range( 0, 2 ):
			tangentBefore = dt.Vector( verticesAfterCpMove[i * 2 + 1] ) - verticesAfterCpMove[i * 2]
			tangentAfter = dt.Vector( verticesAfterRootMove[i * 2 + 1] ) - verticesAfterRootMove[i * 2]

			self.assertAlmostEqual( 0, tangentBefore.distanceTo( tangentAfter ), places = 5 )

		# TODO: Following undo undoes everything including root deletion, but we only need to undo the latest root move. Figure it out and remove the comments.
		## Undo the root move
		#pm.undo()

		#verticesAfterUndo = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )
		#TestUtilities.CheckPointsAllNearEqual( self, verticesAfterCpMove, verticesAfterUndo )

	# Reproduces a bug where a guide would "jump around" when it is moved across mesh triangles, if it was previously created and if its shape was modified
	def test_MoveGuideAfterCreatingAndModifyingItsShape( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 1, rootGenerationMethod = 4, guideLength = 0.5 )

		# Add surface comb operator to make sure all guides are pointing away from center
		guidesFromSurfaceNode = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		surfaceComb = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromSurfaceNode, TestUtilities.SurfaceCombNodeName ) )
		surfaceComb.algorithm.set( 0 )

		# Plant a new guide
		pm.select( editGuidesShape + ".f[0]" );

		# Plant two roots
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + ' -cg 1 0 0 2 0.00277995 0.60664 0.39058 -cv 2 0 0 0 0.306737 -0.340513 0;' )

		# Modify the shape of newly created guide
		pm.select( editGuidesShape + ".ep[4]" );
		pm.move( 0.2, 0, 0, relative = True )
		pm.mel.OxEditGuides( cc = True )

		guideTipBeforeMove = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 4, 1 )

		# Select the root again and move it just enough for its surface dependency to switch faces
		pm.select( editGuidesShape + ".f[4]" );
		pm.move( 0.00643916, 0, 0, relative = True )
		pm.mel.OxEditGuides( cc = True )

		#guideTipAfterMove = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 4, 1 )

		## Since the second move was minor the tip should almost not change
		#epsilon = 0.05
		#self.assertGreaterEqual( epsilon, dt.Vector( guideTipBeforeMove ).distanceTo( guideTipAfterMove ) )

	def test_MovePlantedGuideWithHairFromGuidesOnTop( self ):
		# Generate 1 guide with face-center distribution and add HFG on top
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 2, guideCount = 1 )
		pm.mel.OxAddStrandOperator( editGuidesShape, TestUtilities.HairFromGuidesNodeName )

		# Plant a new guide
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0 0;" )

		# Select and move the first root
		pm.select( editGuidesShape + ".f[0]" )
		pm.move( 0, 0, 0.2, relative = True )
		pm.mel.OxEditGuides( cc = True )

		root1PositionBeforeDeselecting = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 )

		# Select the second root, the position of first root should stay intact
		pm.select( editGuidesShape + ".f[1]" )
		root1PositionAfterDeselecting = pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 0 )

		TestUtilities.CheckPointsNearEqual( self, root1PositionBeforeDeselecting, root1PositionAfterDeselecting )

	def test_DeleteOperatorAfterPlantingSecondGuide( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 2, guideCount = 1 )

		# Plant a new guide
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0 0;" )
		# Should not cause assert failure
		pm.mel.OxDeleteStrandOperator( editGuidesShape )

	# Tests that we can plant strands without issues if input strands are empty
	def test_PlantFirstStrand( self ):
		# Make sure that input guides are empty
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( planeSegmentCount = 2, rootGenerationMethod = 2, guideCount = 0 )

		# Plant two new guides
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0 0;" )
		vertices0 = list( TestUtilities.GetVerticesInObjectCoordinatesByStrandId( editGuidesShape ).items() )[0][1]
		TestUtilities.CheckPointsNotNearEqual( self, vertices0[0], vertices0[1] )
		TestUtilities.CheckPointsNotNearEqual( self, vertices0[0], vertices0[-1] )
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 1 1;" )

		self.assertEqual( 2, pm.mel.OxGetStrandCount( editGuidesShape ) )

	def SetupPerformance( self ):
		plane = pm.polyPlane( w = 20, h = 20, sx = 5, sy = 5 )
		meshShape = plane[0]
		transformNode = meshShape.firstParent2()
		guidesFromMesh = pm.createNode( TestUtilities.GuidesFromMeshNodeName )
		guidesFromMesh.length.set( 5 )
		guidesFromMesh.lengthRandomness.set( 0 )
		guidesFromMesh.distribution.set( 2 )
		meshShape.outMesh >> guidesFromMesh.inputMesh

		guidesShape = pm.createNode( TestUtilities.HairShapeName, parent = transformNode )
		guidesFromMesh.outputGuides >> guidesShape.inputHair
		# Note: `guidesShape.distributionMesh` isn't connected now, it will be connected in `OxAddStrandOperator`

		self.editGuidesShape = pm.PyNode( pm.mel.OxAddStrandOperator( guidesShape, TestUtilities.EditGuidesShapeName ) )
		hairFromGuides = pm.PyNode( pm.mel.OxAddStrandOperator( guidesShape, TestUtilities.HairFromGuidesNodeName ) )
		self.hairShape = pm.ls( type = TestUtilities.HairShapeName )[0]

		hairFromGuides.distribution.set( 2 )
		hairFromGuides.renderCount.set( 20000 )
		hairFromGuides.viewportCountFraction.set( 1 )

		pm.select( self.editGuidesShape )

	def CommandPerformance( self ):
		pm.mel.OxEditGuides( '-eg 37 24 10 0 0 0 0 0.458088 -0.241235 0 0.491121 -0.242218 0 0.491121 -0.242218 0 0.491121 -0.242218 0 0.491121 -0.242218 0 0.491121 -0.242218 0 0.491121 -0.242218 0 0.491121 -0.242218 0 0.491121 -0.242217 26 10 0 0 0 0 0.512256 -0.340531 0 0.587483 -0.345647 0 0.587483 -0.345647 0 0.587483 -0.345647 0 0.587483 -0.345647 0 0.587483 -0.345647 0 0.587483 -0.345647 0 0.587483 -0.345647 0 0.587483 -0.345646 27 10 0 0 0 0 0.48033 -0.276404 0 0.512667 -0.277346 0 0.512667 -0.277346 0 0.512667 -0.277346 0 0.512667 -0.277346 0 0.512667 -0.277346 0 0.512667 -0.277346 0 0.512667 -0.277346 0 0.512667 -0.277345 30 10 0 0 0 -0.31693 0.31693 -0.227298 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 -0.339596 0.339596 -0.228224 31 10 0 0 0 -0.148668 0.148668 -0.0413204 -0.148668 0.148668 -0.0413204 -0.148668 0.148668 -0.0413204 -0.148668 0.148668 -0.0413203 -0.148668 0.148668 -0.0413203 -0.148668 0.148668 -0.0413203 -0.148668 0.148668 -0.0413203 -0.148668 0.148668 -0.0413203 -0.148668 0.148668 -0.0413198 32 10 0 0 0 -0.361539 0.361537 -0.338248 -0.409027 0.409026 -0.342322 -0.409027 0.409026 -0.342322 -0.409027 0.409026 -0.342322 -0.409027 0.409026 -0.342322 -0.409027 0.409026 -0.342322 -0.409027 0.409026 -0.342322 -0.409027 0.409026 -0.342321 -0.409027 0.409026 -0.342321 33 10 0 0 0 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 -0.252172 0.252172 -0.129574 34 10 0 0 0 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289924 -0.125247 0.125247 -0.0289927 -0.125247 0.125247 -0.0289927 84 10 0 0 0 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 0 0.367389 -0.138822 87 10 0 0 0 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 0 0.340065 -0.11624 88 10 0 0 0 0 0.102031 -0.00944978 0 0.102031 -0.00944972 0 0.102031 -0.00944972 0 0.102031 -0.00944972 0 0.102031 -0.00944972 0 0.102031 -0.00944972 0 0.102031 -0.00944972 0 0.102031 -0.00944996 0 0.102031 -0.00944996 89 10 0 0 0 0 0.362138 -0.13425 0 0.362138 -0.13425 0 0.362138 -0.13425 0 0.362138 -0.13425 0 0.362138 -0.13425 0 0.362138 -0.13425 0 0.362138 -0.13425 0 0.362138 -0.134251 0 0.362138 -0.134251 90 10 0 0 0 -0.347948 0.347948 -0.297666 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 -0.371522 0.371522 -0.298667 91 10 0 0 0 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664907 -0.186356 0.186357 -0.0664902 93 10 0 0 0 -0.323095 0.323095 -0.239548 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 -0.347466 0.347466 -0.240618 94 10 0 0 0 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236098 -0.321394 0.321394 -0.236097 -0.321394 0.321394 -0.236097 95 10 0 0 0 -0.351219 0.351219 -0.306693 -0.36939 0.36939 -0.307287 -0.36939 0.36939 -0.307287 -0.36939 0.36939 -0.307287 -0.36939 0.36939 -0.307287 -0.36939 0.36939 -0.307287 -0.36939 0.36939 -0.307287 -0.36939 0.36939 -0.307288 -0.36939 0.36939 -0.307286 145 10 0 0 0 0 0.211048 -0.0416481 0 0.211048 -0.0416481 0 0.211048 -0.0416481 0 0.211048 -0.0416481 0 0.211048 -0.0416481 0 0.211048 -0.0416481 0 0.211048 -0.0416481 0 0.211048 -0.0416479 0 0.211048 -0.0416479 148 10 0 0 0 0 0.273069 -0.0717424 0 0.273069 -0.0717423 0 0.273069 -0.0717423 0 0.273069 -0.0717423 0 0.273069 -0.0717423 0 0.273069 -0.0717423 0 0.273069 -0.0717423 0 0.273069 -0.0717421 0 0.273069 -0.0717421 149 10 0 0 0 0 0.0328063 -0.000969529 0 0.0328063 -0.000969529 0 0.0328063 -0.000969529 0 0.0328063 -0.00096941 0 0.0328063 -0.00096941 0 0.0328063 -0.00096941 0 0.0328063 -0.00096941 0 0.0328063 -0.00096941 0 0.0328063 -0.000968933 154 10 0 0 0 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 -0.229962 0.229963 -0.105137 155 10 0 0 0 -0.203677 0.203677 -0.0805045 -0.203677 0.203677 -0.0805047 -0.203677 0.203677 -0.0805047 -0.203677 0.203677 -0.0805047 -0.203677 0.203677 -0.0805047 -0.203677 0.203677 -0.0805047 -0.203677 0.203677 -0.0805047 -0.203677 0.203677 -0.0805044 -0.203677 0.203677 -0.0805044 204 10 0 0 0 0 0.113478 -0.0117131 0 0.113478 -0.011713 0 0.113478 -0.011713 0 0.113478 -0.011713 0 0.113478 -0.011713 0 0.113478 -0.011713 0 0.113478 -0.011713 0 0.113478 -0.011713 0 0.113478 -0.011713 205 10 0 0 0 0 0.51285 -0.34195 0 0.594772 -0.348024 0 0.594772 -0.348024 0 0.594772 -0.348024 0 0.594772 -0.348023 0 0.594772 -0.348023 0 0.594772 -0.348023 0 0.594772 -0.348023 0 0.594772 -0.348022 207 10 0 0 0 0 0.448714 -0.227993 0 0.481031 -0.228934 0 0.481031 -0.228934 0 0.481031 -0.228934 0 0.481031 -0.228934 0 0.481031 -0.228934 0 0.481031 -0.228934 0 0.481031 -0.228933 0 0.481031 -0.228933 210 10 0 0 0 -0.278497 0.278497 -0.163738 -0.282748 0.282748 -0.16377 -0.282748 0.282748 -0.16377 -0.282748 0.282748 -0.16377 -0.282748 0.282748 -0.16377 -0.282748 0.282748 -0.16377 -0.282748 0.282748 -0.16377 -0.282748 0.282748 -0.163771 -0.282748 0.282748 -0.16377 212 10 0 0 0 -0.348153 0.348152 -0.298218 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298913 -0.36779 0.367789 -0.298912 214 10 0 0 0 -0.184042 0.184043 -0.0647413 -0.184042 0.184043 -0.0647414 -0.184042 0.184043 -0.0647414 -0.184042 0.184043 -0.0647414 -0.184042 0.184043 -0.0647414 -0.184042 0.184043 -0.0647414 -0.184042 0.184043 -0.0647414 -0.184042 0.184043 -0.0647411 -0.184042 0.184043 -0.0647411 215 10 0 0 0 -0.360512 0.360512 -0.334864 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 -0.417524 0.417524 -0.340746 264 10 0 0 0 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 0 0.212375 -0.0421953 265 10 0 0 0 0 0.504245 -0.322363 0 0.569521 -0.326211 0 0.569521 -0.326211 0 0.569521 -0.326211 0 0.569521 -0.326211 0 0.569521 -0.326211 0 0.569521 -0.326211 0 0.569521 -0.326211 0 0.569521 -0.326211 267 10 0 0 0 0 0.422515 -0.194831 0 0.44058 -0.195124 0 0.44058 -0.195124 0 0.44058 -0.195124 0 0.44058 -0.195124 0 0.44058 -0.195124 0 0.44058 -0.195124 0 0.44058 -0.195125 0 0.44058 -0.195124 271 10 0 0 0 -0.27229 0.27229 -0.155108 -0.274499 0.274498 -0.155117 -0.274499 0.274498 -0.155117 -0.274499 0.274498 -0.155117 -0.274499 0.274498 -0.155117 -0.274499 0.274498 -0.155117 -0.274499 0.274498 -0.155117 -0.274499 0.274498 -0.155116 -0.274499 0.274498 -0.155116 272 10 0 0 0 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161903 -0.0941467 0.0941467 -0.0161905 -0.0941467 0.0941467 -0.0161905 273 10 0 0 0 -0.34393 0.34393 -0.287102 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 -0.369381 0.369381 -0.288269 274 10 0 0 0 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557866 -0.171571 0.171571 -0.0557861 275 10 0 0 0 -0.331283 0.331283 -0.256983 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665 -0.350741 0.350741 -0.257665' )
		self.hairShape.boundingBox()
		self.editGuidesShape.boundingBox()
		pm.mel.OxEditGuides( '-eg 300 24 10 0 0 0 -0.0190189 0.46382 -0.250352 -0.0997837 0.576628 -0.267953 -0.244636 0.721481 -0.3071 -0.449955 0.926801 -0.38902 -0.707574 1.184421 -0.525165 -1.01123 1.488077 -0.72826 -1.347571 1.824419 -0.996778 -1.703406 2.180253 -1.316949 -2.069438 2.546287 -1.670809 26 10 0 0 0 -0.0211599 0.515001 -0.348263 -0.0989602 0.665124 -0.374619 -0.244514 0.810678 -0.41416 -0.450002 1.016166 -0.496227 -0.706995 1.273159 -0.631602 -1.009316 1.57548 -0.832412 -1.343971 1.910135 -1.097013 -1.698049 2.264215 -1.411949 -2.062426 2.628592 -1.759895 27 10 0 0 0 -0.0194588 0.484873 -0.285073 -0.100518 0.597301 -0.30264 -0.245604 0.742386 -0.341918 -0.451049 0.947833 -0.423949 -0.708628 1.205413 -0.560044 -1.012089 1.508872 -0.762802 -1.348148 1.844933 -1.030659 -1.703674 2.200459 -1.349901 -2.069393 2.566178 -1.70262 30 10 0 0 0 -0.336895 0.306832 -0.237756 -0.478135 0.328773 -0.256459 -0.688159 0.328773 -0.297687 -0.981373 0.328773 -0.381367 -1.344449 0.328774 -0.516425 -1.768502 0.328773 -0.713065 -2.239212 0.328774 -0.973536 -2.739584 0.328774 -1.287703 -3.256364 0.328774 -1.639346 31 10 0 0 0 -0.186599 0.145331 -0.0528617 -0.315453 0.145332 -0.0680108 -0.531251 0.145332 -0.111635 -0.827961 0.145332 -0.197505 -1.19105 0.145332 -0.332574 -1.611925 0.145332 -0.525492 -2.074453 0.145333 -0.773299 -2.565425 0.145333 -1.068877 -3.072415 0.145333 -1.397269 32 10 0 0 0 -0.380366 0.347057 -0.346951 -0.542864 0.392635 -0.373207 -0.75462 0.392636 -0.415146 -1.047978 0.392636 -0.498915 -1.408997 0.392636 -0.632205 -1.832435 0.392636 -0.828119 -2.302023 0.392637 -1.086809 -2.800588 0.392637 -1.397262 -3.315125 0.392637 -1.743311 33 10 0 0 0 -0.278743 0.244807 -0.142017 -0.40621 0.244807 -0.156838 -0.621233 0.244807 -0.200136 -0.919043 0.244807 -0.286702 -1.287426 0.244807 -0.426401 -1.717754 0.244807 -0.630589 -2.192886 0.244807 -0.898234 -2.695597 0.244808 -1.217308 -3.213201 0.244808 -1.571053 34 10 0 0 0 -0.161539 0.123001 -0.0384309 -0.282056 0.123002 -0.0516605 -0.48974 0.123002 -0.0919394 -0.783156 0.123002 -0.175745 -1.150028 0.123002 -0.31411 -1.581041 0.123002 -0.519138 -2.057348 0.123003 -0.788733 -2.560598 0.123003 -1.108955 -3.079292 0.123003 -1.465518 84 10 0 0 0 -0.0238719 0.380293 -0.151266 -0.114186 0.470606 -0.166147 -0.266412 0.622832 -0.209553 -0.4777 0.834121 -0.296753 -0.738823 1.095244 -0.437254 -1.043578 1.4 -0.642251 -1.379808 1.736231 -0.910511 -1.735371 2.091794 -1.229863 -2.101343 2.457764 -1.583494 87 10 0 0 0 -0.0246024 0.354907 -0.128849 -0.115056 0.445362 -0.143776 -0.267482 0.597787 -0.187302 -0.479074 0.80938 -0.274778 -0.740499 1.070804 -0.415656 -1.045505 1.375811 -0.621092 -1.381926 1.712233 -0.889794 -1.737617 2.067924 -1.209536 -2.103676 2.433983 -1.563485 88 10 0 0 0 -0.0346913 0.135163 -0.0178099 -0.127553 0.228023 -0.0335546 -0.280735 0.381207 -0.0775326 -0.489599 0.59007 -0.162563 -0.744055 0.844527 -0.294861 -1.038929 1.139401 -0.483347 -1.363868 1.46434 -0.726696 -1.707983 1.808455 -1.014277 -2.062594 2.163065 -1.330788 89 10 0 0 0 -0.024579 0.375712 -0.147051 -0.116193 0.467328 -0.162369 -0.271205 0.622339 -0.20745 -0.485964 0.837099 -0.29782 -0.750179 1.101315 -0.44225 -1.057146 1.408279 -0.65113 -1.394581 1.745716 -0.922227 -1.750526 2.101661 -1.242733 -2.116297 2.467435 -1.595647 90 10 0 0 0 -0.369848 0.333048 -0.308709 -0.523584 0.355723 -0.330886 -0.749306 0.355723 -0.378808 -1.05778 0.355723 -0.472319 -1.432777 0.355723 -0.617973 -1.864552 0.355723 -0.823939 -2.336679 0.355723 -1.086685 -2.833431 0.355723 -1.393484 -3.343125 0.355723 -1.728012 91 10 0 0 0 -0.224401 0.180963 -0.0806463 -0.364627 0.180963 -0.098634 -0.59896 0.180963 -0.150474 -0.91845 0.180963 -0.251531 -1.304702 0.180962 -0.407773 -1.745941 0.180962 -0.625763 -2.224971 0.180962 -0.899942 -2.726478 0.180962 -1.216474 -3.239505 0.180962 -1.558849 93 10 0 0 0 -0.346921 0.310212 -0.252148 -0.502842 0.333627 -0.274991 -0.732203 0.333627 -0.324546 -1.046207 0.333627 -0.421797 -1.427748 0.333627 -0.573534 -1.865868 0.333627 -0.787487 -2.34331 0.333627 -1.058982 -2.844275 0.333627 -1.374378 -3.357415 0.333627 -1.717028 94 10 0 0 0 -0.345673 0.30846 -0.248954 -0.483128 0.30846 -0.266227 -0.71035 0.30846 -0.314819 -1.0182 0.308461 -0.407913 -1.390016 0.30846 -0.550679 -1.816651 0.308461 -0.750391 -2.282719 0.308461 -1.003584 -2.77316 0.30846 -1.298161 -3.276377 0.308461 -1.618314 95 10 0 0 0 -0.373958 0.335227 -0.318025 -0.526958 0.352705 -0.339794 -0.756309 0.352705 -0.389345 -1.066659 0.352705 -0.484114 -1.440724 0.352705 -0.628917 -1.868979 0.352705 -0.830581 -2.335986 0.352705 -1.085226 -2.82683 0.352705 -1.380562 -3.330073 0.352705 -1.700767 145 10 0 0 0 -0.0345774 0.239875 -0.0556483 -0.137704 0.342999 -0.0751328 -0.3087 0.513995 -0.130526 -0.537174 0.742472 -0.234152 -0.808841 1.014136 -0.388412 -1.11673 1.322025 -0.598931 -1.450086 1.655379 -0.860561 -1.798773 2.004066 -1.160231 -2.155114 2.360408 -1.481946 148 10 0 0 0 -0.0310287 0.295852 -0.0863523 -0.129368 0.394192 -0.104041 -0.290353 0.555177 -0.152832 -0.506663 0.771488 -0.244642 -0.76648 1.031306 -0.383505 -1.063902 1.328727 -0.576124 -1.38873 1.653553 -0.819244 -1.730673 1.995498 -1.101321 -2.081503 2.346326 -1.406919 149 10 0 0 0 -0.0404692 0.0729604 -0.00630093 -0.139145 0.171635 -0.0241125 -0.29787 0.330361 -0.071481 -0.508659 0.541151 -0.158232 -0.760245 0.792738 -0.287112 -1.047823 1.080315 -0.464197 -1.362193 1.394685 -0.686608 -1.693462 1.725954 -0.943562 -2.033177 2.065669 -1.220137 154 10 0 0 0 -0.265588 0.221884 -0.12096 -0.403988 0.221884 -0.138476 -0.629647 0.221884 -0.18637 -0.932055 0.221884 -0.275887 -1.294554 0.221884 -0.410448 -1.709073 0.221884 -0.596118 -2.161313 0.221884 -0.828995 -2.636685 0.221884 -1.097038 -3.123321 0.221883 -1.384584 155 10 0 0 0 -0.24198 0.197062 -0.0959315 -0.379597 0.197062 -0.113245 -0.603936 0.197062 -0.160556 -0.904693 0.197062 -0.249005 -1.265418 0.197062 -0.382046 -1.678207 0.197062 -0.565786 -2.128845 0.197062 -0.796429 -2.602728 0.197061 -1.062024 -3.087902 0.197061 -1.346937 204 10 0 0 0 -0.0351458 0.146714 -0.0208771 -0.125166 0.236734 -0.03566 -0.269847 0.381415 -0.0747106 -0.462708 0.574276 -0.146271 -0.694103 0.805668 -0.252877 -0.960651 1.07222 -0.400335 -1.254173 1.365744 -0.586661 -1.564859 1.676426 -0.802224 -1.883516 1.995083 -1.032883 205 10 0 0 0 -0.0285769 0.516297 -0.352418 -0.1122 0.678165 -0.383143 -0.257347 0.823314 -0.422456 -0.452827 1.018793 -0.496122 -0.689004 1.254971 -0.607738 -0.962046 1.528015 -0.763873 -1.26313 1.829097 -0.962581 -1.582484 2.148449 -1.19461 -1.910954 2.47692 -1.445442 207 10 0 0 0 -0.0285225 0.45783 -0.242157 -0.119439 0.579881 -0.26341 -0.271526 0.731968 -0.306734 -0.47604 0.936481 -0.387958 -0.722365 1.182805 -0.510742 -1.005738 1.466178 -0.681535 -1.316593 1.777033 -0.897413 -1.644578 2.105018 -1.147201 -1.980886 2.441324 -1.415635 210 10 0 0 0 -0.307209 0.268171 -0.178265 -0.431311 0.272316 -0.192319 -0.627495 0.272316 -0.228112 -0.892334 0.272316 -0.2953 -1.213459 0.272315 -0.397512 -1.586559 0.272315 -0.541438 -1.999941 0.272315 -0.725836 -2.439356 0.272314 -0.941457 -2.891594 0.272314 -1.174331 212 10 0 0 0 -0.371069 0.332493 -0.309793 -0.508566 0.351531 -0.327414 -0.705284 0.351531 -0.363408 -0.970682 0.351532 -0.4309 -1.292358 0.351531 -0.533502 -1.666006 0.351531 -0.677927 -2.079964 0.35153 -0.862967 -2.52 0.35153 -1.079392 -2.972933 0.351529 -1.313241 214 10 0 0 0 -0.221752 0.178844 -0.0786042 -0.345478 0.178844 -0.092557 -0.547098 0.178844 -0.130434 -0.819103 0.178844 -0.201577 -1.148447 0.178844 -0.309723 -1.530165 0.178844 -0.461628 -1.952015 0.178844 -0.65568 -2.399578 0.178843 -0.882103 -2.859831 0.178843 -1.126517 215 10 0 0 0 -0.382498 0.343637 -0.345195 -0.548506 0.398332 -0.373406 -0.740423 0.398332 -0.407608 -0.99967 0.398332 -0.471805 -1.3144 0.398331 -0.569555 -1.68079 0.398332 -0.707497 -2.087595 0.398331 -0.884699 -2.520725 0.398331 -1.092345 -2.966856 0.39833 -1.316826 264 10 0 0 0 -0.0286789 0.236339 -0.0535972 -0.106415 0.314076 -0.0645823 -0.232136 0.439799 -0.093801 -0.401438 0.609097 -0.148041 -0.606905 0.814567 -0.230093 -0.846928 1.054589 -0.345851 -1.11479 1.322449 -0.495029 -1.401163 1.608821 -0.670293 -1.696499 1.904155 -0.859519 265 10 0 0 0 -0.0262928 0.508212 -0.332688 -0.10158 0.64653 -0.355475 -0.229113 0.774065 -0.385568 -0.401234 0.946186 -0.441732 -0.61039 1.155342 -0.527022 -0.854734 1.399688 -0.647569 -1.127277 1.672228 -0.803018 -1.418543 1.963492 -0.985785 -1.718809 2.263761 -1.183121 267 10 0 0 0 -0.0269241 0.433161 -0.20873 -0.107748 0.531579 -0.223523 -0.240709 0.66454 -0.256313 -0.42016 0.843993 -0.317666 -0.637962 1.061792 -0.410871 -0.891674 1.315506 -0.542278 -1.173689 1.597521 -0.711082 -1.474234 1.898064 -0.908883 -1.783848 2.207678 -1.122501 271 10 0 0 0 -0.300417 0.262772 -0.169106 -0.411758 0.264935 -0.180382 -0.588709 0.264935 -0.209316 -0.826834 0.264935 -0.262936 -1.115311 0.264935 -0.343705 -1.451349 0.264935 -0.456858 -1.825064 0.264935 -0.601344 -2.222935 0.264935 -0.76916 -2.632399 0.264935 -0.94924 272 10 0 0 0 -0.13312 0.0927477 -0.0242188 -0.235474 0.0927467 -0.0337288 -0.400759 0.0927467 -0.0588853 -0.623478 0.0927477 -0.105482 -0.894381 0.0927467 -0.176008 -1.212244 0.0927467 -0.275926 -1.568436 0.0927477 -0.405138 -1.94981 0.0927467 -0.556719 -2.342014 0.0927458 -0.718805 273 10 0 0 0 -0.366217 0.32932 -0.298504 -0.496355 0.354089 -0.31453 -0.670551 0.354089 -0.342547 -0.905029 0.354089 -0.394454 -1.189241 0.354089 -0.472656 -1.52063 0.354088 -0.582316 -1.889563 0.354088 -0.722504 -2.28264 0.354087 -0.885462 -2.686779 0.354088 -1.059818 274 10 0 0 0 -0.207573 0.167325 -0.0681565 -0.315968 0.167324 -0.0788333 -0.490772 0.167325 -0.107051 -0.725248 0.167325 -0.158957 -1.008518 0.167324 -0.236599 -1.337828 0.167324 -0.344721 -1.703529 0.167324 -0.482061 -2.094761 0.167323 -0.643182 -2.498211 0.167324 -0.816806 275 10 0 0 0 -0.354822 0.317549 -0.269361 -0.483006 0.336494 -0.284683 -0.663246 0.336493 -0.314734 -0.906292 0.336493 -0.370719 -1.201384 0.336493 -0.45557 -1.545879 0.336493 -0.575277 -1.929764 0.336493 -0.729241 -2.339251 0.336492 -0.909348 -2.76038 0.336493 -1.102561 0 10 0 0 0 -0.0325007 0.032501 -0.00190401 -0.111674 0.111674 -0.0133045 -0.242878 0.242878 -0.0452065 -0.422128 0.422129 -0.106414 -0.641319 0.641321 -0.200935 -0.897733 0.897735 -0.335604 -1.183676 1.183679 -0.510221 -1.489487 1.489491 -0.717063 -1.805936 1.80594 -0.943435 1 10 0 0 0 -0.035512 0.035512 -0.00227427 -0.118452 0.118452 -0.0147977 -0.251276 0.251278 -0.0475173 -0.428449 0.42845 -0.107228 -0.641789 0.641792 -0.196294 -0.889256 0.889258 -0.320383 -1.163992 1.163995 -0.478848 -1.45681 1.456814 -0.664057 -1.758381 1.758385 -0.86359 2 10 0 0 0 -0.0317824 0.0317826 -0.00182104 -0.108945 0.108946 -0.0126443 -0.236557 0.236558 -0.0427738 -0.410835 0.410836 -0.100438 -0.624116 0.624118 -0.189448 -0.874104 0.874107 -0.316455 -1.153521 1.153525 -0.481507 -1.452903 1.452907 -0.677363 -1.762958 1.762963 -0.891785 3 10 0 0 0 -0.0329092 0.0329101 -0.00195289 -0.110654 0.110654 -0.0129409 -0.236271 0.236271 -0.04211 -0.405007 0.405008 -0.0959702 -0.609337 0.609339 -0.177037 -0.84761 0.847612 -0.290897 -1.113418 1.113421 -0.437389 -1.397844 1.397848 -0.60974 -1.691577 1.691582 -0.796402 4 10 0 0 0 -0.0320398 0.03204 -0.00185108 -0.108401 0.108401 -0.0124481 -0.232697 0.232697 -0.0409902 -0.400563 0.400564 -0.0942674 -0.604607 0.604608 -0.175086 -0.843142 0.843144 -0.289231 -1.109694 1.109696 -0.436692 -1.395291 1.395294 -0.610789 -1.690622 1.690627 -0.800012 5 10 0 0 0 -0.036064 0.0360637 -0.00234604 -0.121395 0.121395 -0.0156109 -0.258906 0.258906 -0.0507596 -0.442903 0.442904 -0.115467 -0.664765 0.664767 -0.212551 -0.92194 0.921942 -0.348149 -1.20705 1.207053 -0.521518 -1.510701 1.510704 -0.724605 -1.823758 1.823762 -0.944554 6 10 0 0 0 -0.0529763 2.38419e-007 -0.00253177 -0.182465 2.38419e-007 -0.017833 -0.393365 4.76837e-007 -0.0594203 -0.675337 8.9407e-007 -0.136297 -1.014435 1.43051e-006 -0.251791 -1.40526 1.78814e-006 -0.412508 -1.835648 2.32458e-006 -0.616767 -2.291644 2.68221e-006 -0.854975 -2.760827 3.21865e-006 -1.113026 7 10 0 0 0 -0.0537797 7.15256e-007 -0.00260878 -0.182106 5.96046e-007 -0.017633 -0.389252 4.76837e-007 -0.0576961 -0.666007 1.07288e-006 -0.131537 -0.998723 1.54972e-006 -0.242186 -1.38247 1.90735e-006 -0.39602 -1.805537 2.38419e-006 -0.591497 -2.254116 2.74181e-006 -0.819305 -2.71559 3.09944e-006 -1.065533 8 10 0 0 0 -0.0533478 0 -0.00256729 -0.179746 2.38419e-007 -0.0171375 -0.383577 2.38419e-007 -0.055881 -0.655977 4.76837e-007 -0.127246 -0.983633 9.53674e-007 -0.234155 -1.361948 1.19209e-006 -0.38287 -1.779505 1.66893e-006 -0.571971 -2.22256 2.14577e-006 -0.792348 -2.678334 2.38419e-006 -1.030238 9 10 0 0 0 -0.0531223 4.76837e-007 -0.0025456 -0.179004 7.15256e-007 -0.0169947 -0.38238 4.76837e-007 -0.0555587 -0.654645 7.15256e-007 -0.126848 -0.982509 1.19209e-006 -0.23391 -1.361279 1.66893e-006 -0.383048 -1.779377 2.14577e-006 -0.572767 -2.222919 2.14577e-006 -0.793789 -2.679048 2.86102e-006 -1.032188 10 10 0 0 0 -0.0518402 2.38419e-007 -0.00242329 -0.178875 3.57628e-007 -0.0171425 -0.388876 7.7486e-007 -0.058362 -0.671376 1.07288e-006 -0.13555 -1.011805 1.54972e-006 -0.252073 -1.40449 1.96695e-006 -0.414638 -1.837051 2.40281e-006 -0.621579 -2.295437 2.83122e-006 -0.863249 -2.767262 3.21865e-006 -1.125506 11 10 0 0 0 -0.0541444 1.19209e-007 -0.00264502 -0.184083 -2.38419e-007 -0.0180545 -0.394922 1.19209e-007 -0.0596163 -0.677135 4.76837e-007 -0.136635 -1.016444 8.34465e-007 -0.252291 -1.407291 1.54972e-006 -0.413029 -1.837412 1.90735e-006 -0.616963 -2.292786 2.38419e-006 -0.854279 -2.760931 2.86102e-006 -1.110698 12 10 0 0 0 -0.035941 0.0359414 -0.00232983 -0.125538 0.125539 -0.0169728 -0.27727 0.277271 -0.0600863 -0.487942 0.487943 -0.146732 -0.746582 0.746585 -0.284135 -1.043311 1.043314 -0.475622 -1.367722 1.367726 -0.717878 -1.709384 1.709388 -0.99925 -2.060045 2.06005 -1.304378 13 10 0 0 0 -0.0371634 0.0371642 -0.00249171 -0.129798 0.129799 -0.0181584 -0.286377 0.286378 -0.0641967 -0.502877 0.502879 -0.156182 -0.768017 0.768019 -0.301806 -1.073992 1.073995 -0.508936 -1.408091 1.408095 -0.772264 -1.758771 1.758774 -1.077441 -2.117801 2.117805 -1.407529 14 10 0 0 0 -0.0371543 0.0371552 -0.00249028 -0.129816 0.129817 -0.018167 -0.286522 0.286523 -0.0642833 -0.503311 0.503313 -0.156538 -0.768967 0.76897 -0.302832 -1.075745 1.075747 -0.511377 -1.41085 1.410853 -0.777017 -1.76259 1.762594 -1.085188 -2.122712 2.122716 -1.418782 15 10 0 0 0 -0.0363587 0.0363586 -0.00238466 -0.126579 0.126579 -0.0172348 -0.278664 0.278664 -0.0605576 -0.488898 0.488899 -0.14681 -0.743571 0.743573 -0.279371 -1.035989 1.035991 -0.463948 -1.356218 1.356221 -0.697715 -1.693957 1.69396 -0.969532 -2.040828 2.040832 -1.264315 16 10 0 0 0 -0.0367547 0.0367556 -0.00243711 -0.128702 0.128703 -0.0178692 -0.284713 0.284715 -0.0635586 -0.501199 0.501202 -0.155532 -0.767103 0.767105 -0.302146 -1.074528 1.074531 -0.511838 -1.410393 1.410396 -0.779242 -1.762863 1.762867 -1.089502 -2.12367 2.123675 -1.425333 17 10 0 0 0 -0.038667 0.0386679 -0.00269842 -0.131884 0.131885 -0.0185657 -0.284585 0.284586 -0.0622556 -0.490709 0.49071 -0.144876 -0.739666 0.739668 -0.270684 -1.026443 1.026445 -0.446557 -1.341204 1.341206 -0.669707 -1.67347 1.673473 -0.928884 -2.014522 2.014525 -1.208742 18 10 0 0 0 -0.0544171 -5.96046e-008 -0.00267148 -0.185583 1.19209e-007 -0.0183773 -0.400615 2.98023e-007 -0.0616801 -0.691276 4.76837e-007 -0.143782 -1.042849 8.34465e-007 -0.269176 -1.448306 1.19209e-006 -0.444934 -1.893632 1.49012e-006 -0.668336 -2.363891 1.90735e-006 -0.928089 -2.846728 2.26498e-006 -1.208854 19 10 0 0 0 -0.0547614 2.38419e-007 -0.00270534 -0.187354 1.78814e-007 -0.0187597 -0.405415 3.57628e-007 -0.063344 -0.700599 7.45058e-007 -0.148253 -1.057652 1.10269e-006 -0.278186 -1.468889 1.37091e-006 -0.460209 -1.919777 1.66893e-006 -0.6912 -2.395238 2.05636e-006 -0.95939 -2.88307 2.44379e-006 -1.249123 20 10 0 0 0 -0.0546756 3.57628e-007 -0.00269675 -0.187139 4.17233e-007 -0.0187199 -0.405166 6.55651e-007 -0.0632895 -0.700565 8.04663e-007 -0.148333 -1.058145 1.2517e-006 -0.278708 -1.470163 1.59442e-006 -0.461593 -1.921959 1.89245e-006 -0.69385 -2.398388 2.29478e-006 -0.963649 -2.887307 2.563e-006 -1.255385 21 10 0 0 0 -0.0526218 2.38419e-007 -0.00249767 -0.183809 2.38419e-007 -0.018209 -0.405579 7.15256e-007 -0.0643916 -0.71228 9.53674e-007 -0.156723 -1.088062 1.66893e-006 -0.303097 -1.522046 1.66893e-006 -0.511809 -1.996171 2.14577e-006 -0.777801 -2.493873 2.6226e-006 -1.086507 -3.003508 3.09944e-006 -1.420896 22 10 0 0 0 -0.0538778 0 -0.00261903 -0.185209 5.96046e-008 -0.0183651 -0.402799 3.57628e-007 -0.0627491 -0.69943 5.96046e-007 -0.148569 -1.060293 9.53674e-007 -0.281726 -1.477357 1.2517e-006 -0.470266 -1.935271 1.54972e-006 -0.711248 -2.418365 1.78814e-006 -0.992464 -2.914353 2.20537e-006 -1.297745 23 10 0 0 0 -0.0543566 2.38419e-007 -0.00266528 -0.186508 2.38419e-007 -0.0186117 -0.404801 5.96046e-007 -0.0632955 -0.701485 8.34465e-007 -0.149148 -1.061445 1.13249e-006 -0.281536 -1.476758 1.49012e-006 -0.468097 -1.932416 1.78814e-006 -0.70582 -2.413019 2.08616e-006 -0.982693 -2.906352 2.35438e-006 -1.282778 25 10 0 0 0 -0.0376451 0.0376458 -0.00255704 -0.130387 0.130387 -0.0182602 -0.285511 0.285512 -0.06341 -0.498563 0.498564 -0.15221 -0.758959 0.758961 -0.291798 -1.060307 1.060309 -0.490954 -1.390869 1.390872 -0.746344 -1.739131 1.739134 -1.044861 -2.096506 2.09651 -1.36975 28 10 0 0 0 -0.0364705 0.0364718 -0.00239992 -0.12772 0.127721 -0.0175951 -0.282623 0.282624 -0.0626101 -0.497829 0.497831 -0.153392 -0.762691 0.762693 -0.298657 -1.069971 1.069973 -0.50809 -1.406898 1.406901 -0.777987 -1.761413 1.761416 -1.09421 -2.124982 2.124987 -1.439349 29 10 0 0 0 -0.0373544 0.037354 -0.00251675 -0.130078 0.130077 -0.0182142 -0.286224 0.286225 -0.0639877 -0.501638 0.501639 -0.154962 -0.765398 0.765399 -0.298809 -1.070528 1.07053 -0.504462 -1.404786 1.404788 -0.768152 -1.75648 1.756483 -1.076192 -2.117145 2.117148 -1.411559 35 10 0 0 0 -0.0501032 1.90735e-006 -0.00226331 -0.175747 9.53674e-007 -0.0166581 -0.389272 1.90735e-006 -0.0593299 -0.686124 1.90735e-006 -0.145288 -1.053251 1.90735e-006 -0.28388 -1.484071 2.38419e-006 -0.48867 -1.960981 1.90735e-006 -0.759274 -2.466162 2.38419e-006 -1.083671 -2.986606 2.86102e-006 -1.444855 36 10 0 0 0 -0.0312948 0.0312958 -0.00176644 -0.111725 0.111727 -0.0135353 -0.252246 0.252247 -0.0502942 -0.453003 0.453003 -0.128319 -0.706159 0.706161 -0.259063 -1.007011 1.007013 -0.457381 -1.344169 1.344172 -0.72782 -1.703537 1.703539 -1.058986 -2.074829 2.074831 -1.433087 37 10 0 0 0 -0.031276 0.0312777 -0.00176311 -0.111711 0.111712 -0.0135329 -0.25229 0.252291 -0.0503228 -0.453119 0.45312 -0.12841 -0.70629 0.70629 -0.259169 -1.008009 1.008011 -0.458954 -1.345861 1.345864 -0.731046 -1.70572 1.705722 -1.063785 -2.077337 2.077338 -1.439224 38 10 0 0 0 -0.0302763 0.0302753 -0.00165296 -0.108314 0.108315 -0.0127246 -0.245146 0.245146 -0.0475149 -0.442251 0.442251 -0.122508 -0.693305 0.693306 -0.250764 -0.992888 0.992888 -0.446955 -1.327439 1.32744 -0.711319 -1.684002 1.684006 -1.033711 -2.053566 2.05357 -1.400883 39 10 0 0 0 -0.0310211 0.0310202 -0.00173545 -0.110642 0.110642 -0.0132651 -0.249805 0.249805 -0.0492928 -0.449198 0.449199 -0.126177 -0.701767 0.701768 -0.256222 -1.001747 1.001748 -0.453077 -1.336495 1.336495 -0.717892 -1.694187 1.694188 -1.043771 -2.064406 2.064408 -1.413535 40 10 0 0 0 -0.032016 0.0320158 -0.00184894 -0.113635 0.113636 -0.013973 -0.255594 0.255594 -0.0515146 -0.458167 0.458168 -0.131076 -0.713756 0.713757 -0.264742 -1.016233 1.016234 -0.465816 -1.352102 1.352103 -0.733229 -1.708799 1.7088 -1.056026 -2.077811 2.077813 -1.421053 41 10 0 0 0 -0.0293112 0.0293102 -0.00154853 -0.105598 0.105597 -0.0121231 -0.240582 0.240583 -0.0459518 -0.43567 0.435671 -0.1193 -0.686558 0.68656 -0.247362 -0.988755 0.988756 -0.447956 -1.328119 1.328119 -0.72368 -1.689683 1.689684 -1.062015 -2.062866 2.062868 -1.444052 42 10 0 0 0 -0.0439138 -1.43051e-006 -0.00173903 -0.158592 -4.76837e-007 -0.0137036 -0.361848 -1.43051e-006 -0.0522206 -0.655506 -4.76837e-007 -0.136175 -1.028438 -1.43051e-006 -0.27995 -1.470011 -9.53674e-007 -0.498377 -1.958879 -9.53674e-007 -0.790022 -2.474916 -1.43051e-006 -1.139791 -3.004769 -9.53674e-007 -1.528319 43 10 0 0 0 -0.0402937 -1.43051e-006 -0.00146341 -0.147161 -9.53674e-007 -0.0118396 -0.339903 -9.53674e-007 -0.0463443 -0.623322 -1.43051e-006 -0.124077 -0.988731 -1.90735e-006 -0.261163 -1.426304 -9.53674e-007 -0.474412 -1.91386 -1.90735e-006 -0.763639 -2.429995 -1.43051e-006 -1.113652 -2.960589 -1.90735e-006 -1.504552 44 10 0 0 0 -0.0450258 9.53674e-007 -0.00182843 -0.163062 9.53674e-007 -0.0145113 -0.372702 9.53674e-007 -0.0555823 -0.675231 9.53674e-007 -0.14518 -1.057664 9.53674e-007 -0.297762 -1.507143 4.76837e-007 -0.526805 -2.001173 0 -0.828243 -2.520051 0 -1.185286 -3.051346 0 -1.578458 45 10 0 0 0 -0.0378847 0 -0.00129366 -0.137683 0 -0.0103309 -0.316903 0 -0.0400326 -0.580103 1.90735e-006 -0.106335 -0.923705 9.53674e-007 -0.225338 -1.343575 1.90735e-006 -0.417091 -1.819713 9.53674e-007 -0.686406 -2.329662 1.90735e-006 -1.021522 -2.857272 1.90735e-006 -1.403096 46 10 0 0 0 -0.0404768 1.90735e-006 -0.00147653 -0.146259 9.53674e-007 -0.0116403 -0.334212 1.90735e-006 -0.0444002 -0.608661 1.90735e-006 -0.116923 -0.963948 1.90735e-006 -0.245381 -1.392583 2.86102e-006 -0.447503 -1.873946 2.86102e-006 -0.725693 -2.386466 1.90735e-006 -1.066854 -2.915126 1.90735e-006 -1.451643 47 10 0 0 0 -0.0459023 9.53674e-007 -0.00189972 -0.165724 9.53674e-007 -0.0149748 -0.377545 9.53674e-007 -0.056941 -0.68189 9.53674e-007 -0.147722 -1.065345 9.53674e-007 -0.301277 -1.51511 0 -0.530712 -2.009036 4.76837e-007 -0.831948 -2.527686 0 -1.188393 -3.058635 -4.76837e-007 -1.580439 48 10 0 0 0 -0.0214276 0.0214281 -0.000827312 -0.0794954 0.0794983 -0.00693083 -0.186878 0.186877 -0.0280881 -0.352327 0.352327 -0.0797641 -0.577808 0.577807 -0.180392 -0.862533 0.862534 -0.353187 -1.192596 1.192597 -0.607482 -1.550204 1.550206 -0.933093 -1.922107 1.922109 -1.309706 49 10 0 0 0 -0.0239735 0.0239754 -0.00103617 -0.089242 0.0892429 -0.00875711 -0.210624 0.210628 -0.035943 -0.395061 0.395065 -0.100981 -0.640128 0.640131 -0.222339 -0.940865 0.940865 -0.420463 -1.281139 1.281141 -0.69841 -1.644138 1.644138 -1.041582 -2.018578 2.018579 -1.429112 50 10 0 0 0 -0.0226302 0.0226307 -0.000921726 -0.0834455 0.0834465 -0.00761986 -0.196009 0.196007 -0.0309141 -0.368946 0.368943 -0.0876436 -0.602129 0.60213 -0.196108 -0.893189 0.893189 -0.378556 -1.227357 1.227358 -0.642037 -1.587191 1.587191 -0.974698 -1.960204 1.960203 -1.355989 51 10 0 0 0 -0.0244842 0.0244856 -0.00107932 -0.0909739 0.0909748 -0.00909495 -0.214219 0.214221 -0.0371435 -0.400786 0.400788 -0.103795 -0.647764 0.647766 -0.227325 -0.949858 0.949862 -0.427747 -1.290908 1.290911 -0.7076 -1.654292 1.654297 -1.052096 -2.028915 2.028919 -1.440447 52 10 0 0 0 -0.0205307 0.0205307 -0.000759363 -0.0775557 0.0775547 -0.0066433 -0.186245 0.186245 -0.0283301 -0.356026 0.356026 -0.082895 -0.587779 0.58778 -0.189874 -0.878908 0.878908 -0.372427 -1.213484 1.213486 -0.636851 -1.573357 1.573357 -0.969635 -1.946008 1.946005 -1.349393 53 10 0 0 0 -0.0216613 0.0216646 -0.000845432 -0.0802846 0.0802889 -0.0070653 -0.188561 0.188565 -0.028585 -0.354861 0.354864 -0.0808203 -0.581109 0.581115 -0.182213 -0.866384 0.86639 -0.355827 -1.196764 1.196769 -0.610815 -1.554531 1.554536 -0.936922 -1.926514 1.926517 -1.313872 54 10 0 0 0 -0.0283365 -2.86102e-006 -0.000723362 -0.10729 -9.53674e-007 -0.00636172 -0.258353 -2.86102e-006 -0.0272944 -0.495471 -9.53674e-007 -0.0804381 -0.820507 -1.90735e-006 -0.185446 -1.229997 -1.90735e-006 -0.365555 -1.701145 -2.86102e-006 -0.626727 -2.208061 -3.8147e-006 -0.954955 -2.733027 -3.8147e-006 -1.328701 55 10 0 0 0 -0.0348282 -1.90735e-006 -0.00109315 -0.129632 -1.90735e-006 -0.00924039 -0.305889 -2.86102e-006 -0.0379415 -0.573134 -2.86102e-006 -0.106443 -0.926338 -1.90735e-006 -0.233176 -1.35638 -2.86102e-006 -0.437013 -1.839267 -4.76837e-006 -0.717864 -2.351793 -3.8147e-006 -1.059042 -2.879028 -4.76837e-006 -1.439485 56 10 0 0 0 -0.036026 -9.53674e-007 -0.00116944 -0.133747 -1.90735e-006 -0.00983119 -0.31456 -1.90735e-006 -0.0400782 -0.587116 -1.90735e-006 -0.111532 -0.945189 -1.90735e-006 -0.242322 -1.378789 -1.90735e-006 -0.450552 -1.863832 -1.90735e-006 -0.735231 -2.377571 -1.90735e-006 -1.079329 -2.905488 -2.86102e-006 -1.461835 57 10 0 0 0 -0.0301256 0 -0.000817537 -0.112276 -9.53674e-007 -0.00692368 -0.267234 -9.53674e-007 -0.0289712 -0.508632 9.53674e-007 -0.0841575 -0.837173 -9.53674e-007 -0.191715 -1.248611 0 -0.37396 -1.720171 -9.53674e-007 -0.635794 -2.226561 -9.53674e-007 -0.962852 -2.750563 -9.53674e-007 -1.333841 58 10 0 0 0 -0.0353594 0 -0.00112629 -0.130065 0 -0.00925779 -0.30366 -1.90735e-006 -0.0370762 -0.564453 -9.53674e-007 -0.102092 -0.907873 -2.86102e-006 -0.220951 -1.326996 -1.90735e-006 -0.411843 -1.80031 -2.86102e-006 -0.676513 -2.305923 -2.86102e-006 -1.001853 -2.828487 -2.86102e-006 -1.368809 59 10 0 0 0 -0.030365 0 -0.000830412 -0.114362 0 -0.00721669 -0.273718 -1.90735e-006 -0.0305626 -0.52115 0 -0.0887046 -0.856274 0 -0.201164 -1.273528 -9.53674e-007 -0.389922 -1.749419 0 -0.658823 -2.258861 -9.53674e-007 -0.992769 -2.785112 -1.90735e-006 -1.370276 60 10 0 0 0 -0.0367289 0.0367279 -0.0024333 -0.123212 0.123212 -0.0160635 -0.262391 0.262391 -0.0520997 -0.448421 0.448421 -0.118342 -0.672449 0.67245 -0.217537 -0.931743 0.931744 -0.35575 -1.218695 1.218696 -0.531885 -1.523723 1.523725 -0.73736 -1.837595 1.837597 -0.958832 61 10 0 0 0 -0.0359983 0.0359991 -0.00233769 -0.120431 0.120432 -0.0153213 -0.256007 0.256008 -0.0494552 -0.437087 0.437089 -0.111998 -0.655201 0.655203 -0.205499 -0.908005 0.908008 -0.335822 -1.188285 1.188289 -0.502114 -1.48664 1.486645 -0.696269 -1.793743 1.793748 -0.905389 62 10 0 0 0 -0.0379391 0.0379391 -0.00259662 -0.128342 0.128342 -0.0175076 -0.275119 0.27512 -0.057743 -0.472259 0.47226 -0.132764 -0.710009 0.71001 -0.246061 -0.984511 0.984513 -0.404202 -1.287 1.287003 -0.605297 -1.607386 1.607388 -0.839375 -1.936661 1.936665 -1.091949 63 10 0 0 0 -0.0353298 0.0353284 -0.00225115 -0.118281 0.118281 -0.0147781 -0.251758 0.251758 -0.0478305 -0.430445 0.430445 -0.108629 -0.646104 0.646105 -0.199832 -0.896517 0.896519 -0.327335 -1.17452 1.174522 -0.490373 -1.470661 1.470664 -0.680905 -1.775514 1.775517 -0.886072 64 10 0 0 0 -0.0374818 0.0374823 -0.00253415 -0.126651 0.126652 -0.0170355 -0.271394 0.271395 -0.0561216 -0.465952 0.465953 -0.129042 -0.700868 0.700869 -0.239321 -0.972518 0.972521 -0.393562 -1.272293 1.272295 -0.590074 -1.590106 1.590108 -0.819084 -1.916848 1.916852 -1.066225 65 10 0 0 0 -0.0346026 0.0346031 -0.00215936 -0.115624 0.115625 -0.0141034 -0.245784 0.245785 -0.0454847 -0.419932 0.419933 -0.103057 -0.630151 0.630153 -0.189297 -0.874539 0.87454 -0.309891 -1.146278 1.146281 -0.464252 -1.43609 1.436092 -0.64476 -1.734488 1.73449 -0.838986 66 10 0 0 0 -0.049643 0 -0.00222254 -0.167088 0 -0.0147777 -0.357368 -4.76837e-007 -0.0483799 -0.613466 0 -0.110929 -0.923707 4.76837e-007 -0.205624 -1.28455 4.76837e-007 -0.338764 -1.685281 4.76837e-007 -0.509538 -2.112187 9.53674e-007 -0.709576 -2.552011 9.53674e-007 -0.925725 67 10 0 0 0 -0.051023 0 -0.00234818 -0.172049 2.38419e-007 -0.0156908 -0.368385 7.15256e-007 -0.0515406 -0.632646 7.15256e-007 -0.118416 -0.952504 1.19209e-006 -0.219733 -1.323778 1.66893e-006 -0.362011 -1.735089 1.43051e-006 -0.544117 -2.172415 1.66893e-006 -0.757053 -2.622593 2.14577e-006 -0.98706 68 10 0 0 0 -0.052712 -2.38419e-007 -0.00250602 -0.178433 0 -0.0169182 -0.383059 4.76837e-007 -0.0559753 -0.658758 4.76837e-007 -0.129211 -0.992235 9.53674e-007 -0.24043 -1.37823 1.19209e-006 -0.396424 -1.804327 1.43051e-006 -0.595492 -2.256118 1.66893e-006 -0.827741 -2.720709 1.90735e-006 -1.078671 69 10 0 0 0 -0.0521469 -2.38419e-007 -0.00245214 -0.175909 0 -0.0164127 -0.37658 -2.38419e-007 -0.053921 -0.646323 2.38419e-007 -0.123801 -0.972272 7.15256e-007 -0.229468 -1.349807 9.53674e-007 -0.377459 -1.767186 1.19209e-006 -0.566359 -2.210283 1.66893e-006 -0.786792 -2.666065 1.90735e-006 -1.024692 70 10 0 0 0 -0.0508933 2.38419e-007 -0.00233579 -0.171622 0 -0.0156121 -0.367515 0 -0.0512943 -0.631245 0 -0.117883 -0.950544 0 -0.218807 -1.32127 4.76837e-007 -0.360594 -1.732071 4.76837e-007 -0.542139 -2.168929 4.76837e-007 -0.754478 -2.618667 9.53674e-007 -0.983878 71 10 0 0 0 -0.0510259 0 -0.00234842 -0.171649 4.76837e-007 -0.0156012 -0.366747 4.76837e-007 -0.0509853 -0.628758 4.76837e-007 -0.11665 -0.945436 4.76837e-007 -0.215745 -1.312854 9.53674e-007 -0.354591 -1.719959 1.43051e-006 -0.532116 -2.152909 1.43051e-006 -0.739539 -2.598499 1.90735e-006 -0.963295 72 10 0 0 0 -0.0373573 0.0373573 -0.00251794 -0.127148 0.127148 -0.0172248 -0.27424 0.274241 -0.0576403 -0.47333 0.473331 -0.134272 -0.714783 0.714785 -0.251601 -0.994346 0.994347 -0.41686 -1.302628 1.302629 -0.62808 -1.629133 1.629135 -0.874717 -1.964829 1.964832 -1.141732 73 10 0 0 0 -0.037714 0.037714 -0.00256586 -0.127954 0.127955 -0.0174227 -0.275152 0.275153 -0.0578978 -0.473672 0.473673 -0.134057 -0.713799 0.713801 -0.249931 -0.991426 0.991429 -0.412438 -1.297418 1.297421 -0.619596 -1.621457 1.62146 -0.861082 -1.954533 1.954538 -1.122086 74 10 0 0 0 -0.0380759 0.0380759 -0.00261545 -0.12972 0.129721 -0.0179443 -0.279891 0.279892 -0.0601394 -0.482983 0.482984 -0.140143 -0.728927 0.728929 -0.262496 -1.013084 1.013086 -0.434449 -1.325757 1.32576 -0.653682 -1.656389 1.656392 -0.909226 -1.996097 1.996099 -1.185785 75 10 0 0 0 -0.0379481 0.0379488 -0.00259757 -0.129272 0.129272 -0.0178182 -0.278939 0.27894 -0.0597188 -0.481433 0.481435 -0.139214 -0.726779 0.726782 -0.260889 -1.010395 1.010397 -0.432039 -1.322606 1.322609 -0.650416 -1.652845 1.652848 -0.905097 -1.992198 1.992202 -1.180793 76 10 0 0 0 -0.0377879 0.0377877 -0.00257635 -0.128254 0.128253 -0.0175083 -0.27587 0.275872 -0.0582235 -0.474988 0.474988 -0.134877 -0.715831 0.715833 -0.251536 -0.99423 0.994231 -0.415134 -1.300982 1.300983 -0.623632 -1.625753 1.625755 -0.866636 -1.959556 1.959559 -1.129287 77 10 0 0 0 -0.0377393 0.037739 -0.00256944 -0.128872 0.128872 -0.0177255 -0.278714 0.278714 -0.0597278 -0.481984 0.481984 -0.139883 -0.728728 0.728729 -0.263147 -1.014201 1.014202 -0.437058 -1.328496 1.328497 -0.659328 -1.6609 1.660902 -0.918815 -2.002497 2.0025 -1.200028 78 10 0 0 0 -0.0542603 1.19209e-007 -0.00265574 -0.18448 2.38419e-007 -0.0181327 -0.397041 3.57628e-007 -0.060405 -0.683545 1.19209e-007 -0.13998 -1.029957 4.76837e-007 -0.261206 -1.43032 5.96046e-007 -0.431598 -1.871798 8.34465e-007 -0.649902 -2.340052 5.96046e-007 -0.906492 -2.822718 8.34465e-007 -1.186957 79 10 0 0 0 -0.0529575 2.38419e-007 -0.00252914 -0.1794 2.38419e-007 -0.0171094 -0.384916 4.76837e-007 -0.0565205 -0.660992 2.38419e-007 -0.129972 -0.995744 7.15256e-007 -0.242151 -1.384301 9.53674e-007 -0.400636 -1.814708 9.53674e-007 -0.604919 -2.272911 9.53674e-007 -0.846322 -2.746313 1.19209e-006 -1.111133 80 10 0 0 0 -0.0505652 0 -0.00230551 -0.170529 2.38419e-007 -0.0154123 -0.365122 4.76837e-007 -0.050607 -0.626957 4.76837e-007 -0.116179 -0.943786 9.53674e-007 -0.215377 -1.31149 9.53674e-007 -0.354477 -1.718894 1.19209e-006 -0.532324 -2.152178 1.43051e-006 -0.740162 -2.598154 1.90735e-006 -0.964436 81 10 0 0 0 -0.0511999 2.38419e-007 -0.0023644 -0.172795 4.76837e-007 -0.0158343 -0.370016 4.76837e-007 -0.052019 -0.63512 9.53674e-007 -0.119351 -0.95545 1.43051e-006 -0.221001 -1.326532 1.43051e-006 -0.363107 -1.736935 1.43051e-006 -0.544214 -2.175239 1.66893e-006 -0.758401 -2.628808 2.38419e-006 -0.993151 82 10 0 0 0 -0.0519695 0 -0.00243545 -0.175679 4.76837e-007 -0.0163841 -0.376535 9.53674e-007 -0.053964 -0.646485 7.15256e-007 -0.123959 -0.972381 7.15256e-007 -0.229587 -1.349524 1.66893e-006 -0.377216 -1.768487 1.90735e-006 -0.567925 -2.215559 2.14577e-006 -0.79368 -2.677999 1.90735e-006 -1.041354 83 10 0 0 0 -0.0523529 0 -0.00247192 -0.17757 0 -0.0167675 -0.381751 2.38419e-007 -0.0556486 -0.65708 9.53674e-007 -0.128672 -0.990175 1.19209e-006 -0.239606 -1.375587 1.43051e-006 -0.395036 -1.800857 1.19209e-006 -0.593118 -2.253589 1.66893e-006 -0.826685 -2.722204 2.38419e-006 -1.083844 85 10 0 0 0 -0.0379171 0.0379181 -0.00259399 -0.129712 0.129713 -0.0179741 -0.280661 0.280662 -0.0606256 -0.485057 0.485058 -0.141748 -0.732516 0.732517 -0.26583 -1.019904 1.019905 -0.442626 -1.337746 1.337748 -0.671694 -1.675494 1.675497 -0.943533 -2.024372 2.024374 -1.243725 86 10 0 0 0 -0.0379133 0.0379143 -0.00259399 -0.129639 0.129641 -0.0179508 -0.280394 0.280396 -0.0604873 -0.484444 0.484446 -0.141313 -0.731431 0.731431 -0.264852 -1.018279 1.018283 -0.440834 -1.335554 1.335557 -0.668801 -1.672741 1.672745 -0.939309 -2.021053 2.021058 -1.237966 92 10 0 0 0 -0.05371 0 -0.0026021 -0.187876 -2.38419e-007 -0.0190456 -0.414472 -2.38419e-007 -0.0673568 -0.727112 -4.76837e-007 -0.163677 -1.109582 -4.76837e-007 -0.316295 -1.550929 2.38419e-007 -0.534425 -2.0331 -2.38419e-007 -0.814022 -2.53945 0 -1.14099 -3.0583 -4.76837e-007 -1.497959 96 10 0 0 0 -0.0357208 0.0357199 -0.00230098 -0.126367 0.126366 -0.0172935 -0.282044 0.282044 -0.0627806 -0.50018 0.50018 -0.156302 -0.770291 0.77029 -0.308468 -1.084427 1.084427 -0.530439 -1.428856 1.428856 -0.818824 -1.790923 1.790926 -1.158848 -2.162005 2.162004 -1.532082 97 10 0 0 0 -0.0398831 0.0398846 -0.00287104 -0.13982 0.139823 -0.0211489 -0.308676 0.308678 -0.0750892 -0.540804 0.540806 -0.182455 -0.822753 0.822755 -0.351161 -1.142974 1.142974 -0.584907 -1.487894 1.487897 -0.874566 -1.847188 1.847189 -1.205492 -2.213703 2.213702 -1.561105 98 10 0 0 0 -0.0408087 0.0408087 -0.00300574 -0.14225 0.14225 -0.0218477 -0.31225 0.31225 -0.0765623 -0.544215 0.544215 -0.183757 -0.823298 0.823298 -0.34833 -1.139304 1.139306 -0.573853 -1.480411 1.480411 -0.853843 -1.836331 1.836333 -1.174284 -2.199726 2.199728 -1.518812 99 10 0 0 0 -0.0407724 0.0407724 -0.00300002 -0.142079 0.142079 -0.0217917 -0.311776 0.311775 -0.0762997 -0.543204 0.543203 -0.18294 -0.821732 0.821733 -0.346725 -1.137444 1.137444 -0.571682 -1.478388 1.478388 -0.851274 -1.834235 1.834235 -1.171484 -2.197608 2.197606 -1.51594 100 10 0 0 0 -0.0371294 0.0371294 -0.00248671 -0.130611 0.13061 -0.0184457 -0.289756 0.289758 -0.0660771 -0.51083 0.510831 -0.1624 -0.782581 0.782582 -0.316776 -1.09691 1.096911 -0.539111 -1.440449 1.44045 -0.825216 -1.801025 1.801026 -1.160294 -2.170304 2.170305 -1.526358 101 10 0 0 0 -0.0370131 0.0370121 -0.00247145 -0.131108 0.131108 -0.0186436 -0.29264 0.292639 -0.0677823 -0.518184 0.518183 -0.168473 -0.79579 0.795789 -0.330949 -1.116042 1.11604 -0.564762 -1.464411 1.46441 -0.863569 -1.8283 1.828297 -1.209805 -2.199743 2.199741 -1.584523 102 10 0 0 0 -0.0559387 0 -0.00282288 -0.195853 -2.38419e-007 -0.0207303 -0.431896 -2.38419e-007 -0.0733683 -0.755724 0 -0.177506 -1.147942 -4.76837e-007 -0.339607 -1.595151 -4.76837e-007 -0.565547 -2.078856 -7.15256e-007 -0.847845 -2.583467 -1.19209e-006 -1.170997 -3.098348 -1.66893e-006 -1.51789 103 10 0 0 0 -0.0555391 -2.38419e-007 -0.00278234 -0.195414 -2.38419e-007 -0.0206788 -0.43296 -7.15256e-007 -0.0740254 -0.760772 -1.19209e-006 -0.181048 -1.1595 -1.19209e-006 -0.349748 -1.613763 -7.15256e-007 -0.585479 -2.103415 -1.19209e-006 -0.87858 -2.613121 -1.43051e-006 -1.213131 -3.132758 -1.90735e-006 -1.57217 104 10 0 0 0 -0.0544701 -7.15256e-007 -0.0026772 -0.192352 -7.15256e-007 -0.0200593 -0.427773 -9.53674e-007 -0.0724061 -0.754297 -1.43051e-006 -0.17849 -1.153275 -9.53674e-007 -0.347446 -1.609837 -1.19209e-006 -0.586468 -2.102221 -1.43051e-006 -0.884733 -2.614464 -1.43051e-006 -1.225229 -3.136548 -1.66893e-006 -1.590864 105 10 0 0 0 -0.0519886 4.76837e-007 -0.00243878 -0.185002 0 -0.0185964 -0.41474 4.76837e-007 -0.068323 -0.736809 4.76837e-007 -0.171205 -1.133628 4.76837e-007 -0.337946 -1.590623 4.76837e-007 -0.577595 -2.086075 0 -0.881816 -2.601831 -4.76837e-007 -1.230882 -3.127376 0 -1.606312 106 10 0 0 0 -0.0483837 0 -0.00211024 -0.173901 0 -0.0164747 -0.394131 -9.53674e-007 -0.0619898 -0.707606 -4.76837e-007 -0.158879 -1.098452 -4.76837e-007 -0.319616 -1.552233 -4.76837e-007 -0.554663 -2.046663 -4.76837e-007 -0.856883 -2.563172 -9.53674e-007 -1.20784 -3.090223 -9.53674e-007 -1.587727 107 10 0 0 0 -0.0544786 0 -0.00267696 -0.192169 0 -0.0200102 -0.426934 0 -0.0720506 -0.75203 4.76837e-007 -0.177101 -1.148435 2.38419e-007 -0.34342 -1.601526 -2.38419e-007 -0.577495 -2.091205 -4.76837e-007 -0.870644 -2.601331 -4.76837e-007 -1.206169 -3.121459 -7.15256e-007 -1.566513 108 10 0 0 0 -0.0260978 0.0260925 -0.00122666 -0.0957131 0.0957108 -0.0100191 -0.222821 0.22282 -0.0399044 -0.412687 0.412685 -0.109101 -0.66105 0.661047 -0.234222 -0.962252 0.96225 -0.433129 -1.300579 1.300575 -0.70635 -1.660344 1.660341 -1.038791 -2.031216 2.03121 -1.411176 109 10 0 0 0 -0.0348234 0.0348215 -0.00218678 -0.12351 0.123508 -0.0165303 -0.276815 0.276815 -0.060581 -0.493036 0.493036 -0.152305 -0.761406 0.761404 -0.302154 -1.073434 1.073432 -0.52019 -1.415192 1.415192 -0.801804 -1.774132 1.774132 -1.131603 -2.141809 2.141806 -1.491531 110 10 0 0 0 -0.031074 0.0310755 -0.00174069 -0.113085 0.113087 -0.0139818 -0.25982 0.259821 -0.0541911 -0.472899 0.472898 -0.143018 -0.7432 0.7432 -0.295441 -1.061144 1.061146 -0.524708 -1.410271 1.410273 -0.82558 -1.776453 1.776453 -1.179976 -2.150855 2.150852 -1.567337 111 10 0 0 0 -0.0314202 0.0314217 -0.00178003 -0.114176 0.114178 -0.0142467 -0.261915 0.261916 -0.0550323 -0.475968 0.475969 -0.14475 -0.746939 0.746941 -0.298072 -1.065145 1.065146 -0.527848 -1.414197 1.414199 -0.828518 -1.780102 1.780104 -1.181914 -2.154126 2.154127 -1.567595 112 10 0 0 0 -0.0285044 0.0285053 -0.00146437 -0.10372 0.103722 -0.0117424 -0.239146 0.239147 -0.0457978 -0.438474 0.438477 -0.122629 -0.695663 0.695663 -0.258244 -1.003143 1.003145 -0.468036 -1.344858 1.344858 -0.749537 -1.706186 1.706186 -1.087087 -2.077566 2.077564 -1.46154 113 10 0 0 0 -0.032753 0.032753 -0.00193524 -0.117906 0.117904 -0.0151429 -0.267984 0.267982 -0.0572824 -0.482957 0.482955 -0.147849 -0.752752 0.75275 -0.299592 -1.068089 1.068088 -0.523839 -1.413511 1.413508 -0.814797 -1.775888 1.775886 -1.155857 -2.146835 2.146832 -1.528547 114 10 0 0 0 -0.0447445 4.76837e-007 -0.00180459 -0.159988 0 -0.0138881 -0.361994 4.76837e-007 -0.0519147 -0.651188 -4.76837e-007 -0.133121 -1.015793 -1.43051e-006 -0.269504 -1.447517 -4.76837e-007 -0.47541 -1.929588 -1.43051e-006 -0.754831 -2.441587 -1.43051e-006 -1.094748 -2.969282 -1.90735e-006 -1.476582 115 10 0 0 0 -0.0484772 0 -0.00211859 -0.171967 0 -0.0160177 -0.385485 4.76837e-007 -0.0586871 -0.686754 0 -0.147468 -1.061993 0 -0.293344 -1.50038 0 -0.507638 -1.982157 -9.53674e-007 -0.786547 -2.489134 -9.53674e-007 -1.114909 -3.009001 -1.90735e-006 -1.474559 116 10 0 0 0 -0.0461884 0 -0.00192332 -0.164284 -9.53674e-007 -0.0146189 -0.370062 -9.53674e-007 -0.0541339 -0.663456 -9.53674e-007 -0.137925 -1.032282 -2.38419e-006 -0.278019 -1.466299 -2.38419e-006 -0.486772 -1.945511 -2.86102e-006 -0.761261 -2.453386 -2.86102e-006 -1.091639 -2.977453 -3.8147e-006 -1.462818 117 10 0 0 0 -0.0404186 0 -0.00147247 -0.146391 1.90735e-006 -0.0116737 -0.335867 9.53674e-007 -0.0449822 -0.612804 9.53674e-007 -0.11893 -0.96858 1.90735e-006 -0.247794 -1.395243 1.90735e-006 -0.447538 -1.872232 -9.53674e-007 -0.718273 -2.378799 9.53674e-007 -1.045721 -2.902136 -9.53674e-007 -1.414831 118 10 0 0 0 -0.0481777 -9.53674e-007 -0.0020926 -0.17065 -9.53674e-007 -0.0157602 -0.382296 -9.53674e-007 -0.0576549 -0.681442 -9.53674e-007 -0.145072 -1.054642 -1.43051e-006 -0.289088 -1.491226 -1.90735e-006 -0.501081 -1.971518 -2.38419e-006 -0.777417 -2.477278 -1.90735e-006 -1.103085 -2.998323 -2.86102e-006 -1.465891 119 10 0 0 0 -0.044117 4.76837e-007 -0.00175452 -0.157829 9.53674e-007 -0.0135164 -0.357462 0 -0.0506234 -0.644604 9.53674e-007 -0.130583 -1.008642 -4.76837e-007 -0.266475 -1.440023 -4.76837e-007 -0.471954 -1.918432 -9.53674e-007 -0.745078 -2.426464 -9.53674e-007 -1.075811 -2.951374 -1.90735e-006 -1.449399 120 10 0 0 0 -0.0303688 0.0303674 -0.00166178 -0.101273 0.101272 -0.0107861 -0.215342 0.21534 -0.0347233 -0.368672 0.368673 -0.078789 -0.554972 0.554972 -0.145237 -0.773426 0.773429 -0.23906 -1.018314 1.018315 -0.360216 -1.280701 1.280703 -0.502312 -1.550596 1.550598 -0.654191 121 10 0 0 0 -0.0332088 0.0332091 -0.00198841 -0.111197 0.111197 -0.0130463 -0.236992 0.236992 -0.0423006 -0.406026 0.406026 -0.0963616 -0.610882 0.610883 -0.177882 -0.849942 0.849943 -0.292594 -1.116613 1.116614 -0.44021 -1.40169 1.401691 -0.613527 -1.695656 1.695657 -0.800559 122 10 0 0 0 -0.0325222 0.0325229 -0.00190735 -0.108756 0.108756 -0.0124683 -0.231573 0.231575 -0.0403181 -0.396544 0.396545 -0.0916796 -0.596518 0.596518 -0.169047 -0.830123 0.830125 -0.277951 -1.091073 1.091076 -0.418233 -1.370335 1.370337 -0.583061 -1.65838 1.658384 -0.760857 123 10 0 0 0 -0.0333109 0.0333102 -0.00200081 -0.111556 0.111556 -0.0131326 -0.237783 0.237785 -0.0425938 -0.407401 0.407403 -0.0970492 -0.612953 0.612954 -0.179172 -0.852782 0.852785 -0.29472 -1.120254 1.120256 -0.443386 -1.406137 1.406139 -0.617911 -1.700917 1.70092 -0.806248 124 10 0 0 0 -0.0337753 0.0337741 -0.0020566 -0.113252 0.113252 -0.0135453 -0.241611 0.241611 -0.0440394 -0.414189 0.41419 -0.100521 -0.623358 0.623358 -0.18582 -0.867272 0.867273 -0.305885 -1.139063 1.139064 -0.460316 -1.429369 1.42937 -0.641591 -1.728708 1.72871 -0.837375 125 10 0 0 0 -0.0325851 0.0325823 -0.00191426 -0.109035 0.109033 -0.0125358 -0.232306 0.232301 -0.040596 -0.397974 0.397972 -0.0924172 -0.59889 0.598887 -0.170575 -0.833655 0.833653 -0.280696 -1.095971 1.095969 -0.422703 -1.376717 1.376715 -0.589663 -1.66611 1.666108 -0.769527 126 10 0 0 0 -0.0485926 -4.76837e-007 -0.00212979 -0.162974 -4.76837e-007 -0.0140319 -0.347437 0 -0.0455494 -0.594733 0 -0.103625 -0.893374 -4.76837e-007 -0.190719 -1.240091 9.53674e-007 -0.312191 -1.624736 4.76837e-007 -0.466885 -2.03533 4.76837e-007 -0.648201 -2.459676 4.76837e-007 -0.845187 127 10 0 0 0 -0.0446968 9.53674e-007 -0.00180006 -0.14888 1.90735e-006 -0.0116563 -0.315759 0 -0.0373127 -0.538636 9.53674e-007 -0.0839796 -0.807334 9.53674e-007 -0.15328 -1.119781 2.86102e-006 -0.249468 -1.467355 2.86102e-006 -0.371624 -1.839287 2.86102e-006 -0.514495 -2.222297 2.86102e-006 -0.667628 128 10 0 0 0 -0.0476494 4.76837e-007 -0.00204659 -0.159506 0 -0.0134242 -0.339532 4.76837e-007 -0.0434012 -0.580587 4.76837e-007 -0.0984225 -0.871523 4.76837e-007 -0.180693 -1.209418 4.76837e-007 -0.295263 -1.584547 4.76837e-007 -0.441037 -1.985384 9.53674e-007 -0.611922 -2.39932 4.76837e-007 -0.796939 129 10 0 0 0 -0.0485911 4.76837e-007 -0.00212884 -0.163155 0 -0.0140696 -0.348263 4.76837e-007 -0.0458148 -0.596949 4.76837e-007 -0.104583 -0.897944 4.76837e-007 -0.193188 -1.2482 1.43051e-006 -0.31751 -1.637653 1.43051e-006 -0.476873 -2.052759 9.53674e-007 -0.663202 -2.47991 4.76837e-007 -0.863532 130 10 0 0 0 -0.0487685 0 -0.00214386 -0.163664 4.76837e-007 -0.0141547 -0.349093 0 -0.0460135 -0.597882 9.53674e-007 -0.104833 -0.898555 0 -0.19323 -1.247796 9.53674e-007 -0.316728 -1.635341 9.53674e-007 -0.474226 -2.04795 9.53674e-007 -0.657766 -2.474507 9.53674e-007 -0.857384 131 10 0 0 0 -0.0455923 9.53674e-007 -0.00187325 -0.152061 1.43051e-006 -0.0121713 -0.32281 2.38419e-006 -0.0390613 -0.550996 1.90735e-006 -0.0880859 -0.826157 1.90735e-006 -0.161015 -1.145993 2.38419e-006 -0.262316 -1.501575 2.86102e-006 -0.391018 -1.882055 3.33786e-006 -0.541757 -2.274212 3.33786e-006 -0.703796 132 10 0 0 0 -0.0369453 0.0369449 -0.0024631 -0.124724 0.124722 -0.0165095 -0.26706 0.267058 -0.0542591 -0.458339 0.458338 -0.124566 -0.68956 0.68956 -0.230994 -0.957644 0.957643 -0.380464 -1.254621 1.254621 -0.572359 -1.570833 1.570834 -0.798277 -1.897286 1.897288 -1.044805 133 10 0 0 0 -0.0370512 0.0370522 -0.00247669 -0.125196 0.125198 -0.0166426 -0.268094 0.268095 -0.0547014 -0.459772 0.459774 -0.125324 -0.691997 0.691997 -0.232787 -0.961644 0.961647 -0.384337 -1.260636 1.260638 -0.579546 -1.57922 1.579222 -0.810063 -1.90842 1.908422 -1.062475 134 10 0 0 0 -0.0347109 0.0347106 -0.00217295 -0.116232 0.116231 -0.0142665 -0.247448 0.247447 -0.046174 -0.42316 0.42316 -0.104846 -0.635369 0.63537 -0.192881 -0.882092 0.882093 -0.31612 -1.156568 1.156569 -0.474223 -1.449699 1.4497 -0.659928 -1.752203 1.752205 -0.861048 135 10 0 0 0 -0.0353565 0.0353565 -0.00225472 -0.119065 0.119067 -0.0150142 -0.254665 0.254666 -0.0491608 -0.43701 0.43701 -0.112636 -0.657642 0.657644 -0.208535 -0.913893 0.913894 -0.343003 -1.198112 1.198114 -0.515048 -1.500636 1.500638 -0.716202 -1.811992 1.811995 -0.932999 136 10 0 0 0 -0.0351448 0.0351448 -0.00222778 -0.117882 0.117881 -0.014689 -0.25109 0.25109 -0.0476037 -0.429169 0.429169 -0.107965 -0.644254 0.644253 -0.198635 -0.894478 0.894478 -0.325917 -1.172903 1.172904 -0.489555 -1.470345 1.470345 -0.682206 -1.777566 1.777569 -0.89154 137 10 0 0 0 -0.0357437 0.0357451 -0.00230479 -0.12034 0.120342 -0.0153391 -0.257188 0.257188 -0.0501379 -0.440833 0.440835 -0.114582 -0.662541 0.662543 -0.211518 -0.919361 0.919364 -0.346682 -1.204398 1.2044 -0.519942 -1.509018 1.50902 -0.724703 -1.824078 1.82408 -0.948422 138 10 0 0 0 -0.054388 0 -0.00266886 -0.182331 0 -0.0176017 -0.387673 0 -0.0569435 -0.661061 0 -0.128866 -0.988971 0 -0.23596 -1.366841 0 -0.384262 -1.783509 -4.76837e-007 -0.572353 -2.22548 0 -0.791302 -2.680021 4.76837e-007 -1.027429 139 10 0 0 0 -0.0546875 4.76837e-007 -0.00269818 -0.18516 -2.38419e-007 -0.0182359 -0.39695 4.76837e-007 -0.0601895 -0.681117 2.38419e-007 -0.138366 -1.023565 2.38419e-007 -0.256462 -1.418657 2.38419e-007 -0.421448 -1.854035 2.38419e-007 -0.631913 -2.315661 2.38419e-007 -0.878367 -2.79113 4.76837e-007 -1.14657 140 10 0 0 0 -0.0533142 -1.43051e-006 -0.00256348 -0.178808 -4.76837e-007 -0.0169232 -0.380498 -4.76837e-007 -0.0548271 -0.649495 -9.53674e-007 -0.124294 -0.972726 -9.53674e-007 -0.228005 -1.345967 -4.76837e-007 -0.37206 -1.758335 -4.76837e-007 -0.555332 -2.196423 0 -0.769241 -2.647412 -4.76837e-007 -1.000374 141 10 0 0 0 -0.0491867 1.43051e-006 -0.00218201 -0.163908 4.76837e-007 -0.0141561 -0.347417 9.53674e-007 -0.0453386 -0.591928 9.53674e-007 -0.102039 -0.886086 4.76837e-007 -0.186306 -1.227251 1.90735e-006 -0.303401 -1.606182 1.43051e-006 -0.452689 -2.01018 4.76837e-007 -0.626895 -2.425955 1.43051e-006 -0.813976 142 10 0 0 0 -0.0506868 0 -0.00231624 -0.168943 0 -0.0150485 -0.357991 4.76837e-007 -0.0482023 -0.609565 0 -0.108427 -0.911765 0 -0.19781 -1.261545 1.43051e-006 -0.321744 -1.649237 9.53674e-007 -0.479385 -2.06194 9.53674e-007 -0.663029 -2.486438 4.76837e-007 -0.860194 143 10 0 0 0 -0.0551529 0 -0.00274444 -0.185264 0 -0.0181949 -0.394406 0 -0.0590646 -0.672973 4.76837e-007 -0.13395 -1.007008 4.76837e-007 -0.245589 -1.391512 4.76837e-007 -0.400147 -1.814918 4.76837e-007 -0.596023 -2.263589 2.38419e-007 -0.823955 -2.724942 4.76837e-007 -1.070003 144 10 0 0 0 -0.0393362 0.0393372 -0.00279188 -0.133656 0.133656 -0.0190423 -0.287237 0.287238 -0.0632591 -0.493526 0.493526 -0.146023 -0.742025 0.742027 -0.271303 -1.028149 1.028149 -0.446189 -1.342626 1.342626 -0.668803 -1.675346 1.675347 -0.928999 -2.017693 2.017694 -1.212089 146 10 0 0 0 -0.039238 0.039237 -0.00277781 -0.133336 0.133337 -0.0189509 -0.286606 0.286606 -0.0629805 -0.492561 0.492562 -0.145454 -0.740762 0.740764 -0.270389 -1.026651 1.026652 -0.444924 -1.340974 1.340973 -0.667243 -1.673608 1.673609 -0.927248 -2.015926 2.015929 -1.210268 147 10 0 0 0 -0.0399895 0.0399895 -0.00288486 -0.135615 0.135616 -0.0195961 -0.290765 0.290766 -0.0647602 -0.498392 0.498394 -0.148698 -0.747707 0.747708 -0.274922 -1.034037 1.034038 -0.450119 -1.348163 1.348164 -0.672069 -1.680069 1.680068 -0.930439 -2.021196 2.021197 -1.210487 150 10 0 0 0 -0.0575266 2.38419e-007 -0.00298643 -0.196713 1.78814e-007 -0.0207043 -0.42452 2.98023e-007 -0.0695585 -0.730512 2.38419e-007 -0.161422 -1.097586 -1.78814e-007 -0.299965 -1.517035 -2.38419e-007 -0.491234 -1.974039 -4.76837e-007 -0.730896 -2.453869 -7.15256e-007 -1.006442 -2.944803 -9.53674e-007 -1.30195 151 10 0 0 0 -0.0592461 3.57628e-007 -0.00316811 -0.205069 2.38419e-007 -0.0226471 -0.446665 2.38419e-007 -0.0779295 -0.770757 0 -0.182257 -1.15661 2.38419e-007 -0.338114 -1.59378 -5.96046e-008 -0.55085 -2.066768 -1.19209e-007 -0.814987 -2.561192 -4.17233e-007 -1.117189 -3.066261 -5.36442e-007 -1.441344 152 10 0 0 0 -0.0555029 -5.96046e-007 -0.00277948 -0.188759 -4.76837e-007 -0.0189974 -0.406166 -7.15256e-007 -0.063303 -0.698486 -1.07288e-006 -0.146428 -1.050337 -1.19209e-006 -0.272049 -1.454607 -1.19209e-006 -0.446543 -1.897486 -1.54972e-006 -0.666688 -2.364277 -1.78814e-006 -0.921 -2.842632 -1.90735e-006 -1.194035 153 10 0 0 0 -0.0580769 0 -0.00304341 -0.199065 -2.38419e-007 -0.0212314 -0.430182 -1.78814e-007 -0.0715861 -0.740542 -2.68221e-007 -0.166362 -1.112413 -5.36442e-007 -0.309177 -1.536552 -5.66244e-007 -0.505918 -1.997849 -8.34465e-007 -0.751879 -2.481598 -1.3113e-006 -1.034251 -2.976303 -1.46031e-006 -1.337005 156 10 0 0 0 -0.0383005 0.0383005 -0.00264668 -0.134306 0.134306 -0.0194929 -0.296917 0.296917 -0.0693238 -0.521221 0.521221 -0.168791 -0.794524 0.794524 -0.325279 -1.107969 1.107967 -0.545952 -1.44835 1.448348 -0.824153 -1.804324 1.804322 -1.14475 -2.168159 2.168157 -1.490806 157 10 0 0 0 -0.040947 0.0409431 -0.00302601 -0.1404 0.1404 -0.0211253 -0.30358 0.303575 -0.0713205 -0.522897 0.522893 -0.165963 -0.785849 0.785845 -0.308778 -1.085862 1.08586 -0.50569 -1.412202 1.412198 -0.751975 -1.754433 1.754429 -1.034771 -2.10442 2.104416 -1.338017 158 10 0 0 0 -0.0400524 0.0400524 -0.00289559 -0.137582 0.137588 -0.0202909 -0.298326 0.298328 -0.0689285 -0.5155 0.5155 -0.161543 -0.776997 0.776999 -0.302515 -1.07626 1.07626 -0.498173 -1.402327 1.402328 -0.743891 -1.744551 1.744551 -1.026666 -2.094731 2.094731 -1.330446 159 10 0 0 0 -0.0398874 0.0398855 -0.00287127 -0.139099 0.139095 -0.0208797 -0.305575 0.305572 -0.0732329 -0.533 0.532994 -0.1758 -0.807822 0.807818 -0.334384 -1.121192 1.121186 -0.554913 -1.460516 1.460512 -0.830545 -1.814877 1.814871 -1.146305 -2.176643 2.176638 -1.485317 160 10 0 0 0 -0.0407953 0.0407972 -0.00300407 -0.140551 0.140553 -0.0212147 -0.305216 0.305218 -0.0723772 -0.52741 0.527411 -0.169782 -0.794224 0.794226 -0.317586 -1.098462 1.098464 -0.521685 -1.428895 1.428895 -0.77679 -1.774969 1.774969 -1.06946 -2.128696 2.128698 -1.383367 161 10 0 0 0 -0.0388927 0.0388947 -0.00272942 -0.134951 0.134954 -0.0195944 -0.295514 0.295516 -0.0681176 -0.514928 0.514934 -0.162853 -0.781056 0.78106 -0.309759 -1.086426 1.086428 -0.515825 -1.419135 1.419139 -0.775999 -1.768182 1.768184 -1.07665 -2.125451 2.125453 -1.40121 162 10 0 0 0 -0.0540924 1.19209e-007 -0.00264001 -0.185522 -5.96046e-008 -0.01841 -0.402657 -2.38419e-007 -0.0625999 -0.697836 -3.57628e-007 -0.147506 -1.056017 -4.76837e-007 -0.278387 -1.469383 -6.55651e-007 -0.46277 -1.922937 -1.13249e-006 -0.697499 -2.40126 -1.46031e-006 -0.970479 -2.892061 -1.74344e-006 -1.265732 163 10 0 0 0 -0.0536461 0 -0.00259638 -0.185951 0 -0.0185804 -0.407368 -2.38419e-007 -0.0646108 -0.711203 -4.76837e-007 -0.155057 -1.081816 -8.34465e-007 -0.296742 -1.509687 -1.19209e-006 -0.497943 -1.978022 -1.43051e-006 -0.75466 -2.470669 -1.90735e-006 -1.053427 -2.975677 -2.14577e-006 -1.377448 164 10 0 0 0 -0.0541344 0 -0.00264359 -0.185038 5.96046e-008 -0.018286 -0.400356 -1.78814e-007 -0.0617092 -0.69206 -3.57628e-007 -0.144453 -1.045273 -5.36442e-007 -0.271193 -1.452795 -8.34465e-007 -0.449167 -1.900312 -1.13249e-006 -0.675527 -2.372719 -1.43051e-006 -0.938723 -2.857624 -1.78814e-006 -1.223155 165 10 0 0 0 -0.0548878 0 -0.00271821 -0.18795 0 -0.018889 -0.406999 -1.78814e-007 -0.0638955 -0.703519 -3.57628e-007 -0.149645 -1.061872 -4.76837e-007 -0.28067 -1.474084 -6.25849e-007 -0.463769 -1.925685 -9.23872e-007 -0.695753 -2.401889 -1.2517e-006 -0.965177 -2.890504 -1.54972e-006 -1.256351 166 10 0 0 0 -0.0528507 2.38419e-007 -0.00251961 -0.18391 2.38419e-007 -0.0181999 -0.404596 2.38419e-007 -0.0639124 -0.709003 -2.38419e-007 -0.154733 -1.08164 -4.76837e-007 -0.29824 -1.512556 -9.53674e-007 -0.503148 -1.984324 -1.43051e-006 -0.765315 -2.480438 -1.43051e-006 -1.070848 -2.988943 -2.14577e-006 -1.402652 167 10 0 0 0 -0.0519857 2.38419e-007 -0.00243735 -0.182429 -4.76837e-007 -0.0179687 -0.404288 -2.38419e-007 -0.0641906 -0.712564 0 -0.157568 -1.091048 -7.15256e-007 -0.30644 -1.528646 -9.53674e-007 -0.519722 -2.006965 -1.66893e-006 -0.792696 -2.509284 -1.66893e-006 -1.110939 -3.023892 -1.90735e-006 -1.457162 168 10 0 0 0 -0.0368032 0.0368032 -0.00244331 -0.128476 0.128476 -0.017782 -0.283514 0.283514 -0.0628786 -0.498148 0.49815 -0.153133 -0.761559 0.76156 -0.296533 -1.066586 1.066586 -0.502003 -1.400703 1.400702 -0.765373 -1.752257 1.752258 -1.073017 -2.114606 2.114605 -1.413979 169 10 0 0 0 -0.0360661 0.0360661 -0.0023458 -0.126752 0.126752 -0.0173516 -0.281412 0.281412 -0.0622184 -0.497082 0.49708 -0.15343 -0.763184 0.763184 -0.300303 -1.072131 1.072129 -0.512717 -1.410654 1.410652 -0.786413 -1.766518 1.766516 -1.10667 -2.131239 2.131233 -1.455818 170 10 0 0 0 -0.0372038 0.0372 -0.00249696 -0.129409 0.129409 -0.0180173 -0.284473 0.284472 -0.0631292 -0.498007 0.498005 -0.15237 -0.759026 0.759024 -0.292739 -1.060664 1.060664 -0.492388 -1.39218 1.392179 -0.749891 -1.743199 1.743194 -1.056013 -2.104954 2.104951 -1.394982 171 10 0 0 0 -0.0355225 0.0355263 -0.00227523 -0.124932 0.124932 -0.0168552 -0.277812 0.277814 -0.0606523 -0.491549 0.491552 -0.150081 -0.755774 0.755776 -0.294526 -1.063003 1.063004 -0.503868 -1.400033 1.400033 -0.774004 -1.755982 1.75598 -1.094522 -2.122809 2.122809 -1.451289 172 10 0 0 0 -0.0354166 0.0354185 -0.00226426 -0.124639 0.124639 -0.0167813 -0.277393 0.277393 -0.0605032 -0.491259 0.491262 -0.150049 -0.75596 0.755959 -0.295104 -1.063957 1.063955 -0.505814 -1.401896 1.401894 -0.778109 -1.757551 1.757547 -1.097738 -2.124104 2.124102 -1.453504 173 10 0 0 0 -0.0346785 0.0346813 -0.00216937 -0.122431 0.122431 -0.016207 -0.273333 0.273333 -0.0588304 -0.48536 0.485361 -0.1467 -0.74846 0.748461 -0.289701 -1.055118 1.055119 -0.498039 -1.39419 1.39419 -0.773051 -1.753257 1.753254 -1.103253 -2.12306 2.123058 -1.471372 174 10 0 0 0 -0.0528126 2.38419e-007 -0.00251579 -0.187021 -4.76837e-007 -0.0189695 -0.417463 -2.38419e-007 -0.0690179 -0.739147 -7.15256e-007 -0.171625 -1.134655 -1.19209e-006 -0.337031 -1.59013 -1.43051e-006 -0.574492 -2.084753 -1.43051e-006 -0.877087 -2.600925 -2.38419e-006 -1.227197 -3.127499 -2.14577e-006 -1.605657 175 10 0 0 0 -0.0487061 0 -0.00213909 -0.172571 0 -0.0161238 -0.38596 0 -0.0587394 -0.68714 -1.43051e-006 -0.147462 -1.064798 -9.53674e-007 -0.295568 -1.508165 -1.43051e-006 -0.516359 -1.996777 -1.90735e-006 -0.807528 -2.511541 -2.38419e-006 -1.154138 -3.039682 -2.86102e-006 -1.537328 176 10 0 0 0 -0.0530796 9.53674e-007 -0.00254273 -0.187901 4.76837e-007 -0.0191488 -0.419222 7.15256e-007 -0.0695981 -0.741831 2.38419e-007 -0.172864 -1.138094 0 -0.339039 -1.594027 -4.76837e-007 -0.577156 -2.088791 -4.76837e-007 -0.880029 -2.604868 -9.53674e-007 -1.229899 -3.131224 -1.43051e-006 -1.60771 177 10 0 0 0 -0.0518131 2.38419e-007 -0.00242138 -0.181257 -2.38419e-007 -0.0177119 -0.400415 -2.38419e-007 -0.0627661 -0.703602 -2.38419e-007 -0.152791 -1.076628 -1.19209e-006 -0.29665 -1.511961 -7.15256e-007 -0.507058 -1.991514 -1.19209e-006 -0.782132 -2.497738 -1.66893e-006 -1.108819 -3.018135 -1.90735e-006 -1.46988 178 10 0 0 0 -0.0518723 -2.38419e-007 -0.00242639 -0.181793 0 -0.0178323 -0.402552 -4.76837e-007 -0.063576 -0.711056 -4.76837e-007 -0.157106 -1.092438 -9.53674e-007 -0.308695 -1.535713 -1.43051e-006 -0.529364 -2.021667 -1.66893e-006 -0.815678 -2.532557 -1.90735e-006 -1.152983 -3.056248 -2.14577e-006 -1.523093 179 10 0 0 0 -0.0525341 -4.76837e-007 -0.00248933 -0.183308 -4.76837e-007 -0.0181005 -0.403736 -7.15256e-007 -0.0637009 -0.707527 -9.53674e-007 -0.154119 -1.082373 -9.53674e-007 -0.299633 -1.518585 -9.53674e-007 -0.511153 -1.99819 -1.90735e-006 -0.786314 -2.50388 -2.38419e-006 -1.111829 -3.023329 -2.38419e-006 -1.470373 180 10 0 0 0 -0.0296955 0.029696 -0.00158954 -0.098568 0.0985689 -0.0101943 -0.208534 0.208535 -0.0324066 -0.355173 0.355174 -0.0725617 -0.531847 0.531847 -0.131917 -0.737495 0.737496 -0.214124 -0.966663 0.966665 -0.318452 -1.211864 1.211864 -0.439962 -1.463648 1.463651 -0.569077 181 10 0 0 0 -0.0298481 0.0298486 -0.00160527 -0.0992565 0.0992565 -0.0103445 -0.210396 0.210396 -0.0330427 -0.359023 0.359025 -0.0743394 -0.538593 0.538593 -0.135778 -0.748121 0.748124 -0.2214 -0.982061 0.982062 -0.330652 -1.231852 1.231853 -0.45743 -1.487816 1.487818 -0.591551 182 10 0 0 0 -0.0280447 0.0280437 -0.00141764 -0.0928783 0.092876 -0.00903535 -0.19622 0.19622 -0.0286033 -0.333954 0.333955 -0.0638702 -0.499929 0.49993 -0.115892 -0.693396 0.693395 -0.187934 -0.909298 0.909298 -0.279362 -1.13953 1.139531 -0.384775 -1.374939 1.374939 -0.495575 183 10 0 0 0 -0.0303507 0.0303488 -0.00165939 -0.100821 0.100821 -0.0106719 -0.213403 0.213401 -0.0339742 -0.363518 0.363517 -0.0761359 -0.544326 0.544324 -0.138479 -0.754621 0.75462 -0.224785 -0.989545 0.989544 -0.335074 -1.240901 1.240899 -0.463683 -1.499216 1.499214 -0.600682 184 10 0 0 0 -0.0289001 0.0288978 -0.00150585 -0.0959225 0.095921 -0.00965118 -0.203049 0.203045 -0.0307066 -0.34614 0.346139 -0.0688739 -0.518902 0.518899 -0.125481 -0.720512 0.720511 -0.204226 -0.945702 0.945699 -0.304564 -1.186136 1.186136 -0.420776 -1.432127 1.432127 -0.543183 185 10 0 0 0 -0.0293636 0.0293627 -0.00155377 -0.0977364 0.0977378 -0.0100334 -0.20746 0.20746 -0.0321436 -0.354582 0.354582 -0.0725763 -0.532833 0.532833 -0.133061 -0.74144 0.741441 -0.217863 -0.974998 0.974999 -0.326718 -1.224964 1.224965 -0.453699 -1.481356 1.481359 -0.588341 186 10 0 0 0 -0.04317 -9.53674e-007 -0.00168037 -0.142419 1.90735e-006 -0.0106165 -0.299767 -9.53674e-007 -0.0333657 -0.50846 -9.53674e-007 -0.0740533 -0.758955 0 -0.133732 -1.050001 1.90735e-006 -0.21607 -1.374152 9.53674e-007 -0.320441 -1.719674 0 -0.440959 -2.073003 9.53674e-007 -0.567795 187 10 0 0 0 -0.0470123 -4.76837e-007 -0.00199246 -0.156296 -4.76837e-007 -0.0128469 -0.33083 -4.76837e-007 -0.0409752 -0.563332 -9.53674e-007 -0.0919664 -0.843183 -9.53674e-007 -0.1676 -1.168325 4.76837e-007 -0.272684 -1.530195 4.76837e-007 -0.406703 -1.916434 -4.76837e-007 -0.562932 -2.313631 -4.76837e-007 -0.73006 188 10 0 0 0 -0.0444431 2.86102e-006 -0.00178027 -0.147006 3.8147e-006 -0.0113292 -0.31001 4.76837e-006 -0.0357804 -0.526524 3.8147e-006 -0.0797067 -0.786664 3.8147e-006 -0.144378 -1.088944 5.72205e-006 -0.233811 -1.425582 2.86102e-006 -0.347421 -1.784659 3.8147e-006 -0.47906 -2.152647 4.76837e-006 -0.61841 189 10 0 0 0 -0.037962 0 -0.00129771 -0.125049 0 -0.00816584 -0.263083 0 -0.0255868 -0.446359 9.53674e-007 -0.0566885 -0.666683 0 -0.102246 -0.923414 0 -0.165122 -1.210177 9.53674e-007 -0.244855 -1.516219 1.90735e-006 -0.33675 -1.828239 0 -0.432648 190 10 0 0 0 -0.0456057 0 -0.00187469 -0.150987 9.53674e-007 -0.0119603 -0.31857 9.53674e-007 -0.0378397 -0.541186 0 -0.0843918 -0.808618 9.53674e-007 -0.152995 -1.119162 1.90735e-006 -0.247894 -1.464744 9.53674e-007 -0.368461 -1.833262 9.53674e-007 -0.508282 -2.211174 0 -0.65662 191 10 0 0 0 -0.0463839 0 -0.0019393 -0.153963 0 -0.0124545 -0.3255 0 -0.0396004 -0.55378 0 -0.0886674 -0.828372 0 -0.161273 -1.147385 0 -0.261995 -1.502471 0 -0.390286 -1.881392 0 -0.539564 -2.270683 -9.53674e-007 -0.698771 192 10 0 0 0 -0.0323954 0.0323963 -0.00189185 -0.10712 0.107119 -0.0120347 -0.22578 0.225781 -0.0379856 -0.383235 0.383235 -0.0845628 -0.572202 0.572202 -0.153063 -0.791474 0.791474 -0.24766 -1.035316 1.035316 -0.367642 -1.295129 1.295126 -0.506498 -1.56128 1.56128 -0.653439 193 10 0 0 0 -0.0341463 0.0341458 -0.00210214 -0.113569 0.113571 -0.013576 -0.240406 0.240407 -0.04333 -0.409258 0.409259 -0.0972683 -0.612321 0.61232 -0.177246 -0.847921 0.847921 -0.288249 -1.10976 1.109762 -0.429655 -1.388978 1.388978 -0.594419 -1.676107 1.676107 -0.770823 194 10 0 0 0 -0.0315399 0.0315337 -0.0017941 -0.104162 0.104156 -0.0113695 -0.219389 0.219385 -0.0358059 -0.37225 0.372245 -0.0795906 -0.555717 0.555713 -0.143901 -0.768742 0.768737 -0.232678 -1.005808 1.005801 -0.34524 -1.25844 1.258437 -0.475363 -1.517012 1.517005 -0.612677 195 10 0 0 0 -0.0341587 0.0341582 -0.00210381 -0.113569 0.113568 -0.0135732 -0.240313 0.240312 -0.0432823 -0.408972 0.408973 -0.0970912 -0.611732 0.611732 -0.176813 -0.846931 0.846931 -0.287391 -1.108278 1.108279 -0.428174 -1.386911 1.386913 -0.592106 -1.67336 1.673361 -0.767482 196 10 0 0 0 -0.0355415 0.0355415 -0.00227833 -0.118106 0.118106 -0.0146873 -0.249687 0.249687 -0.0467778 -0.424432 0.424431 -0.104769 -0.634059 0.634058 -0.190478 -0.876573 0.876573 -0.308981 -1.145327 1.145327 -0.459339 -1.431259 1.43126 -0.633939 -1.724921 1.724924 -0.820489 197 10 0 0 0 -0.0342751 0.0342803 -0.00211811 -0.113565 0.113571 -0.0135524 -0.239639 0.239644 -0.0429393 -0.406929 0.406934 -0.0958304 -0.607581 0.607584 -0.173768 -0.839968 0.839972 -0.281401 -1.09783 1.097836 -0.417842 -1.372261 1.372266 -0.575884 -1.653666 1.65367 -0.7438 198 10 0 0 0 -0.0457516 1.90735e-006 -0.00188708 -0.151303 9.53674e-007 -0.0120049 -0.319005 9.53674e-007 -0.0379214 -0.541655 1.90735e-006 -0.0844872 -0.808992 1.90735e-006 -0.153039 -1.119328 1.90735e-006 -0.247796 -1.4645 1.90735e-006 -0.368038 -1.832254 1.90735e-006 -0.507183 -2.208904 1.90735e-006 -0.654353 199 10 0 0 0 -0.0451641 -9.53674e-007 -0.00183916 -0.149273 9.53674e-007 -0.0116799 -0.314598 0 -0.0368497 -0.534045 0 -0.0820284 -0.797527 0 -0.148484 -1.103471 9.53674e-007 -0.240314 -1.44389 9.53674e-007 -0.356829 -1.806639 0 -0.491605 -2.178043 9.53674e-007 -0.634 200 10 0 0 0 -0.0539513 -1.43051e-006 -0.00262594 -0.179681 0 -0.01704 -0.380262 4.76837e-007 -0.0545135 -0.646384 -4.76837e-007 -0.1224 -0.964897 -4.76837e-007 -0.222773 -1.331743 4.76837e-007 -0.361115 -1.736189 4.76837e-007 -0.535795 -2.164802 -4.76837e-007 -0.737892 -2.604495 4.76837e-007 -0.953871 201 10 0 0 0 -0.0464811 0 -0.00194836 -0.153753 1.90735e-006 -0.0124025 -0.324174 2.86102e-006 -0.0391874 -0.550349 2.86102e-006 -0.0873106 -0.821798 1.90735e-006 -0.158143 -1.136675 1.90735e-006 -0.255993 -1.486653 1.90735e-006 -0.38009 -1.859394 1.90735e-006 -0.52369 -2.241224 1.90735e-006 -0.675702 202 10 0 0 0 -0.0483274 4.76837e-007 -0.00210571 -0.160608 1.43051e-006 -0.0135698 -0.339871 9.53674e-007 -0.0432866 -0.57854 9.53674e-007 -0.0971644 -0.865498 -4.76837e-007 -0.177013 -1.198341 0 -0.287758 -1.567911 -4.76837e-007 -0.428513 -1.961323 -4.76837e-007 -0.591807 -2.36492 4.76837e-007 -0.765588 203 10 0 0 0 -0.0498304 -9.53674e-007 -0.00223947 -0.165546 -4.76837e-007 -0.0144238 -0.350057 -4.76837e-007 -0.0459582 -0.595284 4.76837e-007 -0.103011 -0.88961 -4.76837e-007 -0.187382 -1.230269 -9.53674e-007 -0.304084 -1.607767 -9.53674e-007 -0.452039 -2.009099 -9.53674e-007 -0.623441 -2.420705 -9.53674e-007 -0.80587 206 10 0 0 0 -0.0370255 0.037024 -0.00247264 -0.123621 0.12362 -0.0161386 -0.262289 0.262288 -0.0519019 -0.446902 0.446903 -0.117073 -0.66847 0.668469 -0.213871 -0.924289 0.924287 -0.347815 -1.206831 1.20683 -0.517389 -1.506531 1.506529 -0.713776 -1.813987 1.813985 -0.923525 208 10 0 0 0 -0.0375977 0.0375919 -0.00255013 -0.125592 0.12559 -0.0166676 -0.266515 0.266511 -0.0536443 -0.454041 0.45404 -0.12103 -0.678943 0.678939 -0.221084 -0.938309 0.938307 -0.359386 -1.22443 1.224428 -0.534272 -1.527687 1.527683 -0.736676 -1.83875 1.838746 -0.952938 209 10 0 0 0 -0.0367241 0.0367241 -0.00243258 -0.122716 0.122716 -0.0159068 -0.260597 0.260598 -0.0512512 -0.444377 0.444376 -0.115795 -0.665144 0.665144 -0.211823 -0.920245 0.920244 -0.344899 -1.202162 1.20216 -0.513558 -1.501306 1.501302 -0.709019 -1.808253 1.80825 -0.917864 211 10 0 0 0 -0.0479593 4.76837e-007 -0.00207353 -0.160137 4.76837e-007 -0.0135164 -0.340327 0 -0.0435492 -0.581419 -4.76837e-007 -0.098588 -0.872374 -4.76837e-007 -0.180871 -1.211521 -4.76837e-007 -0.296402 -1.589718 0 -0.445007 -1.993841 -9.53674e-007 -0.619343 -2.410322 -4.76837e-007 -0.80722 213 10 0 0 0 -0.0487566 -1.43051e-006 -0.00214386 -0.163146 -4.76837e-007 -0.0140471 -0.347279 -9.53674e-007 -0.0454491 -0.593927 -9.53674e-007 -0.103203 -0.892389 -1.90735e-006 -0.190185 -1.240131 -9.53674e-007 -0.312475 -1.627249 -2.38419e-006 -0.469557 -2.040348 -1.43051e-006 -0.653641 -2.465925 -1.43051e-006 -0.852088 216 10 0 0 0 -0.037775 0.0377808 -0.00257468 -0.12948 0.129482 -0.0179241 -0.280947 0.280949 -0.0608799 -0.48691 0.486914 -0.143361 -0.736991 0.736991 -0.270474 -1.025911 1.025913 -0.44961 -1.343289 1.343289 -0.677773 -1.678267 1.678267 -0.943118 -2.022068 2.022068 -1.229888 217 10 0 0 0 -0.03724 0.03724 -0.00250196 -0.126484 0.126484 -0.0170271 -0.272388 0.272388 -0.0567665 -0.469555 0.469555 -0.13181 -0.708361 0.708361 -0.246248 -0.984767 0.984767 -0.407035 -1.289586 1.289584 -0.612144 -1.612411 1.612411 -0.851138 -1.944151 1.944149 -1.109139 218 10 0 0 0 -0.0376587 0.037653 -0.00255799 -0.12886 0.128854 -0.0177364 -0.279211 0.279205 -0.0600367 -0.483414 0.483408 -0.140993 -0.73122 0.731213 -0.26547 -1.017583 1.017575 -0.44072 -1.332361 1.332354 -0.663903 -1.66479 1.664783 -0.923446 -2.006442 2.006434 -1.204791 219 10 0 0 0 -0.0379391 0.0379372 -0.00259638 -0.129483 0.129479 -0.0178914 -0.279905 0.279902 -0.0602328 -0.483714 0.48371 -0.14085 -0.730702 0.730699 -0.264393 -1.016043 1.016039 -0.438106 -1.329799 1.329795 -0.659362 -1.661303 1.661299 -0.916836 -2.001681 2.001676 -1.195031 220 10 0 0 0 -0.038146 0.0381422 -0.00262499 -0.130591 0.130588 -0.0182273 -0.283019 0.283015 -0.061754 -0.489985 0.489981 -0.145111 -0.741013 0.741007 -0.273332 -1.030824 1.03082 -0.453842 -1.349062 1.349056 -0.683681 -1.684905 1.684899 -0.95103 -2.029589 2.029583 -1.240076 221 10 0 0 0 -0.03755 0.03755 -0.00254369 -0.127666 0.127668 -0.0173593 -0.275113 0.275113 -0.0579765 -0.474387 0.474387 -0.134762 -0.71567 0.71567 -0.251901 -0.994706 0.994706 -0.416408 -1.302124 1.302122 -0.626086 -1.627464 1.627462 -0.870273 -1.961705 1.961703 -1.133926 222 10 0 0 0 -0.0533257 3.57628e-007 -0.00256491 -0.182653 1.19209e-007 -0.0178275 -0.395964 -2.38419e-007 -0.0604106 -0.685566 -2.38419e-007 -0.141864 -1.036707 -4.17233e-007 -0.266907 -1.442026 -7.15256e-007 -0.442517 -1.887341 -1.07288e-006 -0.665904 -2.360674 -1.49012e-006 -0.9306 -2.849688 -1.72853e-006 -1.222509 223 10 0 0 0 -0.0537605 2.38419e-007 -0.00260711 -0.18326 4.76837e-007 -0.017911 -0.395601 1.19209e-007 -0.0600922 -0.68281 0 -0.140091 -1.030748 -4.76837e-007 -0.262541 -1.433157 -4.76837e-007 -0.435069 -1.876675 -1.07288e-006 -0.656058 -2.346601 -1.43051e-006 -0.915286 -2.830462 -1.66893e-006 -1.197854 224 10 0 0 0 -0.0525913 0 -0.00249481 -0.178909 2.38419e-007 -0.0170462 -0.385571 -2.38419e-007 -0.0569153 -0.664618 -7.15256e-007 -0.132081 -1.002045 -9.53674e-007 -0.246291 -1.391771 -9.53674e-007 -0.405925 -1.823772 -1.43051e-006 -0.61217 -2.283915 -1.90735e-006 -0.85642 -2.759455 -1.90735e-006 -1.124741 225 10 0 0 0 -0.0516891 2.38419e-007 -0.00240922 -0.175411 0 -0.016361 -0.377558 0 -0.0544424 -0.650675 -4.76837e-007 -0.126212 -0.981524 -7.15256e-007 -0.23547 -1.364784 -1.19209e-006 -0.38884 -1.78793 -1.19209e-006 -0.584412 -2.236465 -1.19209e-006 -0.812161 -2.698441 -1.43051e-006 -1.059139 226 10 0 0 0 -0.0532227 0 -0.00255537 -0.18075 -2.38419e-007 -0.01739 -0.388771 -2.38419e-007 -0.0578048 -0.670261 0 -0.134398 -1.011711 -2.38419e-007 -0.251714 -1.407501 -4.76837e-007 -0.417408 -1.844805 -1.19209e-006 -0.630316 -2.309084 -1.19209e-006 -0.88077 -2.787701 -1.66893e-006 -1.154251 227 10 0 0 0 -0.0534382 1.19209e-007 -0.00257635 -0.183445 1.78814e-007 -0.0180018 -0.398426 1.19209e-007 -0.0612825 -0.690784 -1.78814e-007 -0.144431 -1.045547 -4.76837e-007 -0.272452 -1.45495 -5.06639e-007 -0.452465 -1.904594 -9.12696e-007 -0.681736 -2.382164 -1.43796e-006 -0.953442 -2.875345 -1.59442e-006 -1.253234 228 10 0 0 0 -0.0392408 0.0392418 -0.00277877 -0.136838 0.136841 -0.0201981 -0.300888 0.300892 -0.0709592 -0.525938 0.52594 -0.17116 -0.799546 0.799545 -0.328067 -1.113411 1.113413 -0.549524 -1.454728 1.454729 -0.830043 -1.81209 1.812088 -1.154889 -2.177565 2.177563 -1.506729 229 10 0 0 0 -0.0383959 0.0383959 -0.00265908 -0.131704 0.131704 -0.0185585 -0.285441 0.285441 -0.0628682 -0.495119 0.495117 -0.148624 -0.750443 0.750441 -0.281968 -1.046 1.045999 -0.471555 -1.371181 1.37118 -0.71541 -1.715006 1.715002 -1.002246 -2.06881 2.068808 -1.316367 230 10 0 0 0 -0.038662 0.0386658 -0.00269723 -0.135091 0.13509 -0.0196941 -0.297735 0.297737 -0.0695474 -0.521674 0.521671 -0.168652 -0.794818 0.794819 -0.324927 -1.108935 1.108936 -0.546863 -1.45102 1.451019 -0.829293 -1.809417 1.809416 -1.157374 -2.176081 2.176081 -1.513533 231 10 0 0 0 -0.0391397 0.0391397 -0.00276375 -0.136353 0.136353 -0.0200434 -0.299617 0.299616 -0.0702952 -0.523536 0.523531 -0.169383 -0.795867 0.795866 -0.324548 -1.108557 1.108553 -0.54381 -1.448948 1.448946 -0.82204 -1.805621 1.805618 -1.144764 -2.170564 2.170559 -1.494712 232 10 0 0 0 -0.0389805 0.0389805 -0.00274229 -0.13629 0.13629 -0.020056 -0.300439 0.300442 -0.0708833 -0.526278 0.526279 -0.171866 -0.801333 0.801332 -0.33077 -1.116993 1.116993 -0.555632 -1.460107 1.460104 -0.840652 -1.819099 1.819098 -1.17062 -2.186079 2.186075 -1.527942 233 10 0 0 0 -0.0384216 0.0384197 -0.00266385 -0.134347 0.134346 -0.0194817 -0.296369 0.296368 -0.0689347 -0.519773 0.519772 -0.167519 -0.792635 0.792635 -0.323405 -1.106756 1.106754 -0.545348 -1.44907 1.449068 -0.828349 -1.80782 1.807817 -1.157548 -2.174912 2.17491 -1.515289 234 10 0 0 0 -0.058794 0 -0.00311971 -0.205309 -3.57628e-007 -0.0227876 -0.451204 -3.57628e-007 -0.080169 -0.786518 -8.34465e-007 -0.192774 -1.186996 -1.07288e-006 -0.363285 -1.638317 -1.43051e-006 -0.594877 -2.123835 -1.78814e-006 -0.880408 -2.629171 -2.14577e-006 -1.205146 -3.144215 -2.26498e-006 -1.552441 235 10 0 0 0 -0.0588436 1.78814e-007 -0.00312495 -0.203617 -1.19209e-007 -0.0223198 -0.443508 -4.47035e-007 -0.076781 -0.767143 -6.55651e-007 -0.180782 -1.154585 -9.83477e-007 -0.338177 -1.594463 -1.19209e-006 -0.554396 -2.070539 -1.40071e-006 -0.823605 -2.568106 -1.96695e-006 -1.132034 -3.076349 -2.11596e-006 -1.463245 236 10 0 0 0 -0.0570068 0 -0.00293231 -0.199596 0 -0.0215425 -0.440006 -2.38419e-007 -0.0762541 -0.769485 -9.53674e-007 -0.1845 -1.16802 -9.53674e-007 -0.352999 -1.621101 -1.43051e-006 -0.58706 -2.109772 -1.66893e-006 -0.878335 -2.618731 -2.14577e-006 -1.211175 -3.137714 -2.38419e-006 -1.568496 237 10 0 0 0 -0.0557384 0 -0.00280356 -0.19611 0 -0.0208294 -0.434507 -4.76837e-007 -0.0745791 -0.763428 -7.15256e-007 -0.182414 -1.16353 -9.53674e-007 -0.352534 -1.620056 -9.53674e-007 -0.591502 -2.112635 -1.43051e-006 -0.890137 -2.625256 -1.90735e-006 -1.231538 -3.147715 -2.38419e-006 -1.598202 238 10 0 0 0 -0.0576172 2.38419e-007 -0.00299549 -0.200701 1.19209e-007 -0.0217371 -0.440252 1.19209e-007 -0.0760373 -0.766657 -3.57628e-007 -0.182035 -1.16008 -7.15256e-007 -0.345341 -1.607468 -9.53674e-007 -0.571525 -2.091063 -1.66893e-006 -0.853625 -2.595699 -1.90735e-006 -1.176834 -3.110802 -2.02656e-006 -1.524276 239 10 0 0 0 -0.0561428 2.38419e-007 -0.00284362 -0.192993 1.19209e-007 -0.0199621 -0.418934 -5.96046e-008 -0.0679815 -0.725 -2.08616e-007 -0.159893 -1.094793 -5.66244e-007 -0.300846 -1.519455 -8.64267e-007 -0.498205 -1.983614 -1.2219e-006 -0.748474 -2.472076 -1.57952e-006 -1.039368 -2.97307 -1.84774e-006 -1.354829 240 10 0 0 0 -0.0286293 0.0286317 -0.001477 -0.0943146 0.0943184 -0.00929832 -0.198353 0.198354 -0.0291357 -0.336315 0.336317 -0.064523 -0.501942 0.501943 -0.116316 -0.694588 0.69459 -0.187705 -0.909405 0.909407 -0.278128 -1.138462 1.138463 -0.38234 -1.372341 1.372341 -0.49153 241 10 0 0 0 -0.0285378 0.0285378 -0.00146842 -0.0939865 0.0939889 -0.00923276 -0.197626 0.197626 -0.0289154 -0.335035 0.335036 -0.0640106 -0.499975 0.499976 -0.115353 -0.691818 0.691817 -0.186102 -0.905731 0.905733 -0.275693 -1.133802 1.133804 -0.378912 -1.366623 1.36663 -0.486998 242 10 0 0 0 -0.0261459 0.0261474 -0.00123191 -0.0858574 0.0858626 -0.00768614 -0.18017 0.180174 -0.0239353 -0.30505 0.305053 -0.0527527 -0.454826 0.454828 -0.0947161 -0.629105 0.629107 -0.152381 -0.82358 0.823581 -0.225233 -1.030827 1.030829 -0.308837 -1.241495 1.241498 -0.395481 243 10 0 0 0 -0.0278015 0.0278053 -0.00139308 -0.0914536 0.0914593 -0.00873446 -0.192142 0.192146 -0.0272918 -0.325577 0.325582 -0.0603225 -0.485708 0.485712 -0.108572 -0.67202 0.672025 -0.175029 -0.879866 0.879872 -0.25916 -1.101471 1.101478 -0.355995 -1.327505 1.327512 -0.457173 244 10 0 0 0 -0.0308151 0.0308132 -0.00171232 -0.101683 0.101689 -0.0108283 -0.214081 0.214084 -0.0340531 -0.363178 0.363181 -0.0756214 -0.542152 0.542155 -0.13663 -0.750088 0.75009 -0.220837 -0.981634 0.981639 -0.327604 -1.228445 1.228447 -0.45094 -1.48089 1.48089 -0.580839 245 10 0 0 0 -0.0268612 0.0268598 -0.00130033 -0.0882607 0.0882611 -0.00812721 -0.185297 0.185296 -0.0253448 -0.313822 0.313822 -0.0559192 -0.468 0.468003 -0.100496 -0.647398 0.647398 -0.161808 -0.847551 0.84755 -0.239327 -1.060884 1.060885 -0.328387 -1.27812 1.278119 -0.421059 246 10 0 0 0 -0.0351896 -2.86102e-006 -0.00111651 -0.115413 -3.8147e-006 -0.00693917 -0.242019 -3.8147e-006 -0.0215569 -0.409645 -1.90735e-006 -0.0474484 -0.610748 -2.86102e-006 -0.0851245 -0.845028 -3.8147e-006 -0.136939 -1.106836 -4.76837e-006 -0.202496 -1.38607 -2.86102e-006 -0.277771 -1.66911 -4.76837e-006 -0.355276 247 10 0 0 0 -0.0369883 2.86102e-006 -0.00123239 -0.121872 1.90735e-006 -0.00775576 -0.256638 1.90735e-006 -0.0243495 -0.436036 3.8147e-006 -0.0541124 -0.652358 9.53674e-007 -0.0979578 -0.906991 1.90735e-006 -0.159748 -1.194541 9.53674e-007 -0.239954 -1.50471 2.86102e-006 -0.334599 -1.823756 2.86102e-006 -0.435348 248 10 0 0 0 -0.0357132 0 -0.0011487 -0.117169 -9.53674e-007 -0.0071528 -0.245758 -9.53674e-007 -0.0222387 -0.416029 -1.90735e-006 -0.0489757 -0.620309 -9.53674e-007 -0.0878959 -0.85825 0 -0.141429 -1.124084 -1.90735e-006 -0.209159 -1.407631 -9.53674e-007 -0.286967 -1.695355 -2.86102e-006 -0.36728 249 10 0 0 0 -0.0370827 9.53674e-007 -0.00123858 -0.122199 9.53674e-007 -0.00779724 -0.25736 -1.90735e-006 -0.0244896 -0.43731 0 -0.054441 -0.654361 -1.90735e-006 -0.0985963 -0.909979 -9.53674e-007 -0.160896 -1.198616 -1.90735e-006 -0.241763 -1.509949 0 -0.337193 -1.830229 -1.90735e-006 -0.438806 250 10 0 0 0 -0.0380192 -1.90735e-006 -0.00130153 -0.124958 -1.90735e-006 -0.00814652 -0.262416 -2.86102e-006 -0.02542 -0.444584 -9.53674e-007 -0.0561359 -0.663246 -1.90735e-006 -0.100977 -0.91785 0 -0.162752 -1.20213 -2.86102e-006 -0.240996 -1.505372 -2.86102e-006 -0.331057 -1.81441 -9.53674e-007 -0.424943 251 10 0 0 0 -0.0407143 -2.86102e-006 -0.00149393 -0.13413 -3.8147e-006 -0.00940418 -0.282146 -2.86102e-006 -0.0294845 -0.478543 -9.53674e-007 -0.0653579 -0.714463 -1.90735e-006 -0.117939 -0.989064 -2.86102e-006 -0.190549 -1.29548 -3.8147e-006 -0.282692 -1.622399 -3.8147e-006 -0.389065 -1.956442 -4.76837e-006 -0.500709 252 10 0 0 0 -0.0301399 0.0301447 -0.0016377 -0.099762 0.0997677 -0.0104327 -0.210684 0.21069 -0.0330381 -0.358486 0.358491 -0.0738599 -0.536625 0.536628 -0.134262 -0.74437 0.744376 -0.218305 -0.976469 0.976474 -0.325638 -1.225504 1.225512 -0.451537 -1.48241 1.482419 -0.586807 253 10 0 0 0 -0.029686 0.0296836 -0.00158906 -0.0982552 0.0982528 -0.0101173 -0.207544 0.207538 -0.0320489 -0.353258 0.353253 -0.0716813 -0.529003 0.528997 -0.130377 -0.734152 0.734148 -0.212149 -0.964544 0.96454 -0.317725 -1.212257 1.212254 -0.442098 -1.467834 1.467829 -0.575749 254 10 0 0 0 -0.0289268 0.0289249 -0.00150824 -0.0955391 0.0955386 -0.00955272 -0.201454 0.201457 -0.0301263 -0.342442 0.342441 -0.0671365 -0.512274 0.512273 -0.121738 -0.710434 0.710434 -0.197598 -0.932022 0.932024 -0.294418 -1.169525 1.169526 -0.407449 -1.414532 1.414534 -0.528744 255 10 0 0 0 -0.0287476 0.0287452 -0.00148964 -0.0948925 0.094892 -0.00942206 -0.199989 0.199986 -0.0296724 -0.339794 0.339793 -0.0660453 -0.508127 0.508124 -0.119635 -0.704481 0.704479 -0.194012 -0.924004 0.924001 -0.288851 -1.158674 1.158673 -0.398872 -1.400387 1.400384 -0.516487 256 10 0 0 0 -0.0295067 0.0295048 -0.00156951 -0.097374 0.0973721 -0.00992298 -0.205116 0.205114 -0.0312274 -0.34828 0.348277 -0.0694327 -0.520441 0.520438 -0.125625 -0.720942 0.720937 -0.203433 -0.944735 0.94473 -0.302398 -1.183655 1.183652 -0.416959 -1.428162 1.428157 -0.537688 257 10 0 0 0 -0.0318546 0.0318532 -0.00182867 -0.105663 0.105663 -0.0117226 -0.223448 0.223448 -0.0372822 -0.380447 0.380451 -0.0835817 -0.569599 0.569598 -0.15222 -0.789789 0.789789 -0.247694 -1.035242 1.03524 -0.369489 -1.29723 1.297228 -0.51108 -1.567247 1.567246 -0.663122 258 10 0 0 0 -0.0410233 9.53674e-007 -0.00151658 -0.136545 1.90735e-006 -0.00979018 -0.290066 9.53674e-007 -0.0314233 -0.496422 1.90735e-006 -0.0711696 -0.747204 1.90735e-006 -0.130993 -1.041937 1.90735e-006 -0.215619 -1.373516 1.90735e-006 -0.32542 -1.730177 1.90735e-006 -0.455024 -2.097637 1.90735e-006 -0.593907 259 10 0 0 0 -0.0439148 0 -0.00173926 -0.146566 0 -0.0113051 -0.311853 0 -0.0364628 -0.534039 0 -0.0828259 -0.80374 0 -0.152683 -1.119722 0 -0.251296 -1.473818 0 -0.378764 -1.853609 0 -0.528858 -2.245506 0 -0.690637 260 10 0 0 0 -0.041028 0 -0.00151682 -0.135903 1.90735e-006 -0.00967693 -0.287352 9.53674e-007 -0.0307193 -0.4904 1.90735e-006 -0.069154 -0.736843 9.53674e-007 -0.126806 -1.0263 9.53674e-007 -0.20817 -1.351873 1.90735e-006 -0.313566 -1.702063 1.90735e-006 -0.437834 -2.063419 0 -0.571413 261 10 0 0 0 -0.0404968 0 -0.00147867 -0.13463 0 -0.00951123 -0.285728 9.53674e-007 -0.0304532 -0.488674 0 -0.0688477 -0.735209 0 -0.126545 -1.024986 1.90735e-006 -0.208106 -1.351112 9.53674e-007 -0.313902 -1.702085 9.53674e-007 -0.438807 -2.06385 9.53674e-007 -0.572738 262 10 0 0 0 -0.0379286 9.53674e-007 -0.00129628 -0.125266 1.90735e-006 -0.00820398 -0.264684 0 -0.0259826 -0.451661 9.53674e-007 -0.0583925 -0.678723 0 -0.106913 -0.946038 1.90735e-006 -0.175452 -1.247643 -9.53674e-007 -0.264451 -1.57278 -9.53674e-007 -0.369531 -1.907479 1.90735e-006 -0.481669 263 10 0 0 0 -0.0398836 -1.90735e-006 -0.00143337 -0.131878 9.53674e-007 -0.00910139 -0.278457 9.53674e-007 -0.0287869 -0.474014 9.53674e-007 -0.0643427 -0.711227 0 -0.117533 -0.990097 9.53674e-007 -0.192596 -1.304176 0 -0.289897 -1.642338 0 -0.404669 -1.991144 1.90735e-006 -0.527817 266 10 0 0 0 -0.031498 0.0314989 -0.00178909 -0.105354 0.105355 -0.0116942 -0.224541 0.224543 -0.0378826 -0.384979 0.384982 -0.0863271 -0.579859 0.579859 -0.159507 -0.808113 0.808115 -0.26291 -1.063705 1.063706 -0.396582 -1.337708 1.337707 -0.554028 -1.620234 1.620234 -0.723577 268 10 0 0 0 -0.0333309 0.0333343 -0.00200343 -0.111303 0.111307 -0.0130563 -0.236622 0.236624 -0.0420839 -0.4049 0.404899 -0.0956337 -0.608793 0.608796 -0.176326 -0.8468 0.846801 -0.289899 -1.112368 1.11237 -0.43608 -1.396263 1.396264 -0.607643 -1.688883 1.688881 -0.792534 269 10 0 0 0 -0.032938 0.032938 -0.00195599 -0.110249 0.11025 -0.0128207 -0.234989 0.234989 -0.0415723 -0.402681 0.402683 -0.0947332 -0.605993 0.605992 -0.174925 -0.843433 0.843433 -0.287889 -1.108465 1.108465 -0.433372 -1.391853 1.391853 -0.604188 -1.684008 1.684007 -0.788349 270 10 0 0 0 -0.0469036 0 -0.00198293 -0.157135 9.53674e-007 -0.0130279 -0.334742 9.53674e-007 -0.0421821 -0.57262 0 -0.0956864 -0.859501 -4.76837e-007 -0.175489 -1.192701 0 -0.2865 -1.564575 -1.43051e-006 -0.429319 -1.962339 -1.90735e-006 -0.597026 -2.372482 -1.43051e-006 -0.777846 276 10 0 0 0 -0.0371418 0.0371418 -0.00248837 -0.125761 0.12576 -0.0168097 -0.27029 0.270292 -0.0557761 -0.465254 0.465254 -0.129023 -0.701324 0.701324 -0.240525 -0.974966 0.974965 -0.397474 -1.277599 1.277598 -0.598815 -1.599186 1.599182 -0.835303 -1.930685 1.930681 -1.092768 277 10 0 0 0 -0.0380402 0.0380402 -0.00261045 -0.129343 0.129346 -0.0178244 -0.278524 0.278528 -0.0594425 -0.47975 0.479751 -0.137861 -0.723083 0.723085 -0.257274 -1.004333 1.004335 -0.424964 -1.314358 1.314358 -0.639328 -1.64292 1.64292 -0.890358 -1.981144 1.981143 -1.163333 278 10 0 0 0 -0.0368214 0.0368185 -0.00244546 -0.124712 0.124711 -0.0165288 -0.267706 0.267703 -0.0546401 -0.459925 0.45992 -0.125689 -0.692724 0.692719 -0.233749 -0.963322 0.963318 -0.38657 -1.263351 1.263346 -0.58351 -1.582821 1.582814 -0.815764 -1.912628 1.912621 -1.069495 279 10 0 0 0 -0.0365372 0.0365343 -0.00240707 -0.123421 0.12342 -0.0161653 -0.264278 0.264277 -0.053107 -0.454041 0.454039 -0.122223 -0.684031 0.684025 -0.227387 -0.951183 0.951179 -0.375633 -1.247379 1.247375 -0.566251 -1.562756 1.562748 -0.790569 -1.888149 1.888144 -1.034874 280 10 0 0 0 -0.0359268 0.0359268 -0.00232744 -0.121506 0.121506 -0.0156705 -0.260763 0.260763 -0.051748 -0.448395 0.448393 -0.119215 -0.675375 0.675375 -0.221339 -0.938438 0.938438 -0.364296 -1.229191 1.229187 -0.546262 -1.539427 1.539427 -0.761016 -1.860252 1.86025 -0.995968 281 10 0 0 0 -0.0336857 0.0336876 -0.00204587 -0.113071 0.11307 -0.0135078 -0.241247 0.241249 -0.0439128 -0.413176 0.413176 -0.0999441 -0.620707 0.620707 -0.183798 -0.861507 0.861509 -0.300409 -1.129593 1.129592 -0.44988 -1.416334 1.416335 -0.625701 -1.71224 1.712238 -0.815852 282 10 0 0 0 -0.0487862 4.76837e-007 -0.00214672 -0.163794 4.76837e-007 -0.0141804 -0.349527 0 -0.0461473 -0.598827 4.76837e-007 -0.105223 -0.900276 0 -0.19412 -1.250927 4.76837e-007 -0.318763 -1.6409 -9.53674e-007 -0.478639 -2.056995 -1.43051e-006 -0.666082 -2.485834 -9.53674e-007 -0.868454 283 10 0 0 0 -0.0538425 -9.53674e-007 -0.00261593 -0.180725 0 -0.0172987 -0.384802 9.53674e-007 -0.0561386 -0.656944 0 -0.127359 -0.983608 -9.53674e-007 -0.233546 -1.360201 -9.53674e-007 -0.380664 -1.775362 -1.43051e-006 -0.567053 -2.215349 -9.53674e-007 -0.783414 -2.667191 -9.53674e-007 -1.01573 284 10 0 0 0 -0.048296 1.43051e-006 -0.00210238 -0.161911 9.53674e-007 -0.0138447 -0.344858 4.76837e-007 -0.0448314 -0.590117 -4.76837e-007 -0.101899 -0.88699 -1.43051e-006 -0.187873 -1.232784 -4.76837e-007 -0.308607 -1.617861 -9.53674e-007 -0.463717 -2.02914 -1.43051e-006 -0.645787 -2.453207 -1.90735e-006 -0.842443 285 10 0 0 0 -0.0564804 0 -0.00287819 -0.190725 0 -0.0193415 -0.407558 -4.76837e-007 -0.063404 -0.696844 -4.76837e-007 -0.144665 -1.043467 -1.19209e-006 -0.266059 -1.44125 -9.53674e-007 -0.433787 -1.877483 -1.90735e-006 -0.645333 -2.338037 -2.38419e-006 -0.890192 -2.810501 -2.38419e-006 -1.153481 286 10 0 0 0 -0.0546455 4.76837e-007 -0.00269341 -0.185526 -2.38419e-007 -0.0183303 -0.398754 0 -0.0608785 -0.685488 -7.15256e-007 -0.140591 -1.031244 -4.76837e-007 -0.261298 -1.429914 -9.53674e-007 -0.429938 -1.868561 -1.19209e-006 -0.644565 -2.33279 -1.43051e-006 -0.894944 -2.810123 -1.66893e-006 -1.166255 287 10 0 0 0 -0.0564117 -2.38419e-007 -0.00287199 -0.190708 -7.15256e-007 -0.0193486 -0.407921 -7.15256e-007 -0.0635707 -0.697975 -1.19209e-006 -0.145301 -1.045706 -1.43051e-006 -0.267583 -1.44483 -1.66893e-006 -0.436692 -1.882528 -1.90735e-006 -0.650103 -2.344624 -2.38419e-006 -0.897259 -2.818739 -2.86102e-006 -1.163234 288 10 0 0 0 -0.0424767 0.0424776 -0.00325704 -0.145674 0.145675 -0.022769 -0.313877 0.313878 -0.0762706 -0.537895 0.537893 -0.175455 -0.804642 0.80464 -0.323169 -1.107594 1.107592 -0.525055 -1.436272 1.43627 -0.77633 -1.780512 1.78051 -1.064233 -2.132385 2.13238 -1.372776 289 10 0 0 0 -0.0406876 0.0406914 -0.00298905 -0.139301 0.139303 -0.0207775 -0.300928 0.30093 -0.0699787 -0.518267 0.51827 -0.162748 -0.77927 0.779273 -0.3031 -1.077814 1.077815 -0.497562 -1.403435 1.403435 -0.742335 -1.745716 1.74572 -1.025264 -2.09642 2.096416 -1.330496 290 10 0 0 0 -0.0423527 0.0423536 -0.00323844 -0.145649 0.145649 -0.0227885 -0.314793 0.314793 -0.0769219 -0.540567 0.540567 -0.177844 -0.809525 0.809521 -0.328468 -1.114794 1.114794 -0.534365 -1.44569 1.445686 -0.790489 -1.792019 1.792015 -1.083833 -2.145939 2.145935 -1.398296 291 10 0 0 0 -0.0400047 0.0400047 -0.00288725 -0.134069 0.134074 -0.0190501 -0.284887 0.284887 -0.0616226 -0.485218 0.485218 -0.13929 -0.72471 0.724707 -0.254471 -0.999479 0.999475 -0.412979 -1.300974 1.300974 -0.612388 -1.619364 1.61936 -0.842519 -1.945688 1.945684 -1.088777 292 10 0 0 0 -0.040163 0.0401592 -0.00291061 -0.134903 0.134899 -0.019309 -0.287069 0.287066 -0.0626793 -0.489359 0.48936 -0.142001 -0.731249 0.731246 -0.259811 -1.008669 1.008664 -0.422021 -1.312902 1.312901 -0.626122 -1.634077 1.634077 -0.861775 -1.963312 1.963308 -1.114258 293 10 0 0 0 -0.0429611 0.0429583 -0.00333095 -0.147423 0.147419 -0.0233335 -0.316473 0.316469 -0.0774052 -0.540745 0.540741 -0.17684 -0.807341 0.807337 -0.324358 -1.109863 1.109856 -0.525509 -1.437912 1.437908 -0.775436 -1.781378 1.781374 -1.061357 -2.132332 2.132324 -1.367296 294 10 0 0 0 -0.0572224 2.38419e-007 -0.00295472 -0.191803 2.38419e-007 -0.0195017 -0.40753 -2.38419e-007 -0.0630963 -0.693939 -4.76837e-007 -0.142614 -1.036093 -9.53674e-007 -0.260479 -1.428225 -1.19209e-006 -0.422494 -1.858023 -1.43051e-006 -0.626033 -2.311518 -1.66893e-006 -0.860675 -2.776157 -2.14577e-006 -1.11168 295 10 0 0 0 -0.0581493 -9.53674e-007 -0.00305247 -0.195042 -1.19209e-006 -0.0201819 -0.414595 -9.53674e-007 -0.0654056 -0.706015 -1.43051e-006 -0.147975 -1.05382 -1.66893e-006 -0.270316 -1.451702 -1.90735e-006 -0.438144 -1.886885 -2.6226e-006 -0.648363 -2.345272 -3.09944e-006 -0.890037 -2.814489 -2.86102e-006 -1.14814 296 10 0 0 0 -0.0593491 -2.38419e-007 -0.00318003 -0.205036 -3.57628e-007 -0.0226219 -0.445578 -5.96046e-007 -0.077396 -0.768997 -7.7486e-007 -0.181241 -1.155415 -1.13249e-006 -0.337644 -1.593622 -1.37091e-006 -0.551708 -2.067352 -1.60933e-006 -0.817054 -2.561909 -1.84774e-006 -1.119519 -3.066471 -2.14577e-006 -1.442569 297 10 0 0 0 -0.0595551 -1.78814e-007 -0.00320101 -0.205647 -2.98023e-007 -0.0227537 -0.446575 -2.98023e-007 -0.0777146 -0.770109 -6.25849e-007 -0.181643 -1.156342 -7.7486e-007 -0.337865 -1.594185 -1.10269e-006 -0.551464 -2.067501 -1.46031e-006 -0.816133 -2.561647 -1.78814e-006 -1.117798 -3.065804 -1.96695e-006 -1.439968 298 10 0 0 0 -0.0593452 -5.96046e-008 -0.00317788 -0.20458 -4.17233e-007 -0.0224977 -0.443777 -5.96046e-007 -0.0766281 -0.76503 -7.45058e-007 -0.17893 -1.148855 -1.01328e-006 -0.332838 -1.584448 -1.2219e-006 -0.543571 -2.055771 -1.60933e-006 -0.805024 -2.548138 -1.81794e-006 -1.103253 -3.050598 -1.99676e-006 -1.421798 299 10 0 0 0 -0.0583115 2.38419e-007 -0.00306773 -0.195456 4.76837e-007 -0.0202618 -0.415276 0 -0.0656008 -0.706934 -4.76837e-007 -0.148316 -1.054945 -9.53674e-007 -0.270825 -1.453028 -1.19209e-006 -0.438859 -1.888433 -1.43051e-006 -0.649359 -2.347088 -1.43051e-006 -0.891422 -2.816628 -1.66893e-006 -1.150037' )
		self.hairShape.boundingBox()
		self.editGuidesShape.boundingBox()
		pm.mel.OxEditGuides( '-eg 300 24 10 0 0 0 -0.0385895 0.468977 -0.260241 -0.191791 0.652061 -0.314146 -0.443386 0.903656 -0.443038 -0.76138 1.221649 -0.672398 -1.117203 1.577474 -0.992536 -1.493971 1.954241 -1.39082 -1.879931 2.340201 -1.842884 -2.269475 2.729746 -2.326652 -2.660489 3.12076 -2.828763 26 10 0 0 0 -0.0427442 0.516917 -0.356519 -0.190597 0.731805 -0.421561 -0.443301 0.984509 -0.551765 -0.761677 1.302884 -0.781872 -1.117353 1.658562 -1.101568 -1.493809 2.035018 -1.498372 -1.879501 2.420711 -1.948442 -2.268929 2.810138 -2.430963 -2.65986 3.201069 -2.931851 27 10 0 0 0 -0.039464 0.488774 -0.294433 -0.19318 0.671735 -0.348451 -0.445081 0.923633 -0.477703 -0.763185 1.241741 -0.707286 -1.118993 1.597548 -1.027376 -1.495704 1.974257 -1.42539 -1.881614 2.36017 -1.877079 -2.271119 2.749675 -2.360445 -2.662106 3.140663 -2.862126 30 10 0 0 0 -0.356694 0.295841 -0.249141 -0.603153 0.316172 -0.307217 -0.967346 0.316172 -0.443242 -1.421511 0.316172 -0.678835 -1.925731 0.316173 -1.001145 -2.458765 0.316172 -1.400118 -3.004972 0.316173 -1.854186 -3.556365 0.316174 -2.34187 -4.109693 0.316173 -2.847728 31 10 0 0 0 -0.223061 0.141312 -0.0667648 -0.465717 0.141312 -0.12256 -0.836897 0.141312 -0.264753 -1.293773 0.141313 -0.504229 -1.797915 0.141313 -0.826368 -2.32947 0.141313 -1.220395 -2.874077 0.141313 -1.666203 -3.424337 0.141313 -2.145239 -3.976833 0.141313 -2.642565 32 10 0 0 0 -0.398709 0.331589 -0.356249 -0.662521 0.373535 -0.424685 -1.02926 0.373535 -0.562934 -1.483934 0.373535 -0.799252 -1.988716 0.373536 -1.122776 -2.522169 0.373535 -1.523187 -3.068443 0.373536 -1.977619 -3.61976 0.373535 -2.464674 -4.172971 0.373536 -2.969233 33 10 0 0 0 -0.304775 0.236619 -0.155848 -0.545523 0.236619 -0.210721 -0.916723 0.236619 -0.352934 -1.376846 0.236619 -0.597155 -1.886118 0.236619 -0.930707 -2.422031 0.236619 -1.33984 -2.969481 0.236619 -1.800841 -3.521357 0.236619 -2.292553 -4.074852 0.236619 -2.800317 34 10 0 0 0 -0.196573 0.120239 -0.0500429 -0.424686 0.120241 -0.099035 -0.784677 0.120241 -0.231448 -1.237638 0.120242 -0.465339 -1.742733 0.120242 -0.789546 -2.276381 0.120242 -1.190634 -2.822507 0.120242 -1.64427 -3.373752 0.120243 -2.130755 -3.927011 0.120243 -2.635849 84 10 0 0 0 -0.0479856 0.392244 -0.165063 -0.21853 0.562792 -0.220148 -0.481573 0.825834 -0.363078 -0.807574 1.151834 -0.60865 -1.168054 1.512313 -0.943412 -1.547166 1.891427 -1.353392 -1.934328 2.278589 -1.814865 -2.324576 2.668838 -2.306763 -2.715953 3.060215 -2.81446 87 10 0 0 0 -0.0493283 0.368723 -0.14294 -0.220108 0.539506 -0.198185 -0.483453 0.802849 -0.341501 -0.809722 1.129118 -0.587641 -1.170373 1.489769 -0.922955 -1.549568 1.868964 -1.333382 -1.936771 2.256167 -1.795176 -2.327037 2.646435 -2.287293 -2.718424 3.037819 -2.795132 88 10 0 0 0 -0.0685053 0.166689 -0.0300422 -0.24343 0.341613 -0.0881591 -0.50818 0.606365 -0.233281 -0.832491 0.930675 -0.475329 -1.189745 1.28793 -0.799842 -1.565944 1.664129 -1.195423 -1.951055 2.04924 -1.641326 -2.339895 2.438081 -2.11785 -2.73019 2.828374 -2.610301 89 10 0 0 0 -0.0493603 0.388265 -0.161278 -0.222472 0.561378 -0.218129 -0.489452 0.828357 -0.366148 -0.818508 1.157412 -0.618247 -1.18071 1.519616 -0.95872 -1.56056 1.899465 -1.37261 -1.947978 2.286882 -1.836195 -2.338287 2.677192 -2.328829 -2.729608 3.068512 -2.835623 90 10 0 0 0 -0.391104 0.316989 -0.320612 -0.658826 0.337664 -0.389815 -1.044728 0.337664 -0.545718 -1.51422 0.337665 -0.804258 -2.0274 0.337665 -1.147003 -2.564048 0.337665 -1.558853 -3.110943 0.337664 -2.016704 -3.661901 0.337664 -2.500931 -4.214404 0.337664 -2.998338 91 10 0 0 0 -0.261075 0.174614 -0.0973065 -0.524053 0.174614 -0.16349 -0.920226 0.174614 -0.329573 -1.39796 0.174614 -0.601555 -1.916325 0.174614 -0.957262 -2.455075 0.174614 -1.377202 -3.002709 0.174614 -1.839272 -3.55396 0.174613 -2.325793 -4.106583 0.174613 -2.824341 93 10 0 0 0 -0.370167 0.296163 -0.265889 -0.640971 0.317463 -0.336828 -1.031202 0.317463 -0.496957 -1.504812 0.317463 -0.762106 -2.020988 0.317463 -1.112222 -2.559382 0.317463 -1.530764 -3.106981 0.317463 -1.99263 -3.6583 0.317462 -2.479707 -4.211031 0.317462 -2.979308 94 10 0 0 0 -0.36933 0.294345 -0.262984 -0.627712 0.294345 -0.326725 -1.01563 0.294345 -0.484585 -1.484816 0.294345 -0.742641 -1.996433 0.294345 -1.081656 -2.531314 0.294345 -1.487061 -3.076841 0.294345 -1.937536 -3.626831 0.294345 -2.414647 -4.17855 0.294345 -2.90503 95 10 0 0 0 -0.395892 0.318018 -0.330219 -0.666509 0.333896 -0.400845 -1.056964 0.333896 -0.561195 -1.528049 0.333896 -0.822267 -2.04061 0.333896 -1.163527 -2.575694 0.333896 -1.569658 -3.121176 0.333896 -2.01989 -3.67106 0.333895 -2.496264 -4.222657 0.333895 -2.985608 145 10 0 0 0 -0.0685272 0.266685 -0.0730354 -0.260698 0.458855 -0.144046 -0.545346 0.743503 -0.31673 -0.883071 1.081229 -0.588516 -1.247099 1.445255 -0.935239 -1.62577 1.823924 -1.342938 -2.011359 2.209514 -1.792268 -2.399931 2.598085 -2.266175 -2.789689 2.987842 -2.752312 148 10 0 0 0 -0.0617256 0.316903 -0.103442 -0.246413 0.501593 -0.16867 -0.521201 0.776381 -0.327201 -0.85186 1.107038 -0.582808 -1.211843 1.467023 -0.91595 -1.588477 1.843657 -1.313598 -1.973093 2.228273 -1.756094 -2.361177 2.616356 -2.225489 -2.750623 3.005801 -2.708195 149 10 0 0 0 -0.0799751 0.111822 -0.0172787 -0.265311 0.297158 -0.0829948 -0.537032 0.568879 -0.23733 -0.862387 0.894234 -0.481547 -1.21695 1.248797 -0.79791 -1.589335 1.621182 -1.176552 -1.970905 2.00275 -1.599996 -2.3567 2.388545 -2.05083 -2.744176 2.776023 -2.514917 154 10 0 0 0 -0.299818 0.212675 -0.138998 -0.559485 0.212675 -0.203417 -0.943584 0.212675 -0.357588 -1.404905 0.212675 -0.603587 -1.907681 0.212674 -0.9228 -2.435266 0.212674 -1.3043 -2.975805 0.212674 -1.73155 -3.522318 0.212674 -2.187272 -4.071274 0.212673 -2.657451 155 10 0 0 0 -0.278741 0.189371 -0.113872 -0.537149 0.18937 -0.177628 -0.919695 0.18937 -0.330318 -1.37966 0.18937 -0.574306 -1.881406 0.18937 -0.891339 -2.408344 0.18937 -1.270886 -2.948551 0.189369 -1.69676 -3.494831 0.189369 -2.151219 -4.043594 0.189369 -2.620166 204 10 0 0 0 -0.0695391 0.178363 -0.0340257 -0.240131 0.348957 -0.089144 -0.493264 0.60209 -0.219858 -0.801207 0.910032 -0.430471 -1.141361 1.250185 -0.70812 -1.503062 1.611886 -1.046915 -1.877394 1.986217 -1.433969 -2.257993 2.366817 -1.851927 -2.641194 2.750018 -2.285191 205 10 0 0 0 -0.0569668 0.518169 -0.363464 -0.216103 0.748821 -0.439314 -0.470038 1.002759 -0.570988 -0.781002 1.313721 -0.787063 -1.125246 1.657965 -1.074978 -1.490993 2.023713 -1.427799 -1.868414 2.401131 -1.829251 -2.251516 2.784235 -2.261895 -2.637009 3.169727 -2.710524 207 10 0 0 0 -0.0570974 0.465358 -0.257522 -0.22893 0.665674 -0.324214 -0.491562 0.928307 -0.46662 -0.809965 1.24671 -0.696784 -1.159831 1.596576 -0.999692 -1.529555 1.966298 -1.367488 -1.909666 2.346411 -1.782784 -2.294639 2.731381 -2.227724 -2.681614 3.118357 -2.687666 210 10 0 0 0 -0.334831 0.256833 -0.194216 -0.568275 0.260692 -0.245658 -0.917011 0.260692 -0.368749 -1.347212 0.260692 -0.572779 -1.826324 0.260691 -0.847101 -2.338024 0.260691 -1.186308 -2.867905 0.26069 -1.574923 -3.406736 0.26069 -1.995197 -3.949291 0.26069 -2.431271 212 10 0 0 0 -0.392896 0.315859 -0.322088 -0.636977 0.333511 -0.378889 -0.985969 0.333511 -0.502186 -1.416106 0.333511 -0.706142 -1.895015 0.333511 -0.980114 -2.406531 0.33351 -1.318895 -2.936311 0.33351 -1.707181 -3.475099 0.33351 -2.127276 -4.017711 0.333509 -2.563614 214 10 0 0 0 -0.258099 0.172729 -0.0949128 -0.494148 0.172729 -0.147555 -0.849882 0.172729 -0.276383 -1.286705 0.172729 -0.488677 -1.771278 0.172729 -0.772516 -2.286985 0.172729 -1.121462 -2.8197 0.172728 -1.519361 -3.360645 0.172728 -1.948348 -3.905045 0.172728 -2.39313 215 10 0 0 0 -0.403297 0.325879 -0.356065 -0.667355 0.376305 -0.425437 -1.010286 0.376306 -0.543912 -1.434896 0.376305 -0.74121 -1.90934 0.376304 -1.007727 -2.417582 0.376304 -1.338929 -2.945005 0.376304 -1.719945 -3.482006 0.376303 -2.133122 -4.022989 0.376303 -2.56227 264 10 0 0 0 -0.0570107 0.258972 -0.0673714 -0.206932 0.408895 -0.10942 -0.435089 0.637054 -0.212724 -0.720074 0.922035 -0.385901 -1.041607 1.243569 -0.622283 -1.389732 1.591693 -0.920424 -1.75407 1.956031 -1.268234 -2.126909 2.328868 -1.648792 -2.503281 2.705242 -2.045203 265 10 0 0 0 -0.0523853 0.510832 -0.343543 -0.197605 0.714728 -0.403136 -0.428854 0.94598 -0.509595 -0.717577 1.234703 -0.688426 -1.042826 1.559948 -0.932415 -1.394114 1.911238 -1.239318 -1.760866 2.277989 -1.59579 -2.135515 2.652637 -1.98427 -2.513439 3.030562 -2.388218 267 10 0 0 0 -0.0537739 0.442365 -0.223798 -0.20923 0.614247 -0.274446 -0.448313 0.853331 -0.389181 -0.74506 1.150078 -0.580695 -1.077396 1.482413 -0.840028 -1.434248 1.839264 -1.163295 -1.805187 2.210204 -1.535975 -2.183147 2.588161 -1.940083 -2.56426 2.969275 -2.360917 271 10 0 0 0 -0.327454 0.252353 -0.184429 -0.540848 0.254392 -0.227051 -0.863728 0.254391 -0.33051 -1.266855 0.25439 -0.503796 -1.720863 0.254389 -0.739168 -2.211576 0.25439 -1.034261 -2.7253 0.254389 -1.378323 -3.25111 0.254389 -1.754525 -3.782194 0.254389 -2.147008 272 10 0 0 0 -0.171179 0.0908833 -0.034903 -0.3699 0.0908823 -0.0716588 -0.674488 0.0908823 -0.162599 -1.058655 0.0908813 -0.31683 -1.496078 0.0908823 -0.529893 -1.973772 0.0908823 -0.801809 -2.476767 0.0908823 -1.121484 -2.993132 0.0908813 -1.47208 -3.514544 0.0908804 -1.835874 273 10 0 0 0 -0.387396 0.313901 -0.31054 -0.614719 0.337148 -0.359711 -0.933445 0.337147 -0.460232 -1.332275 0.337147 -0.629037 -1.782391 0.337147 -0.858958 -2.269802 0.337146 -1.14792 -2.780903 0.337146 -1.48572 -3.30455 0.337145 -1.855704 -3.83358 0.337144 -2.241646 274 10 0 0 0 -0.242443 0.162307 -0.0827756 -0.452503 0.162306 -0.124019 -0.772161 0.162306 -0.225196 -1.170889 0.162306 -0.393897 -1.620317 0.162305 -0.62287 -2.107605 0.162305 -0.911608 -2.618181 0.162304 -1.248178 -3.141581 0.162303 -1.617471 -3.670494 0.162304 -2.00305 275 10 0 0 0 -0.377277 0.302929 -0.282537 -0.605965 0.320672 -0.332098 -0.933315 0.320673 -0.438785 -1.341841 0.320672 -0.617846 -1.801785 0.320672 -0.861805 -2.298094 0.32067 -1.167723 -2.81587 0.320671 -1.521905 -3.345184 0.32067 -1.908736 -3.879536 0.32067 -2.312268 0 10 0 0 0 -0.0642206 0.0642211 -0.00747335 -0.214664 0.214664 -0.0498278 -0.447434 0.447435 -0.157859 -0.740673 0.740675 -0.343735 -1.07249 1.072493 -0.601908 -1.430767 1.43077 -0.92961 -1.804158 1.804163 -1.312543 -2.185133 2.185138 -1.732604 -2.569497 2.569503 -2.173404 1 10 0 0 0 -0.0700221 0.0700226 -0.00889647 -0.226733 0.226733 -0.0550154 -0.461085 0.461086 -0.164702 -0.750815 0.750817 -0.345084 -1.07617 1.076173 -0.589305 -1.427457 1.42746 -0.896191 -1.794548 1.794551 -1.253932 -2.169983 2.169988 -1.645974 -2.549099 2.549104 -2.055981 2 10 0 0 0 -0.0628282 0.0628284 -0.00715113 -0.209779 0.20978 -0.0474859 -0.437445 0.437447 -0.150296 -0.72525 0.725253 -0.327728 -1.052271 1.052274 -0.575457 -1.406925 1.406928 -0.892092 -1.777799 1.777803 -1.264494 -2.157005 2.157009 -1.674969 -2.539999 2.540005 -2.106948 3 10 0 0 0 -0.0650268 0.0650284 -0.00766438 -0.212991 0.212991 -0.0485783 -0.437616 0.437617 -0.148361 -0.718722 0.718724 -0.315844 -1.037314 1.037317 -0.546376 -1.383841 1.383845 -0.840248 -1.747846 1.747851 -1.186898 -2.121345 2.12135 -1.57029 -2.499232 2.499239 -1.974047 4 10 0 0 0 -0.0633367 0.0633373 -0.00726849 -0.208902 0.208903 -0.0468171 -0.431715 0.431716 -0.144823 -0.712034 0.712036 -0.311169 -1.030672 1.030674 -0.541792 -1.377716 1.37772 -0.837037 -1.742427 1.742431 -1.186161 -2.116693 2.116698 -1.572923 -2.495411 2.495417 -1.980877 5 10 0 0 0 -0.0711355 0.0711353 -0.00918424 -0.232113 0.232113 -0.0579712 -0.473206 0.473207 -0.174903 -0.770574 0.770576 -0.367435 -1.103352 1.103354 -0.627763 -1.460954 1.460958 -0.953356 -1.833092 1.833096 -1.330956 -2.212663 2.212667 -1.743354 -2.595521 2.595527 -2.17449 6 10 0 0 0 -0.104315 4.17233e-007 -0.00988138 -0.346082 4.76837e-007 -0.0652462 -0.707729 8.9407e-007 -0.199075 -1.148748 1.43051e-006 -0.416778 -1.636881 2.08616e-006 -0.707065 -2.155718 2.6226e-006 -1.063998 -2.691185 3.03984e-006 -1.47151 -3.234593 3.45707e-006 -1.911529 -3.781366 3.8147e-006 -2.36869 7 10 0 0 0 -0.105806 7.15256e-007 -0.0101679 -0.345628 8.34465e-007 -0.0645974 -0.702389 8.34465e-007 -0.194285 -1.138271 1.66893e-006 -0.405386 -1.621742 2.02656e-006 -0.687266 -2.136914 2.5034e-006 -1.034882 -2.669683 2.98023e-006 -1.432958 -3.211032 3.45707e-006 -1.863679 -3.75603 3.69549e-006 -2.311443 8 10 0 0 0 -0.105003 2.38419e-007 -0.0100134 -0.341769 7.15256e-007 -0.0629919 -0.694339 1.19209e-006 -0.189203 -1.126007 1.43051e-006 -0.395038 -1.605746 1.66893e-006 -0.670427 -2.117999 2.14577e-006 -1.010952 -2.648599 2.6226e-006 -1.40187 -3.188268 3.09944e-006 -1.825515 -3.731797 3.33786e-006 -2.266102 9 10 0 0 0 -0.104608 7.15256e-007 -0.00993752 -0.34066 9.53674e-007 -0.0625798 -0.692776 9.53674e-007 -0.188419 -1.124277 1.43051e-006 -0.394047 -1.60391 1.90735e-006 -0.669258 -2.116021 2.14577e-006 -1.009441 -2.646414 2.6226e-006 -1.399693 -3.185819 2.86102e-006 -1.822267 -3.729034 3.33786e-006 -2.261379 10 10 0 0 0 -0.10215 3.20375e-007 -0.00947118 -0.340291 5.81145e-007 -0.0630999 -0.70199 1.19209e-006 -0.196973 -1.144058 1.63913e-006 -0.41605 -1.633459 2.17557e-006 -0.708681 -2.153421 2.65241e-006 -1.06858 -2.689776 3.12924e-006 -1.479345 -3.233896 3.48687e-006 -1.922759 -3.781303 3.9041e-006 -2.383509 11 10 0 0 0 -0.106519 1.19209e-007 -0.0103075 -0.348979 -1.19209e-007 -0.0660077 -0.710257 5.96046e-007 -0.19952 -1.150969 1.07288e-006 -0.416823 -1.638619 1.54972e-006 -0.70622 -2.15687 2.02656e-006 -1.061632 -2.691741 2.6226e-006 -1.467005 -3.234565 2.98023e-006 -1.904308 -3.780728 3.33786e-006 -2.35814 12 10 0 0 0 -0.070848 0.0708489 -0.0091095 -0.238745 0.238746 -0.0624071 -0.49772 0.497722 -0.200225 -0.818803 0.818805 -0.435697 -1.17435 1.174353 -0.755001 -1.548942 1.548946 -1.143218 -1.932775 1.932781 -1.580516 -2.320633 2.320638 -2.04789 -2.710106 2.710112 -2.530899 13 10 0 0 0 -0.073218 0.0732193 -0.00973523 -0.246224 0.246224 -0.0665121 -0.511263 0.511264 -0.212005 -0.837261 0.837265 -0.457575 -1.196299 1.196302 -0.787684 -1.573579 1.573583 -1.188445 -1.959118 1.959122 -1.637412 -2.348066 2.348071 -2.114999 -2.738345 2.73835 -2.607264 14 10 0 0 0 -0.0732083 0.0732093 -0.0097326 -0.246311 0.246313 -0.0665773 -0.511617 0.511619 -0.212416 -0.837955 0.837957 -0.4587 -1.1973 1.197303 -0.789793 -1.574814 1.574818 -1.191703 -1.960537 1.960542 -1.64201 -2.349622 2.349628 -2.120966 -2.740011 2.740016 -2.614583 15 10 0 0 0 -0.071651 0.071651 -0.00931931 -0.240566 0.240566 -0.0633004 -0.499912 0.499913 -0.201578 -0.820449 0.820451 -0.435957 -1.174165 1.174167 -0.749828 -1.54737 1.547373 -1.131954 -1.930289 1.930292 -1.563464 -2.317531 2.317534 -2.025586 -2.706532 2.706536 -2.503693 16 10 0 0 0 -0.072436 0.0724373 -0.00952625 -0.244347 0.244348 -0.0655469 -0.508784 0.508786 -0.210264 -0.83485 0.834853 -0.455975 -1.194301 1.194304 -0.787406 -1.572018 1.572022 -1.190319 -1.957893 1.957896 -1.641738 -2.347082 2.347085 -2.121756 -2.737549 2.737554 -2.61639 17 10 0 0 0 -0.0761149 0.0761154 -0.0105286 -0.250104 0.250105 -0.0679903 -0.510509 0.510509 -0.207588 -0.828069 0.828071 -0.436108 -1.178256 1.178258 -0.73991 -1.548666 1.548669 -1.11044 -1.929548 1.929551 -1.529987 -2.315255 2.315259 -1.980169 -2.702947 2.70295 -2.446099 18 10 0 0 0 -0.10717 5.96046e-008 -0.0104346 -0.352347 2.98023e-007 -0.0674624 -0.719826 6.55651e-007 -0.206363 -1.168432 8.9407e-007 -0.434208 -1.663462 1.2517e-006 -0.737597 -2.187333 1.60933e-006 -1.108217 -2.726073 2.08616e-006 -1.528123 -3.271624 2.20537e-006 -1.978719 -3.819971 2.44379e-006 -2.445071 19 10 0 0 0 -0.107806 2.68221e-007 -0.0105599 -0.355198 3.8743e-007 -0.0686829 -0.726191 7.15256e-007 -0.210709 -1.178524 1.2219e-006 -0.443717 -1.676696 1.63913e-006 -0.753373 -2.202778 1.96695e-006 -1.130381 -2.742993 2.20537e-006 -1.556284 -3.289583 2.563e-006 -2.012431 -3.838767 2.80142e-006 -2.484096 20 10 0 0 0 -0.107656 4.76837e-007 -0.0105305 -0.354924 6.55651e-007 -0.0685917 -0.726023 9.53674e-007 -0.210713 -1.178641 1.2219e-006 -0.444121 -1.677179 1.63913e-006 -0.75452 -2.203615 2.05636e-006 -1.132572 -2.744097 2.29478e-006 -1.559591 -3.290882 2.65241e-006 -2.01682 -3.840233 2.95043e-006 -2.48958 21 10 0 0 0 -0.103688 0 -0.00976145 -0.348765 2.38419e-007 -0.0667398 -0.724227 7.15256e-007 -0.212819 -1.185902 1.19209e-006 -0.459346 -1.694167 1.66893e-006 -0.790605 -2.22809 1.90735e-006 -1.19264 -2.773625 2.38419e-006 -1.643157 -3.323908 2.6226e-006 -2.122349 -3.87603 3.09944e-006 -2.616239 22 10 0 0 0 -0.106195 1.19209e-007 -0.0102445 -0.352044 2.98023e-007 -0.0676025 -0.723562 8.9407e-007 -0.210101 -1.178346 9.53674e-007 -0.446573 -1.679706 1.54972e-006 -0.762796 -2.208584 1.54972e-006 -1.148262 -2.750853 1.72853e-006 -1.583042 -3.298947 1.96695e-006 -2.047854 -3.849422 2.26498e-006 -2.528444 23 10 0 0 0 -0.10708 3.27826e-007 -0.010417 -0.354036 4.47035e-007 -0.0683231 -0.725812 9.83477e-007 -0.211054 -1.179938 1.40071e-006 -0.446591 -1.680265 1.63913e-006 -0.760662 -2.208214 2.05636e-006 -1.143265 -2.749802 2.23517e-006 -1.575031 -3.297405 2.35438e-006 -2.036912 -3.847429 2.47359e-006 -2.514267 25 10 0 0 0 -0.0742214 0.0742216 -0.0100064 -0.247851 0.247852 -0.0672166 -0.512113 0.512115 -0.211709 -0.836435 0.836437 -0.45378 -1.193807 1.193809 -0.778655 -1.570002 1.570004 -1.174222 -1.954965 1.954968 -1.619107 -2.343592 2.343595 -2.093544 -2.733707 2.733712 -2.583818 28 10 0 0 0 -0.0719438 0.0719457 -0.00939667 -0.243031 0.243032 -0.0648513 -0.506775 0.506778 -0.208681 -0.832408 0.832412 -0.453483 -1.191732 1.191736 -0.784505 -1.569607 1.569612 -1.1882 -1.955757 1.955761 -1.641687 -2.345216 2.34522 -2.124535 -2.735934 2.73594 -2.622502 29 10 0 0 0 -0.0736408 0.0736408 -0.00984776 -0.247126 0.247127 -0.066958 -0.512298 0.5123 -0.212625 -0.838052 0.838054 -0.457677 -1.196812 1.196814 -0.786906 -1.574036 1.574038 -1.187396 -1.95968 1.959683 -1.637129 -2.34877 2.348774 -2.116134 -2.739194 2.739197 -2.610204 35 10 0 0 0 -0.0991364 2.38419e-006 -0.00891578 -0.336813 9.53674e-007 -0.0623254 -0.705858 1.90735e-006 -0.202611 -1.164164 1.43051e-006 -0.444163 -1.672497 1.43051e-006 -0.775576 -2.208357 1.43051e-006 -1.184516 -2.756001 1.90735e-006 -1.646655 -3.308086 2.38419e-006 -2.14021 -3.861774 1.90735e-006 -2.650251 36 10 0 0 0 -0.0620215 0.0620222 -0.00696826 -0.21545 0.215451 -0.0510926 -0.462388 0.46239 -0.174578 -0.777412 0.777413 -0.398225 -1.131959 1.131962 -0.714543 -1.508961 1.508964 -1.113955 -1.895742 1.895745 -1.572337 -2.286101 2.286102 -2.065578 -2.677709 2.677711 -2.577234 37 10 0 0 0 -0.0619867 0.0619907 -0.00696039 -0.215434 0.215435 -0.0510952 -0.462434 0.462437 -0.174651 -0.777459 0.77746 -0.3983 -1.132304 1.132304 -0.715504 -1.509627 1.50963 -1.116481 -1.89655 1.896553 -1.576006 -2.286965 2.286969 -2.069978 -2.678599 2.6786 -2.58206 38 10 0 0 0 -0.0600052 0.0600014 -0.00651956 -0.209099 0.209099 -0.0480874 -0.450804 0.450801 -0.165694 -0.762261 0.76226 -0.382677 -1.114863 1.114862 -0.693315 -1.490383 1.490379 -1.085731 -1.876143 1.876143 -1.536316 -2.265979 2.265978 -2.023346 -2.657363 2.657362 -2.531147 39 10 0 0 0 -0.0614667 0.0614653 -0.00684309 -0.213408 0.213408 -0.0500793 -0.458476 0.458477 -0.171441 -0.772299 0.772301 -0.392823 -1.125978 1.125979 -0.706586 -1.502039 1.502042 -1.101525 -1.888193 1.888194 -1.55503 -2.278284 2.278286 -2.045049 -2.669782 2.669784 -2.55475 40 10 0 0 0 -0.0633996 0.0633984 -0.00728321 -0.218735 0.218735 -0.0525608 -0.467294 0.467295 -0.177909 -0.783693 0.783693 -0.404186 -1.138961 1.138962 -0.722651 -1.515602 1.515602 -1.120325 -1.901769 1.90177 -1.573936 -2.29172 2.291721 -2.062289 -2.683124 2.683125 -2.570407 41 10 0 0 0 -0.0581517 0.0581474 -0.00612116 -0.204278 0.204277 -0.045985 -0.443378 0.443377 -0.160742 -0.753572 0.753572 -0.375413 -1.1071 1.107099 -0.688735 -1.484324 1.484322 -1.089212 -1.871327 1.871326 -1.54938 -2.261805 2.261805 -2.044147 -2.65347 2.653471 -2.556856 42 10 0 0 0 -0.0871668 -9.53674e-007 -0.006881 -0.306458 -4.76837e-007 -0.0519919 -0.663134 -9.53674e-007 -0.181608 -1.11899 0 -0.419613 -1.629287 -1.43051e-006 -0.755538 -2.167231 -9.53674e-007 -1.172316 -2.716111 -1.90735e-006 -1.642012 -3.268828 -1.90735e-006 -2.141474 -3.822862 -2.38419e-006 -2.655929 43 10 0 0 0 -0.0800676 -1.90735e-006 -0.00580049 -0.285548 -9.53674e-007 -0.0451977 -0.627791 -9.53674e-007 -0.163131 -1.074115 -9.53674e-007 -0.387874 -1.579803 -1.43051e-006 -0.713385 -2.116039 -9.53674e-007 -1.123702 -2.664358 -9.53674e-007 -1.58988 -3.216887 -1.90735e-006 -2.087523 -3.770856 -2.38419e-006 -2.601123 44 10 0 0 0 -0.089406 4.76837e-007 -0.00724173 -0.314797 0 -0.0550156 -0.67992 9.53674e-007 -0.191849 -1.142671 9.53674e-007 -0.439991 -1.656785 9.53674e-007 -0.785002 -2.19615 4.76837e-007 -1.207415 -2.74538 0 -1.679379 -3.298109 0 -2.178964 -3.852069 0 -2.69244 45 10 0 0 0 -0.0752325 0 -0.00511789 -0.267385 0 -0.0394061 -0.588695 0 -0.141748 -1.014133 9.53674e-007 -0.340029 -1.506591 0 -0.638433 -2.036944 9.53674e-007 -1.028558 -2.583153 9.53674e-007 -1.482632 -3.134991 0 -1.974025 -3.688743 9.53674e-007 -2.484862 46 10 0 0 0 -0.0803337 1.90735e-006 -0.00583863 -0.283241 0 -0.0442188 -0.61737 1.90735e-006 -0.155927 -1.0551 1.90735e-006 -0.369381 -1.55525 1.90735e-006 -0.683084 -2.088984 1.90735e-006 -1.08446 -2.636429 2.86102e-006 -1.545438 -3.188704 1.90735e-006 -2.0407 -3.742621 9.53674e-007 -2.553622 47 10 0 0 0 -0.091115 4.76837e-007 -0.00752282 -0.319587 -4.76837e-007 -0.0566769 -0.687574 0 -0.196026 -1.152001 4.76837e-007 -0.446705 -1.666873 0 -0.793577 -2.206528 -4.76837e-007 -1.217164 -2.755868 -4.76837e-007 -1.689857 -3.308645 -1.43051e-006 -2.189911 -3.862622 -2.38419e-006 -2.703618 48 10 0 0 0 -0.0426426 0.0426426 -0.00328279 -0.155589 0.155592 -0.0267406 -0.353489 0.353488 -0.102385 -0.631829 0.631829 -0.265899 -0.967198 0.967197 -0.53215 -1.336661 1.336662 -0.898931 -1.720828 1.720833 -1.33842 -2.110247 2.110249 -1.820848 -2.501451 2.501451 -2.325771 49 10 0 0 0 -0.0477238 0.0477257 -0.00411487 -0.174423 0.174421 -0.0338016 -0.394874 0.394876 -0.129529 -0.694549 0.694551 -0.325873 -1.043613 1.043613 -0.62657 -1.41931 1.41931 -1.019812 -1.805753 1.805752 -1.47556 -2.195958 2.195955 -1.966873 -2.587445 2.587442 -2.47642 50 10 0 0 0 -0.0450191 0.0450191 -0.00365901 -0.16306 0.163062 -0.0293334 -0.369753 0.369753 -0.11245 -0.656861 0.65686 -0.288821 -0.998005 0.998005 -0.568915 -1.370184 1.370182 -0.94668 -1.755385 1.755383 -1.393223 -2.145181 2.14518 -1.879779 -2.536539 2.536537 -2.387106 51 10 0 0 0 -0.0487351 0.0487356 -0.00429058 -0.177671 0.17767 -0.0350668 -0.400903 0.400904 -0.133483 -0.702785 0.702785 -0.333544 -1.053093 1.053094 -0.637686 -1.429317 1.429319 -1.033386 -1.815957 1.815956 -1.490649 -2.206234 2.206236 -1.982868 -2.597755 2.597755 -2.492979 52 10 0 0 0 -0.0409336 0.0409336 -0.00302291 -0.152445 0.152443 -0.0258765 -0.353516 0.353514 -0.104162 -0.637212 0.637209 -0.275435 -0.976902 0.976902 -0.551953 -1.348455 1.34845 -0.92712 -1.733273 1.733269 -1.370987 -2.122799 2.122795 -1.854559 -2.514024 2.514021 -2.359872 53 10 0 0 0 -0.0431056 0.043108 -0.00335431 -0.157084 0.157089 -0.0272524 -0.356274 0.356279 -0.103965 -0.635606 0.635609 -0.268893 -0.971572 0.971581 -0.53654 -1.341303 1.341311 -0.904368 -1.725574 1.725584 -1.344542 -2.115036 2.115044 -1.827416 -2.506256 2.506262 -2.33265 54 10 0 0 0 -0.0565529 -2.86102e-006 -0.00288749 -0.211425 0 -0.0249093 -0.492306 -9.53674e-007 -0.101146 -0.890606 -9.53674e-007 -0.269404 -1.368894 0 -0.542321 -1.89338 -1.90735e-006 -0.914696 -2.437389 -1.90735e-006 -1.35757 -2.988251 -1.90735e-006 -1.841057 -3.541449 -2.86102e-006 -2.34545 55 10 0 0 0 -0.0694284 -1.90735e-006 -0.00435591 -0.253771 -1.90735e-006 -0.0358286 -0.573183 -3.8147e-006 -0.136832 -1.003366 -4.76837e-006 -0.340842 -1.499979 -3.8147e-006 -0.647363 -2.03229 -4.76837e-006 -1.043898 -2.578986 -5.72205e-006 -1.500639 -3.130744 -5.72205e-006 -1.991348 -3.68427 -6.67572e-006 -2.499445 56 10 0 0 0 -0.0717869 -9.53674e-007 -0.00465775 -0.261392 -1.90735e-006 -0.0380144 -0.587298 -1.90735e-006 -0.143651 -1.022551 -1.90735e-006 -0.353956 -1.522026 -2.86102e-006 -0.66627 -2.055262 -2.86102e-006 -1.065932 -2.602316 -2.86102e-006 -1.524674 -3.154222 -2.86102e-006 -2.01667 -3.707819 -3.8147e-006 -2.525604 57 10 0 0 0 -0.0601826 0 -0.00326991 -0.22175 9.53674e-007 -0.0272813 -0.510421 1.90735e-006 -0.108166 -0.915544 2.86102e-006 -0.283567 -1.398311 9.53674e-007 -0.564209 -1.924739 9.53674e-007 -0.942239 -2.469234 9.53674e-007 -1.387488 -3.020141 0 -1.871314 -3.573279 0 -2.375095 58 10 0 0 0 -0.0706072 9.53674e-007 -0.00450444 -0.255799 9.53674e-007 -0.0362792 -0.574044 -1.90735e-006 -0.136465 -1.001482 0 -0.337142 -1.49526 -1.90735e-006 -0.63809 -2.02516 -1.90735e-006 -1.026763 -2.570368 -2.86102e-006 -1.475595 -3.121334 -3.8147e-006 -1.959872 -3.674383 -3.8147e-006 -2.462713 59 10 0 0 0 -0.0605764 1.90735e-006 -0.00331092 -0.224874 0 -0.0281627 -0.518714 -9.53674e-007 -0.112232 -0.928406 9.53674e-007 -0.29256 -1.413624 0 -0.577549 -1.940943 9.53674e-007 -0.958242 -2.485982 9.53674e-007 -1.406211 -3.037219 0 -1.89263 -3.590569 -9.53674e-007 -2.398738 60 10 0 0 0 -0.0723672 0.0723665 -0.00950766 -0.235067 0.235067 -0.0593959 -0.477993 0.477993 -0.178356 -0.776773 0.776774 -0.373214 -1.110284 1.110286 -0.635202 -1.468019 1.468021 -0.961206 -1.839869 1.839871 -1.337603 -2.21888 2.218883 -1.747073 -2.600958 2.600962 -2.173496 61 10 0 0 0 -0.0709662 0.0709672 -0.00914085 -0.230223 0.230224 -0.0568416 -0.468327 0.468329 -0.17052 -0.762068 0.762071 -0.357195 -1.091034 1.091038 -0.6091 -1.445154 1.445159 -0.924156 -1.814296 1.814301 -1.289685 -2.191222 2.191227 -1.688729 -2.571511 2.571517 -2.104997 62 10 0 0 0 -0.0747142 0.0747149 -0.0101404 -0.24405 0.24405 -0.0644045 -0.497007 0.497008 -0.19491 -0.806532 0.806535 -0.408371 -1.149646 1.149648 -0.693392 -1.514771 1.514773 -1.043987 -1.891929 1.891932 -1.444154 -2.274908 2.274912 -1.876037 -2.660339 2.660345 -2.324231 63 10 0 0 0 -0.0697103 0.0697088 -0.00881672 -0.226593 0.226592 -0.0550418 -0.462024 0.462024 -0.165865 -0.753325 0.753325 -0.348689 -1.080174 1.080176 -0.596057 -1.432562 1.432565 -0.906078 -1.800291 1.800294 -1.266203 -2.175995 2.175999 -1.659482 -2.55512 2.555122 -2.069535 64 10 0 0 0 -0.0738621 0.0738618 -0.0099079 -0.241268 0.241268 -0.0628774 -0.491798 0.491798 -0.190518 -0.799024 0.799025 -0.399857 -1.1402 1.140201 -0.680023 -1.503885 1.503887 -1.025558 -1.880052 1.880055 -1.420995 -2.262304 2.262307 -1.848456 -2.64712 2.647123 -2.292321 65 10 0 0 0 -0.0683088 0.0683088 -0.00846338 -0.221893 0.221895 -0.0526814 -0.452727 0.452728 -0.158711 -0.739159 0.73916 -0.334062 -1.061523 1.061527 -0.572122 -1.410236 1.410239 -0.871864 -1.775117 1.775121 -1.221593 -2.148557 2.14856 -1.60473 -2.525683 2.525688 -2.00475 66 10 0 0 0 -0.0981174 0 -0.00873315 -0.320973 9.53674e-007 -0.0553898 -0.657455 0 -0.16888 -1.074571 4.76837e-007 -0.357479 -1.542244 9.53674e-007 -0.613163 -2.045295 9.53674e-007 -0.93296 -2.569027 9.53674e-007 -1.303186 -3.103192 1.90735e-006 -1.706069 -3.641804 1.43051e-006 -2.125461 67 10 0 0 0 -0.100709 0 -0.00920451 -0.329282 0 -0.0584038 -0.673222 9.53674e-007 -0.177672 -1.097782 7.15256e-007 -0.374911 -1.571992 1.43051e-006 -0.641041 -2.08018 1.66893e-006 -0.972126 -2.607768 1.90735e-006 -1.353633 -3.145045 1.90735e-006 -1.767857 -3.686433 2.6226e-006 -2.198748 68 10 0 0 0 -0.103933 -4.76837e-007 -0.00980794 -0.340258 -2.38419e-007 -0.0625788 -0.694838 4.76837e-007 -0.190448 -1.130223 4.76837e-007 -0.400921 -1.613908 9.53674e-007 -0.68318 -2.129375 9.53674e-007 -1.031528 -2.662255 1.43051e-006 -1.429978 -3.203534 1.43051e-006 -1.860395 -3.748334 1.90735e-006 -2.307164 69 10 0 0 0 -0.102825 -2.38419e-007 -0.0095979 -0.335761 0 -0.0607902 -0.684983 -2.38419e-007 -0.184273 -1.114399 7.15256e-007 -0.387347 -1.592527 7.15256e-007 -0.659996 -2.103587 1.19209e-006 -0.997698 -2.633226 1.43051e-006 -1.385554 -3.172054 1.66893e-006 -1.805805 -3.714771 1.90735e-006 -2.242616 70 10 0 0 0 -0.100474 4.76837e-007 -0.00916064 -0.32862 4.76837e-007 -0.0581672 -0.67212 4.76837e-007 -0.177088 -1.096333 9.53674e-007 -0.373916 -1.570291 9.53674e-007 -0.639634 -2.078306 1.19209e-006 -0.970327 -2.605772 1.43051e-006 -1.351469 -3.142959 1.43051e-006 -1.765345 -3.684248 1.66893e-006 -2.195809 71 10 0 0 0 -0.100672 0 -0.00919759 -0.328342 4.76837e-007 -0.0579902 -0.670189 4.76837e-007 -0.175617 -1.091964 9.53674e-007 -0.369581 -1.563283 1.43051e-006 -0.631027 -2.06899 1.43051e-006 -0.956577 -2.594656 1.90735e-006 -1.332365 -3.130427 2.38419e-006 -1.740979 -3.670513 2.38419e-006 -2.166349 72 10 0 0 0 -0.0737395 0.0737379 -0.00987512 -0.242828 0.242827 -0.0639727 -0.497676 0.497678 -0.196744 -0.810594 0.810592 -0.416428 -1.157381 1.15738 -0.71099 -1.525578 1.525578 -1.072888 -1.90502 1.905019 -1.484605 -2.289853 2.289854 -1.92859 -2.677077 2.677078 -2.390574 73 10 0 0 0 -0.0743346 0.0743345 -0.0100365 -0.243658 0.243658 -0.0642927 -0.497566 0.497567 -0.195933 -0.808712 0.808714 -0.412349 -1.153596 1.1536 -0.701914 -1.520252 1.520256 -1.05805 -1.898595 1.898599 -1.464084 -2.282478 2.282482 -1.901697 -2.668665 2.66867 -2.355479 74 10 0 0 0 -0.0750341 0.0750351 -0.0102283 -0.24672 0.246722 -0.0660933 -0.504502 0.504505 -0.202439 -0.819946 0.819949 -0.426885 -1.168702 1.168706 -0.726747 -1.538305 1.538309 -1.094077 -1.918729 1.918734 -1.51108 -2.304136 2.304141 -1.959098 -2.691592 2.691597 -2.423015 75 10 0 0 0 -0.0747976 0.0747979 -0.0101627 -0.246016 0.246017 -0.0657082 -0.503313 0.503315 -0.201456 -0.818388 0.81839 -0.425202 -1.166899 1.166902 -0.724395 -1.536351 1.536353 -1.091133 -1.91668 1.916683 -1.507616 -2.302022 2.302024 -1.955165 -2.689425 2.689427 -2.418635 76 10 0 0 0 -0.0744739 0.0744746 -0.0100751 -0.244165 0.244166 -0.0645798 -0.498595 0.498596 -0.196847 -0.81027 0.810273 -0.414234 -1.155609 1.155612 -0.704979 -1.522604 1.522607 -1.06236 -1.901181 1.901185 -1.469595 -2.285237 2.28524 -1.908342 -2.67157 2.671574 -2.363225 77 10 0 0 0 -0.0744438 0.0744445 -0.0100665 -0.245626 0.245626 -0.065586 -0.503703 0.503704 -0.202293 -0.820089 0.820091 -0.428543 -1.169961 1.169962 -0.731468 -1.540489 1.54049 -1.102472 -1.921574 1.921575 -1.523154 -2.307459 2.307461 -1.974654 -2.695515 2.695518 -2.443801 78 10 0 0 0 -0.107355 1.19209e-007 -0.0104707 -0.353534 4.76837e-007 -0.0679921 -0.722982 3.57628e-007 -0.208638 -1.173931 2.38419e-007 -0.439713 -1.671191 7.15256e-007 -0.747528 -2.196815 7.15256e-007 -1.123194 -2.736875 8.34465e-007 -1.548453 -3.283472 4.76837e-007 -2.004647 -3.832733 5.96046e-007 -2.476814 79 10 0 0 0 -0.104723 2.38419e-007 -0.00995874 -0.343869 7.15256e-007 -0.0640649 -0.703015 7.15256e-007 -0.195762 -1.143813 1.19209e-006 -0.413177 -1.633204 1.19209e-006 -0.70579 -2.153629 1.66893e-006 -1.066928 -2.690509 1.19209e-006 -1.479645 -3.235109 1.19209e-006 -1.925415 -3.782958 1.19209e-006 -2.388751 80 10 0 0 0 -0.0999856 0 -0.00907099 -0.3276 0 -0.0578389 -0.670895 1.19209e-006 -0.176599 -1.095068 4.76837e-007 -0.37338 -1.56892 9.53674e-007 -0.638924 -2.076532 1.66893e-006 -0.968712 -2.603873 1.43051e-006 -1.349474 -3.141302 1.43051e-006 -1.764277 -3.683105 1.90735e-006 -2.196976 81 10 0 0 0 -0.101262 7.15256e-007 -0.00930649 -0.331907 1.43051e-006 -0.0594461 -0.679235 1.19209e-006 -0.181406 -1.107265 1.43051e-006 -0.382796 -1.58422 2.14577e-006 -0.653473 -2.094947 2.14577e-006 -0.990393 -2.624669 1.90735e-006 -1.378506 -3.164149 2.38419e-006 -1.801382 -3.707932 2.38419e-006 -2.243176 82 10 0 0 0 -0.102777 -2.38419e-007 -0.00958925 -0.337155 2.38419e-007 -0.0614493 -0.689608 9.53674e-007 -0.187565 -1.122618 7.15256e-007 -0.395062 -1.604829 7.15256e-007 -0.674728 -2.119553 1.43051e-006 -1.021237 -2.65257 1.19209e-006 -1.420156 -3.194485 1.66893e-006 -1.853355 -3.740242 1.43051e-006 -2.305036 83 10 0 0 0 -0.103469 0 -0.00971991 -0.340181 0 -0.0626731 -0.696651 2.38419e-007 -0.192115 -1.134563 7.15256e-007 -0.4058 -1.620502 1.19209e-006 -0.692087 -2.13813 1.43051e-006 -1.045892 -2.672971 1.66893e-006 -1.451154 -3.216246 1.43051e-006 -1.89054 -3.763259 1.66893e-006 -2.349052 85 10 0 0 0 -0.0749364 0.0749383 -0.0102016 -0.248004 0.248006 -0.0670218 -0.508976 0.508978 -0.207332 -0.828362 0.828364 -0.439424 -1.181218 1.181221 -0.750794 -1.554534 1.554536 -1.133395 -1.937949 1.937953 -1.568014 -2.325787 2.325791 -2.035221 -2.71542 2.715423 -2.519965 86 10 0 0 0 -0.0749326 0.0749307 -0.0102007 -0.247889 0.24789 -0.0669434 -0.508615 0.508616 -0.206945 -0.827711 0.827712 -0.438468 -1.180293 1.180294 -0.749044 -1.553401 1.553403 -1.130759 -1.936676 1.936678 -1.56449 -2.324417 2.324419 -2.030851 -2.713974 2.713977 -2.514767 92 10 0 0 0 -0.106307 0 -0.0102657 -0.359212 2.38419e-007 -0.0711687 -0.746256 7.15256e-007 -0.228178 -1.218936 2.38419e-007 -0.491818 -1.735576 4.76837e-007 -0.843106 -2.27484 4.76837e-007 -1.265106 -2.823519 2.38419e-007 -1.733523 -3.375656 0 -2.227544 -3.929052 0 -2.734169 96 10 0 0 0 -0.0707631 0.0707607 -0.00908673 -0.242398 0.242399 -0.0649174 -0.510303 0.510303 -0.214156 -0.841919 0.841919 -0.47188 -1.206684 1.206682 -0.821191 -1.588218 1.588218 -1.244446 -1.97653 1.97653 -1.715927 -2.367291 2.36729 -2.21444 -2.758908 2.758905 -2.726226 97 10 0 0 0 -0.0789127 0.0789165 -0.0113255 -0.266144 0.266147 -0.078485 -0.549299 0.5493 -0.248954 -0.890103 0.890105 -0.528202 -1.258643 1.258646 -0.89141 -1.640715 1.640716 -1.317789 -2.028487 2.028489 -1.784435 -2.418557 2.418559 -2.27417 -2.809497 2.809498 -2.775176 98 10 0 0 0 -0.080708 0.0807095 -0.0118508 -0.270413 0.270414 -0.0809228 -0.554989 0.554993 -0.253498 -0.8959 0.895904 -0.533007 -1.26357 1.263574 -0.892912 -1.644786 1.64479 -1.314327 -2.032017 2.032021 -1.77637 -2.421725 2.421726 -2.261944 -2.812377 2.812379 -2.758986 99 10 0 0 0 -0.0806365 0.080636 -0.0118296 -0.270144 0.270141 -0.0807474 -0.554488 0.554486 -0.252977 -0.895261 0.895257 -0.532146 -1.262882 1.262878 -0.891865 -1.64411 1.644106 -1.313341 -2.03136 2.031356 -1.775546 -2.421087 2.421083 -2.261335 -2.811756 2.811752 -2.758621 100 10 0 0 0 -0.0735159 0.0735178 -0.0098151 -0.250003 0.250002 -0.0690365 -0.522286 0.522287 -0.224136 -0.856344 0.856344 -0.487366 -1.222042 1.222044 -0.840014 -1.603759 1.603761 -1.26432 -1.992002 1.992001 -1.735147 -2.382626 2.382626 -2.231838 -2.774098 2.774098 -2.741167 101 10 0 0 0 -0.0733242 0.0733194 -0.00976288 -0.250895 0.250892 -0.0697587 -0.525852 0.52585 -0.228529 -0.86267 0.862665 -0.49816 -1.230211 1.230206 -0.85758 -1.612764 1.612761 -1.286862 -2.0013 2.001295 -1.760418 -2.392015 2.39201 -2.258281 -2.783502 2.783495 -2.767806 102 10 0 0 0 -0.110744 2.38419e-007 -0.0111493 -0.373084 0 -0.0769913 -0.769438 4.76837e-007 -0.243258 -1.246787 0 -0.514596 -1.763976 0 -0.867277 -2.301874 -2.38419e-007 -1.283876 -2.849077 -7.15256e-007 -1.743465 -3.400179 -9.53674e-007 -2.228816 -3.952787 -1.19209e-006 -2.727227 103 10 0 0 0 -0.109944 -2.38419e-007 -0.0109872 -0.372272 -2.38419e-007 -0.0768223 -0.770771 -4.76837e-007 -0.245284 -1.251532 -9.53674e-007 -0.522431 -1.772 -9.53674e-007 -0.88368 -2.312131 -1.19209e-006 -1.309235 -2.860486 -1.43051e-006 -1.775628 -3.412109 -1.90735e-006 -2.265197 -3.965101 -2.14577e-006 -2.767434 104 10 0 0 0 -0.107867 0 -0.0105729 -0.36696 0 -0.0746882 -0.763238 -2.38419e-007 -0.240879 -1.243421 -4.76837e-007 -0.517028 -1.764224 -4.76837e-007 -0.879179 -2.304988 -7.15256e-007 -1.307389 -2.853822 -1.43051e-006 -1.776788 -3.405778 -1.43051e-006 -2.269195 -3.958982 -1.43051e-006 -2.773676 105 10 0 0 0 -0.103045 4.76837e-007 -0.00964105 -0.354081 0 -0.0695916 -0.743826 0 -0.229242 -1.220928 4.76837e-007 -0.500164 -1.74087 0 -0.860008 -2.281718 9.53674e-007 -1.288584 -2.830952 0 -1.760571 -3.383256 -4.76837e-007 -2.256109 -3.936755 -4.76837e-007 -2.763911 106 10 0 0 0 -0.0960474 9.53674e-007 -0.00836492 -0.334506 4.76837e-007 -0.0621445 -0.712686 -4.76837e-007 -0.210732 -1.182923 -4.76837e-007 -0.470452 -1.699615 -4.76837e-007 -0.821874 -2.23918 -4.76837e-007 -1.245091 -2.787958 -1.43051e-006 -1.714152 -3.340319 -1.43051e-006 -2.210211 -3.893965 -1.90735e-006 -2.719746 107 10 0 0 0 -0.107908 0 -0.0105796 -0.36666 4.76837e-007 -0.0745159 -0.761802 -2.38419e-007 -0.239553 -1.240311 2.38419e-007 -0.512847 -1.75946 0 -0.870603 -2.299002 -7.15256e-007 -1.293733 -2.847143 -9.53674e-007 -1.758833 -3.398849 -1.19209e-006 -2.249114 -3.951931 -1.66893e-006 -2.752275 108 10 0 0 0 -0.0520792 0.0520782 -0.00490332 -0.187767 0.187767 -0.0390944 -0.418908 0.418911 -0.14544 -0.726256 0.726257 -0.354995 -1.07862 1.078619 -0.664942 -1.454958 1.454958 -1.061182 -1.841223 1.841223 -1.515536 -2.231153 2.231153 -2.003605 -2.622411 2.62241 -2.509447 109 10 0 0 0 -0.0692415 0.0692434 -0.00869787 -0.238493 0.238495 -0.0629057 -0.504261 0.504261 -0.209343 -0.833828 0.833832 -0.462551 -1.196501 1.196499 -0.80461 -1.576399 1.576399 -1.218762 -1.96364 1.96364 -1.680874 -2.353725 2.353725 -2.170773 -2.744877 2.744875 -2.674933 110 10 0 0 0 -0.0617638 0.0617657 -0.00690889 -0.218735 0.218737 -0.0531886 -0.47486 0.474862 -0.187503 -0.800725 0.800725 -0.43279 -1.163309 1.163308 -0.774551 -1.544047 1.544045 -1.193281 -1.931984 1.931983 -1.661363 -2.322606 2.322604 -2.158017 -2.714197 2.714194 -2.669319 111 10 0 0 0 -0.0624456 0.0624475 -0.00706434 -0.220735 0.220737 -0.0541604 -0.478155 0.478157 -0.190058 -0.804744 0.804747 -0.436875 -1.167569 1.167573 -0.779451 -1.54832 1.548323 -1.198264 -1.936217 1.93622 -1.665986 -2.326827 2.32683 -2.162497 -2.718396 2.718399 -2.673441 112 10 0 0 0 -0.0568314 0.0568352 -0.00584459 -0.2026 0.202602 -0.0455067 -0.445344 0.445347 -0.164263 -0.761177 0.761179 -0.389456 -1.118159 1.118162 -0.713128 -1.496396 1.496397 -1.118624 -1.883344 1.883347 -1.578361 -2.273526 2.273527 -2.069413 -2.664884 2.664886 -2.576795 113 10 0 0 0 -0.0651112 0.0651131 -0.00768542 -0.227758 0.227758 -0.057538 -0.488308 0.488307 -0.197316 -0.815548 0.815547 -0.445513 -1.178067 1.178065 -0.787053 -1.558379 1.558379 -1.203453 -1.946004 1.946004 -1.668776 -2.336371 2.336368 -2.162135 -2.727754 2.72775 -2.669909 114 10 0 0 0 -0.0892334 4.76837e-007 -0.0072124 -0.311831 4.76837e-007 -0.0537558 -0.670702 4.76837e-007 -0.18522 -1.12542 0 -0.4216 -1.632524 -1.43051e-006 -0.750248 -2.168337 0 -1.159014 -2.716373 -1.43051e-006 -1.623474 -3.268829 -9.53674e-007 -2.120413 -3.822826 -9.53674e-007 -2.634412 115 10 0 0 0 -0.0964661 -4.76837e-007 -0.00843787 -0.332756 -4.76837e-007 -0.0611933 -0.704981 0 -0.204328 -1.168304 -4.76837e-007 -0.453334 -1.679802 0 -0.792069 -2.21653 -4.76837e-007 -1.204217 -2.763985 -9.53674e-007 -1.665244 -3.315584 -1.90735e-006 -2.154612 -3.868883 -2.38419e-006 -2.660168 116 10 0 0 0 -0.092041 9.53674e-007 -0.00767803 -0.319405 -9.53674e-007 -0.0563323 -0.682798 -4.76837e-007 -0.191665 -1.140757 -1.43051e-006 -0.43271 -1.649696 -2.38419e-006 -0.765509 -2.185369 -2.86102e-006 -1.173768 -2.732698 -2.86102e-006 -1.634068 -3.284693 -3.33786e-006 -2.126814 -3.83844 -2.86102e-006 -2.637587 117 10 0 0 0 -0.0806503 -9.53674e-007 -0.00588608 -0.286491 1.90735e-006 -0.0454271 -0.62771 9.53674e-007 -0.162561 -1.070662 9.53674e-007 -0.382804 -1.571989 9.53674e-007 -0.698961 -2.104847 0 -1.097342 -2.651031 -9.53674e-007 -1.551284 -3.202353 0 -2.038383 -3.755764 -9.53674e-007 -2.545143 118 10 0 0 0 -0.095911 -9.53674e-007 -0.00834155 -0.330629 0 -0.06036 -0.701077 -9.53674e-007 -0.201898 -1.163309 -4.76837e-007 -0.44926 -1.674212 -4.76837e-007 -0.786592 -2.210613 -9.53674e-007 -1.197528 -2.757883 -1.90735e-006 -1.657485 -3.309601 -9.53674e-007 -2.147861 -3.863117 -1.90735e-006 -2.655869 119 10 0 0 0 -0.0879688 -4.76837e-007 -0.00700879 -0.307653 9.53674e-007 -0.0522892 -0.662952 0 -0.180756 -1.11558 4.76837e-007 -0.414178 -1.621937 -9.53674e-007 -0.741168 -2.156639 -9.53674e-007 -1.145932 -2.703611 -1.43051e-006 -1.604215 -3.255499 -1.90735e-006 -2.096035 -3.809235 -2.38419e-006 -2.606643 120 10 0 0 0 -0.0602074 0.0602067 -0.00656247 -0.196646 0.196644 -0.0411458 -0.405035 0.405034 -0.125757 -0.668301 0.668302 -0.268971 -0.969166 0.969167 -0.467313 -1.299401 1.299405 -0.721987 -1.648779 1.648781 -1.023543 -2.008566 2.008569 -1.356056 -2.372316 2.37232 -1.701824 121 10 0 0 0 -0.0656738 0.0656731 -0.00781822 -0.214296 0.214297 -0.049113 -0.439413 0.439413 -0.14938 -0.720562 0.720564 -0.316926 -1.038441 1.038443 -0.546063 -1.383612 1.383614 -0.836371 -1.745771 1.745773 -1.176699 -2.117012 2.117015 -1.550592 -2.49209 2.492093 -1.940992 122 10 0 0 0 -0.0643396 0.0643406 -0.00750208 -0.209888 0.209887 -0.0470397 -0.430672 0.430673 -0.143085 -0.707131 0.707133 -0.303948 -1.020571 1.020572 -0.52461 -1.361972 1.361974 -0.805335 -1.721101 1.721104 -1.135736 -2.089833 2.089836 -1.499677 -2.462624 2.462627 -1.880028 123 10 0 0 0 -0.0658703 0.0658696 -0.00786495 -0.21494 0.214939 -0.0494184 -0.440672 0.440673 -0.150297 -0.72248 0.722481 -0.318798 -1.040972 1.040971 -0.549132 -1.386653 1.386655 -0.840779 -1.749219 1.74922 -1.182478 -2.120795 2.120798 -1.557748 -2.49618 2.496182 -1.949549 124 10 0 0 0 -0.0667839 0.066783 -0.00808644 -0.218073 0.218072 -0.0509386 -0.447023 0.447024 -0.155048 -0.732367 0.732369 -0.328767 -1.054224 1.054224 -0.565794 -1.402749 1.402751 -0.865027 -1.767578 1.76758 -1.214571 -2.14097 2.140972 -1.597506 -2.517955 2.517957 -1.996827 125 10 0 0 0 -0.0645161 0.0645123 -0.00754285 -0.210747 0.210744 -0.0474672 -0.432828 0.432822 -0.144761 -0.710934 0.710931 -0.307947 -1.02609 1.026087 -0.531846 -1.368939 1.368937 -0.816199 -1.729028 1.729026 -1.149687 -2.098253 2.098252 -1.515537 -2.471224 2.471223 -1.896664 126 10 0 0 0 -0.0962191 -1.43051e-006 -0.00839579 -0.314466 -1.43051e-006 -0.0530599 -0.644046 0 -0.161379 -1.052973 -4.76837e-007 -0.340875 -1.512216 -4.76837e-007 -0.583798 -2.008339 4.76837e-007 -0.889349 -2.526669 0 -1.244964 -3.056855 0 -1.634552 -3.592389 -9.53674e-007 -2.042307 127 10 0 0 0 -0.0886374 9.53674e-007 -0.0071156 -0.288862 1.90735e-006 -0.0444515 -0.592394 9.53674e-007 -0.134699 -0.972289 9.53674e-007 -0.28489 -1.40401 1.90735e-006 -0.490788 -1.875774 3.8147e-006 -0.752947 -2.373523 3.8147e-006 -1.061747 -2.886024 1.90735e-006 -1.402862 -3.404808 2.86102e-006 -1.75966 128 10 0 0 0 -0.0943909 4.76837e-007 -0.00807643 -0.308216 0 -0.0508747 -0.63136 4.76837e-007 -0.154523 -1.033148 4.76837e-007 -0.326402 -1.485757 4.76837e-007 -0.559797 -1.976148 4.76837e-007 -0.854279 -2.489768 9.53674e-007 -1.19809 -3.016042 1.43051e-006 -1.575668 -3.547956 4.76837e-007 -1.970883 129 10 0 0 0 -0.0961642 -9.53674e-007 -0.00838542 -0.314484 -9.53674e-007 -0.0530803 -0.644612 -4.76837e-007 -0.161805 -1.054886 -4.76837e-007 -0.34277 -1.516028 -4.76837e-007 -0.588506 -2.013283 4.76837e-007 -0.896311 -2.532385 4.76837e-007 -1.253942 -3.063236 4.76837e-007 -1.645672 -3.599182 4.76837e-007 -2.054927 130 10 0 0 0 -0.09655 -9.53674e-007 -0.00845277 -0.315643 1.43051e-006 -0.0534793 -0.64659 4.76837e-007 -0.16281 -1.057219 9.53674e-007 -0.344166 -1.518034 9.53674e-007 -0.589414 -2.015402 9.53674e-007 -0.897445 -2.534813 1.43051e-006 -1.25589 -3.065728 1.90735e-006 -1.647824 -3.601966 1.90735e-006 -2.058154 131 10 0 0 0 -0.0903845 9.53674e-007 -0.00740099 -0.294684 1.90735e-006 -0.0463305 -0.604073 2.86102e-006 -0.140451 -0.990527 1.90735e-006 -0.296888 -1.428607 2.38419e-006 -0.510791 -1.906127 3.8147e-006 -0.782412 -2.40886 3.8147e-006 -1.101536 -2.92579 3.8147e-006 -1.453559 -3.448847 3.8147e-006 -1.821892 132 10 0 0 0 -0.0731373 0.0731354 -0.00971383 -0.239803 0.2398 -0.0621917 -0.490341 0.490338 -0.189841 -0.798249 0.798248 -0.400398 -1.140476 1.140475 -0.683182 -1.505263 1.505262 -1.032572 -1.882427 1.882425 -1.432768 -2.265579 2.265578 -1.865736 -2.651269 2.651268 -2.315788 133 10 0 0 0 -0.0732841 0.0732861 -0.00975281 -0.240228 0.240231 -0.0624154 -0.490954 0.490956 -0.190286 -0.799151 0.799155 -0.401356 -1.142037 1.142038 -0.685798 -1.507588 1.507591 -1.037918 -1.885468 1.885471 -1.441639 -2.269272 2.269275 -1.878733 -2.655593 2.655597 -2.333527 134 10 0 0 0 -0.0687761 0.0687747 -0.00858057 -0.224549 0.224547 -0.0541245 -0.45948 0.459478 -0.16442 -0.750824 0.750824 -0.347311 -1.077941 1.077943 -0.595248 -1.43062 1.430621 -0.9061 -1.798619 1.79862 -1.267249 -2.174606 2.174607 -1.661839 -2.554034 2.554036 -2.073482 135 10 0 0 0 -0.069931 0.0699317 -0.00887322 -0.22897 0.228971 -0.0564365 -0.468987 0.468986 -0.17219 -0.766038 0.766038 -0.364203 -1.098485 1.098485 -0.623785 -1.455303 1.455304 -0.946953 -1.826555 1.826557 -1.32089 -2.205297 2.205299 -1.728966 -2.587284 2.587286 -2.154843 136 10 0 0 0 -0.0695829 0.0695829 -0.00878489 -0.22718 0.22718 -0.0554508 -0.464406 0.464407 -0.168187 -0.758162 0.758162 -0.35488 -1.087653 1.087653 -0.607927 -1.442471 1.442472 -0.925048 -1.812274 1.812276 -1.293166 -2.189815 2.189815 -1.6952 -2.570715 2.570718 -2.114856 137 10 0 0 0 -0.0707283 0.070729 -0.00907862 -0.231509 0.23151 -0.057741 -0.473631 0.473632 -0.175809 -0.772358 0.77236 -0.370579 -1.106329 1.106331 -0.633612 -1.464706 1.464709 -0.961634 -1.83729 1.837291 -1.341101 -2.217183 2.217186 -1.755224 -2.60029 2.600292 -2.187891 138 10 0 0 0 -0.107677 0 -0.010535 -0.350163 0 -0.0662481 -0.710106 0 -0.19862 -1.148232 4.76837e-007 -0.412579 -1.632348 9.53674e-007 -0.695607 -2.146902 9.53674e-007 -1.041697 -2.678394 4.76837e-007 -1.435514 -3.218173 4.76837e-007 -1.85961 -3.761339 0 -2.298494 139 10 0 0 0 -0.108214 -2.38419e-007 -0.0106413 -0.354923 -7.15256e-007 -0.0684245 -0.723466 2.38419e-007 -0.208267 -1.172262 0 -0.436373 -1.66688 0 -0.738955 -2.190105 0 -1.107755 -2.728253 -2.38419e-007 -1.525331 -3.273322 -2.38419e-007 -1.973451 -3.821222 -2.38419e-007 -2.437093 140 10 0 0 0 -0.10557 -1.43051e-006 -0.0101219 -0.343824 -4.76837e-007 -0.0638046 -0.698879 -4.76837e-007 -0.192068 -1.132794 -1.43051e-006 -0.400695 -1.613757 -9.53674e-007 -0.678191 -2.126221 -1.43051e-006 -1.019217 -2.6564 -1.43051e-006 -1.408782 -3.195322 -9.53674e-007 -1.829412 -3.73786 -1.43051e-006 -2.265407 141 10 0 0 0 -0.0975161 2.38419e-006 -0.00862551 -0.316992 1.43051e-006 -0.0538163 -0.646493 1.43051e-006 -0.162078 -1.05418 1.90735e-006 -0.34023 -1.511621 1.43051e-006 -0.580527 -2.005235 2.38419e-006 -0.881156 -2.52096 9.53674e-007 -1.230143 -3.048336 4.76837e-007 -1.611011 -3.580559 1.43051e-006 -2.007251 142 10 0 0 0 -0.100458 0 -0.0091579 -0.326173 1.43051e-006 -0.0570772 -0.663634 1.43051e-006 -0.171313 -1.079136 9.53674e-007 -0.358084 -1.543357 9.53674e-007 -0.608451 -2.042287 2.38419e-006 -0.919646 -2.562008 1.90735e-006 -1.278909 -3.092505 1.90735e-006 -1.669496 -3.627439 9.53674e-007 -2.075096 143 10 0 0 0 -0.109165 0 -0.0108312 -0.355361 -7.15256e-007 -0.0683602 -0.720368 -4.76837e-007 -0.205094 -1.163527 -2.38419e-007 -0.425607 -1.651951 2.38419e-007 -0.716427 -2.16975 2.38419e-007 -1.070669 -2.70358 -4.76837e-007 -1.472381 -3.245133 -7.15256e-007 -1.903988 -3.789832 -4.76837e-007 -2.350253 144 10 0 0 0 -0.0778198 0.0778232 -0.0110098 -0.255786 0.255789 -0.07129 -0.521362 0.521366 -0.217479 -0.843601 0.843605 -0.455284 -1.197348 1.197351 -0.769243 -1.57011 1.570114 -1.149472 -1.952458 1.952462 -1.577513 -2.339131 2.339133 -2.035018 -2.727539 2.727542 -2.507398 146 10 0 0 0 -0.0776272 0.0776243 -0.0109538 -0.255212 0.25521 -0.0709598 -0.520381 0.52038 -0.216623 -0.842327 0.842327 -0.453836 -1.195906 1.195905 -0.767307 -1.568596 1.568595 -1.147232 -1.950926 1.950925 -1.575148 -2.337599 2.337598 -2.032666 -2.726019 2.726018 -2.505162 147 10 0 0 0 -0.0791054 0.0791059 -0.0113795 -0.259329 0.259328 -0.0732949 -0.526912 0.526913 -0.222107 -0.850235 0.850236 -0.462122 -1.204281 1.204282 -0.776959 -1.576908 1.57691 -1.15662 -1.958976 1.958977 -1.58298 -2.345334 2.345334 -2.038043 -2.733411 2.73341 -2.507378 150 10 0 0 0 -0.113796 4.17233e-007 -0.0117795 -0.374771 1.19209e-007 -0.0768919 -0.76139 2.38419e-007 -0.233488 -1.225455 1.19209e-007 -0.483617 -1.730565 -2.38419e-007 -0.807856 -2.259997 -2.98023e-007 -1.195054 -2.80185 -7.15256e-007 -1.627981 -3.34928 -8.9407e-007 -2.088883 -3.898974 -1.43051e-006 -2.563941 151 10 0 0 0 -0.11709 2.98023e-007 -0.0124793 -0.3888 2.38419e-007 -0.0834559 -0.791081 1.19209e-007 -0.255852 -1.269053 1.19209e-007 -0.528236 -1.784303 0 -0.876045 -2.320116 -2.38419e-007 -1.284814 -2.865592 -4.17233e-007 -1.735021 -3.415219 -8.34465e-007 -2.209631 -3.966601 -1.3113e-006 -2.69723 152 10 0 0 0 -0.109921 -4.76837e-007 -0.0109829 -0.361291 -4.76837e-007 -0.071104 -0.735578 -4.76837e-007 -0.216108 -1.188504 -9.53674e-007 -0.449952 -1.685771 -1.3113e-006 -0.757784 -2.210322 -1.43051e-006 -1.130339 -2.748984 -1.66893e-006 -1.549934 -3.294177 -2.14577e-006 -1.998674 -3.842038 -2.38419e-006 -2.462091 153 10 0 0 0 -0.114847 0 -0.0120001 -0.378706 -2.98023e-007 -0.0786588 -0.769142 -2.98023e-007 -0.238989 -1.236624 -2.98023e-007 -0.494373 -1.744322 -5.66244e-007 -0.824353 -2.275292 -8.9407e-007 -1.216469 -2.818034 -1.16229e-006 -1.653389 -3.366078 -1.54972e-006 -2.117901 -3.916265 -1.72853e-006 -2.596407 156 10 0 0 0 -0.0759087 0.0759125 -0.0104705 -0.256726 0.256726 -0.0728195 -0.532163 0.532166 -0.232258 -0.866296 0.866298 -0.495659 -1.23018 1.230182 -0.841882 -1.610012 1.610012 -1.255681 -1.996861 1.996862 -1.714613 -2.386547 2.386549 -2.199949 -2.77733 2.777332 -2.698778 157 10 0 0 0 -0.0809822 0.0809784 -0.0119321 -0.267185 0.267185 -0.0783082 -0.542991 0.542985 -0.238251 -0.873422 0.873418 -0.493358 -1.232385 1.232381 -0.82323 -1.607841 1.60784 -1.215362 -1.99168 1.991674 -1.652686 -2.379253 2.379248 -2.11759 -2.768332 2.768326 -2.596473 158 10 0 0 0 -0.0792837 0.0792837 -0.0114325 -0.262469 0.262475 -0.0755358 -0.53554 0.535542 -0.231705 -0.864273 0.864273 -0.483101 -1.222483 1.222483 -0.810592 -1.598042 1.598042 -1.203202 -1.982033 1.982035 -1.641546 -2.369736 2.369736 -2.107569 -2.758929 2.758927 -2.58762 159 10 0 0 0 -0.0789528 0.0789509 -0.011336 -0.264894 0.26489 -0.0775073 -0.544609 0.544605 -0.242986 -0.880988 0.880984 -0.511594 -1.24554 1.245534 -0.860156 -1.625061 1.625055 -1.272288 -2.011558 2.011552 -1.728439 -2.400958 2.400952 -2.210653 -2.791489 2.791483 -2.70609 160 10 0 0 0 -0.0806971 0.080699 -0.0118483 -0.267427 0.267429 -0.0786238 -0.544961 0.544963 -0.240998 -0.877534 0.877534 -0.500857 -1.238381 1.238381 -0.836826 -1.615203 1.615203 -1.235364 -1.999959 1.999957 -1.678822 -2.388169 2.388169 -2.149367 -2.777758 2.77776 -2.633638 161 10 0 0 0 -0.0770731 0.077075 -0.010797 -0.258006 0.25801 -0.0732322 -0.531072 0.531074 -0.229397 -0.86195 0.861952 -0.485484 -1.22307 1.223072 -0.822342 -1.600915 1.600917 -1.225881 -1.986493 1.986495 -1.675139 -2.375335 2.375337 -2.151659 -2.765464 2.765466 -2.642076 162 10 0 0 0 -0.107359 1.11759e-007 -0.0104722 -0.356699 -1.71363e-007 -0.0695683 -0.732796 -3.42727e-007 -0.216231 -1.190798 -4.32134e-007 -0.457341 -1.693529 -7.59959e-007 -0.776459 -2.222542 -9.98378e-007 -1.162342 -2.76445 -1.2964e-006 -1.595518 -3.312065 -1.83284e-006 -2.05749 -3.862007 -2.13087e-006 -2.534258 163 10 0 0 0 -0.106514 -1.19209e-007 -0.0103063 -0.357447 2.38419e-007 -0.0702059 -0.739017 -4.76837e-007 -0.221974 -1.204193 -3.57628e-007 -0.473794 -1.713547 -7.15256e-007 -0.807533 -2.247406 -1.07288e-006 -1.20935 -2.792559 -1.54972e-006 -1.657902 -3.342443 -1.90735e-006 -2.134276 -3.894299 -2.14577e-006 -2.625817 164 10 0 0 0 -0.107412 0 -0.0104823 -0.35569 0 -0.0690467 -0.72909 -3.57628e-007 -0.213247 -1.183702 -4.76837e-007 -0.449475 -1.683298 -7.7486e-007 -0.762035 -2.209978 -1.13249e-006 -1.140812 -2.750298 -1.43051e-006 -1.567157 -3.296789 -1.90735e-006 -2.022769 -3.8458 -2.26498e-006 -2.493301 165 10 0 0 0 -0.108775 8.9407e-008 -0.010753 -0.360036 5.96046e-008 -0.070819 -0.737185 -2.38419e-007 -0.218451 -1.195107 -3.57628e-007 -0.459444 -1.697142 -7.7486e-007 -0.777088 -2.225365 -1.04308e-006 -1.160529 -2.766649 -1.37091e-006 -1.590973 -3.31382 -1.66893e-006 -2.050376 -3.863388 -2.17557e-006 -2.52458 166 10 0 0 0 -0.104992 7.15256e-007 -0.0100114 -0.354073 4.76837e-007 -0.0689776 -0.73521 0 -0.220335 -1.201248 -2.38419e-007 -0.473482 -1.711857 -4.76837e-007 -0.810132 -2.246759 -9.53674e-007 -1.215614 -2.792629 -1.66893e-006 -1.667889 -3.343079 -1.90735e-006 -2.148295 -3.895483 -2.38419e-006 -2.644753 167 10 0 0 0 -0.103265 4.76837e-007 -0.00968111 -0.351039 -2.38419e-007 -0.0679945 -0.73291 -4.76837e-007 -0.220045 -1.200912 -2.38419e-007 -0.47624 -1.71356 -7.15256e-007 -0.8177 -2.249996 -9.53674e-007 -1.228769 -2.796898 -1.43051e-006 -1.686645 -3.348 -1.66893e-006 -2.171993 -3.900717 -2.38419e-006 -2.671469 168 10 0 0 0 -0.0731792 0.073185 -0.00972426 -0.247875 0.247881 -0.0676811 -0.516628 0.516632 -0.218038 -0.846141 0.846148 -0.471134 -1.207459 1.207465 -0.808651 -1.585944 1.585948 -1.215412 -1.97224 1.972243 -1.669991 -2.361833 2.361836 -2.154314 -2.75282 2.752824 -2.656018 169 10 0 0 0 -0.0716972 0.0716991 -0.00933051 -0.244454 0.244456 -0.0659357 -0.512183 0.512185 -0.214939 -0.841919 0.841919 -0.468516 -1.204078 1.204082 -0.808849 -1.583424 1.583424 -1.220051 -1.970242 1.97024 -1.678724 -2.360022 2.36002 -2.165107 -2.751057 2.751055 -2.667484 170 10 0 0 0 -0.0740204 0.0740108 -0.00994933 -0.249882 0.249876 -0.0687301 -0.519141 0.519137 -0.219759 -0.84829 0.848285 -0.47206 -1.208837 1.208831 -0.807035 -1.587003 1.586997 -1.212183 -1.973321 1.973313 -1.666942 -2.363032 2.363023 -2.152537 -2.754064 2.754054 -2.654849 171 10 0 0 0 -0.0706978 0.0706959 -0.00906897 -0.241652 0.241646 -0.0644307 -0.507935 0.507935 -0.211543 -0.836886 0.836885 -0.463414 -1.198633 1.198631 -0.802362 -1.577739 1.577734 -1.212312 -1.964657 1.964653 -1.671794 -2.354771 2.354765 -2.162036 -2.746136 2.746128 -2.669494 172 10 0 0 0 -0.070467 0.070467 -0.00901306 -0.240987 0.240987 -0.0640793 -0.506911 0.506911 -0.210721 -0.835752 0.835754 -0.462356 -1.197573 1.197571 -0.801538 -1.576802 1.576797 -1.212133 -1.963662 1.963659 -1.671144 -2.353666 2.353662 -2.160132 -2.744967 2.744965 -2.666622 173 10 0 0 0 -0.0690784 0.0690823 -0.00865674 -0.237413 0.237417 -0.0622478 -0.501641 0.501642 -0.206694 -0.829617 0.82962 -0.456468 -1.191034 1.191038 -0.79431 -1.570532 1.570538 -1.206324 -1.957993 1.957998 -1.670279 -2.348489 2.348492 -2.165293 -2.740094 2.740097 -2.676859 174 10 0 0 0 -0.105324 0 -0.0100749 -0.361776 -7.15256e-007 -0.0728065 -0.758074 -1.19209e-006 -0.239018 -1.240093 -1.66893e-006 -0.51835 -1.763092 -1.90735e-006 -0.886517 -2.30559 -2.38419e-006 -1.322336 -2.855707 -2.38419e-006 -1.800357 -3.408585 -2.14577e-006 -2.301407 -3.962435 -2.86102e-006 -2.813443 175 10 0 0 0 -0.0970879 0 -0.00854874 -0.334938 0 -0.0620407 -0.708857 0 -0.206711 -1.175318 -9.53674e-007 -0.460507 -1.690559 -9.53674e-007 -0.808297 -2.23023 -1.90735e-006 -1.231951 -2.779652 -1.90735e-006 -1.705183 -3.332558 -2.38419e-006 -2.206547 -3.886671 -2.38419e-006 -2.722108 176 10 0 0 0 -0.105852 9.53674e-007 -0.0101786 -0.363354 9.53674e-007 -0.0734571 -0.760677 9.53674e-007 -0.240715 -1.243311 4.76837e-007 -0.521121 -1.766575 0 -0.890028 -2.309137 4.76837e-007 -1.326141 -2.859245 -4.76837e-007 -1.804096 -3.412094 -7.15256e-007 -2.304878 -3.965916 -1.19209e-006 -2.816543 177 10 0 0 0 -0.103188 9.53674e-007 -0.00966692 -0.350712 4.76837e-007 -0.0678548 -0.73204 0 -0.219393 -1.199614 -4.76837e-007 -0.474923 -1.71315 -9.53674e-007 -0.818529 -2.251154 -1.19209e-006 -1.235545 -2.799511 -9.53674e-007 -1.701957 -3.351801 -1.43051e-006 -2.197359 -3.905533 -1.90735e-006 -2.707949 178 10 0 0 0 -0.103329 -7.15256e-007 -0.00969291 -0.351665 0 -0.0682876 -0.736153 -9.53674e-007 -0.222829 -1.208885 -9.53674e-007 -0.486552 -1.726433 -1.43051e-006 -0.840154 -2.266544 -1.90735e-006 -1.265625 -2.815868 -2.6226e-006 -1.738193 -3.36858 -2.86102e-006 -2.237591 -3.922468 -3.33786e-006 -2.750138 179 10 0 0 0 -0.104614 -4.76837e-007 -0.00993824 -0.354431 -7.15256e-007 -0.069275 -0.737318 -1.19209e-006 -0.222288 -1.206369 -9.53674e-007 -0.48013 -1.720999 -1.19209e-006 -0.826405 -2.25938 -1.43051e-006 -1.244892 -2.807817 -1.66893e-006 -1.711802 -3.360088 -2.14577e-006 -2.207048 -3.913771 -2.6226e-006 -2.717037 180 10 0 0 0 -0.0589428 0.0589428 -0.00628829 -0.191895 0.191896 -0.0390732 -0.394087 0.394091 -0.118314 -0.64897 0.648972 -0.251124 -0.940336 0.940337 -0.434047 -1.261002 1.261002 -0.668681 -1.601452 1.601454 -0.947059 -1.953341 1.953341 -1.255646 -2.309835 2.30984 -1.577832 181 10 0 0 0 -0.0592318 0.0592322 -0.00635076 -0.193136 0.193136 -0.0396204 -0.397194 0.397195 -0.120454 -0.654455 0.654457 -0.256156 -0.948032 0.948033 -0.442567 -1.270411 1.270414 -0.680654 -1.612694 1.612696 -0.963579 -1.966195 1.966197 -1.276824 -2.324177 2.324178 -1.603596 182 10 0 0 0 -0.0557041 0.0557013 -0.00561309 -0.181305 0.1813 -0.0347748 -0.37295 0.37295 -0.10537 -0.615593 0.615593 -0.224016 -0.894381 0.894378 -0.388165 -1.203063 1.203062 -0.600106 -1.532501 1.532501 -0.853037 -1.873783 1.873781 -1.13346 -2.219513 2.219511 -1.425222 183 10 0 0 0 -0.0602274 0.0602221 -0.00656605 -0.196069 0.196065 -0.0408392 -0.402342 0.402339 -0.123589 -0.662016 0.662013 -0.262274 -0.958193 0.958189 -0.452863 -1.283314 1.28331 -0.696591 -1.628011 1.628009 -0.985676 -1.983793 1.983791 -1.305693 -2.344124 2.34412 -1.639957 184 10 0 0 0 -0.057375 0.0573711 -0.00595832 -0.18696 0.186957 -0.0370555 -0.384686 0.384681 -0.112556 -0.634579 0.634576 -0.239452 -0.92052 0.920517 -0.414066 -1.235873 1.235872 -0.638343 -1.571703 1.571699 -0.90566 -1.91917 1.91917 -1.202039 -2.2711 2.271097 -1.510748 185 10 0 0 0 -0.0582714 0.0582702 -0.00614548 -0.190269 0.190272 -0.0384481 -0.392084 0.392082 -0.117364 -0.647398 0.647397 -0.250699 -0.93964 0.939642 -0.434998 -1.261316 1.261318 -0.671662 -1.602587 1.602588 -0.952063 -1.954666 1.954669 -1.261207 -2.311358 2.311363 -1.583989 186 10 0 0 0 -0.0857811 -2.86102e-006 -0.00666308 -0.277754 0 -0.040884 -0.567901 -1.90735e-006 -0.122672 -0.931683 -2.86102e-006 -0.258343 -1.345625 -1.90735e-006 -0.443366 -1.799854 0 -0.679047 -2.281268 -1.90735e-006 -0.957329 -2.777925 -2.86102e-006 -1.263935 -3.28035 -1.90735e-006 -1.582403 187 10 0 0 0 -0.0932617 -9.53674e-007 -0.00788355 -0.3031 -9.53674e-007 -0.0490365 -0.619386 -4.76837e-007 -0.14786 -1.013066 -9.53674e-007 -0.311422 -1.457383 -9.53674e-007 -0.533475 -1.939848 0 -0.813586 -2.446488 9.53674e-007 -1.141199 -2.966213 0 -1.500468 -3.49134 0 -1.874686 188 10 0 0 0 -0.088254 2.86102e-006 -0.00705409 -0.286137 3.8147e-006 -0.0434904 -0.585012 4.76837e-006 -0.130735 -0.958954 3.8147e-006 -0.275424 -1.383442 3.8147e-006 -0.472579 -1.847733 5.72205e-006 -0.723051 -2.338379 3.8147e-006 -1.018015 -2.843655 3.8147e-006 -1.34262 -3.354635 4.76837e-006 -1.680137 189 10 0 0 0 -0.0754995 0 -0.00515342 -0.245034 -9.53674e-007 -0.0316529 -0.503824 0 -0.0956085 -0.832419 9.53674e-007 -0.203206 -1.21081 0 -0.351989 -1.631423 1.90735e-006 -0.544602 -2.082069 1.90735e-006 -0.775261 -2.550056 2.86102e-006 -1.031433 -3.023945 1.90735e-006 -1.297038 190 10 0 0 0 -0.0905323 0 -0.00742555 -0.293484 1.90735e-006 -0.0458224 -0.59938 9.53674e-007 -0.137623 -0.980993 9.53674e-007 -0.289429 -1.412976 1.90735e-006 -0.495653 -1.88401 1.90735e-006 -0.756641 -2.380505 9.53674e-007 -1.062927 -2.890993 0 -1.399294 -3.407005 0 -1.748997 191 10 0 0 0 -0.0920401 -2.86102e-006 -0.00767684 -0.298866 -1.90735e-006 -0.0476112 -0.610693 -1.90735e-006 -0.143377 -0.999289 -2.86102e-006 -0.301899 -1.438502 -1.90735e-006 -0.517259 -1.916327 -1.90735e-006 -0.789397 -2.418929 -1.90735e-006 -1.108238 -2.935042 -1.90735e-006 -1.458199 -3.45665 -3.8147e-006 -1.822533 192 10 0 0 0 -0.0643177 0.0643201 -0.00749683 -0.208233 0.208233 -0.0461197 -0.424784 0.424786 -0.138153 -0.694597 0.694596 -0.289916 -0.999735 0.999737 -0.495588 -1.332294 1.332294 -0.755416 -1.682734 1.682731 -1.059917 -2.042982 2.042979 -1.393924 -2.407051 2.40705 -1.740798 193 10 0 0 0 -0.0677156 0.0677178 -0.00831532 -0.219928 0.219931 -0.0517147 -0.448664 0.448666 -0.155606 -0.732289 0.732295 -0.326774 -1.051287 1.051288 -0.558097 -1.396461 1.396461 -0.848412 -1.757929 1.757932 -1.186441 -2.128104 2.128105 -1.556021 -2.501859 2.50186 -1.94053 194 10 0 0 0 -0.062645 0.0626378 -0.00711036 -0.202801 0.202793 -0.0436707 -0.414111 0.414106 -0.130892 -0.678141 0.678138 -0.27509 -0.977589 0.977584 -0.471051 -1.304964 1.30496 -0.719537 -1.650858 1.650852 -1.011733 -2.007017 2.007013 -1.332895 -2.367132 2.367127 -1.666464 195 10 0 0 0 -0.0677481 0.0677471 -0.00832319 -0.219944 0.21994 -0.051712 -0.448551 0.448549 -0.155471 -0.731945 0.731944 -0.326299 -1.050643 1.050641 -0.557039 -1.395503 1.395502 -0.84654 -1.756672 1.756671 -1.183566 -2.12655 2.126549 -1.551975 -2.499998 2.499996 -1.935149 196 10 0 0 0 -0.0704613 0.0704622 -0.00900984 -0.228239 0.22824 -0.055788 -0.463833 0.463833 -0.166783 -0.754011 0.754013 -0.347861 -1.078564 1.078565 -0.590406 -1.428008 1.42801 -0.892157 -1.792628 1.79263 -1.240957 -2.165195 2.165197 -1.620363 -2.540976 2.540977 -2.013989 197 10 0 0 0 -0.0680027 0.0680065 -0.00838709 -0.220036 0.22004 -0.0516784 -0.447601 0.447607 -0.154389 -0.729128 0.729132 -0.322476 -1.04539 1.045394 -0.548491 -1.38763 1.387634 -0.831311 -1.746211 1.746217 -1.159974 -2.113541 2.113547 -1.5186 -2.484343 2.484345 -1.890701 198 10 0 0 0 -0.0908594 3.8147e-006 -0.00748014 -0.294215 1.90735e-006 -0.0460336 -0.600275 2.86102e-006 -0.137942 -0.9816 2.86102e-006 -0.289474 -1.412808 2.86102e-006 -0.494743 -1.882734 2.86102e-006 -0.753965 -2.377926 2.86102e-006 -1.057674 -2.887021 2.86102e-006 -1.390825 -3.40265 1.90735e-006 -1.739571 199 10 0 0 0 -0.0897074 9.53674e-007 -0.00728965 -0.290477 9.53674e-007 -0.0448346 -0.592938 9.53674e-007 -0.134388 -0.970329 9.53674e-007 -0.282245 -1.397706 9.53674e-007 -0.482849 -1.864195 9.53674e-007 -0.736693 -2.356416 1.90735e-006 -1.034643 -2.862865 9.53674e-007 -1.361829 -3.375756 9.53674e-007 -1.703875 200 10 0 0 0 -0.106838 -1.43051e-006 -0.0103704 -0.345401 -4.76837e-007 -0.0641986 -0.698304 0 -0.190684 -1.127964 0 -0.394055 -1.603597 -9.53674e-007 -0.662529 -2.110768 0 -0.991326 -2.636175 -4.76837e-007 -1.366353 -3.170769 -9.53674e-007 -1.77074 -3.709269 -9.53674e-007 -2.189692 201 10 0 0 0 -0.0922794 0 -0.00771713 -0.298686 1.90735e-006 -0.047483 -0.608837 2.86102e-006 -0.142118 -0.994511 2.86102e-006 -0.297801 -1.429881 1.90735e-006 -0.508257 -1.90347 1.90735e-006 -0.773374 -2.401777 1.90735e-006 -1.083305 -2.913584 9.53674e-007 -1.422765 -3.430932 9.53674e-007 -1.775854 202 10 0 0 0 -0.0959187 0 -0.0083437 -0.311259 4.76837e-007 -0.0517755 -0.634403 0 -0.155423 -1.034457 0 -0.325492 -1.483784 -9.53674e-007 -0.554326 -1.96969 -9.53674e-007 -0.840554 -2.479446 -9.53674e-007 -1.175226 -3.002077 -9.53674e-007 -1.542369 -3.530178 -9.53674e-007 -1.925435 203 10 0 0 0 -0.0988359 -2.38419e-006 -0.00886273 -0.320147 -4.76837e-007 -0.0548456 -0.65079 -4.76837e-007 -0.163951 -1.058257 -4.76837e-007 -0.341864 -1.514174 -4.76837e-007 -0.579957 -2.00544 -1.43051e-006 -0.87609 -2.518427 -4.76837e-007 -1.218371 -3.043503 -4.76837e-007 -1.592434 -3.573893 -9.53674e-007 -1.982678 206 10 0 0 0 -0.0733767 0.0733771 -0.00977695 -0.238128 0.238126 -0.0609936 -0.483265 0.483262 -0.182433 -0.783133 0.783131 -0.379103 -1.116195 1.116191 -0.640064 -1.472322 1.472318 -0.961127 -1.842339 1.842337 -1.3301 -2.219616 2.219614 -1.730845 -2.599936 2.599932 -2.147281 208 10 0 0 0 -0.0744858 0.0744781 -0.0100765 -0.241616 0.241611 -0.0628643 -0.489785 0.489778 -0.187761 -0.792596 0.792592 -0.389407 -1.12819 1.128186 -0.656181 -1.486293 1.486288 -0.983335 -1.857631 1.857625 -1.357624 -2.235951 2.235946 -1.763544 -2.617196 2.617188 -2.185129 209 10 0 0 0 -0.0728025 0.0727968 -0.00962281 -0.236534 0.236532 -0.0601795 -0.480608 0.480604 -0.180421 -0.779592 0.779586 -0.375613 -1.111974 1.111968 -0.635047 -1.467783 1.467775 -0.955147 -1.83778 1.837774 -1.324036 -2.215016 2.215009 -1.72458 -2.595279 2.595271 -2.140701 211 10 0 0 0 -0.0952463 0 -0.00822496 -0.310586 9.53674e-007 -0.0516555 -0.635335 -9.53674e-007 -0.156454 -1.039412 -9.53674e-007 -0.330743 -1.494323 -1.43051e-006 -0.567399 -1.986564 -9.53674e-007 -0.865387 -2.501674 -1.90735e-006 -1.212849 -3.028872 -2.38419e-006 -1.593168 -3.56128 -1.90735e-006 -1.990024 213 10 0 0 0 -0.0968122 -1.90735e-006 -0.00850058 -0.31607 -1.43051e-006 -0.0535967 -0.647046 -1.43051e-006 -0.16295 -1.057734 -9.53674e-007 -0.344368 -1.519 -1.90735e-006 -0.590287 -2.01645 -9.53674e-007 -0.898485 -2.535546 -2.86102e-006 -1.256098 -3.065919 -2.38419e-006 -1.646288 -3.601159 -2.86102e-006 -2.052972 216 10 0 0 0 -0.0750628 0.0750628 -0.0102363 -0.249669 0.249665 -0.0681283 -0.513705 0.513703 -0.212335 -0.835949 0.835947 -0.450143 -1.190177 1.190171 -0.765512 -1.563341 1.563337 -1.147466 -1.945895 1.945889 -1.576739 -2.332875 2.33287 -2.036738 -2.721719 2.721713 -2.513244 217 10 0 0 0 -0.073945 0.073945 -0.00993109 -0.243972 0.243974 -0.0646633 -0.500153 0.500153 -0.199047 -0.81385 0.81385 -0.420194 -1.160666 1.160666 -0.714829 -1.528433 1.528433 -1.075102 -1.907304 1.907302 -1.483837 -2.291416 2.291414 -1.922947 -2.677686 2.677683 -2.377365 218 10 0 0 0 -0.0748377 0.0748301 -0.0101731 -0.248626 0.24862 -0.0674945 -0.511334 0.511325 -0.209997 -0.832148 0.83214 -0.444929 -1.185139 1.185127 -0.756677 -1.557391 1.557384 -1.134769 -1.939343 1.939334 -1.560453 -2.325936 2.325927 -2.017341 -2.714492 2.71448 -2.491082 219 10 0 0 0 -0.0753059 0.075304 -0.0103022 -0.249249 0.249246 -0.067732 -0.511248 0.511244 -0.209335 -0.830896 0.830891 -0.441944 -1.182768 1.182764 -0.750491 -1.55423 1.554226 -1.125289 -1.935654 1.93565 -1.547905 -2.321623 2.321617 -2.000025 -2.70945 2.709442 -2.467129 220 10 0 0 0 -0.0757151 0.0757074 -0.0104149 -0.251209 0.251204 -0.0689343 -0.515787 0.515781 -0.213833 -0.83819 0.838186 -0.451975 -1.192439 1.192431 -0.767402 -1.565617 1.565611 -1.149418 -1.948207 1.948199 -1.578911 -2.33501 2.335002 -2.037445 -2.723532 2.723524 -2.510912 221 10 0 0 0 -0.0745411 0.074543 -0.0100929 -0.245993 0.245996 -0.065799 -0.50405 0.504052 -0.20248 -0.819539 0.819541 -0.427011 -1.167826 1.167828 -0.725593 -1.536658 1.53666 -1.089929 -1.916271 1.916271 -1.502546 -2.30093 2.30093 -1.945316 -2.687671 2.687671 -2.403391 222 10 0 0 0 -0.106049 3.57628e-007 -0.0102153 -0.352819 5.96046e-008 -0.0680299 -0.725927 -4.17233e-007 -0.211963 -1.181091 -5.36442e-007 -0.448979 -1.681324 -9.53674e-007 -0.762854 -2.208893 -1.3113e-006 -1.144301 -2.750229 -1.54972e-006 -1.574977 -3.298002 -2.02656e-006 -2.037865 -3.848465 -2.38419e-006 -2.518362 223 10 0 0 0 -0.107077 2.38419e-007 -0.0104162 -0.355125 7.15256e-007 -0.0688665 -0.728928 -1.19209e-007 -0.213431 -1.184353 0 -0.45082 -1.684874 -7.15256e-007 -0.765296 -2.212353 -9.53674e-007 -1.146466 -2.753321 -1.19209e-006 -1.575547 -3.300352 -1.66893e-006 -2.034153 -3.849834 -1.66893e-006 -2.507773 224 10 0 0 0 -0.104649 0 -0.00994492 -0.346405 2.38419e-007 -0.0653057 -0.711178 -4.76837e-007 -0.201836 -1.157246 -9.53674e-007 -0.426235 -1.649914 -1.43051e-006 -0.72504 -2.172018 -1.43051e-006 -1.090731 -2.709898 -2.14577e-006 -1.507255 -3.255156 -2.38419e-006 -1.956335 -3.803532 -2.6226e-006 -2.422859 225 10 0 0 0 -0.102798 0 -0.00959283 -0.33967 -4.76837e-007 -0.0626211 -0.697727 -4.76837e-007 -0.193396 -1.137436 -9.53674e-007 -0.409396 -1.624586 -9.53674e-007 -0.697882 -2.142092 -1.66893e-006 -1.051374 -2.676149 -2.14577e-006 -1.453882 -3.218395 -2.38419e-006 -1.888565 -3.764231 -2.6226e-006 -2.34065 226 10 0 0 0 -0.105957 -2.38419e-007 -0.0101981 -0.35004 -2.38419e-007 -0.0666885 -0.718031 -4.76837e-007 -0.206039 -1.168177 -2.38419e-007 -0.436004 -1.664682 -9.53674e-007 -0.742313 -2.189541 -1.19209e-006 -1.115753 -2.728905 -1.19209e-006 -1.538157 -3.274908 -1.66893e-006 -1.991137 -3.823635 -1.90735e-006 -2.459867 227 10 0 0 0 -0.106272 1.17347e-007 -0.0102594 -0.35421 1.78814e-007 -0.0686543 -0.72945 0 -0.214531 -1.186975 -4.17233e-007 -0.454946 -1.689268 -7.45058e-007 -0.773131 -2.218342 -9.83477e-007 -1.159208 -2.760717 -1.46031e-006 -1.594477 -3.309208 -1.93715e-006 -2.061695 -3.860224 -2.23517e-006 -2.546363 228 10 0 0 0 -0.0781975 0.0781956 -0.0111173 -0.2641 0.264101 -0.0772647 -0.545852 0.545849 -0.245683 -0.88554 0.885538 -0.522189 -1.253524 1.253521 -0.883269 -1.635754 1.635754 -1.3106 -2.023837 2.02383 -1.779971 -2.414157 2.414154 -2.272771 -2.805319 2.805315 -2.77704 229 10 0 0 0 -0.0764303 0.0764265 -0.0106141 -0.254468 0.254464 -0.0709472 -0.523363 0.523359 -0.221491 -0.851086 0.851078 -0.470718 -1.210318 1.210314 -0.801453 -1.58732 1.587314 -1.20085 -1.972627 1.972621 -1.648148 -2.361431 2.361425 -2.124325 -2.751612 2.751606 -2.615305 230 10 0 0 0 -0.0770426 0.0770473 -0.0107899 -0.260942 0.260941 -0.0754222 -0.540998 0.541 -0.24139 -0.879846 0.879843 -0.515863 -1.24759 1.247591 -0.876047 -1.629862 1.629864 -1.303638 -2.018146 2.018145 -1.774861 -2.408624 2.408623 -2.269656 -2.79992 2.799919 -2.776043 231 10 0 0 0 -0.0779839 0.0779858 -0.0110564 -0.263228 0.263227 -0.076701 -0.54405 0.544049 -0.243776 -0.882946 0.882943 -0.518365 -1.250402 1.250401 -0.877469 -1.632353 1.63235 -1.303123 -2.020333 2.020333 -1.771588 -2.410613 2.410613 -2.263835 -2.801747 2.801744 -2.767727 232 10 0 0 0 -0.0776806 0.0776825 -0.0109707 -0.2631 0.263102 -0.0767472 -0.544938 0.544941 -0.245297 -0.885119 0.88512 -0.523008 -1.253639 1.253641 -0.886139 -1.636271 1.636271 -1.315878 -2.024606 2.024607 -1.787566 -2.415091 2.415087 -2.282397 -2.806372 2.806372 -2.788551 233 10 0 0 0 -0.0765648 0.0765648 -0.0106544 -0.2596 0.259602 -0.0746422 -0.538898 0.538898 -0.239522 -0.877341 0.877342 -0.513026 -1.244946 1.244946 -0.87268 -1.627205 1.627205 -1.3002 -2.015553 2.015554 -1.77198 -2.406091 2.40609 -2.267493 -2.797438 2.797436 -2.774662 234 10 0 0 0 -0.117107 0 -0.0124826 -0.393978 -2.38419e-007 -0.0863907 -0.80682 -3.57628e-007 -0.270188 -1.295994 -7.15256e-007 -0.562401 -1.819336 -8.34465e-007 -0.931531 -2.359865 -1.54972e-006 -1.358748 -2.908041 -1.66893e-006 -1.824057 -3.459377 -1.66893e-006 -2.311279 -4.011925 -2.14577e-006 -2.809111 235 10 0 0 0 -0.117199 1.19209e-007 -0.012502 -0.391485 -1.78814e-007 -0.0849323 -0.799095 -6.55651e-007 -0.263 -1.283178 -9.83477e-007 -0.545963 -1.802948 -1.43051e-006 -0.90535 -2.341461 -1.57952e-006 -1.324359 -2.888521 -1.87755e-006 -1.783123 -3.439173 -2.23517e-006 -2.265018 -3.991196 -2.35438e-006 -2.758025 236 10 0 0 0 -0.113594 -2.38419e-007 -0.0117368 -0.384184 -2.38419e-007 -0.082088 -0.791943 -7.15256e-007 -0.260319 -1.279295 -1.19209e-006 -0.549172 -1.80345 -1.43051e-006 -0.920589 -2.345167 -1.66893e-006 -1.352928 -2.894255 -2.14577e-006 -1.823949 -3.44627 -2.38419e-006 -2.316892 -3.99938 -2.86102e-006 -2.820385 237 10 0 0 0 -0.111103 2.38419e-007 -0.0112234 -0.377987 -2.38419e-007 -0.0795267 -0.783392 -9.53674e-007 -0.255229 -1.270153 -7.15256e-007 -0.543005 -1.79471 -1.19209e-006 -0.915579 -2.337154 -1.43051e-006 -1.35114 -2.886793 -1.66893e-006 -1.82583 -3.439198 -1.90735e-006 -2.32229 -3.99262 -2.6226e-006 -2.82921 238 10 0 0 0 -0.114786 1.19209e-007 -0.0119869 -0.386232 -3.57628e-007 -0.0828167 -0.793287 -3.57628e-007 -0.260288 -1.279144 -7.15256e-007 -0.546428 -1.801977 -9.53674e-007 -0.914135 -2.342628 -1.3113e-006 -1.341872 -2.891066 -1.66893e-006 -1.808784 -3.442669 -1.90735e-006 -2.298175 -3.995461 -2.02656e-006 -2.798405 239 10 0 0 0 -0.111835 2.08616e-007 -0.0113723 -0.372705 1.49012e-007 -0.0764294 -0.763939 -8.9407e-008 -0.237552 -1.235315 -4.47035e-007 -0.499089 -1.747835 -6.55651e-007 -0.840247 -2.282928 -9.53674e-007 -1.246415 -2.82844 -1.40071e-006 -1.696803 -3.378336 -1.72853e-006 -2.173267 -3.929926 -1.96695e-006 -2.662575 240 10 0 0 0 -0.0569229 0.0569224 -0.00586295 -0.184444 0.184443 -0.0359478 -0.378086 0.378085 -0.108133 -0.622419 0.622417 -0.228664 -0.902142 0.902139 -0.394152 -1.211161 1.211159 -0.606699 -1.540541 1.540541 -0.859506 -1.881527 1.881529 -1.139201 -2.226706 2.226704 -1.429527 241 10 0 0 0 -0.0567379 0.0567408 -0.00582814 -0.183834 0.183837 -0.0357065 -0.376844 0.376846 -0.107384 -0.620417 0.620418 -0.227066 -0.899326 0.899328 -0.391394 -1.207546 1.207545 -0.602498 -1.536167 1.536166 -0.853652 -1.876421 1.876419 -1.131546 -2.220863 2.220864 -1.419968 242 10 0 0 0 -0.052042 0.0520372 -0.00489497 -0.168556 0.168555 -0.0298929 -0.346251 0.346248 -0.089978 -0.571959 0.571956 -0.190831 -0.832098 0.832094 -0.330096 -1.12184 1.121834 -0.510495 -1.432915 1.432911 -0.726778 -1.756367 1.756361 -0.967054 -2.083849 2.083844 -1.21577 243 10 0 0 0 -0.0553131 0.0553136 -0.00553536 -0.17915 0.179153 -0.0338626 -0.367434 0.367432 -0.101828 -0.605497 0.605501 -0.215467 -0.878639 0.878644 -0.371739 -1.181248 1.181252 -0.573036 -1.504629 1.504633 -0.81317 -1.839973 1.839978 -1.079361 -2.179644 2.179648 -1.355832 244 10 0 0 0 -0.0612278 0.0612264 -0.00678968 -0.198235 0.198241 -0.0416749 -0.405174 0.405179 -0.125009 -0.664375 0.664376 -0.2631 -0.959015 0.959018 -0.451215 -1.281986 1.281989 -0.690509 -1.62398 1.623984 -0.972709 -1.976618 1.976618 -1.283443 -2.333328 2.333329 -1.606284 245 10 0 0 0 -0.0534439 0.0534463 -0.00516582 -0.173101 0.173102 -0.0315642 -0.355347 0.355347 -0.0949676 -0.58639 0.58639 -0.201211 -0.852167 0.85217 -0.347662 -1.147528 1.147528 -0.536928 -1.464008 1.464006 -0.763361 -1.7927 1.7927 -1.014672 -2.12558 2.125579 -1.275226 246 10 0 0 0 -0.0700855 -3.8147e-006 -0.00443959 -0.227061 -3.8147e-006 -0.0270786 -0.467085 -3.8147e-006 -0.0816038 -0.773088 -1.90735e-006 -0.173473 -1.12708 -1.90735e-006 -0.300855 -1.523156 -3.8147e-006 -0.466839 -1.950212 -3.8147e-006 -0.667056 -2.396374 -3.8147e-006 -0.891581 -2.85029 -3.8147e-006 -1.126817 247 10 0 0 0 -0.0736732 3.8147e-006 -0.00490642 -0.239511 2.86102e-006 -0.0302362 -0.493597 4.76837e-006 -0.0917451 -0.818069 6.67572e-006 -0.196347 -1.194502 3.8147e-006 -0.343318 -1.616571 3.8147e-006 -0.537627 -2.071959 3.8147e-006 -0.774962 -2.547531 5.72205e-006 -1.043331 -3.031395 5.72205e-006 -1.325914 248 10 0 0 0 -0.0711174 9.53674e-007 -0.00457168 -0.230396 1.90735e-006 -0.0278926 -0.473774 1.90735e-006 -0.084039 -0.783739 1.90735e-006 -0.17855 -1.14195 0 -0.309454 -1.542246 1.90735e-006 -0.479776 -1.973358 0 -0.684926 -2.422979 0 -0.914166 -2.880547 0 -1.154645 249 10 0 0 0 -0.0738592 1.90735e-006 -0.00493026 -0.240134 1.90735e-006 -0.0303969 -0.494885 -9.53674e-007 -0.0922477 -0.820262 1.90735e-006 -0.197501 -1.197678 -1.90735e-006 -0.345382 -1.620771 0 -0.540891 -2.077097 -1.90735e-006 -0.779576 -2.55353 -1.90735e-006 -1.049379 -3.038235 -1.90735e-006 -1.33345 250 10 0 0 0 -0.0756664 -4.76837e-006 -0.00517583 -0.245145 -4.76837e-006 -0.0316586 -0.503358 -3.8147e-006 -0.0953104 -0.830782 -2.86102e-006 -0.202052 -1.207499 -3.8147e-006 -0.349286 -1.626217 -1.90735e-006 -0.539716 -2.074984 -3.8147e-006 -0.767779 -2.541274 -4.76837e-006 -1.021317 -3.015759 -2.86102e-006 -1.287897 251 10 0 0 0 -0.0809774 -1.90735e-006 -0.00593066 -0.2624 -3.8147e-006 -0.0363894 -0.537838 -1.90735e-006 -0.109474 -0.885206 -9.53674e-007 -0.231467 -1.282677 -1.90735e-006 -0.398874 -1.721566 -1.90735e-006 -0.613815 -2.189239 -1.90735e-006 -0.869495 -2.674069 -2.86102e-006 -1.153795 -3.166769 -3.8147e-006 -1.452661 252 10 0 0 0 -0.0599442 0.0599437 -0.00650501 -0.194817 0.194818 -0.0402763 -0.399618 0.399617 -0.121746 -0.6572 0.657202 -0.257847 -0.951342 0.951341 -0.445154 -1.27537 1.27537 -0.68662 -1.619574 1.619574 -0.974428 -1.975561 1.975562 -1.295064 -2.336857 2.336859 -1.632511 253 10 0 0 0 -0.0590534 0.0590463 -0.00631213 -0.192038 0.192033 -0.0391135 -0.39432 0.394309 -0.118426 -0.649288 0.64928 -0.251344 -0.941502 0.94149 -0.435591 -1.263813 1.263807 -0.673548 -1.606907 1.606899 -0.958518 -1.962048 1.962039 -1.276606 -2.322529 2.322519 -1.611366 254 10 0 0 0 -0.0575523 0.0575485 -0.0059936 -0.186932 0.186933 -0.0369887 -0.383825 0.383827 -0.111806 -0.632427 0.632425 -0.2372 -0.917109 0.917106 -0.409932 -1.232447 1.232445 -0.634177 -1.569044 1.569041 -0.903296 -1.918203 1.918201 -1.204257 -2.27313 2.273128 -1.521711 255 10 0 0 0 -0.0571995 0.0571942 -0.00591969 -0.185715 0.185712 -0.0364895 -0.381275 0.38127 -0.110221 -0.628262 0.628256 -0.233761 -0.911127 0.911117 -0.403806 -1.2244 1.224391 -0.624158 -1.559145 1.559134 -0.888972 -1.90645 1.906443 -1.184916 -2.259546 2.259538 -1.496982 256 10 0 0 0 -0.0586777 0.0586767 -0.0062325 -0.190304 0.1903 -0.0383449 -0.390011 0.39001 -0.115493 -0.641434 0.64143 -0.24418 -0.928541 0.928541 -0.420552 -1.244764 1.244761 -0.646486 -1.581528 1.581526 -0.915998 -1.930632 1.930628 -1.2168 -2.285017 2.285014 -1.532647 257 10 0 0 0 -0.0632973 0.0632954 -0.00725842 -0.205702 0.205706 -0.045047 -0.421013 0.421013 -0.135922 -0.690105 0.69011 -0.286735 -0.995163 0.995162 -0.492254 -1.328888 1.328887 -0.754726 -1.681305 1.681307 -1.064841 -2.044104 2.044105 -1.407324 -2.411606 2.411607 -1.766608 258 10 0 0 0 -0.0817108 1.90735e-006 -0.00604177 -0.267628 4.76837e-006 -0.0380731 -0.553244 3.8147e-006 -0.117114 -0.915953 3.8147e-006 -0.251857 -1.332365 4.76837e-006 -0.439658 -1.791891 3.8147e-006 -0.683 -2.280405 3.8147e-006 -0.97399 -2.785185 3.8147e-006 -1.297513 -3.296528 3.8147e-006 -1.635881 259 10 0 0 0 -0.0874004 0 -0.00691843 -0.286266 9.53674e-007 -0.043731 -0.59012 0 -0.134191 -0.972786 9.53674e-007 -0.286993 -1.408365 0 -0.497715 -1.884567 0 -0.767132 -2.386908 -9.53674e-007 -1.085418 -2.903591 0 -1.436823 -3.426302 0 -1.804186 260 10 0 0 0 -0.0816498 9.53674e-007 -0.00603271 -0.265978 3.8147e-006 -0.0375013 -0.548061 2.86102e-006 -0.114444 -0.906321 2.86102e-006 -0.24539 -1.318096 2.86102e-006 -0.428007 -1.773314 3.8147e-006 -0.665098 -2.258125 3.8147e-006 -0.949362 -2.760019 3.8147e-006 -1.266706 -3.269274 1.90735e-006 -1.600221 261 10 0 0 0 -0.0806618 0 -0.00588727 -0.263992 -9.53674e-007 -0.0370083 -0.545698 0 -0.113727 -0.903879 0 -0.244607 -1.315725 0 -0.427303 -1.771099 9.53674e-007 -0.664618 -2.256096 0 -0.949215 -2.75798 0 -1.266537 -3.266804 0 -1.599068 262 10 0 0 0 -0.0755424 2.86102e-006 -0.00516009 -0.246078 3.8147e-006 -0.0319815 -0.508751 3.8147e-006 -0.0980017 -0.845018 2.86102e-006 -0.211329 -1.234655 2.86102e-006 -0.370873 -1.669476 2.86102e-006 -0.580637 -2.136465 2.86102e-006 -0.835256 -2.622492 1.90735e-006 -1.121702 -3.116312 2.86102e-006 -1.422732 263 10 0 0 0 -0.0793905 -1.90735e-006 -0.00570226 -0.258412 0 -0.0353341 -0.532082 0 -0.107417 -0.880177 0 -0.229992 -1.281685 -9.53674e-007 -0.401576 -1.727514 -1.90735e-006 -0.625654 -2.204195 -1.90735e-006 -0.895869 -2.698984 -1.90735e-006 -1.198783 -3.201553 -1.90735e-006 -1.517552 266 10 0 0 0 -0.0626945 0.0626945 -0.00712085 -0.205669 0.205667 -0.0452187 -0.424122 0.424124 -0.139044 -0.698784 0.698784 -0.2974 -1.01074 1.010739 -0.515305 -1.350893 1.350892 -0.792951 -1.708879 1.708878 -1.119745 -2.076431 2.076429 -1.479194 -2.44788 2.447876 -1.853925 268 10 0 0 0 -0.0662251 0.0662274 -0.00795078 -0.216185 0.216188 -0.0500214 -0.443501 0.443501 -0.152481 -0.727167 0.727167 -0.323708 -1.047289 1.04729 -0.557262 -1.394146 1.394147 -0.852001 -1.757423 1.757424 -1.196132 -2.129423 2.129424 -1.573145 -2.505196 2.505194 -1.966736 269 10 0 0 0 -0.0655022 0.0655003 -0.00777674 -0.214571 0.214572 -0.0493307 -0.441147 0.441146 -0.151046 -0.72418 0.724179 -0.321341 -1.043756 1.043756 -0.55381 -1.39016 1.390157 -0.847345 -1.753067 1.753066 -1.190211 -2.124783 2.12478 -1.566048 -2.500263 2.500259 -1.958291 270 10 0 0 0 -0.0934792 -4.76837e-007 -0.00792003 -0.306858 0 -0.0505309 -0.630651 -4.76837e-007 -0.154644 -1.033449 -9.53674e-007 -0.327582 -1.487263 -9.53674e-007 -0.562678 -1.978121 -1.90735e-006 -0.858037 -2.492054 -1.90735e-006 -1.202613 -3.018234 -2.86102e-006 -1.579905 -3.549627 -1.90735e-006 -1.973396 276 10 0 0 0 -0.0739555 0.0739536 -0.00993365 -0.244175 0.24417 -0.0647964 -0.500961 0.50096 -0.199922 -0.815426 0.815424 -0.422507 -1.163 1.162999 -0.719174 -1.531425 1.531424 -1.081924 -1.910852 1.910848 -1.493575 -2.295462 2.295458 -1.936028 -2.682215 2.682208 -2.394182 277 10 0 0 0 -0.0757885 0.0757895 -0.0104364 -0.25075 0.250753 -0.0685809 -0.513823 0.513824 -0.211549 -0.834173 0.834174 -0.445556 -1.186417 1.186419 -0.755159 -1.558041 1.558041 -1.13061 -1.939554 1.939552 -1.553729 -2.325588 2.325588 -2.006343 -2.713455 2.713454 -2.47382 278 10 0 0 0 -0.0733109 0.0733061 -0.00975865 -0.241905 0.241904 -0.0635254 -0.495754 0.495749 -0.195091 -0.80722 0.807217 -0.412092 -1.152534 1.152528 -0.702764 -1.51959 1.519587 -1.060383 -1.898266 1.898261 -1.468135 -2.282475 2.282469 -1.907866 -2.66901 2.669002 -2.364269 279 10 0 0 0 -0.0727577 0.0727549 -0.00961053 -0.239578 0.23958 -0.0621927 -0.49114 0.491141 -0.191046 -0.800459 0.800458 -0.404126 -1.143826 1.143821 -0.689793 -1.509247 1.509245 -1.041438 -1.886648 1.886643 -1.44278 -2.269819 2.269814 -1.875872 -2.655416 2.655412 -2.325269 280 10 0 0 0 -0.0715141 0.0715122 -0.00928092 -0.235874 0.235875 -0.0602474 -0.484421 0.484423 -0.185579 -0.79039 0.790388 -0.392696 -1.130232 1.130232 -0.669588 -1.492933 1.492931 -1.011737 -1.868347 1.868344 -1.403669 -2.250229 2.250228 -1.828963 -2.634956 2.634953 -2.272207 281 10 0 0 0 -0.0671253 0.0671272 -0.00817037 -0.220596 0.220595 -0.0523198 -0.453417 0.453416 -0.160403 -0.742418 0.742419 -0.339666 -1.067127 1.067126 -0.58254 -1.417452 1.417452 -0.886719 -1.783398 1.783393 -1.240248 -2.157629 2.157627 -1.626857 -2.535425 2.53542 -2.030162 282 10 0 0 0 -0.097271 1.43051e-006 -0.00858307 -0.31967 0 -0.0550396 -0.65654 0 -0.168826 -1.073935 4.76837e-007 -0.357744 -1.54121 0 -0.612806 -2.043285 0 -0.930536 -2.565788 -9.53674e-007 -1.297323 -3.098757 -1.90735e-006 -1.696072 -3.636108 -1.43051e-006 -2.110583 283 10 0 0 0 -0.107317 0 -0.010464 -0.350908 4.76837e-007 -0.0667142 -0.713549 9.53674e-007 -0.201397 -1.15432 4.76837e-007 -0.418775 -1.64015 -9.53674e-007 -0.704865 -2.15542 -4.76837e-007 -1.052718 -2.686958 -9.53674e-007 -1.446686 -3.226423 -9.53674e-007 -1.869498 -3.769062 -1.43051e-006 -2.305941 284 10 0 0 0 -0.0962391 9.53674e-007 -0.00839829 -0.31571 9.53674e-007 -0.0535874 -0.648136 -4.76837e-007 -0.164019 -1.060947 -1.43051e-006 -0.347785 -1.52434 -1.90735e-006 -0.596894 -2.023466 -1.90735e-006 -0.90849 -2.543818 -2.86102e-006 -1.269433 -3.075159 -2.38419e-006 -1.66276 -3.611147 -2.86102e-006 -2.072169 285 10 0 0 0 -0.11253 -2.38419e-007 -0.0115157 -0.368927 0 -0.0742201 -0.748383 -4.76837e-007 -0.223997 -1.204979 -4.76837e-007 -0.463071 -1.703703 -1.19209e-006 -0.773848 -2.228304 -9.53674e-007 -1.146549 -2.766397 -1.90735e-006 -1.563908 -3.31078 -2.14577e-006 -2.008603 -3.857655 -2.6226e-006 -2.466339 286 10 0 0 0 -0.108873 9.53674e-007 -0.0107716 -0.359425 2.38419e-007 -0.0704781 -0.734461 4.76837e-007 -0.216167 -1.189398 0 -0.452859 -1.6884 -4.76837e-007 -0.764207 -2.214136 -9.53674e-007 -1.140202 -2.753548 -9.53674e-007 -1.562802 -3.299231 -1.43051e-006 -2.014087 -3.847435 -1.90735e-006 -2.479572 287 10 0 0 0 -0.112392 -4.76837e-007 -0.0114883 -0.368889 -9.53674e-007 -0.0742445 -0.748888 -9.53674e-007 -0.22453 -1.206263 -9.53674e-007 -0.46473 -1.705772 -1.43051e-006 -0.777112 -2.231016 -1.66893e-006 -1.15167 -2.769602 -1.90735e-006 -1.570959 -3.314379 -2.14577e-006 -2.017603 -3.86161 -2.38419e-006 -2.477347 288 10 0 0 0 -0.0845833 0.0845795 -0.0130295 -0.279957 0.279951 -0.0866075 -0.566246 0.566243 -0.261749 -0.904238 0.904231 -0.534165 -1.267935 1.267928 -0.879736 -1.646074 1.646068 -1.28475 -2.031162 2.031156 -1.730493 -2.419292 2.419286 -2.200319 -2.808607 2.8086 -2.68162 289 10 0 0 0 -0.0810452 0.0810499 -0.0119535 -0.268696 0.268697 -0.0794317 -0.54759 0.547594 -0.243741 -0.881454 0.881461 -0.506532 -1.243315 1.243316 -0.845844 -1.62081 1.620813 -1.247657 -2.005749 2.005753 -1.692387 -2.393925 2.393929 -2.162616 -2.783363 2.783363 -2.645206 290 10 0 0 0 -0.084341 0.084341 -0.0129547 -0.279858 0.279858 -0.0866516 -0.567421 0.567417 -0.263705 -0.906826 0.906822 -0.539532 -1.271639 1.271635 -0.889011 -1.65053 1.650522 -1.297843 -2.0361 2.036097 -1.747077 -2.42458 2.424576 -2.220106 -2.814178 2.814175 -2.704492 291 10 0 0 0 -0.0796967 0.0796995 -0.0115532 -0.259398 0.259412 -0.0730909 -0.524099 0.524109 -0.218151 -0.842621 0.842627 -0.44854 -1.191174 1.191175 -0.74784 -1.558834 1.558837 -1.107702 -1.93681 1.936816 -1.511903 -2.319735 2.319734 -1.943424 -2.704613 2.704612 -2.387719 292 10 0 0 0 -0.0800209 0.0800095 -0.0116457 -0.260923 0.260916 -0.0740606 -0.527485 0.527478 -0.22153 -0.847927 0.847923 -0.455721 -1.198156 1.198153 -0.759657 -1.567118 1.56711 -1.124468 -1.946064 1.946056 -1.533585 -2.329748 2.329741 -1.969916 -2.715309 2.715298 -2.41903 293 10 0 0 0 -0.0855179 0.0855064 -0.0133212 -0.282972 0.282961 -0.0886029 -0.569542 0.56953 -0.264161 -0.90716 0.907148 -0.535681 -1.270374 1.270363 -0.8796 -1.648108 1.648092 -1.282585 -2.032879 2.032867 -1.726159 -2.420763 2.420752 -2.193775 -2.809877 2.809862 -2.672986 294 10 0 0 0 -0.113989 -4.76837e-007 -0.0118195 -0.370745 -2.38419e-007 -0.0747098 -0.748108 -7.15256e-007 -0.222541 -1.201168 -9.53674e-007 -0.45657 -1.696091 -1.43051e-006 -0.759752 -2.217421 -1.90735e-006 -1.123326 -2.752901 -2.14577e-006 -1.530879 -3.29512 -1.90735e-006 -1.965437 -3.840008 -2.86102e-006 -2.41264 295 10 0 0 0 -0.115757 -9.53674e-007 -0.0121942 -0.376154 -4.76837e-007 -0.0769999 -0.757698 -7.15256e-007 -0.228741 -1.214256 -1.43051e-006 -0.467761 -1.711737 -1.19209e-006 -0.776018 -2.234724 -1.66893e-006 -1.14415 -2.77125 -2.14577e-006 -1.555548 -3.314367 -2.38419e-006 -1.994208 -3.860388 -2.38419e-006 -2.447276 296 10 0 0 0 -0.11817 -2.98023e-007 -0.0127137 -0.393126 -2.98023e-007 -0.0855258 -0.798679 -6.55651e-007 -0.261386 -1.278845 -8.9407e-007 -0.537504 -1.795288 -1.07288e-006 -0.888295 -2.331486 -1.37091e-006 -1.298471 -2.877205 -1.54972e-006 -1.749949 -3.427114 -1.96695e-006 -2.226516 -3.978691 -1.96695e-006 -2.715694 297 10 0 0 0 -0.118563 -1.49012e-007 -0.0127988 -0.394156 -2.68221e-007 -0.0859737 -0.800131 -4.17233e-007 -0.262285 -1.28042 -6.85453e-007 -0.538621 -1.796848 -1.04308e-006 -0.889373 -2.332993 -1.2219e-006 -1.299366 -2.878597 -1.40071e-006 -1.750237 -3.428425 -1.63913e-006 -2.226233 -3.979935 -1.81794e-006 -2.714871 298 10 0 0 0 -0.118155 -5.96046e-008 -0.012709 -0.392349 -4.17233e-007 -0.0850883 -0.796402 -6.55651e-007 -0.259351 -1.275061 -9.53674e-007 -0.5329 -1.790331 -1.16229e-006 -0.880755 -2.325752 -1.37091e-006 -1.288097 -2.871037 -1.66893e-006 -1.737314 -3.420652 -1.84774e-006 -2.211841 -3.971991 -2.08616e-006 -2.699081 299 10 0 0 0 -0.116087 2.38419e-007 -0.0122629 -0.376976 2.38419e-007 -0.0773299 -0.759012 0 -0.229536 -1.215994 -4.76837e-007 -0.469164 -1.713821 -9.53674e-007 -0.778126 -2.23708 -1.43051e-006 -1.147018 -2.773813 -1.43051e-006 -1.559186 -3.316923 -1.90735e-006 -1.997813 -3.862926 -1.43051e-006 -2.450797' )
		self.hairShape.boundingBox()
		self.editGuidesShape.boundingBox()
		pm.mel.OxEditGuides( '-eg 300 24 10 0 0 0 -0.0585606 0.473458 -0.270869 -0.273265 0.715077 -0.374578 -0.592018 1.033832 -0.605428 -0.957896 1.399709 -0.958724 -1.341359 1.783174 -1.393633 -1.731361 2.173174 -1.882553 -2.123319 2.565132 -2.400981 -2.515836 2.957648 -2.934068 -2.908529 3.350341 -3.474614 26 10 0 0 0 -0.064538 0.517939 -0.365238 -0.272479 0.786003 -0.480859 -0.592295 1.105819 -0.713801 -0.958474 1.471997 -1.06819 -1.341941 1.855467 -1.503137 -1.731895 2.24542 -1.991513 -2.12382 2.637344 -2.509209 -2.516332 3.029858 -3.042189 -2.909024 3.422548 -3.582599 27 10 0 0 0 -0.0598445 0.491935 -0.304439 -0.275182 0.733601 -0.408475 -0.594198 1.052616 -0.639839 -0.960145 1.418564 -0.993386 -1.343607 1.802027 -1.428303 -1.733599 2.192019 -1.917116 -2.125553 2.583973 -2.435424 -2.518064 2.976483 -2.968397 -2.910756 3.369174 -3.508833 30 10 0 0 0 -0.3761 0.284009 -0.261397 -0.710079 0.302138 -0.373363 -1.168798 0.302138 -0.61552 -1.689186 0.302138 -0.976555 -2.232285 0.302139 -1.415129 -2.784142 0.302139 -1.906684 -3.338638 0.302138 -2.427933 -3.893846 0.302139 -2.963851 -4.44925 0.302138 -3.506441 31 10 0 0 0 -0.257886 0.136664 -0.0828401 -0.593655 0.136666 -0.195788 -1.057922 0.136665 -0.446224 -1.579891 0.136666 -0.811542 -2.123037 0.136666 -1.25033 -2.67466 0.136666 -1.73989 -3.228979 0.136666 -2.258416 -3.784099 0.136666 -2.791997 -4.339454 0.136666 -3.332618 32 10 0 0 0 -0.416325 0.315284 -0.36605 -0.763965 0.352462 -0.489858 -1.224834 0.352462 -0.735186 -1.745863 0.352462 -1.097946 -2.289433 0.352463 -1.538724 -2.841467 0.352463 -2.031824 -3.396001 0.352463 -2.553709 -3.951212 0.352463 -3.089704 -4.506612 0.352462 -3.632066 33 10 0 0 0 -0.330059 0.227647 -0.171004 -0.663949 0.227648 -0.282532 -1.128862 0.227648 -0.533954 -1.653008 0.227648 -0.905353 -2.198133 0.227648 -1.353755 -2.750708 0.227648 -1.851829 -3.30541 0.227648 -2.376613 -3.860677 0.227648 -2.914237 -4.416095 0.227648 -3.457508 34 10 0 0 0 -0.230227 0.116997 -0.0636728 -0.548388 0.116999 -0.1638 -1.002985 0.116999 -0.400005 -1.522368 0.117 -0.758371 -2.065444 0.117 -1.19683 -2.617212 0.117001 -1.687633 -3.171601 0.117001 -2.207217 -3.726763 0.117001 -2.741886 -4.282151 0.117001 -3.283764 84 10 0 0 0 -0.0721583 0.403121 -0.180146 -0.308619 0.639581 -0.292065 -0.637913 0.968877 -0.544679 -1.008876 1.339841 -0.917439 -1.394463 1.725427 -1.36675 -1.785227 2.116193 -1.86531 -2.17747 2.508433 -2.390303 -2.570104 2.901067 -2.927982 -2.962845 3.293808 -3.471213 87 10 0 0 0 -0.0740027 0.381399 -0.158442 -0.310737 0.618135 -0.27065 -0.640289 0.947688 -0.523826 -1.011391 1.318789 -0.897155 -1.397037 1.704435 -1.346884 -1.787819 2.095218 -1.845703 -2.180068 2.487467 -2.370844 -2.572701 2.8801 -2.908607 -2.965444 3.27284 -3.451885 88 10 0 0 0 -0.101103 0.196347 -0.0457827 -0.342955 0.4382 -0.163552 -0.673936 0.76918 -0.419866 -1.044366 1.13961 -0.790479 -1.429142 1.524387 -1.234061 -1.819424 1.914669 -1.726362 -2.211441 2.306687 -2.246027 -2.603963 2.69921 -2.779374 -2.99663 3.091875 -3.318575 89 10 0 0 0 -0.0741558 0.399665 -0.176862 -0.314009 0.639519 -0.292436 -0.646557 0.972068 -0.552247 -1.01907 1.344579 -0.931422 -1.405221 1.730731 -1.384921 -1.79614 2.12165 -1.88565 -2.188416 2.513925 -2.411519 -2.581053 2.906564 -2.949366 -2.973788 3.299296 -3.492226 90 10 0 0 0 -0.411385 0.299963 -0.333232 -0.770744 0.317958 -0.465492 -1.247573 0.317958 -0.735956 -1.77649 0.317958 -1.121546 -2.322978 0.317958 -1.577138 -2.875741 0.317957 -2.077064 -3.430365 0.317957 -2.600468 -3.98553 0.317957 -3.135201 -4.540857 0.317957 -3.674873 91 10 0 0 0 -0.296054 0.16739 -0.116267 -0.656286 0.16739 -0.248886 -1.140471 0.16739 -0.532033 -1.672845 0.16739 -0.928775 -2.220597 0.16739 -1.391542 -2.773645 0.16739 -1.894352 -3.328323 0.167389 -2.418687 -3.883498 0.167389 -2.953644 -4.438824 0.167389 -3.493207 93 10 0 0 0 -0.392483 0.281095 -0.280627 -0.755169 0.299574 -0.415754 -1.235082 0.299574 -0.691442 -1.76571 0.299574 -1.082449 -2.31293 0.299574 -1.542124 -2.865939 0.299573 -2.044542 -3.420628 0.299573 -2.569056 -3.975816 0.299573 -3.104416 -4.531156 0.299572 -3.644494 94 10 0 0 0 -0.392012 0.279201 -0.278037 -0.747095 0.279201 -0.406325 -1.225581 0.279201 -0.679581 -1.754524 0.279201 -1.065253 -2.300682 0.279201 -1.519052 -2.853111 0.279201 -2.015737 -3.407545 0.2792 -2.536007 -3.962605 0.2792 -3.068107 -4.517859 0.2792 -3.605392 95 10 0 0 0 -0.416668 0.299828 -0.343108 -0.780802 0.313559 -0.479306 -1.261093 0.313559 -0.755645 -1.79085 0.313559 -1.14387 -2.337143 0.313559 -1.5984 -2.889535 0.313559 -2.094745 -3.443933 0.313558 -2.614438 -3.998967 0.313558 -3.145917 -4.554203 0.313558 -3.682563 145 10 0 0 0 -0.101463 0.291237 -0.0934622 -0.3627 0.552472 -0.234104 -0.70738 0.89715 -0.523137 -1.08322 1.272993 -0.917052 -1.469536 1.659307 -1.371795 -1.860021 2.04979 -1.866608 -2.251941 2.441712 -2.384244 -2.64435 2.83412 -2.913876 -3.036925 3.226695 -3.449139 148 10 0 0 0 -0.091794 0.336051 -0.12279 -0.345206 0.589464 -0.253839 -0.684022 0.928281 -0.528233 -1.057593 1.301851 -0.911948 -1.44318 1.687438 -1.361255 -1.833453 2.077713 -1.853459 -2.225325 2.469582 -2.370057 -2.617713 2.86197 -2.899115 -3.010269 3.254527 -3.433637 149 10 0 0 0 -0.117875 0.148763 -0.0334275 -0.372077 0.402965 -0.165421 -0.708295 0.739183 -0.433646 -1.078863 1.109751 -0.804808 -1.462359 1.493249 -1.239933 -1.851522 1.882411 -1.719698 -2.242824 2.27371 -2.226142 -2.634879 2.665765 -2.746692 -3.027189 3.058075 -3.273448 154 10 0 0 0 -0.33225 0.202465 -0.158996 -0.688264 0.202465 -0.288059 -1.162264 0.202465 -0.553847 -1.68577 0.202465 -0.923436 -2.227839 0.202464 -1.357316 -2.778177 0.202464 -1.836924 -3.33165 0.202464 -2.344398 -3.886221 0.202463 -2.866879 -4.441164 0.202463 -3.396347 155 10 0 0 0 -0.313558 0.180718 -0.134053 -0.668176 0.180718 -0.261954 -1.140998 0.180718 -0.525822 -1.663798 0.180717 -0.893439 -2.205555 0.180717 -1.325937 -2.755753 0.180717 -1.80452 -3.309168 0.180716 -2.311373 -3.863708 0.180716 -2.83336 -4.418629 0.180715 -3.362385 204 10 0 0 0 -0.102758 0.208081 -0.050794 -0.339595 0.444919 -0.163114 -0.660149 0.765471 -0.397521 -1.020412 1.125736 -0.731584 -1.397699 1.503019 -1.132374 -1.783678 1.888998 -1.584565 -2.173412 2.278732 -2.070426 -2.564604 2.669924 -2.575202 -2.956324 3.061644 -3.088849 205 10 0 0 0 -0.0847769 0.518485 -0.374915 -0.307226 0.803787 -0.508855 -0.628729 1.125294 -0.745171 -0.991053 1.487617 -1.086051 -1.370304 1.866867 -1.496769 -1.757546 2.254109 -1.95888 -2.147984 2.644547 -2.453139 -2.539619 3.036182 -2.96526 -2.931676 3.42824 -3.485823 207 10 0 0 0 -0.0853138 0.471188 -0.273866 -0.323217 0.734015 -0.401677 -0.651471 1.062268 -0.652045 -1.017661 1.428456 -1.006471 -1.399052 1.809849 -1.428892 -1.787374 2.19817 -1.900443 -2.178354 2.589149 -2.40201 -2.570284 2.981079 -2.919845 -2.962543 3.373339 -3.445271 210 10 0 0 0 -0.361042 0.244625 -0.211391 -0.686441 0.248072 -0.316674 -1.133057 0.248071 -0.541808 -1.640272 0.248071 -0.870706 -2.173586 0.24807 -1.270635 -2.719652 0.24807 -1.723938 -3.270992 0.24807 -2.211192 -3.824347 0.248069 -2.717365 -4.378412 0.248069 -3.232288 212 10 0 0 0 -0.413365 0.298486 -0.334929 -0.746575 0.314209 -0.446225 -1.193099 0.314209 -0.671236 -1.700038 0.314209 -0.999516 -2.233117 0.314208 -1.398646 -2.779052 0.314207 -1.851255 -3.330335 0.314207 -2.338015 -3.883666 0.314207 -2.843915 -4.437737 0.314206 -3.3589 214 10 0 0 0 -0.292711 0.165787 -0.113428 -0.622383 0.165787 -0.221817 -1.074814 0.165787 -0.454962 -1.585855 0.165786 -0.792623 -2.121452 0.165786 -1.200607 -2.668724 0.165786 -1.660586 -3.220697 0.165785 -2.153169 -3.774431 0.165785 -2.663772 -4.328783 0.165784 -3.182796 215 10 0 0 0 -0.422678 0.307495 -0.367319 -0.768829 0.352323 -0.490657 -1.210363 0.352324 -0.709033 -1.714136 0.352323 -1.030378 -2.245352 0.352323 -1.423295 -2.790293 0.352322 -1.870761 -3.341038 0.352322 -2.353369 -3.894037 0.352321 -2.855699 -4.447838 0.352321 -3.367136 264 10 0 0 0 -0.0847359 0.280099 -0.0833218 -0.297197 0.49256 -0.171583 -0.596205 0.791569 -0.366817 -0.942326 1.137686 -0.659603 -1.311127 1.506489 -1.023823 -1.692429 1.887792 -1.445734 -2.079542 2.274902 -1.906786 -2.469076 2.664434 -2.390453 -2.859488 3.05485 -2.884401 265 10 0 0 0 -0.0779724 0.512108 -0.354785 -0.284018 0.770953 -0.464037 -0.586208 1.073151 -0.664629 -0.935099 1.42204 -0.964852 -1.305931 1.792868 -1.33707 -1.688469 2.17541 -1.766284 -2.076267 2.563205 -2.23309 -2.466209 2.953146 -2.721349 -2.857002 3.343938 -3.220308 267 10 0 0 0 -0.0802364 0.450032 -0.239844 -0.299574 0.68412 -0.341815 -0.609011 0.99356 -0.555117 -0.963226 1.347775 -0.87045 -1.337532 1.722078 -1.257388 -1.72208 2.106628 -1.699423 -2.111019 2.495568 -2.176956 -2.501736 2.886281 -2.674821 -2.893152 3.277699 -3.183121 271 10 0 0 0 -0.353125 0.241159 -0.200891 -0.655304 0.243012 -0.290263 -1.07926 0.243012 -0.486787 -1.569313 0.243011 -0.780636 -2.090673 0.24301 -1.144295 -2.629583 0.243011 -1.564884 -3.176836 0.24301 -2.024754 -3.727612 0.24301 -2.50757 -4.279739 0.24301 -3.001493 272 10 0 0 0 -0.207989 0.0885859 -0.0480766 -0.491974 0.0885849 -0.126143 -0.897638 0.0885839 -0.302121 -1.373785 0.0885839 -0.571442 -1.885944 0.0885849 -0.911749 -2.419056 0.088583 -1.310988 -2.962429 0.0885839 -1.750834 -3.51026 0.088583 -2.214067 -4.05983 0.088583 -2.688282 273 10 0 0 0 -0.407257 0.297868 -0.323053 -0.718891 0.31896 -0.419174 -1.138813 0.318959 -0.610987 -1.625792 0.318959 -0.899161 -2.144989 0.318958 -1.257037 -2.682533 0.318957 -1.672276 -3.228952 0.318956 -2.127504 -3.779173 0.318956 -2.606241 -4.330847 0.318955 -3.096245 274 10 0 0 0 -0.275846 0.156583 -0.0994489 -0.574665 0.156583 -0.186657 -0.995485 0.156582 -0.379511 -1.482409 0.156582 -0.667588 -2.001783 0.15658 -1.025929 -2.539507 0.156581 -1.441869 -3.08592 0.15658 -1.897045 -3.636139 0.156579 -2.375797 -4.187811 0.156579 -2.865773 275 10 0 0 0 -0.398407 0.287614 -0.29634 -0.7136 0.303647 -0.394688 -1.141582 0.303646 -0.596023 -1.635384 0.303645 -0.897015 -2.159885 0.303646 -1.269435 -2.700663 0.303645 -1.697714 -3.248936 0.303645 -2.163601 -3.80043 0.303644 -2.652105 -4.353155 0.303644 -3.151655 0 10 0 0 0 -0.0948395 0.0948402 -0.0164325 -0.30602 0.306021 -0.103536 -0.60814 0.608142 -0.304003 -0.960243 0.960246 -0.613209 -1.335127 1.33513 -1.002732 -1.720828 1.720833 -1.452882 -2.110807 2.110812 -1.941546 -2.502334 2.502339 -2.451764 -2.894402 2.894408 -2.972597 1 10 0 0 0 -0.103126 0.103126 -0.0194842 -0.321683 0.321684 -0.113405 -0.624491 0.624493 -0.315044 -0.97314 0.973143 -0.614614 -1.343957 1.343961 -0.986782 -1.726758 1.726762 -1.417565 -2.114939 2.114944 -1.88784 -2.505295 2.5053 -2.381059 -2.896456 2.896461 -2.88536 2 10 0 0 0 -0.0928356 0.0928359 -0.0157357 -0.299686 0.299686 -0.0989908 -0.596945 0.596947 -0.291346 -0.945344 0.945347 -0.590234 -1.317976 1.317979 -0.969912 -1.70258 1.702584 -1.412322 -2.092047 2.092052 -1.895282 -2.483316 2.483321 -2.401245 -2.875226 2.875232 -2.918675 3 10 0 0 0 -0.0960181 0.0960205 -0.0168509 -0.304113 0.304115 -0.1012 -0.598203 0.598205 -0.288429 -0.941429 0.941431 -0.573736 -1.309334 1.309338 -0.93453 -1.690817 1.690822 -1.357482 -2.078468 2.078474 -1.823073 -2.468634 2.468639 -2.31394 -2.859749 2.859756 -2.817563 4 10 0 0 0 -0.0935804 0.0935813 -0.0159935 -0.29871 0.298711 -0.0977499 -0.591155 0.591157 -0.28237 -0.934009 0.934012 -0.566734 -1.302133 1.302136 -0.928354 -1.683945 1.683949 -1.353217 -2.071866 2.071871 -1.821164 -2.462234 2.46224 -2.31454 -2.853523 2.85353 -2.820821 5 10 0 0 0 -0.104786 0.104786 -0.0201287 -0.328759 0.32876 -0.11927 -0.638117 0.638118 -0.332426 -0.992236 0.992238 -0.64748 -1.366973 1.366976 -1.036344 -1.752151 1.752154 -1.482732 -2.141703 2.141708 -1.966601 -2.532919 2.532924 -2.471735 -2.924732 2.924737 -2.987192 6 10 0 0 0 -0.153354 4.76837e-007 -0.021585 -0.485978 7.15256e-007 -0.132165 -0.940668 1.3113e-006 -0.368505 -1.453501 1.84774e-006 -0.710414 -1.990738 2.26498e-006 -1.124483 -2.539201 2.74181e-006 -1.591555 -3.091947 3.15905e-006 -2.091305 -3.646195 3.39746e-006 -2.608769 -4.200953 3.63588e-006 -3.134561 7 10 0 0 0 -0.155408 7.15256e-007 -0.0221786 -0.485862 1.19209e-006 -0.131145 -0.936483 1.19209e-006 -0.361765 -1.446215 2.02656e-006 -0.696379 -1.981432 2.38419e-006 -1.102988 -2.528791 2.86102e-006 -1.56347 -3.080959 3.09944e-006 -2.057764 -3.634873 3.57628e-006 -2.570633 -4.189392 3.8147e-006 -3.092261 8 10 0 0 0 -0.154306 2.38419e-007 -0.0218595 -0.481309 1.19209e-006 -0.128293 -0.928343 1.43051e-006 -0.353995 -1.435452 1.66893e-006 -0.682653 -1.968971 2.14577e-006 -1.083293 -2.515391 2.6226e-006 -1.538511 -3.067054 2.86102e-006 -2.028419 -3.620668 3.09944e-006 -2.537556 -4.174969 3.09944e-006 -3.055797 9 10 0 0 0 -0.153797 7.15256e-007 -0.0217126 -0.480088 9.53674e-007 -0.127627 -0.926729 9.53674e-007 -0.352799 -1.433599 1.43051e-006 -0.680924 -1.966905 1.90735e-006 -1.08083 -2.513152 2.14577e-006 -1.535111 -3.064686 2.6226e-006 -2.023942 -3.618208 2.86102e-006 -2.532007 -4.172435 3.09944e-006 -3.049165 10 10 0 0 0 -0.150305 4.02331e-007 -0.0207179 -0.479144 7.89762e-007 -0.128494 -0.934013 1.49012e-006 -0.365088 -1.447551 1.90735e-006 -0.708699 -1.985365 2.44379e-006 -1.124976 -2.534175 2.83122e-006 -1.594219 -3.087111 3.24845e-006 -2.095886 -3.641473 3.54648e-006 -2.615042 -4.196315 3.78489e-006 -3.142432 11 10 0 0 0 -0.156442 3.57628e-007 -0.0224813 -0.489959 2.38419e-007 -0.133731 -0.944284 8.34465e-007 -0.36955 -1.456727 1.54972e-006 -0.710528 -1.993576 2.02656e-006 -1.123128 -2.541753 2.38419e-006 -1.588444 -3.09431 2.74181e-006 -2.086362 -3.648428 3.09944e-006 -2.601982 -4.203084 3.33786e-006 -3.125933 12 10 0 0 0 -0.104311 0.104312 -0.019943 -0.336249 0.336251 -0.127112 -0.66052 0.660522 -0.369078 -1.027978 1.02798 -0.728187 -1.411508 1.411512 -1.163534 -1.801179 1.801183 -1.648694 -2.192899 2.192904 -2.16241 -2.585266 2.58527 -2.690762 -2.977835 2.977841 -3.225835 13 10 0 0 0 -0.107718 0.107719 -0.0212942 -0.345726 0.345726 -0.13487 -0.674951 0.674952 -0.387335 -1.044948 1.04495 -0.756219 -1.429612 1.429615 -1.199041 -1.819811 1.819815 -1.690324 -2.211757 2.211761 -2.20848 -2.60423 2.604234 -2.740134 -2.996866 2.99687 -3.277925 14 10 0 0 0 -0.107716 0.107718 -0.0212935 -0.345891 0.345893 -0.135049 -0.675374 0.675376 -0.388074 -1.045547 1.04555 -0.757658 -1.430299 1.430303 -1.201093 -1.820542 1.820546 -1.692914 -2.212513 2.212517 -2.211594 -2.605001 2.605005 -2.743703 -2.997645 2.99765 -3.281894 15 10 0 0 0 -0.105452 0.105452 -0.0203906 -0.338549 0.338549 -0.128763 -0.663057 0.663059 -0.371221 -1.030188 1.030191 -0.729113 -1.413138 1.413141 -1.160811 -1.80251 1.802513 -1.64274 -2.194094 2.194098 -2.153962 -2.586391 2.586395 -2.680393 -2.978917 2.978921 -3.213842 16 10 0 0 0 -0.106612 0.106613 -0.0208504 -0.343395 0.343396 -0.133112 -0.672183 0.672184 -0.384629 -1.042217 1.042219 -0.753661 -1.427006 1.427009 -1.197345 -1.817292 1.817296 -1.689695 -2.209284 2.209288 -2.208837 -2.601784 2.601787 -2.741323 -2.994435 2.994439 -3.279843 17 10 0 0 0 -0.111839 0.111839 -0.0229903 -0.350947 0.350948 -0.137756 -0.676425 0.676425 -0.382232 -1.041984 1.041987 -0.734383 -1.423503 1.423506 -1.157547 -1.812064 1.812068 -1.631356 -2.20323 2.203234 -2.135728 -2.595294 2.595298 -2.656419 -2.987657 2.987661 -3.1847 18 10 0 0 0 -0.157547 1.78814e-007 -0.0228068 -0.494869 5.36442e-007 -0.136938 -0.954636 8.9407e-007 -0.380634 -1.471361 1.13249e-006 -0.732137 -2.01085 1.37091e-006 -1.155051 -2.560372 1.72853e-006 -1.62895 -3.11358 1.90735e-006 -2.1335 -3.668052 1.96695e-006 -2.654346 -4.222943 2.08616e-006 -3.182759 19 10 0 0 0 -0.158416 3.27826e-007 -0.0230647 -0.498193 5.06639e-007 -0.139082 -0.960752 8.9407e-007 -0.386935 -1.479482 1.3113e-006 -0.743588 -2.020133 1.66893e-006 -1.171322 -2.570219 1.96695e-006 -1.649112 -3.123696 2.02656e-006 -2.15666 -3.67831 2.20537e-006 -2.679893 -4.233299 2.32458e-006 -3.21038 20 10 0 0 0 -0.158227 5.36442e-007 -0.0230084 -0.497941 7.59959e-007 -0.138976 -0.960638 1.16229e-006 -0.387037 -1.479533 1.38581e-006 -0.744124 -2.020333 1.71363e-006 -1.172488 -2.570512 2.04146e-006 -1.650937 -3.124038 2.19047e-006 -2.159043 -3.678679 2.36928e-006 -2.682729 -4.233685 2.48849e-006 -3.213621 21 10 0 0 0 -0.152564 2.38419e-007 -0.0213585 -0.489746 4.76837e-007 -0.135382 -0.955941 1.19209e-006 -0.38877 -1.479529 1.66893e-006 -0.758591 -2.023669 2.14577e-006 -1.202103 -2.575563 2.14577e-006 -1.693975 -3.129898 2.38419e-006 -2.212731 -3.684962 2.38419e-006 -2.744913 -4.240248 2.6226e-006 -3.283179 22 10 0 0 0 -0.156259 2.98023e-007 -0.0224283 -0.494802 4.17233e-007 -0.137495 -0.958371 8.9407e-007 -0.386871 -1.47875 1.07288e-006 -0.747883 -2.020718 1.49012e-006 -1.181316 -2.571542 1.66893e-006 -1.66452 -3.125388 1.78814e-006 -2.176525 -3.680202 1.90735e-006 -2.703391 -4.235351 1.78814e-006 -3.237678 23 10 0 0 0 -0.157463 3.8743e-007 -0.022782 -0.497009 5.06639e-007 -0.138622 -0.960444 1.10269e-006 -0.387795 -1.480288 1.49012e-006 -0.747379 -2.021808 1.66893e-006 -1.178843 -2.572383 1.84774e-006 -1.660178 -3.126107 2.05636e-006 -2.170657 -3.680857 2.02656e-006 -2.696288 -4.235939 2.08616e-006 -3.228897 25 10 0 0 0 -0.109257 0.109258 -0.0219195 -0.348417 0.348419 -0.136739 -0.677463 0.677464 -0.388813 -1.046825 1.046827 -0.755206 -1.431022 1.431025 -1.194875 -1.821012 1.821015 -1.683683 -2.212879 2.212882 -2.200202 -2.605316 2.60532 -2.730707 -2.997947 2.997951 -3.26825 28 10 0 0 0 -0.105992 0.105994 -0.0206045 -0.342145 0.342149 -0.132194 -0.670617 0.670621 -0.383034 -1.040479 1.040483 -0.751381 -1.425226 1.42523 -1.194767 -1.815542 1.815546 -1.687495 -2.20757 2.207575 -2.207415 -2.600093 2.600099 -2.740778 -2.992763 2.992768 -3.280232 29 10 0 0 0 -0.108401 0.108401 -0.0215694 -0.347283 0.347283 -0.136088 -0.676847 0.676849 -0.389294 -1.046794 1.046795 -0.757972 -1.431383 1.431386 -1.200294 -1.821576 1.821578 -1.6915 -2.213535 2.213538 -2.209966 -2.606022 2.606025 -2.742059 -2.998669 2.998672 -3.280328 35 10 0 0 0 -0.146504 2.38419e-006 -0.0196642 -0.476813 4.76837e-007 -0.128524 -0.939478 1.90735e-006 -0.376533 -1.462466 1.43051e-006 -0.74467 -2.007201 1.90735e-006 -1.191106 -2.559722 1.90735e-006 -1.688673 -3.114433 1.43051e-006 -2.213645 -3.669715 2.38419e-006 -2.75173 -4.225147 2.38419e-006 -3.295601 36 10 0 0 0 -0.0918949 0.0918941 -0.015415 -0.307639 0.30764 -0.106696 -0.622745 0.622747 -0.3305 -0.986905 0.986906 -0.677682 -1.36995 1.369951 -1.109976 -1.760029 1.760031 -1.599822 -2.152126 2.152126 -2.12131 -2.544736 2.544739 -2.65796 -2.937479 2.937479 -3.201355 37 10 0 0 0 -0.0918491 0.0918522 -0.0153997 -0.30762 0.30762 -0.106703 -0.62274 0.622741 -0.330533 -0.986962 0.986963 -0.677937 -1.370147 1.370145 -1.111096 -1.7603 1.7603 -1.601834 -2.152419 2.152419 -2.123819 -2.545033 2.545033 -2.660717 -2.937776 2.937778 -3.204223 38 10 0 0 0 -0.0889339 0.0889301 -0.0144238 -0.299192 0.299191 -0.100698 -0.609722 0.609718 -0.315983 -0.971996 0.971994 -0.656693 -1.35414 1.354138 -1.083519 -1.743707 1.743702 -1.56752 -2.135601 2.135597 -2.084594 -2.528137 2.528134 -2.618468 -2.920856 2.920853 -3.160344 39 10 0 0 0 -0.0910637 0.0910597 -0.0151331 -0.304926 0.304923 -0.104674 -0.618477 0.618476 -0.325548 -0.982105 0.982104 -0.670884 -1.364646 1.364646 -1.100085 -1.754446 1.754447 -1.586699 -2.146436 2.146436 -2.105762 -2.539009 2.53901 -2.641104 -2.931746 2.931746 -3.183826 40 10 0 0 0 -0.0938494 0.0938501 -0.0160878 -0.311839 0.311842 -0.109473 -0.628445 0.628448 -0.336147 -0.993636 0.993639 -0.686977 -1.3769 1.376902 -1.120628 -1.766829 1.766832 -1.608713 -2.15883 2.158833 -2.127997 -2.551396 2.551399 -2.663005 -2.944125 2.944126 -3.205482 41 10 0 0 0 -0.0862837 0.0862789 -0.013567 -0.292881 0.292879 -0.0965974 -0.600648 0.600647 -0.3069 -0.962177 0.962177 -0.645118 -1.344722 1.344719 -1.074343 -1.734756 1.734755 -1.56364 -2.126849 2.126846 -2.084982 -2.519455 2.519453 -2.621473 -2.912193 2.912189 -3.164625 42 10 0 0 0 -0.129341 -1.43051e-006 -0.0152655 -0.43795 -4.76837e-007 -0.108867 -0.890656 -1.43051e-006 -0.342399 -1.411728 -9.53674e-007 -0.705266 -1.956481 -1.43051e-006 -1.151808 -2.50913 -1.43051e-006 -1.650606 -3.063889 -1.43051e-006 -2.176431 -3.61918 -1.90735e-006 -2.714827 -4.174612 -1.90735e-006 -3.258659 43 10 0 0 0 -0.118999 -9.53674e-007 -0.0128953 -0.410368 0 -0.0954325 -0.850292 0 -0.311709 -1.365831 4.76837e-007 -0.660235 -1.908712 -9.53674e-007 -1.0978 -2.460845 -1.43051e-006 -1.591771 -3.015465 -1.43051e-006 -2.115096 -3.570711 -1.90735e-006 -2.652131 -4.126125 -2.38419e-006 -3.195163 44 10 0 0 0 -0.132682 0 -0.0160768 -0.448932 -9.53674e-007 -0.114872 -0.908465 -9.53674e-007 -0.358225 -1.43269 -9.53674e-007 -0.72985 -1.978442 -9.53674e-007 -1.181497 -2.531284 -9.53674e-007 -1.682222 -3.086055 -9.53674e-007 -2.208244 -3.64133 -1.43051e-006 -2.746144 -4.196754 -1.90735e-006 -3.28959 45 10 0 0 0 -0.111789 1.90735e-006 -0.0113628 -0.385428 0 -0.0834279 -0.8047 9.53674e-007 -0.274493 -1.307474 2.86102e-006 -0.593699 -1.845209 1.90735e-006 -1.009673 -2.395844 1.90735e-006 -1.491458 -2.950089 1.90735e-006 -2.008864 -3.505241 1.90735e-006 -2.543238 -4.060629 1.90735e-006 -3.085133 46 10 0 0 0 -0.119255 2.86102e-006 -0.0129509 -0.40666 0 -0.0930698 -0.838793 9.53674e-007 -0.299478 -1.34951 9.53674e-007 -0.636377 -1.890488 9.53674e-007 -1.065498 -2.44209 9.53674e-007 -1.55489 -2.996593 9.53674e-007 -2.076255 -3.551819 9.53674e-007 -2.612672 -4.107234 0 -3.15569 47 10 0 0 0 -0.135154 9.53674e-007 -0.01669 -0.45513 4.76837e-007 -0.118089 -0.917085 4.76837e-007 -0.365036 -1.442277 9.53674e-007 -0.73944 -1.98836 4.76837e-007 -1.192847 -2.541306 -4.76837e-007 -1.694595 -3.096107 -4.76837e-007 -2.221217 -3.651395 -9.53674e-007 -2.759486 -4.20682 -1.90735e-006 -3.303015 48 10 0 0 0 -0.0635486 0.0635509 -0.0073173 -0.226706 0.226709 -0.0574999 -0.493608 0.493608 -0.205415 -0.833392 0.833394 -0.482159 -1.207442 1.207445 -0.867966 -1.594918 1.594921 -1.332058 -1.986299 1.986303 -1.839788 -2.37868 2.378683 -2.368638 -2.771331 2.77133 -2.907015 49 10 0 0 0 -0.0711145 0.0711155 -0.00917816 -0.253165 0.253164 -0.0724337 -0.543832 0.543835 -0.254272 -0.898451 0.898455 -0.5708 -1.278708 1.278711 -0.986882 -1.668071 1.668072 -1.468709 -2.059947 2.059948 -1.985492 -2.452472 2.452471 -2.518839 -2.845178 2.845177 -3.060041 50 10 0 0 0 -0.0670509 0.0670528 -0.00815177 -0.237049 0.237051 -0.0628645 -0.513574 0.513575 -0.22382 -0.859755 0.859755 -0.516773 -1.236579 1.236578 -0.915341 -1.624933 1.624931 -1.387179 -2.016554 2.016552 -1.899045 -2.409008 2.409008 -2.430075 -2.801689 2.801687 -2.969708 51 10 0 0 0 -0.0726051 0.0726051 -0.00956869 -0.257596 0.257593 -0.0750237 -0.550998 0.550997 -0.261156 -0.907041 0.907041 -0.581965 -1.287837 1.287837 -1.001023 -1.677358 1.677359 -1.484545 -2.069282 2.069279 -2.002226 -2.46182 2.461819 -2.536064 -2.85453 2.854528 -3.077508 52 10 0 0 0 -0.0611162 0.0611162 -0.00676274 -0.222963 0.222964 -0.0561059 -0.493847 0.493846 -0.209305 -0.837518 0.837516 -0.495749 -1.213443 1.213443 -0.890069 -1.601475 1.60147 -1.35899 -1.99301 1.993007 -1.869311 -2.385444 2.38544 -2.399652 -2.778116 2.778111 -2.939242 53 10 0 0 0 -0.0642323 0.0642385 -0.00747609 -0.228782 0.228789 -0.0585632 -0.497025 0.497032 -0.208242 -0.837549 0.837553 -0.486797 -1.211903 1.211912 -0.873956 -1.599476 1.599482 -1.33886 -1.990883 1.990892 -1.847044 -2.383275 2.38328 -2.376161 -2.775927 2.775932 -2.91471 54 10 0 0 0 -0.0845318 -9.53674e-007 -0.00647092 -0.310011 9.53674e-007 -0.0542839 -0.690015 -9.53674e-007 -0.204576 -1.174664 9.53674e-007 -0.488548 -1.706116 0 -0.882233 -2.255011 -1.90735e-006 -1.352021 -2.808798 0 -1.863256 -3.363802 -1.90735e-006 -2.394043 -3.919125 -1.90735e-006 -2.933428 55 10 0 0 0 -0.10358 -1.90735e-006 -0.00974131 -0.368471 -2.86102e-006 -0.0769553 -0.787396 -4.76837e-006 -0.267621 -1.293143 -4.76837e-006 -0.59326 -1.832456 -4.76837e-006 -1.015455 -2.383562 -4.76837e-006 -1.50084 -2.937892 -5.72205e-006 -2.019509 -3.493039 -5.72205e-006 -2.55378 -4.048409 -5.72205e-006 -3.094979 56 10 0 0 0 -0.107039 -9.53674e-007 -0.0104094 -0.378703 -3.8147e-006 -0.0813625 -0.803668 -1.90735e-006 -0.27908 -1.312031 -1.90735e-006 -0.610561 -1.852312 -1.90735e-006 -1.036735 -2.403646 -3.8147e-006 -1.523929 -2.95804 -4.76837e-006 -2.043593 -3.513208 -5.72205e-006 -2.578446 -4.068588 -5.72205e-006 -3.120017 57 10 0 0 0 -0.0900278 -9.53674e-007 -0.00734448 -0.325525 9.53674e-007 -0.0597253 -0.715707 9.53674e-007 -0.219805 -1.206539 1.90735e-006 -0.515119 -1.740415 -9.53674e-007 -0.916991 -2.28996 0 -1.39105 -2.843865 -9.53674e-007 -1.903816 -3.398883 -9.53674e-007 -2.434854 -3.954196 -9.53674e-007 -2.973962 58 10 0 0 0 -0.105506 0 -0.0101101 -0.372785 1.90735e-006 -0.0786283 -0.792833 -9.53674e-007 -0.270589 -1.298227 9.53674e-007 -0.595452 -1.836758 -9.53674e-007 -1.014525 -2.38728 -1.90735e-006 -1.495452 -2.941322 -1.90735e-006 -2.010031 -3.496338 -1.90735e-006 -2.541078 -4.05163 -2.86102e-006 -3.079505 59 10 0 0 0 -0.0904903 2.86102e-006 -0.00741696 -0.328705 0 -0.0610828 -0.722211 -1.90735e-006 -0.224473 -1.214333 0 -0.522232 -1.748724 -1.90735e-006 -0.925902 -2.298427 -1.90735e-006 -1.401026 -2.852422 -9.53674e-007 -1.914954 -3.407485 0 -2.44711 -3.962827 -1.90735e-006 -2.987283 60 10 0 0 0 -0.106474 0.106473 -0.020795 -0.332213 0.332213 -0.12168 -0.642731 0.642731 -0.336943 -0.997203 0.997204 -0.65304 -1.371788 1.371789 -1.04122 -1.756667 1.756669 -1.485523 -2.145937 2.145939 -1.966386 -2.536912 2.536915 -2.467911 -2.928504 2.928507 -2.979266 61 10 0 0 0 -0.104482 0.104483 -0.0200107 -0.326112 0.326114 -0.11687 -0.632292 0.632295 -0.324361 -0.983486 0.98349 -0.630984 -1.355914 1.355918 -1.009806 -1.739576 1.739581 -1.445989 -2.128184 2.128189 -1.920251 -2.518764 2.518769 -2.41633 -2.910064 2.910069 -2.922808 62 10 0 0 0 -0.109845 0.109845 -0.0221602 -0.343513 0.343513 -0.131128 -0.662781 0.662782 -0.36299 -1.023775 1.023777 -0.699437 -1.402476 1.402478 -1.1073 -1.789563 1.789566 -1.568161 -2.179985 2.179988 -2.062203 -2.57163 2.571634 -2.574522 -2.963704 2.963708 -3.09545 63 10 0 0 0 -0.102735 0.102733 -0.0193336 -0.321622 0.321622 -0.113567 -0.625372 0.625373 -0.316825 -0.974793 0.974794 -0.618507 -1.345971 1.345972 -0.992142 -1.728834 1.728836 -1.42331 -2.116947 2.11695 -1.892965 -2.507199 2.507201 -2.384883 -2.898241 2.898243 -2.887409 64 10 0 0 0 -0.108666 0.108667 -0.0216777 -0.340177 0.340177 -0.128403 -0.657427 0.657428 -0.356322 -1.017015 1.017017 -0.688193 -1.394849 1.394851 -1.091683 -1.781489 1.781492 -1.548947 -2.171685 2.171688 -2.040199 -2.563202 2.563205 -2.550251 -2.955184 2.955187 -3.069156 65 10 0 0 0 -0.100731 0.100733 -0.0185752 -0.315607 0.315609 -0.10905 -0.615063 0.615065 -0.305032 -0.961128 0.961131 -0.597681 -1.330005 1.33001 -0.962186 -1.7115 1.711504 -1.385208 -2.09882 2.098824 -1.847984 -2.488568 2.488573 -2.334018 -2.87923 2.879233 -2.831179 66 10 0 0 0 -0.144836 0 -0.0192122 -0.456397 9.53674e-007 -0.114797 -0.890095 0 -0.323153 -1.388218 4.76837e-007 -0.632709 -1.91607 9.53674e-007 -1.01502 -2.459305 4.76837e-007 -1.454222 -3.009177 9.53674e-007 -1.930516 -3.561684 1.43051e-006 -2.427954 -4.115203 9.53674e-007 -2.935988 67 10 0 0 0 -0.148441 0 -0.0201986 -0.466497 -4.76837e-007 -0.120251 -0.906451 9.53674e-007 -0.336569 -1.409071 4.76837e-007 -0.655449 -1.939831 9.53674e-007 -1.046881 -2.484719 7.15256e-007 -1.494092 -3.035538 1.43051e-006 -1.977261 -3.588635 1.43051e-006 -2.480611 -4.142537 1.90735e-006 -2.99334 68 10 0 0 0 -0.152997 -2.38419e-007 -0.0214821 -0.480114 -2.38419e-007 -0.127998 -0.929221 7.15256e-007 -0.356528 -1.438443 7.15256e-007 -0.689968 -1.973382 1.19209e-006 -1.095581 -2.520552 1.19209e-006 -1.554977 -3.072581 1.66893e-006 -2.048038 -3.626392 1.66893e-006 -2.559582 -4.180822 1.66893e-006 -3.079779 69 10 0 0 0 -0.151387 2.38419e-007 -0.0210232 -0.47438 4.76837e-007 -0.124564 -0.91868 4.76837e-007 -0.346593 -1.424117 1.19209e-006 -0.671549 -1.956538 1.43051e-006 -1.068454 -2.502307 1.66893e-006 -1.520192 -3.053593 1.90735e-006 -2.007009 -3.606969 2.14577e-006 -2.513407 -4.16109 2.38419e-006 -3.02905 70 10 0 0 0 -0.148125 4.76837e-007 -0.0201101 -0.465749 7.15256e-007 -0.119861 -0.905408 9.53674e-007 -0.335798 -1.407874 1.43051e-006 -0.65435 -1.938554 1.66893e-006 -1.045528 -2.483403 1.43051e-006 -1.492537 -3.034201 1.43051e-006 -1.975543 -3.587282 1.43051e-006 -2.478717 -4.141178 1.43051e-006 -2.991362 71 10 0 0 0 -0.148324 0 -0.0201664 -0.465049 4.76837e-007 -0.119292 -0.902815 9.53674e-007 -0.332792 -1.403396 1.43051e-006 -0.647388 -1.932654 1.43051e-006 -1.034043 -2.476644 1.90735e-006 -1.47683 -3.026962 1.43051e-006 -1.956282 -3.579758 2.38419e-006 -2.456532 -4.133452 2.38419e-006 -2.966661 72 10 0 0 0 -0.10866 0.108659 -0.0216753 -0.342824 0.342823 -0.131162 -0.664509 0.66451 -0.367846 -1.027936 1.027936 -0.712489 -1.408369 1.408368 -1.12954 -1.79643 1.796429 -1.598732 -2.187373 2.187373 -2.099807 -2.57936 2.579362 -2.618863 -2.971714 2.971715 -3.146859 73 10 0 0 0 -0.109379 0.109378 -0.0219682 -0.343272 0.343272 -0.131172 -0.663528 0.663529 -0.364992 -1.025534 1.025535 -0.704806 -1.404993 1.404995 -1.116614 -1.792535 1.792538 -1.581271 -2.183206 2.183208 -2.078546 -2.574994 2.574997 -2.59353 -2.967165 2.967168 -3.11679 74 10 0 0 0 -0.110382 0.110383 -0.0223821 -0.347092 0.347094 -0.134564 -0.670673 0.670676 -0.375112 -1.035195 1.035199 -0.723571 -1.416236 1.416241 -1.144011 -1.804609 1.804614 -1.61605 -2.195696 2.1957 -2.119214 -2.587716 2.587721 -2.638978 -2.98005 2.980056 -3.166433 75 10 0 0 0 -0.110059 0.11006 -0.0222481 -0.346295 0.346297 -0.133928 -0.669558 0.66956 -0.373818 -1.033922 1.033925 -0.721724 -1.414892 1.414895 -1.141752 -1.803232 1.803235 -1.613492 -2.194303 2.194305 -2.116431 -2.586315 2.586318 -2.636016 -2.97865 2.978654 -3.163506 76 10 0 0 0 -0.109574 0.109575 -0.0220496 -0.343894 0.343895 -0.1317 -0.664581 0.664583 -0.366379 -1.026888 1.026891 -0.707209 -1.406531 1.406535 -1.119991 -1.794171 1.794175 -1.585467 -2.184888 2.184892 -2.083391 -2.576704 2.576708 -2.598904 -2.968894 2.968899 -3.122632 77 10 0 0 0 -0.109625 0.109627 -0.0220701 -0.346106 0.346106 -0.134007 -0.670328 0.67033 -0.375875 -1.035611 1.035614 -0.727033 -1.417194 1.417198 -1.150562 -1.805854 1.805858 -1.625317 -2.197081 2.197085 -2.13062 -2.589198 2.589201 -2.652545 -2.981636 2.981639 -3.183028 78 10 0 0 0 -0.158487 1.19209e-007 -0.0230852 -0.49949 7.15256e-007 -0.140053 -0.963726 5.96046e-007 -0.390441 -1.483693 3.57628e-007 -0.750353 -2.025135 5.96046e-007 -1.181477 -2.575623 5.96046e-007 -1.662168 -3.129292 3.57628e-007 -2.171983 -3.684004 2.38419e-007 -2.696945 -4.23905 0 -3.228729 79 10 0 0 0 -0.154566 2.38419e-007 -0.0219338 -0.486607 7.15256e-007 -0.132078 -0.941904 7.15256e-007 -0.369286 -1.456138 1.19209e-006 -0.714589 -1.99456 1.19209e-006 -1.133238 -2.543724 1.19209e-006 -1.604764 -3.096846 7.15256e-007 -2.108372 -3.651312 7.15256e-007 -2.62916 -4.206223 4.76837e-007 -3.157949 80 10 0 0 0 -0.147632 -9.53674e-007 -0.0199744 -0.465421 -7.15256e-007 -0.119841 -0.905874 9.53674e-007 -0.336806 -1.409017 -2.38419e-007 -0.656799 -1.940011 4.76837e-007 -1.048991 -2.485021 7.15256e-007 -1.496817 -3.036017 4.76837e-007 -1.981358 -3.589314 0 -2.486877 -4.143437 7.15256e-007 -3.002547 81 10 0 0 0 -0.149529 9.53674e-007 -0.0205011 -0.471214 1.43051e-006 -0.123109 -0.915537 1.43051e-006 -0.34517 -1.4212 1.66893e-006 -0.670621 -1.954037 1.90735e-006 -1.068928 -2.500278 1.90735e-006 -1.523173 -3.051923 1.90735e-006 -2.012922 -3.60561 1.66893e-006 -2.522939 -4.160017 1.90735e-006 -3.042804 82 10 0 0 0 -0.151736 -9.53674e-007 -0.0211223 -0.478024 2.38419e-007 -0.127035 -0.926908 7.15256e-007 -0.355262 -1.436135 4.76837e-007 -0.688715 -1.971461 7.15256e-007 -1.095715 -2.519004 1.19209e-006 -1.557258 -3.071366 4.76837e-007 -2.053329 -3.625445 7.15256e-007 -2.568397 -4.180111 2.38419e-007 -3.092535 83 10 0 0 0 -0.152659 0 -0.0213855 -0.481508 2.38419e-007 -0.129169 -0.933702 4.76837e-007 -0.36198 -1.445388 1.19209e-006 -0.701156 -1.981973 1.43051e-006 -1.112776 -2.530234 1.43051e-006 -1.578595 -3.082943 1.43051e-006 -2.077968 -3.637226 1.19209e-006 -2.595939 -4.192051 1.19209e-006 -3.123034 85 10 0 0 0 -0.110537 0.110537 -0.0224466 -0.349989 0.34999 -0.137586 -0.677359 0.677359 -0.386054 -1.045062 1.045063 -0.746082 -1.428303 1.428304 -1.1796 -1.817939 1.81794 -1.664375 -2.209702 2.209703 -2.178892 -2.60212 2.602121 -2.708757 -2.994736 2.994737 -3.245711 86 10 0 0 0 -0.110531 0.110529 -0.0224446 -0.349854 0.349855 -0.137443 -0.677022 0.677022 -0.385485 -1.044576 1.044576 -0.744955 -1.427724 1.427725 -1.17788 -1.817312 1.817312 -1.662145 -2.209054 2.209055 -2.176255 -2.60146 2.601461 -2.70579 -2.994071 2.994071 -3.242454 92 10 0 0 0 -0.157024 -2.38419e-007 -0.0226524 -0.505899 7.15256e-007 -0.145856 -0.98358 9.53674e-007 -0.417751 -1.513847 4.76837e-007 -0.807605 -2.061246 4.76837e-007 -1.268307 -2.614459 4.76837e-007 -1.772883 -3.169302 2.38419e-007 -2.300333 -3.724581 0 -2.838352 -4.279984 -4.76837e-007 -3.380882 96 10 0 0 0 -0.104681 0.104678 -0.0200872 -0.342941 0.342942 -0.133934 -0.67613 0.676129 -0.395195 -1.049655 1.049654 -0.778703 -1.436415 1.436414 -1.236922 -1.827575 1.827576 -1.741201 -2.219932 2.21993 -2.269288 -2.612598 2.612597 -2.808427 -3.005347 3.005343 -3.352245 97 10 0 0 0 -0.116454 0.116452 -0.0249722 -0.372517 0.372519 -0.159215 -0.716506 0.716506 -0.446466 -1.093735 1.093733 -0.846982 -1.481341 1.481343 -1.312194 -1.872504 1.872504 -1.816499 -2.264749 2.264747 -2.341548 -2.657327 2.657323 -2.876906 -3.050014 3.05001 -3.417094 98 10 0 0 0 -0.119008 0.119012 -0.0261065 -0.377831 0.377832 -0.163733 -0.72289 0.722893 -0.453753 -1.100242 1.100245 -0.85486 -1.48767 1.487674 -1.318553 -1.8787 1.878704 -1.820855 -2.270879 2.270885 -2.344298 -2.663419 2.663423 -2.878274 -3.056079 3.056083 -3.417187 99 10 0 0 0 -0.118908 0.118906 -0.0260604 -0.377539 0.377535 -0.16345 -0.722505 0.722503 -0.453227 -1.09985 1.099846 -0.854301 -1.487293 1.487287 -1.318093 -1.878335 1.878332 -1.820572 -2.27052 2.270514 -2.344187 -2.663065 2.663061 -2.878315 -3.055726 3.055721 -3.417292 100 10 0 0 0 -0.108659 0.10866 -0.0216748 -0.352626 0.352622 -0.141795 -0.689243 0.689241 -0.410963 -1.063993 1.063992 -0.799882 -1.451088 1.451086 -1.260789 -1.842305 1.842302 -1.76595 -2.234661 2.234657 -2.293994 -2.627314 2.627312 -2.832658 -3.020051 3.020047 -3.375822 101 10 0 0 0 -0.108426 0.108419 -0.0215783 -0.353632 0.353628 -0.143097 -0.691952 0.691948 -0.4163 -1.06761 1.067604 -0.809358 -1.45502 1.455014 -1.272877 -1.846298 1.846292 -1.779015 -2.238653 2.238647 -2.307081 -2.631304 2.631298 -2.845373 -3.024035 3.024028 -3.387991 102 10 0 0 0 -0.163517 7.15256e-007 -0.0246084 -0.522711 7.15256e-007 -0.156346 -1.005663 9.53674e-007 -0.437313 -1.536622 7.15256e-007 -0.829389 -2.083337 4.76837e-007 -1.286227 -2.635983 0 -1.78498 -3.190531 -4.76837e-007 -2.307108 -3.74567 -7.15256e-007 -2.841184 -4.300995 -9.53674e-007 -3.380684 103 10 0 0 0 -0.162345 0 -0.024249 -0.521626 0 -0.15606 -1.006375 -2.38419e-007 -0.440212 -1.53904 -7.15256e-007 -0.837936 -2.086832 -9.53674e-007 -1.300938 -2.639902 -1.43051e-006 -1.804005 -3.194593 -1.66893e-006 -2.328567 -3.749783 -1.90735e-006 -2.863971 -4.305138 -2.38419e-006 -3.404577 104 10 0 0 0 -0.159371 0 -0.0233501 -0.515079 4.76837e-007 -0.152158 -0.998352 0 -0.433687 -1.530814 -4.76837e-007 -0.830725 -2.078716 -7.15256e-007 -1.294384 -2.631911 -9.53674e-007 -1.798773 -3.186659 -1.43051e-006 -2.324378 -3.741881 -1.66893e-006 -2.860679 -4.297252 -1.90735e-006 -3.401936 105 10 0 0 0 -0.152456 4.76837e-007 -0.0213287 -0.499041 0 -0.142692 -0.977555 0 -0.415998 -1.508572 4.76837e-007 -0.808259 -2.056159 -4.76837e-007 -1.270059 -2.609338 0 -1.77428 -3.164129 -9.53674e-007 -2.300726 -3.719384 -9.53674e-007 -2.837968 -4.274777 -1.90735e-006 -3.380147 106 10 0 0 0 -0.142405 -4.76837e-007 -0.0185609 -0.474401 -4.76837e-007 -0.128672 -0.94396 -1.43051e-006 -0.387316 -1.471461 -9.53674e-007 -0.76856 -2.017836 -1.43051e-006 -1.223541 -2.570705 -1.90735e-006 -1.724526 -3.125433 -2.38419e-006 -2.249834 -3.680687 -2.38419e-006 -2.787116 -4.236092 -2.86102e-006 -3.32969 107 10 0 0 0 -0.159459 0 -0.0233747 -0.514715 2.38419e-007 -0.151807 -0.996902 0 -0.431431 -1.528402 0 -0.825278 -2.075702 -2.38419e-007 -1.285418 -2.628631 -9.53674e-007 -1.787021 -3.1833 -1.19209e-006 -2.311211 -3.738499 -1.66893e-006 -2.846889 -4.293863 -1.90735e-006 -3.387798 108 10 0 0 0 -0.0777602 0.0777588 -0.0109916 -0.272779 0.272778 -0.0842831 -0.575346 0.57535 -0.285513 -0.935692 0.93569 -0.619831 -1.317751 1.317749 -1.04614 -1.707487 1.707485 -1.53203 -2.09942 2.099419 -2.049876 -2.491942 2.491941 -2.583159 -2.884626 2.884624 -3.123405 109 10 0 0 0 -0.102803 0.102806 -0.0193609 -0.339079 0.339079 -0.131081 -0.670692 0.670691 -0.388795 -1.04283 1.042831 -0.766383 -1.428648 1.428648 -1.217402 -1.819357 1.819357 -1.71521 -2.211527 2.211525 -2.238349 -2.60411 2.604111 -2.773902 -2.996813 2.996811 -3.315026 110 10 0 0 0 -0.0917597 0.0917635 -0.0153677 -0.312541 0.312545 -0.11141 -0.635567 0.635571 -0.35081 -1.005219 1.005221 -0.718336 -1.390591 1.390594 -1.166096 -1.7813 1.781299 -1.66386 -2.173515 2.173516 -2.188205 -2.566136 2.566136 -2.725345 -2.958871 2.958871 -3.268182 111 10 0 0 0 -0.0927577 0.0927649 -0.0157111 -0.315162 0.315169 -0.113321 -0.639239 0.639246 -0.354881 -1.009243 1.00925 -0.7238 -1.394691 1.394697 -1.172091 -1.785423 1.785427 -1.670218 -2.17764 2.177647 -2.194566 -2.570261 2.570265 -2.731708 -2.962994 2.962999 -3.274475 112 10 0 0 0 -0.0847383 0.0847397 -0.0130785 -0.292559 0.292559 -0.0971847 -0.605449 0.60545 -0.316819 -0.970526 0.970528 -0.667247 -1.354278 1.354279 -1.104009 -1.744497 1.744495 -1.595537 -2.136561 2.136562 -2.116304 -2.52913 2.529129 -2.651221 -2.921836 2.921837 -3.192475 113 10 0 0 0 -0.0967031 0.0967035 -0.0170977 -0.324557 0.324557 -0.120095 -0.651455 0.651454 -0.367559 -1.022227 1.022225 -0.739549 -1.40786 1.40786 -1.189202 -1.79859 1.79859 -1.687312 -2.190794 2.190792 -2.211205 -2.583395 2.583391 -2.747588 -2.976116 2.976112 -3.289604 114 10 0 0 0 -0.132965 1.43051e-006 -0.0161452 -0.447838 9.53674e-007 -0.113991 -0.903881 1.43051e-006 -0.352267 -1.424482 9.53674e-007 -0.713872 -1.96814 4.76837e-007 -1.15506 -2.520329 9.53674e-007 -1.649535 -3.074968 0 -2.173216 -3.630238 0 -2.710918 -4.185669 0 -3.254886 115 10 0 0 0 -0.143343 -1.43051e-006 -0.0188091 -0.474087 -4.76837e-007 -0.127991 -0.940407 -9.53674e-007 -0.381573 -1.46549 -1.43051e-006 -0.755669 -2.010721 -1.43051e-006 -1.204625 -2.56316 -1.90735e-006 -1.70141 -3.117737 -2.38419e-006 -2.223999 -3.672922 -2.86102e-006 -2.759284 -4.228304 -2.86102e-006 -3.301077 116 10 0 0 0 -0.137014 1.43051e-006 -0.0171616 -0.457531 -9.53674e-007 -0.118942 -0.917362 -4.76837e-007 -0.362729 -1.439957 -9.53674e-007 -0.729773 -1.984312 -1.90735e-006 -1.174337 -2.536444 -1.90735e-006 -1.668309 -3.090994 -1.90735e-006 -2.190448 -3.646225 -2.38419e-006 -2.726962 -4.201645 -2.38419e-006 -3.270231 117 10 0 0 0 -0.120338 -9.53674e-007 -0.01319 -0.414282 0 -0.0973241 -0.855312 0 -0.315038 -1.369652 0 -0.660604 -1.911085 0 -1.091703 -2.462395 -1.90735e-006 -1.578707 -3.016628 -2.86102e-006 -2.095942 -3.571716 -1.90735e-006 -2.628755 -4.127079 -1.90735e-006 -3.169551 118 10 0 0 0 -0.142588 -9.53674e-007 -0.0186093 -0.471653 0 -0.126551 -0.936869 -9.53674e-007 -0.378436 -1.461559 -9.53674e-007 -0.751391 -2.006651 -9.53674e-007 -1.199627 -2.559021 -1.43051e-006 -1.695782 -3.113577 -1.90735e-006 -2.217995 -3.668787 -1.90735e-006 -2.753966 -4.224194 -2.38419e-006 -3.296669 119 10 0 0 0 -0.131081 -4.76837e-007 -0.0156846 -0.442272 1.90735e-006 -0.11102 -0.895445 9.53674e-007 -0.34521 -1.41518 9.53674e-007 -0.704506 -1.958525 -9.53674e-007 -1.144236 -2.510344 -1.43051e-006 -1.635449 -3.064795 -9.53674e-007 -2.155989 -3.619995 -1.43051e-006 -2.69164 -4.175407 -2.86102e-006 -3.234452 120 10 0 0 0 -0.0892305 0.0892305 -0.0145199 -0.283268 0.28327 -0.0870218 -0.561199 0.5612 -0.249958 -0.890099 0.890103 -0.501717 -1.246434 1.246437 -0.823411 -1.619589 1.619593 -1.20532 -2.001375 2.001378 -1.630019 -2.387091 2.387094 -2.080275 -2.774347 2.774351 -2.542527 121 10 0 0 0 -0.0970383 0.0970383 -0.0172167 -0.306096 0.306098 -0.102421 -0.60033 0.600332 -0.289879 -0.942656 0.942658 -0.572911 -1.309032 1.309036 -0.928024 -1.68903 1.689032 -1.342701 -2.075474 2.075476 -1.798434 -2.464636 2.464638 -2.27818 -2.854794 2.854796 -2.768948 122 10 0 0 0 -0.0951166 0.0951164 -0.0165308 -0.300323 0.300321 -0.0983515 -0.590283 0.590282 -0.279092 -0.929101 0.929101 -0.553495 -1.292941 1.292941 -0.899569 -1.67133 1.671331 -1.305846 -2.056791 2.056793 -1.754251 -2.445298 2.445299 -2.227538 -2.834929 2.83493 -2.712254 123 10 0 0 0 -0.0973196 0.0973201 -0.0173175 -0.306931 0.306931 -0.103013 -0.60176 0.601761 -0.291426 -0.944564 0.944566 -0.575665 -1.311281 1.311282 -0.932023 -1.691486 1.69149 -1.347843 -2.078057 2.078061 -1.804566 -2.467304 2.467308 -2.285191 -2.857532 2.857534 -2.776795 124 10 0 0 0 -0.0986557 0.0986528 -0.0178039 -0.3111 0.311099 -0.106052 -0.609153 0.609151 -0.299708 -0.954606 0.954606 -0.590753 -1.323221 1.32322 -0.954247 -1.704607 1.704608 -1.376645 -2.091874 2.091874 -1.838981 -2.481561 2.481562 -2.324321 -2.872123 2.872122 -2.820151 125 10 0 0 0 -0.0954514 0.0954452 -0.0166481 -0.301855 0.301851 -0.0995126 -0.59358 0.593574 -0.282998 -0.934044 0.934038 -0.561405 -1.299126 1.299121 -0.911844 -1.678209 1.678205 -1.321692 -2.063953 2.063949 -1.772144 -2.452599 2.452595 -2.246753 -2.842372 2.842367 -2.733075 126 10 0 0 0 -0.142303 -9.53674e-007 -0.0185339 -0.448925 -9.53674e-007 -0.110813 -0.876931 -9.53674e-007 -0.312175 -1.369919 -9.53674e-007 -0.611594 -1.894128 -9.53674e-007 -0.983175 -2.43529 -4.76837e-007 -1.413095 -2.984052 -9.53674e-007 -1.882037 -3.535941 -9.53674e-007 -2.373867 -4.089011 -1.90735e-006 -2.87694 127 10 0 0 0 -0.131359 9.53674e-007 -0.0157523 -0.415357 1.90735e-006 -0.0938277 -0.817623 9.53674e-007 -0.266205 -1.289406 0 -0.5284 -1.798038 0 -0.860491 -2.328894 1.90735e-006 -1.252229 -2.871028 1.90735e-006 -1.686405 -3.418344 0 -2.146624 -3.967655 1.90735e-006 -2.619127 128 10 0 0 0 -0.139676 4.76837e-007 -0.017844 -0.440812 0 -0.106539 -0.862633 4.76837e-007 -0.300559 -1.350678 9.53674e-007 -0.590682 -1.871388 9.53674e-007 -0.952583 -2.41034 4.76837e-007 -1.373331 -2.957748 9.53674e-007 -1.834096 -3.508748 4.76837e-007 -2.318669 -4.061121 4.76837e-007 -2.814827 129 10 0 0 0 -0.142151 -9.53674e-007 -0.0184939 -0.448626 -9.53674e-007 -0.110675 -0.877005 -4.76837e-007 -0.312487 -1.370917 -9.53674e-007 -0.613694 -1.89565 -9.53674e-007 -0.986776 -2.436913 -4.76837e-007 -1.41713 -2.985745 -4.76837e-007 -1.886518 -3.537709 -4.76837e-007 -2.379008 -4.090837 -9.53674e-007 -2.882678 130 10 0 0 0 -0.142764 -9.53674e-007 -0.0186546 -0.450396 1.90735e-006 -0.111604 -0.879707 9.53674e-007 -0.314549 -1.373829 9.53674e-007 -0.616166 -1.898732 4.76837e-007 -0.989734 -2.440314 9.53674e-007 -1.421468 -2.989347 9.53674e-007 -1.892153 -3.541392 9.53674e-007 -2.385356 -4.094594 9.53674e-007 -2.889821 131 10 0 0 0 -0.133888 0 -0.0163739 -0.423062 1.43051e-006 -0.0975661 -0.831278 1.90735e-006 -0.276288 -1.308117 1.90735e-006 -0.546769 -1.820598 1.90735e-006 -0.88784 -2.354104 2.86102e-006 -1.288428 -2.898012 2.86102e-006 -1.730826 -3.446605 2.86102e-006 -2.198704 -3.996998 2.86102e-006 -2.678707 132 10 0 0 0 -0.108066 0.108066 -0.021435 -0.340334 0.340334 -0.128945 -0.659526 0.659525 -0.360655 -1.020935 1.020936 -0.698478 -1.400087 1.400088 -1.108678 -1.787495 1.787496 -1.572184 -2.178114 2.178114 -2.06878 -2.569883 2.569883 -2.583417 -2.962048 2.962045 -3.106427 133 10 0 0 0 -0.108198 0.108199 -0.0214878 -0.340405 0.340408 -0.128936 -0.659578 0.659579 -0.360605 -1.021141 1.021143 -0.698934 -1.40057 1.400572 -1.110584 -1.788209 1.788211 -1.576051 -2.178984 2.178986 -2.074767 -2.570867 2.570868 -2.59156 -2.963117 2.963117 -3.116829 134 10 0 0 0 -0.101775 0.101776 -0.0189685 -0.320751 0.320752 -0.113284 -0.625793 0.625793 -0.31878 -0.976544 0.976547 -0.62416 -1.348686 1.348689 -1.001775 -1.732121 1.732124 -1.436506 -2.120559 2.120562 -1.909172 -2.511015 2.511019 -2.403639 -2.902204 2.902207 -2.908367 135 10 0 0 0 -0.103292 0.103292 -0.0195478 -0.325697 0.325698 -0.117158 -0.634686 0.634685 -0.329647 -0.988604 0.988602 -0.644108 -1.362831 1.362829 -1.030696 -1.74749 1.74749 -1.473483 -2.136702 2.136702 -1.953742 -2.527702 2.527702 -2.455629 -2.91935 2.919349 -2.96797 136 10 0 0 0 -0.102877 0.102874 -0.0193886 -0.32374 0.323739 -0.11551 -0.630736 0.630735 -0.324442 -0.983078 0.983078 -0.634326 -1.356403 1.356402 -1.016971 -1.740612 1.740612 -1.45673 -2.129537 2.129535 -1.934063 -2.520319 2.520319 -2.43287 -2.911773 2.911772 -2.941882 137 10 0 0 0 -0.104498 0.104498 -0.0200163 -0.329236 0.329238 -0.119911 -0.640309 0.640311 -0.336189 -0.995562 0.995563 -0.654612 -1.370893 1.370896 -1.046176 -1.756289 1.756293 -1.494114 -2.145907 2.145909 -1.978679 -2.537145 2.537147 -2.48417 -2.928965 2.928966 -2.99976 138 10 0 0 0 -0.159039 0 -0.0232508 -0.495789 0 -0.136943 -0.95194 0 -0.375373 -1.464352 4.76837e-007 -0.716276 -2.000296 9.53674e-007 -1.125528 -2.547537 9.53674e-007 -1.585332 -3.09939 9.53674e-007 -2.07685 -3.652985 4.76837e-007 -2.585745 -4.207177 0 -3.102408 139 10 0 0 0 -0.159757 0 -0.0234653 -0.501455 -7.15256e-007 -0.140975 -0.96497 2.38419e-007 -0.39027 -1.483737 0 -0.747026 -2.024145 0 -1.173738 -2.573982 -4.76837e-007 -1.649787 -3.127281 -4.76837e-007 -2.155317 -3.681767 -4.76837e-007 -2.676425 -4.236646 -9.53674e-007 -3.204541 140 10 0 0 0 -0.155992 -1.43051e-006 -0.0223484 -0.487693 -9.53674e-007 -0.13224 -0.939733 -9.53674e-007 -0.364837 -1.449762 -1.43051e-006 -0.700138 -1.984481 -9.53674e-007 -1.104966 -2.531175 -1.43051e-006 -1.561688 -3.082789 -1.43051e-006 -2.051191 -3.636269 -1.90735e-006 -2.558759 -4.190394 -1.90735e-006 -3.074481 141 10 0 0 0 -0.144367 1.90735e-006 -0.0190847 -0.453185 9.53674e-007 -0.112825 -0.882058 9.53674e-007 -0.315237 -1.375038 0 -0.614643 -1.898543 4.76837e-007 -0.984231 -2.438745 1.90735e-006 -1.410082 -2.986665 0 -1.873835 -3.537804 0 -2.359486 -4.090117 0 -2.855105 142 10 0 0 0 -0.148637 9.53674e-007 -0.0202522 -0.46516 1.90735e-006 -0.11924 -0.90142 1.90735e-006 -0.330817 -1.399636 1.90735e-006 -0.640562 -1.926484 1.90735e-006 -1.019843 -2.468581 2.86102e-006 -1.453855 -3.017554 1.90735e-006 -1.924153 -3.569345 2.38419e-006 -2.415145 -4.122146 1.90735e-006 -2.915462 143 10 0 0 0 -0.16118 -4.76837e-007 -0.0238949 -0.502319 -7.15256e-007 -0.140968 -0.962747 -4.76837e-007 -0.385639 -1.478088 -4.76837e-007 -0.73367 -2.015818 0 -1.14962 -2.56401 -2.38419e-007 -1.615015 -3.116363 -7.15256e-007 -2.111004 -3.670255 -1.43051e-006 -2.623583 -4.224669 -1.19209e-006 -3.143563 144 10 0 0 0 -0.114839 0.114841 -0.024268 -0.360587 0.360589 -0.146399 -0.692378 0.692381 -0.404513 -1.061823 1.061826 -0.771225 -1.445412 1.445414 -1.206938 -1.834891 1.834894 -1.690007 -2.226444 2.226447 -2.200684 -2.618689 2.61869 -2.72568 -3.011151 3.011154 -3.257051 146 10 0 0 0 -0.114557 0.114556 -0.024146 -0.359855 0.359856 -0.145766 -0.691321 0.691322 -0.403159 -1.060617 1.06062 -0.769291 -1.444151 1.444153 -1.204667 -1.833619 1.83362 -1.687614 -2.225177 2.225176 -2.198296 -2.617421 2.617422 -2.723357 -3.009888 3.009887 -3.254825 147 10 0 0 0 -0.116705 0.116704 -0.025081 -0.3651 0.365098 -0.150241 -0.69846 0.698459 -0.41188 -1.068402 1.068401 -0.780549 -1.452018 1.452018 -1.216443 -1.841408 1.841408 -1.698565 -2.232882 2.232881 -2.207865 -2.625065 2.625063 -2.731367 -3.017478 3.017476 -3.261158 150 10 0 0 0 -0.16783 4.76837e-007 -0.0259567 -0.525311 1.19209e-007 -0.156247 -1.001144 1.78814e-007 -0.425053 -1.525949 0 -0.798342 -2.068931 -5.96046e-007 -1.236366 -2.619756 -5.96046e-007 -1.719576 -3.173472 -1.07288e-006 -2.22998 -3.728173 -1.43051e-006 -2.754777 -4.283208 -1.90735e-006 -3.286227 151 10 0 0 0 -0.17248 2.38419e-007 -0.0274526 -0.541826 1.19209e-007 -0.168008 -1.02904 0 -0.456608 -1.56061 -2.38419e-007 -0.850686 -2.107005 -5.96046e-007 -1.305772 -2.65926 -7.15256e-007 -1.800852 -3.213534 -1.07288e-006 -2.318695 -3.768502 -1.60933e-006 -2.848712 -4.323707 -1.90735e-006 -3.384542 152 10 0 0 0 -0.162357 -7.15256e-007 -0.0242531 -0.509112 -7.15256e-007 -0.145754 -0.975568 -9.53674e-007 -0.399545 -1.49507 -1.3113e-006 -0.758226 -2.035453 -1.66893e-006 -1.18483 -2.585114 -2.02656e-006 -1.659675 -3.138278 -2.14577e-006 -2.16373 -3.692681 -2.74181e-006 -2.683485 -4.247504 -3.09944e-006 -3.210525 153 10 0 0 0 -0.169312 0 -0.0264275 -0.529968 -2.38419e-007 -0.159409 -1.008619 -3.27826e-007 -0.432942 -1.535117 -4.76837e-007 -0.811181 -2.078889 -8.9407e-007 -1.252917 -2.630057 -1.2517e-006 -1.738784 -3.183921 -1.57952e-006 -2.251015 -3.738702 -1.96695e-006 -2.777244 -4.293786 -2.29478e-006 -3.309943 156 10 0 0 0 -0.11224 0.112244 -0.0231584 -0.361071 0.361074 -0.14882 -0.699071 0.699075 -0.421264 -1.073051 1.073053 -0.806757 -1.459213 1.459215 -1.260324 -1.849915 1.849916 -1.75806 -2.242036 2.242038 -2.280097 -2.634583 2.634584 -2.814264 -3.027257 3.027257 -3.353946 157 10 0 0 0 -0.119404 0.1194 -0.0262836 -0.374023 0.374025 -0.158784 -0.712275 0.712273 -0.431819 -1.08449 1.08449 -0.809748 -1.469004 1.469002 -1.251544 -1.858747 1.858745 -1.737499 -2.250401 2.250397 -2.249944 -2.642695 2.642694 -2.776349 -3.035205 3.035203 -3.3092 158 10 0 0 0 -0.117027 0.117027 -0.0252238 -0.368379 0.368381 -0.153832 -0.704612 0.704614 -0.422092 -1.076068 1.076066 -0.796852 -1.460455 1.460453 -1.237809 -1.850248 1.850248 -1.72436 -2.241935 2.241935 -2.237487 -2.634254 2.634253 -2.764481 -3.026781 3.026775 -3.297857 159 10 0 0 0 -0.116547 0.116547 -0.0250129 -0.371 0.370998 -0.157303 -0.712083 0.712081 -0.437239 -1.087 1.087 -0.826918 -1.473265 1.473265 -1.281287 -1.86389 1.863888 -1.777958 -2.255956 2.255952 -2.298734 -2.64847 2.648466 -2.831743 -3.041125 3.041121 -3.370319 160 10 0 0 0 -0.119013 0.119011 -0.0261083 -0.374249 0.374249 -0.159348 -0.713752 0.71375 -0.435405 -1.086959 1.086957 -0.817538 -1.4721 1.472096 -1.263653 -1.862157 1.862153 -1.753214 -2.253957 2.253953 -2.268502 -2.646333 2.646329 -2.797163 -3.038895 3.038891 -3.331969 161 10 0 0 0 -0.113919 0.113921 -0.0238724 -0.362892 0.362894 -0.149701 -0.699438 0.699438 -0.418695 -1.07214 1.072138 -0.798661 -1.45753 1.457532 -1.246577 -1.847856 1.847854 -1.739395 -2.239802 2.2398 -2.257587 -2.632261 2.632261 -2.788719 -3.024882 3.02488 -3.325767 162 10 0 0 0 -0.158937 8.19564e-008 -0.0232205 -0.504408 -2.01166e-007 -0.143698 -0.973562 -4.61936e-007 -0.401704 -1.496244 -6.10948e-007 -0.768988 -2.038836 -9.98378e-007 -1.205228 -2.589701 -1.2666e-006 -1.688742 -3.143478 -1.83284e-006 -2.199886 -3.698223 -2.16067e-006 -2.72549 -4.253311 -2.5779e-006 -3.258167 163 10 0 0 0 -0.157766 -1.19209e-007 -0.0228723 -0.505207 2.38419e-007 -0.144924 -0.978624 -3.57628e-007 -0.409759 -1.504818 0 -0.787088 -2.049587 -8.34465e-007 -1.2337 -2.601536 -8.34465e-007 -1.72605 -3.155827 -1.19209e-006 -2.244156 -3.710848 -1.78814e-006 -2.775338 -4.266125 -2.26498e-006 -3.313229 164 10 0 0 0 -0.158974 5.96046e-008 -0.0232307 -0.503023 0 -0.142584 -0.969858 -3.57628e-007 -0.396967 -1.490694 -7.15256e-007 -0.759207 -2.032135 -1.01328e-006 -1.190327 -2.582417 -1.49012e-006 -1.66951 -3.13591 -1.84774e-006 -2.177259 -3.690504 -2.26498e-006 -2.700153 -4.245468 -2.68221e-006 -3.230067 165 10 0 0 0 -0.160787 2.08616e-007 -0.023776 -0.507854 1.19209e-007 -0.145527 -0.977324 -2.38419e-007 -0.404024 -1.499699 -3.8743e-007 -0.770458 -2.041941 -8.64267e-007 -1.205117 -2.592585 -1.3411e-006 -1.686948 -3.146244 -1.60933e-006 -2.196657 -3.700928 -2.08616e-006 -2.721122 -4.255957 -2.5332e-006 -3.252434 166 10 0 0 0 -0.155622 4.76837e-007 -0.0222421 -0.501171 7.15256e-007 -0.142783 -0.974451 0 -0.407396 -1.501114 -4.76837e-007 -0.786124 -2.046291 -7.15256e-007 -1.234801 -2.598461 -1.19209e-006 -1.729113 -3.152859 -1.66893e-006 -2.248813 -3.707956 -1.66893e-006 -2.781803 -4.263287 -2.14577e-006 -3.321484 167 10 0 0 0 -0.15308 0 -0.021506 -0.496895 -4.76837e-007 -0.140674 -0.970369 -9.53674e-007 -0.405602 -1.497673 -9.53674e-007 -0.786246 -2.043344 -1.66893e-006 -1.237466 -2.595779 -1.90735e-006 -1.734223 -3.150307 -2.6226e-006 -2.256001 -3.705452 -2.6226e-006 -2.790238 -4.260786 -3.09944e-006 -3.330122 168 10 0 0 0 -0.10858 0.108582 -0.0216424 -0.351511 0.351517 -0.140611 -0.685822 0.685822 -0.40441 -1.058304 1.058306 -0.783464 -1.443914 1.443914 -1.232958 -1.834415 1.834417 -1.728017 -2.226504 2.226505 -2.249235 -2.619064 2.619066 -2.784037 -3.011778 3.011778 -3.325552 169 10 0 0 0 -0.10639 0.106386 -0.0207597 -0.346827 0.346823 -0.136973 -0.680074 0.680068 -0.398355 -1.052468 1.052463 -0.777035 -1.438235 1.43823 -1.227671 -1.828874 1.828869 -1.724514 -2.220997 2.220989 -2.246577 -2.613555 2.61355 -2.781257 -3.006263 3.006258 -3.32243 170 10 0 0 0 -0.109858 0.109848 -0.022163 -0.354373 0.354365 -0.142903 -0.689296 0.689292 -0.408126 -1.061766 1.061758 -0.787118 -1.447254 1.447248 -1.235712 -1.837796 1.837788 -1.731318 -2.229922 2.229914 -2.253474 -2.622511 2.622503 -2.789186 -3.015235 3.015224 -3.331188 171 10 0 0 0 -0.105022 0.105019 -0.0202179 -0.343718 0.34371 -0.134533 -0.676207 0.676205 -0.394215 -1.048391 1.048388 -0.772007 -1.434088 1.434082 -1.222118 -1.824699 1.82469 -1.718592 -2.216861 2.216856 -2.241632 -2.609468 2.609461 -2.778219 -3.002208 3.002199 -3.321223 172 10 0 0 0 -0.10466 0.104654 -0.020081 -0.342746 0.34274 -0.133738 -0.674857 0.674853 -0.392571 -1.046928 1.046926 -0.769889 -1.432619 1.432615 -1.219962 -1.823231 1.823225 -1.71648 -2.21537 2.215366 -2.238975 -2.607964 2.60796 -2.774956 -3.00069 3.000685 -3.317565 173 10 0 0 0 -0.102733 0.102732 -0.0193338 -0.338578 0.338579 -0.130595 -0.669669 0.669668 -0.387148 -1.041388 1.041389 -0.763018 -1.426964 1.426964 -1.212242 -1.81768 1.81768 -1.710152 -2.209919 2.209919 -2.235021 -2.602558 2.602554 -2.773103 -2.995314 2.99531 -3.317159 174 10 0 0 0 -0.156645 2.38419e-007 -0.0225414 -0.512304 -2.38419e-007 -0.151307 -0.997736 -9.53674e-007 -0.436684 -1.531908 -1.19209e-006 -0.83959 -2.080816 -1.90735e-006 -1.309461 -2.634483 -1.90735e-006 -1.819272 -3.189459 -2.14577e-006 -2.34945 -3.744789 -2.14577e-006 -2.889097 -4.300217 -2.38419e-006 -3.432725 175 10 0 0 0 -0.14448 0 -0.0191152 -0.477793 0 -0.130211 -0.946354 -4.76837e-007 -0.387282 -1.473555 -9.53674e-007 -0.767612 -2.020222 -9.53674e-007 -1.224194 -2.573362 -1.43051e-006 -1.727996 -3.128252 -1.43051e-006 -2.256389 -3.683588 -1.90735e-006 -2.796315 -4.23904 -1.90735e-006 -3.341223 176 10 0 0 0 -0.157417 7.15256e-007 -0.0227699 -0.514271 4.76837e-007 -0.152533 -1.000404 2.38419e-007 -0.439174 -1.534822 0 -0.842936 -2.083796 -2.38419e-007 -1.313216 -2.637472 -2.38419e-007 -1.823123 -3.192444 -9.53674e-007 -2.353244 -3.747768 -1.43051e-006 -2.892775 -4.303194 -1.43051e-006 -3.436266 177 10 0 0 0 -0.153324 7.15256e-007 -0.021576 -0.498114 2.38419e-007 -0.141514 -0.972301 -4.76837e-007 -0.407608 -1.500303 -7.15256e-007 -0.790379 -2.046749 -9.53674e-007 -1.245735 -2.599707 -1.43051e-006 -1.747618 -3.154518 -1.66893e-006 -2.274409 -3.709822 -2.14577e-006 -2.81321 -4.265259 -2.6226e-006 -3.357341 178 10 0 0 0 -0.153559 -7.15256e-007 -0.0216423 -0.499374 -7.15256e-007 -0.142395 -0.976528 -1.43051e-006 -0.413401 -1.506981 -1.19209e-006 -0.803841 -2.054571 -1.90735e-006 -1.26567 -2.607904 -2.14577e-006 -1.77159 -3.162825 -2.86102e-006 -2.300571 -3.718161 -2.6226e-006 -2.840409 -4.273603 -3.33786e-006 -3.384828 179 10 0 0 0 -0.155399 -2.38419e-007 -0.0221766 -0.502834 -4.76837e-007 -0.144222 -0.978479 -1.19209e-006 -0.412719 -1.507378 -9.53674e-007 -0.798249 -2.054209 -1.43051e-006 -1.255747 -2.607265 -1.66893e-006 -1.758665 -3.162094 -1.90735e-006 -2.285846 -3.717396 -2.38419e-006 -2.824696 -4.272831 -2.38419e-006 -3.368683 180 10 0 0 0 -0.0874634 0.0874634 -0.0139434 -0.277185 0.277187 -0.0830307 -0.548862 0.548867 -0.237309 -0.871241 0.871244 -0.475394 -1.221861 1.221865 -0.780407 -1.590704 1.590709 -1.144768 -1.969454 1.969458 -1.552886 -2.352994 2.352999 -1.988289 -2.738373 2.738376 -2.436099 181 10 0 0 0 -0.0878696 0.0878725 -0.014076 -0.278812 0.278813 -0.0841138 -0.552269 0.552275 -0.24082 -0.876073 0.876076 -0.481818 -1.227581 1.227585 -0.78933 -1.597007 1.597011 -1.155956 -1.976297 1.976303 -1.566887 -2.360271 2.360274 -2.005087 -2.746038 2.746041 -2.455724 182 10 0 0 0 -0.0827436 0.0827384 -0.0124619 -0.262852 0.262845 -0.0742931 -0.52297 0.522967 -0.213536 -0.834505 0.834501 -0.430658 -1.17602 1.176017 -0.711666 -1.537921 1.537918 -1.051126 -1.911498 1.911494 -1.434866 -2.290853 2.290848 -1.846118 -2.672385 2.672381 -2.269348 183 10 0 0 0 -0.0893345 0.0893288 -0.0145533 -0.282795 0.282791 -0.0865932 -0.558939 0.558934 -0.247014 -0.885432 0.885428 -0.493624 -1.239371 1.239367 -0.808148 -1.610624 1.61062 -1.182087 -1.991161 1.991159 -1.59972 -2.376064 2.376061 -2.044188 -2.762646 2.762642 -2.500982 184 10 0 0 0 -0.0851707 0.0851679 -0.0132155 -0.270498 0.270496 -0.0789258 -0.537148 0.537143 -0.226509 -0.854561 0.854558 -0.454743 -1.200864 1.200861 -0.748022 -1.566429 1.566428 -1.100198 -1.942875 1.942871 -1.496934 -2.3246 2.324598 -1.921297 -2.708328 2.708323 -2.357911 185 10 0 0 0 -0.0864592 0.086457 -0.0136213 -0.274892 0.274894 -0.0817072 -0.545963 0.545961 -0.235163 -0.86818 0.868178 -0.472926 -1.218662 1.218661 -0.777543 -1.587203 1.587201 -1.14075 -1.965688 1.965686 -1.547517 -2.348937 2.348935 -1.981086 -2.734108 2.734108 -2.427429 186 10 0 0 0 -0.127398 -5.72205e-006 -0.014806 -0.40152 -9.53674e-007 -0.087141 -0.790948 -1.90735e-006 -0.246483 -1.249946 -2.86102e-006 -0.489048 -1.747025 -1.90735e-006 -0.796502 -2.268763 0 -1.16118 -2.804005 -2.86102e-006 -1.567886 -3.345723 -3.8147e-006 -2.000212 -3.890043 -3.8147e-006 -2.444617 187 10 0 0 0 -0.138203 -9.53674e-007 -0.0174637 -0.434978 -1.43051e-006 -0.103374 -0.851094 -9.53674e-007 -0.290842 -1.334004 -9.53674e-007 -0.57173 -1.850369 -9.53674e-007 -0.922325 -2.386071 0 -1.33069 -2.931209 4.76837e-007 -1.779155 -3.480464 -4.76837e-007 -2.25128 -4.031246 -4.76837e-007 -2.734166 188 10 0 0 0 -0.130962 4.76837e-006 -0.0156555 -0.412623 5.72205e-006 -0.0923481 -0.81119 5.72205e-006 -0.260881 -1.278711 5.72205e-006 -0.516325 -1.782963 5.72205e-006 -0.838702 -2.310134 6.67572e-006 -1.218945 -2.849389 5.72205e-006 -1.640919 -3.394258 4.76837e-006 -2.088032 -3.94121 5.72205e-006 -2.546195 189 10 0 0 0 -0.112315 9.53674e-007 -0.0114696 -0.356786 0 -0.068151 -0.711742 9.53674e-007 -0.196333 -1.139456 3.8147e-006 -0.397342 -1.610696 1.90735e-006 -0.658658 -2.112808 1.90735e-006 -0.976462 -2.633375 2.86102e-006 -1.337982 -3.163347 3.8147e-006 -1.726889 -3.6967 1.90735e-006 -2.126958 190 10 0 0 0 -0.134276 -9.53674e-007 -0.0164709 -0.422411 1.90735e-006 -0.0970299 -0.828143 0 -0.273082 -1.301732 0 -0.538198 -1.810641 9.53674e-007 -0.870923 -2.341043 9.53674e-007 -1.261203 -2.882505 0 -1.692417 -3.429007 -9.53674e-007 -2.148082 -3.977375 -1.90735e-006 -2.614551 191 10 0 0 0 -0.136442 -1.90735e-006 -0.017015 -0.429441 -9.53674e-007 -0.10056 -0.841145 -1.90735e-006 -0.283096 -1.320205 -9.53674e-007 -0.557327 -1.833579 -9.53674e-007 -0.90054 -2.367216 -1.90735e-006 -1.301586 -2.910962 -1.90735e-006 -1.743203 -3.459203 -1.90735e-006 -2.208898 -4.009109 -3.8147e-006 -2.685435 192 10 0 0 0 -0.0954046 0.0954061 -0.0166321 -0.299684 0.299684 -0.0976534 -0.586781 0.586781 -0.274007 -0.921502 0.921501 -0.538754 -1.280993 1.280993 -0.870322 -1.655648 1.655646 -1.258811 -2.038161 2.038157 -1.687824 -2.424275 2.424271 -2.141048 -2.811727 2.811724 -2.604928 193 10 0 0 0 -0.100296 0.100298 -0.0184114 -0.31497 0.314971 -0.108702 -0.614161 0.614162 -0.304244 -0.95944 0.959443 -0.594836 -1.327211 1.327212 -0.955114 -1.707645 1.707645 -1.372175 -2.094107 2.094107 -1.828065 -2.483153 2.483153 -2.306605 -2.873143 2.873143 -2.795396 194 10 0 0 0 -0.0929775 0.0929685 -0.0157857 -0.292476 0.292466 -0.0927565 -0.574223 0.574215 -0.26117 -0.904325 0.904323 -0.515564 -1.260212 1.260205 -0.835876 -1.632299 1.63229 -1.213264 -2.013018 2.013011 -1.6319 -2.397785 2.397777 -2.075432 -2.784182 2.784173 -2.530767 195 10 0 0 0 -0.100347 0.100346 -0.01843 -0.315006 0.315002 -0.108706 -0.61407 0.614068 -0.304032 -0.959166 0.959165 -0.59415 -1.326753 1.32675 -0.953742 -1.707042 1.70704 -1.370006 -2.093397 2.093396 -1.825064 -2.482355 2.482355 -2.302752 -2.87227 2.872268 -2.790647 196 10 0 0 0 -0.104291 0.104291 -0.0199354 -0.325868 0.325868 -0.116742 -0.631455 0.631455 -0.323192 -0.981145 0.981145 -0.625618 -1.351659 1.351657 -0.996558 -1.73361 1.73361 -1.422244 -2.120898 2.1209 -1.884747 -2.510434 2.510434 -2.368425 -2.900783 2.900781 -2.861533 197 10 0 0 0 -0.10075 0.100754 -0.0185816 -0.315219 0.315222 -0.108681 -0.613106 0.61311 -0.302066 -0.956533 0.956537 -0.587877 -1.322426 1.322427 -0.941228 -1.70133 1.701334 -1.35016 -2.086643 2.086647 -1.79749 -2.474779 2.474781 -2.267323 -2.864002 2.864004 -2.74774 198 10 0 0 0 -0.134805 2.86102e-006 -0.0166035 -0.423509 1.90735e-006 -0.0975058 -0.829195 1.90735e-006 -0.273509 -1.302075 2.86102e-006 -0.537468 -1.809928 9.53674e-007 -0.867799 -2.339285 1.90735e-006 -1.254755 -2.879895 1.90735e-006 -1.682322 -3.425977 9.53674e-007 -2.135711 -3.974366 9.53674e-007 -2.602319 199 10 0 0 0 -0.133131 9.53674e-007 -0.0161865 -0.418551 9.53674e-007 -0.0951102 -0.820624 1.90735e-006 -0.267287 -1.290464 9.53674e-007 -0.526378 -1.795996 9.53674e-007 -0.851542 -2.323737 9.53674e-007 -1.233517 -2.863204 9.53674e-007 -1.656344 -3.408433 0 -2.105266 -3.956082 0 -2.567417 200 10 0 0 0 -0.157845 -2.38419e-006 -0.0228959 -0.489669 -9.53674e-007 -0.132878 -0.939035 -1.43051e-006 -0.361766 -1.44548 -9.53674e-007 -0.688946 -1.97699 -9.53674e-007 -1.082823 -2.52142 -1.90735e-006 -1.527752 -3.071546 -2.38419e-006 -2.005815 -3.624043 -2.38419e-006 -2.503147 -4.177446 -2.86102e-006 -3.009862 201 10 0 0 0 -0.136858 0 -0.0171206 -0.429406 9.53674e-007 -0.100384 -0.839208 2.86102e-006 -0.280833 -1.3155 2.86102e-006 -0.550404 -1.825949 1.90735e-006 -0.886679 -2.357087 1.90735e-006 -1.279341 -2.898862 9.53674e-007 -1.71193 -3.445632 0 -2.16906 -3.994454 0 -2.638384 202 10 0 0 0 -0.142166 4.76837e-007 -0.0184991 -0.445902 9.53674e-007 -0.108881 -0.868141 1.43051e-006 -0.303388 -1.354659 4.76837e-007 -0.590719 -1.872709 4.76837e-007 -0.945612 -2.409349 -4.76837e-007 -1.357433 -2.955206 -4.76837e-007 -1.809638 -3.50506 -4.76837e-007 -2.285808 -4.056433 -4.76837e-007 -2.773305 203 10 0 0 0 -0.146358 -2.86102e-006 -0.0196257 -0.457405 -9.53674e-007 -0.114863 -0.886726 -1.43051e-006 -0.31782 -1.378534 -9.53674e-007 -0.61498 -1.90021 -9.53674e-007 -0.979496 -2.438696 -1.90735e-006 -1.398393 -2.985462 -1.43051e-006 -1.855524 -3.535908 -1.43051e-006 -2.33589 -4.08777 -2.38419e-006 -2.827504 206 10 0 0 0 -0.108521 0.108521 -0.0216178 -0.33844 0.338438 -0.126708 -0.652162 0.652158 -0.347898 -1.007538 1.007536 -0.666692 -1.381623 1.381619 -1.052641 -1.765739 1.765736 -1.49179 -2.154418 2.154413 -1.966741 -2.544965 2.544961 -2.462402 -2.936198 2.936193 -2.967813 208 10 0 0 0 -0.110103 0.110095 -0.0222651 -0.342845 0.342838 -0.130268 -0.659209 0.6592 -0.356473 -1.016437 1.016429 -0.680907 -1.391713 1.391707 -1.072216 -1.776469 1.77646 -1.515645 -2.165481 2.165471 -1.993864 -2.556237 2.556228 -2.492318 -2.947633 2.947622 -3.000288 209 10 0 0 0 -0.107703 0.107697 -0.0212862 -0.336419 0.336415 -0.125158 -0.649187 0.649183 -0.344563 -1.003963 1.003956 -0.66156 -1.3778 1.377794 -1.046419 -1.761856 1.761848 -1.485183 -2.150535 2.150525 -1.960105 -2.541074 2.541064 -2.455669 -2.932298 2.93229 -2.960953 211 10 0 0 0 -0.141257 -9.53674e-007 -0.0182576 -0.445153 0 -0.108743 -0.869525 -1.43051e-006 -0.305758 -1.35962 -1.90735e-006 -0.599684 -1.881326 -2.86102e-006 -0.964287 -2.420568 -2.86102e-006 -1.386199 -2.968005 -2.86102e-006 -1.847137 -3.51892 -3.8147e-006 -2.330995 -4.071153 -3.8147e-006 -2.825865 213 10 0 0 0 -0.143537 -1.43051e-006 -0.0188627 -0.452388 -1.43051e-006 -0.112624 -0.882728 -9.53674e-007 -0.316826 -1.377381 -4.76837e-007 -0.619479 -1.902272 -1.43051e-006 -0.993019 -2.443439 -9.53674e-007 -1.422948 -2.992005 -1.90735e-006 -1.890656 -3.543639 -1.43051e-006 -2.380329 -4.096433 -2.38419e-006 -2.880544 216 10 0 0 0 -0.111258 0.111256 -0.0227458 -0.353912 0.353909 -0.141404 -0.684513 0.684507 -0.39688 -1.053474 1.053471 -0.761702 -1.436797 1.436791 -1.195726 -1.82614 1.826135 -1.677339 -2.217649 2.217643 -2.187247 -2.609916 2.60991 -2.712975 -3.002446 3.002437 -3.246339 217 10 0 0 0 -0.109549 0.109545 -0.0220381 -0.346472 0.34647 -0.13445 -0.670403 0.670399 -0.375708 -1.034601 1.034599 -0.723033 -1.415068 1.415064 -1.140267 -1.802964 1.80296 -1.607996 -2.193738 2.193734 -2.106719 -2.585556 2.585552 -2.622271 -2.977741 2.977736 -3.145879 218 10 0 0 0 -0.11094 0.110934 -0.0226122 -0.352682 0.352678 -0.14026 -0.682278 0.68227 -0.393528 -1.050528 1.050524 -0.755637 -1.433431 1.433423 -1.187023 -1.822555 1.822547 -1.666392 -2.213971 2.213964 -2.174761 -2.606197 2.60619 -2.699328 -2.998696 2.998687 -3.231738 219 10 0 0 0 -0.111506 0.111506 -0.0228482 -0.35302 0.353022 -0.140248 -0.681658 0.681656 -0.391436 -1.049078 1.049078 -0.750406 -1.431463 1.431465 -1.178671 -1.820343 1.820341 -1.655571 -2.211615 2.211613 -2.161538 -2.603709 2.603705 -2.682978 -2.996096 2.996094 -3.211901 220 10 0 0 0 -0.1121 0.112092 -0.0230969 -0.355446 0.355442 -0.142528 -0.68614 0.686134 -0.398208 -1.054998 1.054995 -0.762656 -1.438245 1.438239 -1.196179 -1.827552 1.827546 -1.67744 -2.219028 2.219025 -2.186823 -2.61124 2.611235 -2.711008 -3.003708 3.0037 -3.242478 221 10 0 0 0 -0.110392 0.11039 -0.0223857 -0.34893 0.348928 -0.136534 -0.67433 0.674326 -0.380844 -1.039479 1.039475 -0.731526 -1.420503 1.420499 -1.151851 -1.808684 1.808681 -1.622135 -2.199604 2.199598 -2.122889 -2.591508 2.591504 -2.640127 -2.983747 2.983743 -3.165081 222 10 0 0 0 -0.15731 4.76837e-007 -0.0227364 -0.500717 1.78814e-007 -0.141583 -0.968439 -1.78814e-007 -0.397341 -1.4902 -5.96046e-007 -0.762089 -2.032249 -1.01328e-006 -1.195887 -2.583017 -1.3113e-006 -1.678664 -3.136845 -1.72853e-006 -2.190454 -3.691694 -2.08616e-006 -2.718007 -4.246876 -2.563e-006 -3.253105 223 10 0 0 0 -0.159042 3.57628e-007 -0.0232515 -0.504852 5.96046e-007 -0.143999 -0.974178 -1.19209e-007 -0.402275 -1.496853 -2.38419e-007 -0.76954 -2.039421 -8.34465e-007 -1.20568 -2.590277 -1.19209e-006 -1.689104 -3.144053 -1.3113e-006 -2.200215 -3.698797 -1.54972e-006 -2.725752 -4.253857 -1.78814e-006 -3.257776 224 10 0 0 0 -0.155331 0 -0.0221563 -0.492983 2.38419e-007 -0.136541 -0.954201 -7.15256e-007 -0.38239 -1.471275 -1.19209e-006 -0.734782 -2.010805 -1.43051e-006 -1.157861 -2.560299 -1.66893e-006 -1.631585 -3.113544 -2.14577e-006 -2.136497 -3.668074 -2.6226e-006 -2.658275 -4.22303 -2.86102e-006 -3.188066 225 10 0 0 0 -0.152542 0 -0.0213518 -0.48402 -4.76837e-007 -0.131077 -0.939308 -7.15256e-007 -0.368268 -1.452591 -1.19209e-006 -0.711262 -1.989563 -1.19209e-006 -1.124335 -2.537527 -1.66893e-006 -1.588353 -3.089941 -2.14577e-006 -2.084914 -3.644001 -2.6226e-006 -2.599768 -4.198652 -2.6226e-006 -3.1236 226 10 0 0 0 -0.157314 -4.76837e-007 -0.0227385 -0.498177 0 -0.139596 -0.96283 -4.76837e-007 -0.390617 -1.48274 -7.15256e-007 -0.750379 -2.023876 -1.19209e-006 -1.180178 -2.574097 -1.66893e-006 -1.658917 -3.127602 -1.43051e-006 -2.166787 -3.682219 -1.90735e-006 -2.69006 -4.2372 -2.38419e-006 -3.22037 227 10 0 0 0 -0.157635 1.02445e-007 -0.0228335 -0.502411 2.98023e-008 -0.142761 -0.971796 -1.19209e-007 -0.401132 -1.494747 -6.85453e-007 -0.769161 -2.037502 -1.01328e-006 -1.206139 -2.588615 -1.2815e-006 -1.691593 -3.142607 -1.69873e-006 -2.205523 -3.697542 -2.14577e-006 -2.734777 -4.252773 -2.5332e-006 -3.271341 228 10 0 0 0 -0.116134 0.116131 -0.0248307 -0.37283 0.372831 -0.15985 -0.717472 0.71747 -0.448782 -1.095064 1.09506 -0.851056 -1.482935 1.482931 -1.318535 -1.874286 1.874283 -1.825843 -2.266633 2.266632 -2.353676 -2.659266 2.659264 -2.891252 -3.051983 3.051979 -3.433022 229 10 0 0 0 -0.113427 0.113423 -0.0236606 -0.360719 0.360715 -0.147551 -0.696092 0.696088 -0.413804 -1.068462 1.068457 -0.792386 -1.453873 1.453867 -1.240436 -1.844295 1.844287 -1.73441 -2.236318 2.236312 -2.254235 -2.628821 2.628816 -2.787037 -3.021483 3.021474 -3.325513 230 10 0 0 0 -0.114455 0.11446 -0.0241051 -0.368882 0.368883 -0.156367 -0.712321 0.712324 -0.44222 -1.089536 1.089537 -0.842662 -1.477329 1.477331 -1.309467 -1.868687 1.868689 -1.816919 -2.261065 2.261063 -2.34552 -2.653707 2.653706 -2.883841 -3.046436 3.046434 -3.426338 231 10 0 0 0 -0.11582 0.11582 -0.024694 -0.371776 0.371773 -0.158803 -0.715792 0.715788 -0.446134 -1.093075 1.093073 -0.846906 -1.480823 1.480818 -1.313311 -1.87214 1.872139 -1.820032 -2.264484 2.264482 -2.347657 -2.657116 2.65711 -2.885152 -3.049829 3.049824 -3.426891 232 10 0 0 0 -0.115389 0.115395 -0.0245082 -0.371536 0.37154 -0.15885 -0.716189 0.716194 -0.447819 -1.093933 1.093937 -0.850842 -1.481907 1.481909 -1.319269 -1.873313 1.873316 -1.82746 -2.26569 2.265692 -2.356039 -2.658332 2.658335 -2.894207 -3.051061 3.05106 -3.436491 233 10 0 0 0 -0.113763 0.113762 -0.0238041 -0.367203 0.367203 -0.154885 -0.710101 0.710098 -0.439363 -1.087139 1.087135 -0.838933 -1.474886 1.474882 -1.305354 -1.866243 1.86624 -1.812764 -2.258624 2.258621 -2.341591 -2.651283 2.651276 -2.880176 -3.044021 3.044014 -3.422939 234 10 0 0 0 -0.173683 -3.57628e-007 -0.0278472 -0.551647 -3.57628e-007 -0.176234 -1.047579 -5.96046e-007 -0.4814 -1.584219 -1.19209e-006 -0.893224 -2.133036 -1.07288e-006 -1.362529 -2.686275 -1.66893e-006 -1.867394 -3.240961 -1.90735e-006 -2.391891 -3.796116 -2.02656e-006 -2.926335 -4.351427 -2.26498e-006 -3.465465 235 10 0 0 0 -0.173813 2.38419e-007 -0.0278889 -0.549404 -1.19209e-007 -0.174086 -1.042953 -6.55651e-007 -0.474585 -1.578331 -1.01328e-006 -0.881765 -2.126501 -1.2815e-006 -1.347036 -2.67948 -1.49012e-006 -1.849134 -3.23406 -1.78814e-006 -2.371791 -3.78916 -2.02656e-006 -2.904814 -4.344421 -2.17557e-006 -3.442318 236 10 0 0 0 -0.168629 -2.38419e-007 -0.0262098 -0.540073 -4.76837e-007 -0.168641 -1.03331 -1.19209e-006 -0.468541 -1.569683 -1.66893e-006 -0.879366 -2.118906 -1.43051e-006 -1.351257 -2.672436 -1.90735e-006 -1.859445 -3.22728 -2.38419e-006 -2.386875 -3.78252 -2.14577e-006 -2.923714 -4.337883 -2.6226e-006 -3.464598 237 10 0 0 0 -0.165037 2.38419e-007 -0.0250801 -0.532387 -4.76837e-007 -0.163867 -1.023975 -1.19209e-006 -0.460613 -1.56004 -1.43051e-006 -0.870301 -2.109329 -1.19209e-006 -1.342656 -2.662977 -1.90735e-006 -1.852174 -3.217886 -2.14577e-006 -2.380925 -3.773165 -2.14577e-006 -2.918903 -4.328552 -2.6226e-006 -3.460829 238 10 0 0 0 -0.170338 3.57628e-007 -0.026757 -0.542795 -3.57628e-007 -0.170101 -1.035809 -3.57628e-007 -0.469574 -1.57181 -5.96046e-007 -0.879033 -2.120781 -1.07288e-006 -1.349326 -2.674162 -1.3113e-006 -1.855779 -3.228935 -1.66893e-006 -2.381869 -3.784142 -1.90735e-006 -2.917674 -4.339478 -2.02656e-006 -3.457661 239 10 0 0 0 -0.166012 1.19209e-007 -0.0253836 -0.526749 -2.98023e-008 -0.158434 -1.009266 -2.98023e-007 -0.438636 -1.53961 -7.45058e-007 -0.828725 -2.086071 -1.07288e-006 -1.284165 -2.63858 -1.43051e-006 -1.781631 -3.193035 -1.78814e-006 -2.302202 -3.748096 -2.02656e-006 -2.834328 -4.303347 -2.32458e-006 -3.371495 240 10 0 0 0 -0.0846157 0.0846181 -0.0130417 -0.267694 0.267697 -0.0770617 -0.530689 0.530691 -0.219929 -0.844193 0.844195 -0.440714 -1.186693 1.186696 -0.724189 -1.548925 1.548925 -1.064758 -1.922489 1.922493 -1.448443 -2.30167 2.301674 -1.858792 -2.682924 2.682926 -2.280429 241 10 0 0 0 -0.0843525 0.0843577 -0.012964 -0.26688 0.266887 -0.0765734 -0.529182 0.529187 -0.218559 -0.842003 0.842007 -0.438067 -1.183895 1.183898 -0.720016 -1.545633 1.545639 -1.058936 -1.918814 1.918818 -1.440953 -2.297676 2.29768 -1.849649 -2.678638 2.678643 -2.269636 242 10 0 0 0 -0.0774765 0.0774755 -0.0109107 -0.245949 0.245951 -0.0645947 -0.491001 0.491001 -0.185936 -0.787151 0.787154 -0.376484 -1.114395 1.114393 -0.624681 -1.464281 1.46428 -0.927652 -1.828081 1.828082 -1.273587 -2.199127 2.199127 -1.646676 -2.572872 2.572871 -2.031148 243 10 0 0 0 -0.082283 0.0822883 -0.0123246 -0.260553 0.260564 -0.0728273 -0.517647 0.517649 -0.208317 -0.825434 0.82544 -0.41866 -1.162918 1.162925 -0.689874 -1.521141 1.521147 -1.017404 -1.891598 1.891602 -1.388122 -2.268351 2.268355 -1.78633 -2.647831 2.647837 -2.198264 244 10 0 0 0 -0.0909195 0.090919 -0.0150852 -0.286392 0.2864 -0.0887477 -0.563601 0.563608 -0.250667 -0.889746 0.889749 -0.496532 -1.242458 1.242465 -0.807498 -1.612255 1.612262 -1.175593 -1.991352 1.991358 -1.585481 -2.37491 2.374916 -2.021011 -2.760376 2.760381 -2.469445 245 10 0 0 0 -0.0795383 0.0795426 -0.0115082 -0.252232 0.252235 -0.0680676 -0.502508 0.502511 -0.195413 -0.803774 0.803775 -0.394427 -1.135567 1.135571 -0.652546 -1.489208 1.489211 -0.966196 -1.856041 1.856042 -1.32298 -2.229668 2.229667 -1.706921 -2.606026 2.606028 -2.103256 246 10 0 0 0 -0.104442 -2.86102e-006 -0.00990748 -0.332262 -4.76837e-006 -0.0587692 -0.665767 -2.86102e-006 -0.170006 -1.071623 -2.86102e-006 -0.346188 -1.52273 -1.90735e-006 -0.577484 -2.007894 -2.86102e-006 -0.862375 -2.515129 -2.86102e-006 -1.191316 -3.035049 -3.8147e-006 -1.551103 -3.560429 -4.76837e-006 -1.926044 247 10 0 0 0 -0.109769 1.90735e-006 -0.0109527 -0.349731 9.53674e-007 -0.0654492 -0.699817 3.8147e-006 -0.189634 -1.124777 3.8147e-006 -0.387347 -1.595768 1.90735e-006 -0.648272 -2.099838 1.90735e-006 -0.970251 -2.623642 2.86102e-006 -1.340685 -3.1576 3.8147e-006 -1.742836 -3.695581 3.8147e-006 -2.159778 248 10 0 0 0 -0.105958 1.90735e-006 -0.0101984 -0.336892 3.8147e-006 -0.0604682 -0.674332 4.76837e-006 -0.174688 -1.084135 3.8147e-006 -0.355142 -1.53884 1.90735e-006 -0.591498 -2.027027 2.86102e-006 -0.881882 -2.536542 1.90735e-006 -1.215994 -3.058217 9.53674e-007 -1.580512 -3.585275 9.53674e-007 -1.960427 249 10 0 0 0 -0.110042 1.90735e-006 -0.0110061 -0.350594 1.90735e-006 -0.0657845 -0.701509 9.53674e-007 -0.190643 -1.127352 1.90735e-006 -0.389407 -1.599121 -1.90735e-006 -0.651578 -2.103817 -9.53674e-007 -0.974919 -2.628082 -1.90735e-006 -1.346665 -3.162394 -1.90735e-006 -1.750053 -3.700687 -2.86102e-006 -2.168194 250 10 0 0 0 -0.112637 -4.76837e-006 -0.0115364 -0.357273 -3.8147e-006 -0.0682993 -0.711815 -1.90735e-006 -0.196136 -1.138496 -2.86102e-006 -0.395905 -1.608329 -4.76837e-006 -0.654984 -2.10903 -2.86102e-006 -0.969831 -2.628474 -2.86102e-006 -1.32836 -3.158249 -4.76837e-006 -1.716636 -3.692703 -4.76837e-006 -2.120527 251 10 0 0 0 -0.120418 -5.72205e-006 -0.0132043 -0.380874 -5.72205e-006 -0.0780411 -0.754562 -5.72205e-006 -0.2225 -1.199379 -4.76837e-006 -0.44522 -1.684823 -6.67572e-006 -0.730617 -2.197866 -6.67572e-006 -1.073031 -2.727108 -6.67572e-006 -1.459634 -3.264835 -7.62939e-006 -1.875583 -3.806242 -8.58307e-006 -2.306546 252 10 0 0 0 -0.0890999 0.0890965 -0.0144763 -0.281935 0.281932 -0.0860183 -0.556927 0.556925 -0.244834 -0.881777 0.881775 -0.488003 -1.234533 1.234529 -0.799076 -1.605267 1.60526 -1.170902 -1.985609 1.985607 -1.587474 -2.370558 2.370554 -2.032245 -2.757383 2.757379 -2.490968 253 10 0 0 0 -0.0878048 0.0877962 -0.0140541 -0.278215 0.278209 -0.0836787 -0.550587 0.550577 -0.238898 -0.873611 0.873604 -0.478304 -1.225212 1.225201 -0.786064 -1.595123 1.595119 -1.154629 -1.975008 1.975003 -1.568697 -2.359646 2.359639 -2.011339 -2.746204 2.746197 -2.46795 254 10 0 0 0 -0.0856018 0.0855966 -0.0133483 -0.271286 0.271287 -0.0793281 -0.537884 0.537885 -0.226847 -0.854996 0.854993 -0.45449 -1.201344 1.20134 -0.747886 -1.56753 1.567527 -1.102298 -1.944759 1.944757 -1.502819 -2.327429 2.327426 -1.932785 -2.712378 2.712374 -2.377579 255 10 0 0 0 -0.0850849 0.0850787 -0.0131857 -0.269623 0.269619 -0.0782995 -0.534771 0.534762 -0.223932 -0.850477 0.850469 -0.448881 -1.195383 1.195374 -0.738502 -1.560394 1.560383 -1.08869 -1.936804 1.936792 -1.485276 -2.318825 2.318815 -1.911363 -2.703218 2.703208 -2.352378 256 10 0 0 0 -0.0872288 0.0872316 -0.0138681 -0.275793 0.275792 -0.0820508 -0.545366 0.545366 -0.233502 -0.864887 0.864885 -0.46586 -1.212362 1.212362 -0.76226 -1.578695 1.578696 -1.117207 -1.9559 1.955898 -1.517588 -2.338533 2.338529 -1.947321 -2.723368 2.723365 -2.391339 257 10 0 0 0 -0.0939827 0.0939784 -0.0161326 -0.296446 0.296447 -0.095602 -0.58218 0.582178 -0.2699 -0.916027 0.916029 -0.53266 -1.275532 1.275527 -0.864253 -1.650846 1.650845 -1.255742 -2.034208 2.034207 -1.690037 -2.421228 2.421226 -2.150295 -2.809708 2.809705 -2.623374 258 10 0 0 0 -0.121662 9.53674e-007 -0.013485 -0.388704 2.86102e-006 -0.0818741 -0.774349 3.8147e-006 -0.23753 -1.233192 3.8147e-006 -0.479868 -1.732323 2.86102e-006 -0.791472 -2.256962 3.8147e-006 -1.164281 -2.795128 2.86102e-006 -1.581928 -3.33955 9.53674e-007 -2.026829 -3.886309 9.53674e-007 -2.483904 259 10 0 0 0 -0.129976 0 -0.0154183 -0.413723 1.90735e-006 -0.0933447 -0.818134 9.53674e-007 -0.267989 -1.292749 1.90735e-006 -0.534782 -1.803661 0 -0.872142 -2.336143 9.53674e-007 -1.26925 -2.879442 0 -1.708743 -3.427555 0 -2.173683 -3.977482 -9.53674e-007 -2.650358 260 10 0 0 0 -0.121478 9.53674e-007 -0.0134439 -0.386106 3.8147e-006 -0.0805166 -0.767944 1.90735e-006 -0.232537 -1.223211 9.53674e-007 -0.469696 -1.71953 9.53674e-007 -0.77563 -2.242239 1.90735e-006 -1.142986 -2.779186 9.53674e-007 -1.555958 -3.322914 9.53674e-007 -1.997488 -3.869303 -9.53674e-007 -2.452535 261 10 0 0 0 -0.120115 9.53674e-007 -0.0131404 -0.383738 0 -0.0796723 -0.765219 9.53674e-007 -0.231354 -1.220345 -9.53674e-007 -0.468312 -1.716568 0 -0.77406 -2.239199 0 -1.141202 -2.776072 -9.53674e-007 -1.553899 -3.319635 -9.53674e-007 -1.994638 -3.865726 -9.53674e-007 -2.448095 262 10 0 0 0 -0.112531 1.90735e-006 -0.0115163 -0.359097 6.67572e-006 -0.0692282 -0.719738 4.76837e-006 -0.202196 -1.155832 2.86102e-006 -0.413568 -1.636703 2.86102e-006 -0.690907 -2.148289 2.86102e-006 -1.029847 -2.677542 2.86102e-006 -1.416482 -3.215565 1.90735e-006 -1.833573 -3.756994 2.86102e-006 -2.264644 263 10 0 0 0 -0.118165 -9.53674e-007 -0.0127122 -0.375686 9.53674e-007 -0.0759997 -0.748236 9.53674e-007 -0.219428 -1.194836 0 -0.444546 -1.684242 0 -0.737187 -2.202065 -9.53674e-007 -1.091493 -2.735695 -9.53674e-007 -1.492508 -3.277039 -9.53674e-007 -1.923206 -3.821444 0 -2.368004 266 10 0 0 0 -0.0932312 0.0932279 -0.0158715 -0.296925 0.296921 -0.0963874 -0.586496 0.586498 -0.276532 -0.925158 0.925156 -0.550549 -1.288706 1.288702 -0.895608 -1.666731 1.666728 -1.300052 -2.051821 2.051819 -1.745816 -2.439972 2.43997 -2.215808 -2.829241 2.829235 -2.696613 268 10 0 0 0 -0.0982647 0.0982695 -0.0176625 -0.310539 0.310547 -0.105756 -0.608643 0.608649 -0.299499 -0.953831 0.953835 -0.589847 -1.321899 1.321904 -0.951252 -1.7027 1.702707 -1.370347 -2.089497 2.089502 -1.828869 -2.478857 2.478864 -2.310632 -2.86919 2.869195 -2.803562 269 10 0 0 0 -0.09729 0.0972867 -0.0173063 -0.308592 0.308594 -0.104521 -0.606009 0.606009 -0.297129 -0.950705 0.950705 -0.586204 -1.318426 1.318426 -0.946296 -1.699024 1.699022 -1.364241 -2.085697 2.085698 -1.821789 -2.474977 2.474976 -2.302735 -2.865232 2.865229 -2.794686 270 10 0 0 0 -0.139114 0 -0.0176978 -0.442112 9.53674e-007 -0.107597 -0.866829 -4.76837e-007 -0.305021 -1.357138 -1.43051e-006 -0.59935 -1.879059 -4.76837e-007 -0.964537 -2.418322 -1.43051e-006 -1.386527 -2.965755 -2.38419e-006 -1.847447 -3.516649 -2.86102e-006 -2.331166 -4.068836 -2.38419e-006 -2.825633 276 10 0 0 0 -0.109844 0.109838 -0.0221602 -0.348211 0.3482 -0.136119 -0.673981 0.673974 -0.38121 -1.03953 1.039526 -0.733326 -1.420849 1.420841 -1.155328 -1.809208 1.809198 -1.627178 -2.200218 2.200211 -2.129322 -2.592175 2.592168 -2.647759 -2.984455 2.984443 -3.173655 277 10 0 0 0 -0.112581 0.112587 -0.023303 -0.356611 0.356618 -0.143496 -0.687492 0.687498 -0.399591 -1.056204 1.056209 -0.763454 -1.439257 1.43926 -1.195757 -1.828423 1.828426 -1.675566 -2.219822 2.219822 -2.18354 -2.611977 2.611977 -2.70635 -3.004379 3.004379 -3.235832 278 10 0 0 0 -0.108885 0.10888 -0.0217651 -0.344936 0.344932 -0.133248 -0.667948 0.667944 -0.372622 -1.031765 1.031756 -0.718608 -1.412197 1.41219 -1.135657 -1.800192 1.800185 -1.604239 -2.191072 2.191066 -2.104459 -2.582985 2.582977 -2.621875 -2.97525 2.97524 -3.14733 279 10 0 0 0 -0.108088 0.108086 -0.0214411 -0.342083 0.342086 -0.130755 -0.663408 0.663411 -0.366719 -1.026045 1.026047 -0.708655 -1.405697 1.405694 -1.121472 -1.793221 1.793221 -1.585993 -2.183842 2.183841 -2.082587 -2.575584 2.575583 -2.596737 -2.967714 2.967711 -3.11904 280 10 0 0 0 -0.106207 0.106211 -0.0206891 -0.337021 0.337025 -0.126703 -0.655224 0.655231 -0.356469 -1.015383 1.015383 -0.690167 -1.393314 1.393316 -1.094147 -1.779976 1.779979 -1.551594 -2.170174 2.170177 -2.042855 -2.561718 2.561722 -2.553414 -2.953747 2.953748 -3.073357 281 10 0 0 0 -0.0998631 0.0998631 -0.01825 -0.317366 0.317362 -0.111173 -0.621374 0.621369 -0.314872 -0.970816 0.970812 -0.616616 -1.34178 1.341779 -0.989387 -1.724312 1.724306 -1.418489 -2.112137 2.112131 -1.8856 -2.50218 2.502175 -2.374978 -2.893042 2.893037 -2.874948 282 10 0 0 0 -0.144758 4.76837e-007 -0.0191925 -0.459518 -9.53674e-007 -0.116961 -0.897337 0 -0.330528 -1.39786 -4.76837e-007 -0.645006 -1.926636 -1.43051e-006 -1.030152 -2.470001 -1.43051e-006 -1.469966 -3.019781 -1.90735e-006 -1.945604 -3.572142 -2.86102e-006 -2.441647 -4.125433 -2.38419e-006 -2.947118 283 10 0 0 0 -0.159472 -4.76837e-007 -0.0233808 -0.500021 0 -0.139995 -0.960068 9.53674e-007 -0.384103 -1.474463 4.76837e-007 -0.729798 -2.01111 -9.53674e-007 -1.141653 -2.558456 -9.53674e-007 -1.602036 -3.110231 -9.53674e-007 -2.092865 -3.663708 -9.53674e-007 -2.600403 -4.217777 -1.43051e-006 -3.115321 284 10 0 0 0 -0.143158 9.53674e-007 -0.0187609 -0.453917 1.43051e-006 -0.113804 -0.887527 0 -0.322047 -1.385033 -4.76837e-007 -0.630363 -1.911943 -1.43051e-006 -1.009831 -2.45429 -1.43051e-006 -1.444957 -3.003517 -1.90735e-006 -1.91691 -3.555552 -2.86102e-006 -2.410019 -4.108612 -3.33786e-006 -2.912998 285 10 0 0 0 -0.167048 -4.76837e-007 -0.025709 -0.5226 -4.76837e-007 -0.154386 -0.996065 -4.76837e-007 -0.419297 -1.518963 -9.53674e-007 -0.787181 -2.060551 -1.90735e-006 -1.218946 -2.610456 -1.90735e-006 -1.695473 -3.163572 -2.6226e-006 -2.199015 -3.717855 -2.86102e-006 -2.716975 -4.27253 -3.33786e-006 -3.24126 286 10 0 0 0 -0.161701 9.53674e-007 -0.0240521 -0.510504 2.38419e-007 -0.147198 -0.980797 2.38419e-007 -0.407003 -1.503172 0 -0.773439 -2.045137 -4.76837e-007 -1.206873 -2.595507 -7.15256e-007 -1.686704 -3.148968 -1.19209e-006 -2.194071 -3.703503 -1.43051e-006 -2.715961 -4.25839 -1.90735e-006 -3.244318 287 10 0 0 0 -0.166845 -1.19209e-006 -0.025646 -0.52253 -1.66893e-006 -0.154435 -0.996466 -1.66893e-006 -0.420113 -1.519814 -1.90735e-006 -0.78926 -2.061726 -2.38419e-006 -1.22244 -2.611818 -2.6226e-006 -1.70028 -3.165041 -2.86102e-006 -2.204956 -3.719391 -3.09944e-006 -2.723913 -4.274118 -3.09944e-006 -3.249136 288 10 0 0 0 -0.125368 0.12536 -0.0290486 -0.392569 0.392558 -0.177355 -0.740107 0.740099 -0.473919 -1.116676 1.116665 -0.871265 -1.503103 1.503096 -1.326846 -1.893539 1.893528 -1.821083 -2.285408 2.285397 -2.337616 -2.677757 2.67775 -2.865667 -3.070286 3.070274 -3.398957 289 10 0 0 0 -0.120256 0.120264 -0.0266731 -0.378952 0.378958 -0.164143 -0.721748 0.721759 -0.448372 -1.09697 1.096978 -0.839401 -1.48312 1.483128 -1.292911 -1.873554 1.873566 -1.787108 -2.265459 2.265471 -2.304437 -2.657854 2.657862 -2.83361 -3.050404 3.050415 -3.367937 290 10 0 0 0 -0.125019 0.125019 -0.0288838 -0.392334 0.392334 -0.177342 -0.74085 0.74085 -0.476544 -1.118055 1.118055 -0.876944 -1.504827 1.504824 -1.335237 -1.89543 1.895426 -1.831614 -2.287373 2.287373 -2.349794 -2.679773 2.679773 -2.879172 -3.072327 3.072323 -3.41357 291 10 0 0 0 -0.118296 0.118303 -0.0257884 -0.367687 0.367703 -0.152107 -0.699331 0.699344 -0.409894 -1.066505 1.066518 -0.767942 -1.447725 1.447733 -1.189388 -1.835562 1.83557 -1.656568 -2.226095 2.226108 -2.15209 -2.617668 2.617673 -2.66304 -3.009611 3.00962 -3.181163 292 10 0 0 0 -0.118769 0.118762 -0.025996 -0.369612 0.369603 -0.154004 -0.702808 0.7028 -0.415272 -1.071085 1.071076 -0.777465 -1.452999 1.452991 -1.202956 -1.84123 1.841227 -1.673661 -2.231989 2.231977 -2.172133 -2.623697 2.62369 -2.685619 -3.015734 3.015726 -3.205712 293 10 0 0 0 -0.126688 0.126673 -0.0296779 -0.395678 0.395666 -0.180356 -0.74295 0.742939 -0.476206 -1.119165 1.11915 -0.871861 -1.50535 1.505335 -1.325594 -1.895655 1.895639 -1.818166 -2.287451 2.287432 -2.33329 -2.679764 2.679749 -2.860243 -3.072271 3.072252 -3.39286 294 10 0 0 0 -0.169142 -7.15256e-007 -0.026374 -0.524786 -2.38419e-007 -0.155127 -0.996046 -7.15256e-007 -0.41648 -1.516594 -1.43051e-006 -0.777942 -2.056419 -1.66893e-006 -1.202237 -2.605249 -2.14577e-006 -1.67161 -3.157732 -2.14577e-006 -2.168792 -3.711599 -2.14577e-006 -2.681074 -4.265985 -2.6226e-006 -3.200599 295 10 0 0 0 -0.171612 -1.19209e-006 -0.0271709 -0.531027 -9.53674e-007 -0.159095 -1.004955 -1.43051e-006 -0.424764 -1.52688 -1.66893e-006 -0.789966 -2.067377 -1.66893e-006 -1.217039 -2.616499 -1.90735e-006 -1.688287 -3.169147 -2.14577e-006 -2.187101 -3.723186 -2.14577e-006 -2.701632 -4.277737 -2.38419e-006 -3.223791 296 10 0 0 0 -0.175158 -2.98023e-007 -0.0283357 -0.55048 -2.98023e-007 -0.174288 -1.040966 -6.55651e-007 -0.468947 -1.573542 -8.9407e-007 -0.86637 -2.120159 -1.01328e-006 -1.322665 -2.67247 -1.43051e-006 -1.818244 -3.226797 -1.49012e-006 -2.336884 -3.781797 -1.78814e-006 -2.867671 -4.337023 -1.84774e-006 -3.40404 297 10 0 0 0 -0.175716 -1.19209e-007 -0.0285203 -0.55171 -3.27826e-007 -0.175088 -1.042494 -5.06639e-007 -0.470312 -1.57515 -7.7486e-007 -0.868009 -2.121782 -1.04308e-006 -1.324383 -2.67407 -1.2517e-006 -1.819791 -3.22838 -1.46031e-006 -2.338141 -3.783371 -1.60933e-006 -2.868693 -4.338587 -1.69873e-006 -3.404859 298 10 0 0 0 -0.175131 -5.96046e-008 -0.0283249 -0.549614 -4.17233e-007 -0.173507 -1.039087 -7.45058e-007 -0.466273 -1.571009 -1.01328e-006 -0.861513 -2.117264 -1.19209e-006 -1.315839 -2.669416 -1.37091e-006 -1.809984 -3.223679 -1.49012e-006 -2.327647 -3.778648 -1.72853e-006 -2.857668 -4.333847 -1.90735e-006 -3.393371 299 10 0 0 0 -0.172106 -4.76837e-007 -0.0273294 -0.53216 -2.38419e-007 -0.159798 -1.00659 -4.76837e-007 -0.426286 -1.528843 -1.19209e-006 -0.792376 -2.069538 -9.53674e-007 -1.220305 -2.618773 -1.43051e-006 -1.6923 -3.17145 -1.43051e-006 -2.191365 -3.725483 -1.90735e-006 -2.705844 -4.280029 -1.90735e-006 -3.22795' )
		self.hairShape.boundingBox()
		self.editGuidesShape.boundingBox()
		pm.mel.OxEditGuides( '-eg 300 24 10 0 0 0 -0.0787601 0.477178 -0.282177 -0.343032 0.764976 -0.442794 -0.699093 1.121036 -0.763655 -1.082575 1.504519 -1.198695 -1.473042 1.894986 -1.693285 -1.865358 2.287301 -2.220238 -2.258073 2.680017 -2.761995 -2.650877 3.072821 -3.31019 -3.043699 3.465644 -3.861279 26 10 0 0 0 -0.0863197 0.518033 -0.374342 -0.343332 0.827418 -0.546668 -0.700187 1.184273 -0.869946 -1.083844 1.567932 -1.306099 -1.474327 1.958416 -1.800934 -1.866637 2.350724 -2.327742 -2.259349 2.743436 -2.869286 -2.652154 3.136241 -3.417561 -3.044978 3.529066 -3.968686 27 10 0 0 0 -0.080411 0.494287 -0.315025 -0.345343 0.782216 -0.476181 -0.701587 1.138461 -0.797601 -1.085108 1.521981 -1.232872 -1.475573 1.912448 -1.727506 -1.867888 2.304762 -2.254442 -2.260605 2.69748 -2.796176 -2.653405 3.09028 -3.344347 -3.046229 3.483103 -3.895413 30 10 0 0 0 -0.394888 0.271429 -0.274428 -0.797272 0.287069 -0.447251 -1.306509 0.28707 -0.780725 -1.850274 0.287071 -1.222437 -2.402814 0.28707 -1.720197 -2.95774 0.287071 -2.249296 -3.513167 0.28707 -2.792872 -4.068691 0.287071 -3.342463 -4.624235 0.287071 -3.894424 31 10 0 0 0 -0.290897 0.131453 -0.100868 -0.697803 0.131454 -0.278177 -1.210711 0.131454 -0.620265 -1.755319 0.131454 -1.066074 -2.307943 0.131455 -1.564634 -2.862851 0.131455 -2.093344 -3.418262 0.131455 -2.636259 -3.973779 0.131455 -3.185347 -4.52932 0.131455 -3.736932 32 10 0 0 0 -0.432997 0.298332 -0.376239 -0.845789 0.330231 -0.561354 -1.356444 0.330231 -0.898108 -1.900653 0.330231 -1.34196 -2.453395 0.330231 -1.841677 -3.00837 0.330231 -2.371862 -3.563806 0.330231 -2.915874 -4.11933 0.330231 -3.465587 -4.674875 0.330231 -4.017531 33 10 0 0 0 -0.354366 0.217957 -0.18737 -0.759624 0.217959 -0.362913 -1.273116 0.217958 -0.706414 -1.818698 0.217959 -1.15717 -2.371897 0.217959 -1.661599 -2.92698 0.217959 -2.194249 -3.482439 0.21796 -2.739427 -4.037969 0.217959 -3.289683 -4.593513 0.217959 -3.841885 34 10 0 0 0 -0.26237 0.113316 -0.07915 -0.651569 0.113317 -0.238263 -1.157815 0.113317 -0.565001 -1.70107 0.113317 -1.004298 -2.253528 0.113318 -1.501262 -2.808396 0.113318 -2.029202 -3.363796 0.113318 -2.571628 -3.919312 0.113318 -3.120547 -4.474853 0.113318 -3.672157 84 10 0 0 0 -0.0961699 0.412803 -0.196401 -0.383081 0.699713 -0.372476 -0.746558 1.063192 -0.717295 -1.132485 1.44912 -1.169105 -1.523691 1.840326 -1.674101 -1.916202 2.232836 -2.206998 -2.308972 2.625607 -2.752263 -2.701791 3.018423 -3.302537 -3.094624 3.411257 -3.854724 87 10 0 0 0 -0.098423 0.392824 -0.175232 -0.385595 0.679999 -0.351703 -0.749242 1.043646 -0.697108 -1.135227 1.42963 -1.149359 -1.526454 1.820857 -1.654615 -1.918964 2.213367 -2.187644 -2.311737 2.60614 -2.732971 -2.704554 2.998959 -3.283275 -3.097386 3.391788 -3.835478 88 10 0 0 0 -0.132184 0.223912 -0.0646049 -0.424627 0.516356 -0.249221 -0.789415 0.881144 -0.598618 -1.175291 1.26702 -1.05005 -1.56635 1.658079 -1.552824 -1.958798 2.050526 -2.0836 -2.351546 2.443275 -2.627345 -2.744358 2.836087 -3.176511 -3.137183 3.22891 -3.727698 89 10 0 0 0 -0.0987406 0.409797 -0.193676 -0.389265 0.70032 -0.375287 -0.754833 1.065889 -0.727468 -1.141398 1.452455 -1.184159 -1.532766 1.843823 -1.691714 -1.925308 2.236362 -2.22568 -2.31808 2.629135 -2.77129 -2.710901 3.021957 -3.321607 -3.10373 3.414786 -3.873667 90 10 0 0 0 -0.430404 0.282212 -0.346388 -0.85779 0.297288 -0.547323 -1.378839 0.297288 -0.910132 -1.926411 0.297288 -1.371842 -2.480015 0.297288 -1.880884 -3.035139 0.297288 -2.414578 -3.590587 0.297288 -2.959252 -4.146107 0.297287 -3.508512 -4.701643 0.297287 -4.05956 91 10 0 0 0 -0.32901 0.159397 -0.137241 -0.759768 0.159398 -0.341953 -1.284942 0.159398 -0.716308 -1.833701 0.159397 -1.185222 -2.387583 0.159397 -1.697679 -2.942747 0.159397 -2.232324 -3.498199 0.159397 -2.777133 -4.053719 0.159397 -3.326315 -4.609255 0.159396 -3.877209 93 10 0 0 0 -0.413546 0.265213 -0.296161 -0.843946 0.280637 -0.500775 -1.366702 0.280638 -0.868267 -1.914847 0.280637 -1.333381 -2.468608 0.280637 -1.844304 -3.023767 0.280637 -2.378854 -3.579223 0.280637 -2.923796 -4.134744 0.280636 -3.473164 -4.690281 0.280636 -4.02426 94 10 0 0 0 -0.413387 0.263236 -0.293907 -0.83948 0.263236 -0.492969 -1.361609 0.263236 -0.858725 -1.90926 0.263236 -1.320907 -2.462804 0.263235 -1.829209 -3.017875 0.263235 -2.361514 -3.573299 0.263235 -2.905002 -4.128808 0.263235 -3.453368 -4.684337 0.263235 -4.003655 95 10 0 0 0 -0.435986 0.280947 -0.356486 -0.868444 0.29236 -0.563486 -1.391566 0.29236 -0.931997 -1.939475 0.29236 -1.395693 -2.493022 0.292359 -1.904041 -3.048076 0.292359 -2.435956 -3.603494 0.292359 -2.979075 -4.158998 0.292359 -3.52713 -4.714529 0.292358 -4.077137 145 10 0 0 0 -0.132999 0.313316 -0.116478 -0.443135 0.62345 -0.331046 -0.815428 0.995743 -0.709288 -1.203098 1.383415 -1.175049 -1.594442 1.774755 -1.682208 -1.986877 2.167192 -2.212576 -2.379594 2.559909 -2.754379 -2.772385 2.9527 -3.30161 -3.165201 3.345514 -3.851153 148 10 0 0 0 -0.120925 0.353157 -0.144095 -0.424448 0.656679 -0.34696 -0.794019 1.026248 -0.714151 -1.18116 1.413392 -1.175473 -1.572423 1.804655 -1.681328 -1.964851 2.197081 -2.211534 -2.357571 2.589802 -2.753547 -2.750363 2.982593 -3.300827 -3.143177 3.375407 -3.850197 149 10 0 0 0 -0.153596 0.183261 -0.0540915 -0.457887 0.487552 -0.258282 -0.825749 0.855416 -0.618914 -1.211551 1.241216 -1.069777 -1.60216 1.631827 -1.566275 -1.994341 2.024008 -2.089762 -2.386971 2.416634 -2.627124 -2.779718 2.809381 -3.170965 -3.172505 3.202168 -3.717508 154 10 0 0 0 -0.362507 0.19143 -0.180609 -0.788827 0.19143 -0.379944 -1.307338 0.19143 -0.736031 -1.852272 0.19143 -1.183467 -2.404439 0.19143 -1.67774 -2.959023 0.191429 -2.200476 -3.514287 0.191429 -2.738059 -4.069736 0.191429 -3.282645 -4.625234 0.191428 -3.830208 155 10 0 0 0 -0.346052 0.171263 -0.156107 -0.771085 0.171263 -0.353906 -1.288866 0.171262 -0.708103 -1.833491 0.171262 -1.154 -2.385571 0.171262 -1.647489 -2.940126 0.171261 -2.169732 -3.49538 0.171261 -2.707056 -4.050825 0.171261 -3.251461 -4.606319 0.17126 -3.798876 204 10 0 0 0 -0.134423 0.235589 -0.0707169 -0.421765 0.522933 -0.247447 -0.779076 0.880241 -0.572123 -1.159945 1.261115 -0.991615 -1.548384 1.649549 -1.464243 -1.939756 2.04092 -1.971899 -2.332088 2.433256 -2.499267 -2.724712 2.825878 -3.036465 -3.117424 3.218591 -3.577957 205 10 0 0 0 -0.111639 0.517332 -0.386584 -0.383667 0.842836 -0.583408 -0.741758 1.200927 -0.910523 -1.123694 1.582867 -1.336129 -1.512838 1.972011 -1.815703 -1.904542 2.363713 -2.329023 -2.297009 2.756179 -2.860502 -2.689701 3.148871 -3.400856 -3.082457 3.541625 -3.945066 207 10 0 0 0 -0.11277 0.475285 -0.290918 -0.400624 0.784179 -0.485413 -0.762925 1.14648 -0.82622 -1.146584 1.530138 -1.26238 -1.536437 1.919991 -1.749604 -1.928405 2.311957 -2.26811 -2.32097 2.704524 -2.803033 -2.713703 3.097254 -3.345775 -3.106483 3.490035 -3.891864 210 10 0 0 0 -0.385561 0.231727 -0.229538 -0.782692 0.2347 -0.396608 -1.283905 0.2347 -0.712521 -1.821688 0.2347 -1.128679 -2.370998 0.234699 -1.601169 -2.924555 0.234699 -2.109609 -3.479433 0.234698 -2.637752 -4.034704 0.234698 -3.175573 -4.590094 0.234698 -3.717651 212 10 0 0 0 -0.43227 0.280634 -0.348124 -0.834795 0.294177 -0.521015 -1.335794 0.294177 -0.836487 -1.873369 0.294177 -1.25185 -2.422564 0.294176 -1.723568 -2.976068 0.294176 -2.231448 -3.530935 0.294175 -2.759236 -4.086199 0.294175 -3.296887 -4.64159 0.294174 -3.83886 214 10 0 0 0 -0.325251 0.158134 -0.133836 -0.727118 0.158134 -0.305796 -1.232246 0.158134 -0.630078 -1.771851 0.158134 -1.053462 -2.32197 0.158134 -1.531492 -2.875839 0.158133 -2.043803 -3.430841 0.158133 -2.574598 -3.986174 0.158132 -3.1144 -4.541599 0.158132 -3.658057 215 10 0 0 0 -0.440474 0.288757 -0.378789 -0.8508 0.327395 -0.561811 -1.348436 0.327395 -0.870384 -1.884491 0.327394 -1.280045 -2.433014 0.327394 -1.747488 -2.986256 0.327394 -2.252378 -3.541008 0.327394 -2.778085 -4.09622 0.327393 -3.314104 -4.651573 0.327394 -3.85467 264 10 0 0 0 -0.111597 0.29957 -0.101195 -0.37484 0.562811 -0.24438 -0.717014 0.904986 -0.527028 -1.090715 1.278685 -0.911291 -1.475855 1.663823 -1.357405 -1.865849 2.053819 -1.846257 -2.257587 2.445557 -2.360271 -2.649904 2.83787 -2.887251 -3.042402 3.23037 -3.419766 265 10 0 0 0 -0.102785 0.5121 -0.366263 -0.35853 0.813982 -0.531813 -0.703211 1.15867 -0.82086 -1.078476 1.533932 -1.212109 -1.464478 1.919931 -1.664469 -1.854855 2.31031 -2.158035 -2.246756 2.702209 -2.675151 -2.63916 3.09461 -3.204555 -3.031731 3.487182 -3.739806 267 10 0 0 0 -0.106001 0.45613 -0.256639 -0.376257 0.739207 -0.417892 -0.726292 1.089244 -0.721271 -1.104387 1.467339 -1.126062 -1.491783 1.854732 -1.58947 -1.882772 2.245722 -2.091207 -2.274933 2.637882 -2.614261 -2.667484 3.030432 -3.148509 -3.060158 3.423108 -3.688039 271 10 0 0 0 -0.377195 0.229347 -0.218264 -0.751493 0.230978 -0.363281 -1.23661 0.230978 -0.648089 -1.765509 0.230977 -1.033619 -2.310175 0.230977 -1.479726 -2.861683 0.230977 -1.968355 -3.415682 0.230976 -2.482374 -3.97052 0.230976 -3.009665 -4.525631 0.230975 -3.542987 272 10 0 0 0 -0.243233 0.0858917 -0.0635161 -0.598318 0.0858898 -0.191803 -1.068964 0.0858879 -0.452171 -1.58983 0.0858889 -0.814482 -2.130312 0.0858898 -1.241512 -2.679726 0.0858879 -1.714682 -3.232599 0.0858889 -2.215715 -3.786743 0.0858879 -2.731679 -4.341343 0.0858879 -3.254659 273 10 0 0 0 -0.425646 0.281438 -0.335876 -0.805815 0.300016 -0.486751 -1.287804 0.300015 -0.766026 -1.814947 0.300015 -1.146194 -2.358652 0.300015 -1.587603 -2.9097 0.300014 -2.072542 -3.463484 0.300014 -2.583784 -4.018206 0.300013 -3.108901 -4.573235 0.300012 -3.640302 274 10 0 0 0 -0.307472 0.150244 -0.117914 -0.678667 0.150243 -0.260121 -1.161343 0.150243 -0.540601 -1.688629 0.150242 -0.921202 -2.232557 0.150241 -1.363683 -2.783714 0.150242 -1.849501 -3.337524 0.150241 -2.361036 -3.892252 0.15024 -2.886335 -4.447285 0.15024 -3.417813 275 10 0 0 0 -0.418026 0.271817 -0.310579 -0.80279 0.285861 -0.465633 -1.290867 0.28586 -0.755819 -1.821861 0.285859 -1.148011 -2.367781 0.285859 -1.600553 -2.9198 0.285859 -2.093538 -3.474022 0.285858 -2.610613 -4.028986 0.285858 -3.140514 -4.584188 0.285857 -3.676275 0 10 0 0 0 -0.124099 0.1241 -0.0284486 -0.384699 0.3847 -0.168292 -0.729375 0.729378 -0.457319 -1.10703 1.107033 -0.859919 -1.495047 1.495051 -1.328717 -1.886517 1.886522 -1.837965 -2.278957 2.278962 -2.368568 -2.671655 2.671661 -2.909359 -3.064424 3.06443 -3.454602 1 10 0 0 0 -0.134504 0.134505 -0.0335789 -0.40242 0.402422 -0.18283 -0.746986 0.746989 -0.471572 -1.122269 1.122273 -0.862911 -1.50849 1.508494 -1.31693 -1.899092 1.899097 -1.813326 -2.29115 2.291156 -2.333912 -2.683658 2.683663 -2.866744 -3.076304 3.07631 -3.404984 2 10 0 0 0 -0.12156 0.12156 -0.0272668 -0.377594 0.377596 -0.161475 -0.718567 0.718569 -0.441137 -1.094298 1.094301 -0.834539 -1.481557 1.481561 -1.296813 -1.872775 1.87278 -1.801975 -2.265132 2.265137 -2.33009 -2.657799 2.657804 -2.869296 -3.050551 3.050558 -3.413407 3 10 0 0 0 -0.125613 0.125616 -0.0291677 -0.382941 0.382943 -0.164953 -0.721265 0.721267 -0.438168 -1.093937 1.09394 -0.818018 -1.479217 1.479222 -1.265136 -1.869552 1.869557 -1.758085 -2.261548 2.261554 -2.277319 -2.654051 2.654058 -2.809946 -3.046706 3.046713 -3.348658 4 10 0 0 0 -0.122518 0.122519 -0.0277106 -0.37673 0.376731 -0.159716 -0.713869 0.713871 -0.430111 -1.086453 1.086457 -0.80959 -1.471898 1.471902 -1.257879 -1.862358 1.862363 -1.752418 -2.254422 2.254427 -2.273136 -2.64696 2.646966 -2.807065 -3.039642 3.039649 -3.34699 5 10 0 0 0 -0.136679 0.136679 -0.0347099 -0.410422 0.410423 -0.191804 -0.760133 0.760135 -0.494287 -1.138472 1.138474 -0.900303 -1.526284 1.526287 -1.367291 -1.917578 1.917582 -1.873658 -2.309927 2.309932 -2.401543 -2.702577 2.702583 -2.939948 -3.095314 3.09532 -3.482951 6 10 0 0 0 -0.199592 5.36442e-007 -0.0370913 -0.601823 9.53674e-007 -0.209434 -1.107915 1.54972e-006 -0.535833 -1.648553 2.02656e-006 -0.96351 -2.199414 2.6226e-006 -1.446991 -2.753658 2.92063e-006 -1.964402 -3.308831 3.33786e-006 -2.499359 -3.864248 3.69549e-006 -3.042506 -4.419732 3.93391e-006 -3.589122 7 10 0 0 0 -0.202083 5.96046e-007 -0.0380566 -0.602279 1.19209e-006 -0.208275 -1.105641 1.3113e-006 -0.528736 -1.644809 2.26498e-006 -0.950352 -2.194978 2.5034e-006 -1.428735 -2.748954 2.98023e-006 -1.942428 -3.304026 3.21865e-006 -2.474788 -3.859397 3.69549e-006 -3.016027 -4.414854 3.8147e-006 -3.561078 8 10 0 0 0 -0.200754 7.15256e-007 -0.0375403 -0.597623 1.43051e-006 -0.204333 -1.098476 1.66893e-006 -0.519495 -1.636343 2.14577e-006 -0.935978 -2.185899 2.38419e-006 -1.410105 -2.739631 3.09944e-006 -1.920685 -3.294606 3.33786e-006 -2.450846 -3.849933 3.57628e-006 -2.990462 -4.405362 3.8147e-006 -3.534177 9 10 0 0 0 -0.20018 9.53674e-007 -0.0373183 -0.596412 9.53674e-007 -0.20346 -1.096928 1.19209e-006 -0.517924 -1.634593 1.43051e-006 -0.933627 -2.184015 1.90735e-006 -1.406858 -2.737675 2.38419e-006 -1.916558 -3.292615 2.86102e-006 -2.445968 -3.847925 2.86102e-006 -2.984993 -4.403343 3.33786e-006 -3.52819 10 10 0 0 0 -0.195822 4.61936e-007 -0.0356551 -0.594654 9.23872e-007 -0.204461 -1.100875 1.62423e-006 -0.531147 -1.641848 2.13087e-006 -0.960249 -2.192905 2.6077e-006 -1.445242 -2.747232 2.99513e-006 -1.963874 -3.302438 3.44217e-006 -2.499736 -3.857871 3.68059e-006 -3.043581 -4.413363 4.03821e-006 -3.590789 11 10 0 0 0 -0.203396 1.19209e-007 -0.0385717 -0.606595 4.76837e-007 -0.211933 -1.112415 9.53674e-007 -0.537729 -1.6528 1.54972e-006 -0.964346 -2.203488 2.02656e-006 -1.446516 -2.757647 2.38419e-006 -1.962694 -3.312779 2.74181e-006 -2.496601 -3.868176 2.98023e-006 -3.038873 -4.423646 3.21865e-006 -3.58473 12 10 0 0 0 -0.136009 0.136011 -0.0343599 -0.417504 0.417506 -0.202406 -0.777209 0.777213 -0.534657 -1.161534 1.161537 -0.975186 -1.552119 1.552124 -1.471344 -1.944409 1.944414 -1.997613 -2.337105 2.337111 -2.538316 -2.729897 2.729902 -3.085365 -3.12271 3.122717 -3.635028 13 10 0 0 0 -0.140316 0.140317 -0.0366488 -0.427844 0.427845 -0.213657 -0.790752 0.790754 -0.556524 -1.176119 1.176122 -1.004251 -1.567011 1.567015 -1.504599 -1.959395 1.9594 -2.033505 -2.352121 2.352125 -2.575783 -2.744922 2.744926 -3.123863 -3.137742 3.137747 -3.674306 14 10 0 0 0 -0.140329 0.14033 -0.0366559 -0.428058 0.428059 -0.21397 -0.791136 0.791137 -0.557415 -1.176567 1.176569 -1.00561 -1.567479 1.567481 -1.506238 -1.95987 1.959874 -2.035362 -2.352598 2.352602 -2.577827 -2.745402 2.745406 -3.126049 -3.138224 3.138227 -3.676606 15 10 0 0 0 -0.137432 0.137432 -0.0351067 -0.420034 0.420034 -0.204769 -0.779873 0.779874 -0.537452 -1.164058 1.16406 -0.977053 -1.554486 1.554489 -1.471173 -1.94672 1.946723 -1.995956 -2.339396 2.3394 -2.535693 -2.73218 2.732184 -3.082094 -3.124991 3.124996 -3.631255 16 10 0 0 0 -0.138943 0.138947 -0.0359102 -0.425314 0.425317 -0.211168 -0.787949 0.787951 -0.553102 -1.173321 1.173325 -1.000879 -1.564244 1.564248 -1.501657 -1.956644 1.956648 -2.031016 -2.349374 2.349379 -2.573651 -2.742179 2.742184 -3.121995 -3.135 3.135005 -3.67265 17 10 0 0 0 -0.145448 0.145449 -0.0394831 -0.433968 0.433969 -0.218004 -0.79443 0.794432 -0.552708 -1.1779 1.177903 -0.987671 -1.567852 1.567856 -1.476034 -1.959895 1.959899 -1.996278 -2.352502 2.352506 -2.532821 -2.745257 2.745261 -3.076926 -3.13805 3.138054 -3.624221 18 10 0 0 0 -0.204993 3.57628e-007 -0.0392029 -0.612335 6.55651e-007 -0.216983 -1.121759 1.01328e-006 -0.550887 -1.663951 1.43051e-006 -0.985319 -2.21541 1.49012e-006 -1.473532 -2.769845 1.96695e-006 -1.993825 -3.325078 2.08616e-006 -2.530442 -3.880517 2.20537e-006 -3.074602 -4.436011 2.38419e-006 -3.621941 19 10 0 0 0 -0.206039 3.8743e-007 -0.0396193 -0.615706 6.25849e-007 -0.219922 -1.126965 1.16229e-006 -0.55809 -1.670036 1.43051e-006 -0.996533 -2.221858 1.84774e-006 -1.487786 -2.776418 2.02656e-006 -2.010099 -3.331694 2.26498e-006 -2.548031 -3.887151 2.47359e-006 -3.093105 -4.442655 2.5928e-006 -3.641168 20 10 0 0 0 -0.205831 5.66244e-007 -0.0395361 -0.615489 8.9407e-007 -0.219829 -1.126838 1.37091e-006 -0.55821 -1.669984 1.62423e-006 -0.997 -2.221854 1.98185e-006 -1.48866 -2.776435 2.22027e-006 -2.011317 -3.331718 2.41399e-006 -2.549492 -3.887178 2.5928e-006 -3.094736 -4.442684 2.72691e-006 -3.642938 21 10 0 0 0 -0.198754 4.76837e-007 -0.0367694 -0.606029 2.38419e-007 -0.214476 -1.119652 1.19209e-006 -0.558291 -1.664765 1.66893e-006 -1.006638 -2.217601 2.14577e-006 -1.507284 -2.772529 2.38419e-006 -2.036424 -3.32793 2.6226e-006 -2.578913 -3.883438 3.09944e-006 -3.127156 -4.438971 3.09944e-006 -3.677733 22 10 0 0 0 -0.203514 3.57628e-007 -0.0386187 -0.612401 5.96046e-007 -0.218071 -1.124518 1.01328e-006 -0.558271 -1.668374 1.2517e-006 -1.000411 -2.220618 1.60933e-006 -1.495394 -2.775341 1.90735e-006 -2.020547 -3.330675 2.02656e-006 -2.56042 -3.886157 2.08616e-006 -3.106906 -4.44168 2.08616e-006 -3.656414 23 10 0 0 0 -0.204945 4.76837e-007 -0.0391839 -0.614576 6.55651e-007 -0.219447 -1.126477 1.2517e-006 -0.559132 -1.670064 1.66893e-006 -0.999989 -2.222162 1.90735e-006 -1.493659 -2.77683 2.08616e-006 -2.017825 -3.332145 2.29478e-006 -2.55703 -3.887619 2.32458e-006 -3.103007 -4.443135 2.563e-006 -3.651827 25 10 0 0 0 -0.142373 0.142373 -0.0377702 -0.431304 0.431306 -0.216923 -0.794217 0.794217 -0.559798 -1.179349 1.179352 -1.005875 -1.57012 1.570121 -1.504498 -1.962466 1.962469 -2.032318 -2.355181 2.355182 -2.573988 -2.747979 2.747982 -3.121726 -3.1408 3.140802 -3.672227 28 10 0 0 0 -0.13827 0.138272 -0.0355515 -0.424249 0.424253 -0.210222 -0.786694 0.7867 -0.55152 -1.171967 1.171973 -0.998573 -1.562865 1.562871 -1.498993 -1.955267 1.955273 -2.028434 -2.348002 2.348009 -2.57131 -2.740808 2.740814 -3.119896 -3.133629 3.133636 -3.670814 29 10 0 0 0 -0.141271 0.141271 -0.0371655 -0.429853 0.429854 -0.215782 -0.793003 0.793005 -0.559477 -1.178336 1.178339 -1.006958 -1.569202 1.569205 -1.506938 -1.961583 1.961587 -2.035785 -2.35431 2.354313 -2.578165 -2.747113 2.747116 -3.126372 -3.139933 3.139936 -3.676963 35 10 0 0 0 -0.191694 2.38419e-006 -0.0341189 -0.593481 9.53674e-007 -0.205997 -1.105241 2.38419e-006 -0.545343 -1.650261 1.90735e-006 -0.99322 -2.203322 2.38419e-006 -1.496181 -2.758388 2.86102e-006 -2.028384 -3.313844 1.90735e-006 -2.57352 -3.869375 2.38419e-006 -3.123868 -4.424921 1.43051e-006 -3.676216 36 10 0 0 0 -0.120659 0.120659 -0.026855 -0.386629 0.386629 -0.173554 -0.740097 0.740099 -0.486697 -1.122629 1.12263 -0.915841 -1.51291 1.512912 -1.408127 -1.90523 1.905231 -1.935192 -2.297962 2.297964 -2.478112 -2.690778 2.690781 -3.027484 -3.083603 3.083604 -3.579363 37 10 0 0 0 -0.120608 0.12061 -0.0268321 -0.386596 0.386596 -0.173554 -0.740034 0.740035 -0.486609 -1.122631 1.122633 -0.91614 -1.51296 1.51296 -1.409028 -1.905297 1.905296 -1.936548 -2.298035 2.298038 -2.479675 -2.690847 2.690848 -3.029132 -3.083675 3.083676 -3.581041 38 10 0 0 0 -0.116835 0.116833 -0.0251396 -0.377033 0.377033 -0.164478 -0.727375 0.727372 -0.468709 -1.109074 1.109074 -0.892906 -1.499008 1.499006 -1.381058 -1.891184 1.891181 -1.904401 -2.283881 2.283878 -2.445089 -2.676684 2.676682 -2.993267 -3.069509 3.069507 -3.544547 39 10 0 0 0 -0.119568 0.119563 -0.0263591 -0.383578 0.383574 -0.170526 -0.736056 0.736053 -0.480808 -1.11838 1.118377 -0.908693 -1.50845 1.508447 -1.39844 -1.900698 1.900697 -1.923603 -2.293415 2.293414 -2.465397 -2.686222 2.686219 -3.014247 -3.079054 3.079052 -3.565886 40 10 0 0 0 -0.123105 0.123105 -0.0279846 -0.391394 0.391397 -0.177729 -0.746116 0.746118 -0.494564 -1.129267 1.129271 -0.927517 -1.519624 1.519629 -1.420757 -1.911911 1.911915 -1.946862 -2.304632 2.304637 -2.488915 -2.69744 2.697443 -3.037774 -3.090266 3.09027 -3.589432 41 10 0 0 0 -0.113493 0.113489 -0.0236907 -0.369759 0.369755 -0.158173 -0.717756 0.717754 -0.455981 -1.099013 1.09901 -0.877626 -1.489067 1.489063 -1.367193 -1.881355 1.881351 -1.89334 -2.274083 2.27408 -2.435796 -2.666895 2.666891 -2.984854 -3.059721 3.059717 -3.536468 42 10 0 0 0 -0.170057 -1.43051e-006 -0.0266666 -0.550282 -4.76837e-007 -0.177165 -1.055258 -1.43051e-006 -0.501113 -1.599057 -4.76837e-007 -0.942977 -2.151925 -9.53674e-007 -1.443972 -2.706955 -9.53674e-007 -1.975353 -3.262401 -1.43051e-006 -2.51995 -3.817928 -9.53674e-007 -3.069845 -4.373472 -4.76837e-007 -3.621777 43 10 0 0 0 -0.15679 -9.53674e-007 -0.0225847 -0.518915 -4.76837e-007 -0.156824 -1.014755 -4.76837e-007 -0.461807 -1.555847 -4.76837e-007 -0.891424 -2.108031 -1.43051e-006 -1.385871 -2.662917 -1.43051e-006 -1.914135 -3.218333 -1.90735e-006 -2.457262 -3.773851 -2.38419e-006 -3.006404 -4.329392 -2.86102e-006 -3.557956 44 10 0 0 0 -0.17443 -9.53674e-007 -0.0280936 -0.562523 -4.76837e-007 -0.186122 -1.071837 -1.43051e-006 -0.519777 -1.616856 -1.43051e-006 -0.967644 -2.169973 -1.43051e-006 -1.471198 -2.725029 -1.90735e-006 -2.003198 -3.280476 -1.43051e-006 -2.547684 -3.836 -2.38419e-006 -3.097397 -4.391544 -2.38419e-006 -3.649285 45 10 0 0 0 -0.147319 1.90735e-006 -0.0198867 -0.489655 0 -0.137896 -0.969316 9.53674e-007 -0.413153 -1.503965 2.86102e-006 -0.817738 -2.054315 1.90735e-006 -1.297418 -2.608811 1.90735e-006 -1.818701 -3.164155 1.90735e-006 -2.358868 -3.719659 1.90735e-006 -2.906777 -4.275194 9.53674e-007 -3.457726 46 10 0 0 0 -0.156956 3.8147e-006 -0.0226328 -0.514214 0 -0.152738 -1.004288 1.90735e-006 -0.446625 -1.543147 1.90735e-006 -0.866999 -2.094728 1.90735e-006 -1.356212 -2.649497 1.90735e-006 -1.882199 -3.204895 1.90735e-006 -2.424532 -3.760412 1.90735e-006 -2.973478 -4.315954 9.53674e-007 -3.524987 47 10 0 0 0 -0.177574 1.90735e-006 -0.0291426 -0.56944 1.43051e-006 -0.190892 -1.080416 1.90735e-006 -0.528397 -1.625903 1.90735e-006 -0.978663 -2.179144 9.53674e-007 -1.483549 -2.734229 9.53674e-007 -2.016235 -3.289681 9.53674e-007 -2.561075 -3.845209 4.76837e-007 -3.11092 -4.400753 0 -3.662846 48 10 0 0 0 -0.0840616 0.0840607 -0.0128672 -0.291734 0.291733 -0.0968437 -0.60648 0.606478 -0.319965 -0.975375 0.975376 -0.684541 -1.361867 1.361872 -1.14065 -1.753342 1.753344 -1.649986 -2.145895 2.145897 -2.184324 -2.538659 2.53866 -2.729199 -2.931471 2.931471 -3.278548 49 10 0 0 0 -0.0940161 0.0940208 -0.0161436 -0.323852 0.323854 -0.121149 -0.658813 0.658817 -0.386463 -1.036231 1.036236 -0.787891 -1.425284 1.425291 -1.266531 -1.81734 1.817343 -1.787042 -2.210007 2.210012 -2.326454 -2.602806 2.602808 -2.87391 -2.995627 2.99563 -3.4247 50 10 0 0 0 -0.0886264 0.0886278 -0.0143211 -0.304322 0.304323 -0.105557 -0.627563 0.627563 -0.345401 -1.000293 1.000292 -0.725495 -1.387978 1.387978 -1.191414 -1.779737 1.779736 -1.705786 -2.172349 2.17235 -2.242501 -2.565131 2.565128 -2.78854 -2.957952 2.957949 -3.338508 51 10 0 0 0 -0.0959525 0.0959558 -0.0168254 -0.329079 0.329079 -0.125228 -0.66622 0.66622 -0.395624 -1.044406 1.044406 -0.800882 -1.433674 1.433676 -1.281712 -1.825777 1.825777 -1.803294 -2.218459 2.218456 -2.34321 -2.611258 2.611256 -2.890886 -3.004081 3.00408 -3.441782 52 10 0 0 0 -0.0809989 0.0810013 -0.011935 -0.287803 0.287807 -0.0951533 -0.606311 0.606314 -0.325513 -0.977489 0.977489 -0.699154 -1.364741 1.36474 -1.161381 -1.756409 1.756407 -1.674124 -2.149016 2.149014 -2.210545 -2.541799 2.541796 -2.756731 -2.934617 2.934612 -3.306859 53 10 0 0 0 -0.08495 0.0849571 -0.0131457 -0.294242 0.29425 -0.0985591 -0.610151 0.610161 -0.323889 -0.979503 0.979511 -0.690238 -1.366131 1.366142 -1.147427 -1.757641 1.757648 -1.657318 -2.150201 2.150209 -2.191932 -2.542966 2.542973 -2.736952 -2.935778 2.935783 -3.286363 54 10 0 0 0 -0.112157 -9.53674e-007 -0.011441 -0.401107 9.53674e-007 -0.092495 -0.849663 9.53674e-007 -0.320271 -1.374565 1.90735e-006 -0.693837 -1.922536 1.90735e-006 -1.157899 -2.476568 9.53674e-007 -1.672343 -3.031822 1.90735e-006 -2.209596 -3.587305 0 -2.756056 -4.142836 -9.53674e-007 -3.306074 55 10 0 0 0 -0.137071 0 -0.0171738 -0.470984 -1.90735e-006 -0.128718 -0.95128 -2.86102e-006 -0.405062 -1.487608 -3.8147e-006 -0.815725 -2.038583 -3.8147e-006 -1.300082 -2.593224 -3.8147e-006 -1.823782 -3.14859 -3.8147e-006 -2.364892 -3.704098 -5.72205e-006 -2.913035 -4.259634 -5.72205e-006 -3.463926 56 10 0 0 0 -0.141547 -1.90735e-006 -0.0183346 -0.482846 -3.8147e-006 -0.135535 -0.967426 -1.90735e-006 -0.419382 -1.504954 -2.86102e-006 -0.834563 -2.056251 -2.86102e-006 -1.321448 -2.61095 -3.8147e-006 -1.846169 -3.166329 -4.76837e-006 -2.387773 -3.721839 -3.8147e-006 -2.936185 -4.277373 -4.76837e-006 -3.487255 57 10 0 0 0 -0.119516 0 -0.0130084 -0.42116 2.86102e-006 -0.102029 -0.879589 2.86102e-006 -0.343758 -1.408345 3.8147e-006 -0.728849 -1.957342 1.90735e-006 -1.199291 -2.511564 1.90735e-006 -1.716412 -3.066843 1.90735e-006 -2.254458 -3.62233 1.90735e-006 -2.801008 -4.177856 1.90735e-006 -3.350869 58 10 0 0 0 -0.139818 9.53674e-007 -0.0178823 -0.477643 3.8147e-006 -0.132395 -0.959915 9.53674e-007 -0.412167 -1.496307 9.53674e-007 -0.823069 -2.047014 0 -1.305388 -2.601494 0 -1.826335 -3.156797 0 -2.365136 -3.712278 -9.53674e-007 -2.911543 -4.267799 -1.90735e-006 -3.460969 59 10 0 0 0 -0.119965 1.90735e-006 -0.0131052 -0.423665 -2.86102e-006 -0.103467 -0.883277 -4.76837e-006 -0.346937 -1.412275 -1.90735e-006 -0.732775 -1.961378 -3.8147e-006 -1.203895 -2.515637 -2.86102e-006 -1.721511 -3.070938 -2.86102e-006 -2.260166 -3.626429 -3.8147e-006 -2.807288 -4.181961 -3.8147e-006 -3.357683 60 10 0 0 0 -0.138704 0.138703 -0.0357819 -0.413898 0.413898 -0.19488 -0.764053 0.764053 -0.498592 -1.142286 1.142288 -0.904082 -1.529881 1.529883 -1.369182 -1.921022 1.921025 -1.873182 -2.313285 2.313288 -2.398716 -2.705884 2.705888 -2.934933 -3.098585 3.098588 -3.475838 61 10 0 0 0 -0.136216 0.136217 -0.0344682 -0.407251 0.407252 -0.187877 -0.754237 0.754239 -0.482964 -1.130774 1.130777 -0.880152 -1.517552 1.517557 -1.338524 -1.908362 1.908366 -1.837705 -2.300491 2.300496 -2.359923 -2.693026 2.693032 -2.893719 -3.085686 3.085692 -3.432624 62 10 0 0 0 -0.142954 0.142956 -0.0380902 -0.426117 0.426118 -0.208574 -0.782402 0.782404 -0.530111 -1.16378 1.163782 -0.952464 -1.552802 1.552805 -1.430776 -1.944494 1.944498 -1.943977 -2.336971 2.336976 -2.475773 -2.729669 2.729674 -3.016514 -3.12243 3.122435 -3.561106 63 10 0 0 0 -0.134077 0.134077 -0.0333585 -0.40237 0.402369 -0.183106 -0.747467 0.747467 -0.473222 -1.122938 1.122939 -0.865424 -1.509131 1.509132 -1.319243 -1.899662 1.899664 -1.814695 -2.291665 2.291667 -2.334061 -2.684135 2.684138 -2.865649 -3.076753 3.076756 -3.402652 64 10 0 0 0 -0.141523 0.141524 -0.0373046 -0.422614 0.422615 -0.204764 -0.777489 0.777491 -0.522058 -1.158161 1.158163 -0.94043 -1.546869 1.546872 -1.415655 -1.938447 1.938451 -1.926763 -2.330881 2.330885 -2.457179 -2.72356 2.723563 -2.996937 -3.116308 3.116312 -3.540725 65 10 0 0 0 -0.131559 0.131561 -0.0320809 -0.395706 0.395708 -0.176424 -0.737534 0.737536 -0.458213 -1.111159 1.111162 -0.842155 -1.496409 1.496414 -1.289062 -1.886526 1.88653 -1.779345 -2.278346 2.278351 -2.294961 -2.670724 2.67073 -2.823707 -3.063282 3.063288 -3.358316 66 10 0 0 0 -0.18931 4.76837e-007 -0.0332491 -0.571437 9.53674e-007 -0.185541 -1.062916 0 -0.482078 -1.59614 9.53674e-007 -0.881703 -2.1435 9.53674e-007 -1.342185 -2.696327 9.53674e-007 -1.84275 -3.250919 1.43051e-006 -2.365603 -3.806072 9.53674e-007 -2.899997 -4.361402 1.43051e-006 -3.439729 67 10 0 0 0 -0.193717 0 -0.0348676 -0.582137 -4.76837e-007 -0.193217 -1.077971 4.76837e-007 -0.498194 -1.613436 4.76837e-007 -0.905694 -2.161869 9.53674e-007 -1.372574 -2.715146 9.53674e-007 -1.877872 -3.269938 1.19209e-006 -2.404308 -3.825178 1.19209e-006 -2.941136 -4.380552 1.43051e-006 -3.482512 68 10 0 0 0 -0.199372 -9.53674e-007 -0.0370061 -0.596762 -4.76837e-007 -0.204331 -1.099023 4.76837e-007 -0.522449 -1.637748 4.76837e-007 -0.94229 -2.187691 1.19209e-006 -1.419076 -2.741565 7.15256e-007 -1.931447 -3.296591 1.19209e-006 -2.46276 -3.851942 1.66893e-006 -3.003164 -4.407382 1.43051e-006 -3.547439 69 10 0 0 0 -0.19732 2.38419e-007 -0.0362214 -0.590494 4.76837e-007 -0.199278 -1.089288 7.15256e-007 -0.510196 -1.626112 1.43051e-006 -0.922701 -2.175138 1.66893e-006 -1.39333 -2.728645 1.90735e-006 -1.901228 -3.283527 1.90735e-006 -2.429453 -3.838811 2.14577e-006 -2.967626 -4.394214 2.38419e-006 -3.510115 70 10 0 0 0 -0.193344 7.15256e-007 -0.0347282 -0.581392 9.53674e-007 -0.192715 -1.077059 9.53674e-007 -0.49736 -1.612464 1.90735e-006 -0.904644 -2.160874 1.90735e-006 -1.371388 -2.714144 2.38419e-006 -1.876595 -3.268931 2.38419e-006 -2.402964 -3.824169 2.38419e-006 -2.939704 -4.379545 2.38419e-006 -3.48112 71 10 0 0 0 -0.193491 0 -0.0347842 -0.580351 4.76837e-007 -0.191614 -1.074431 1.43051e-006 -0.49315 -1.608768 1.90735e-006 -0.896633 -2.156602 1.90735e-006 -1.359882 -2.709621 2.38419e-006 -1.862415 -3.264306 2.86102e-006 -2.386887 -3.819497 2.86102e-006 -2.922305 -4.37484 2.86102e-006 -3.462519 72 10 0 0 0 -0.141724 0.141723 -0.0374134 -0.425972 0.425971 -0.209501 -0.784213 0.784213 -0.537087 -1.166831 1.16683 -0.966747 -1.556474 1.556473 -1.451614 -1.948413 1.948413 -1.969641 -2.340991 2.340991 -2.505059 -2.733741 2.733742 -3.048916 -3.126538 3.126538 -3.596505 73 10 0 0 0 -0.142462 0.14246 -0.0378177 -0.426007 0.426007 -0.208866 -0.783005 0.783005 -0.532589 -1.164859 1.16486 -0.9577 -1.55414 1.554142 -1.438681 -1.945945 1.945947 -1.954001 -2.338468 2.33847 -2.487311 -2.731184 2.731187 -3.029142 -3.123955 3.123958 -3.574591 74 10 0 0 0 -0.143726 0.143728 -0.0385183 -0.430096 0.430099 -0.213774 -0.789379 0.789383 -0.544671 -1.17241 1.172413 -0.976872 -1.562221 1.562224 -1.463601 -1.954223 1.954227 -1.982964 -2.346818 2.346822 -2.518991 -2.739565 2.73957 -3.062746 -3.132356 3.132361 -3.60982 75 10 0 0 0 -0.143342 0.143342 -0.0383037 -0.429304 0.429306 -0.21295 -0.788409 0.788409 -0.543273 -1.171381 1.171383 -0.975105 -1.561172 1.561175 -1.461611 -1.953167 1.95317 -1.980837 -2.345758 2.345762 -2.516777 -2.738506 2.738508 -3.060464 -3.131299 3.131302 -3.607607 76 10 0 0 0 -0.142703 0.142705 -0.0379524 -0.426663 0.426666 -0.209612 -0.783952 0.783955 -0.534232 -1.165947 1.16595 -0.960167 -1.55529 1.555294 -1.441795 -1.947119 1.947124 -1.957564 -2.339649 2.339654 -2.491177 -2.732371 2.732376 -3.033223 -3.125144 3.12515 -3.578849 77 10 0 0 0 -0.142884 0.142887 -0.0380521 -0.42931 0.42931 -0.213389 -0.789182 0.789185 -0.546181 -1.172612 1.172615 -0.980889 -1.562616 1.562619 -1.469833 -1.954688 1.954692 -1.990764 -2.347308 2.347312 -2.527836 -2.740073 2.740076 -3.072674 -3.132877 3.13288 -3.621036 78 10 0 0 0 -0.206968 2.38419e-007 -0.0399908 -0.619772 8.34465e-007 -0.223747 -1.133344 7.15256e-007 -0.567438 -1.677513 7.15256e-007 -1.01109 -2.229815 7.15256e-007 -1.506595 -2.784542 7.15256e-007 -2.031836 -3.339872 4.76837e-007 -2.571608 -3.895349 4.76837e-007 -3.117815 -4.450862 0 -3.666598 79 10 0 0 0 -0.201864 4.76837e-007 -0.0379714 -0.605426 1.19209e-006 -0.211715 -1.113004 1.19209e-006 -0.541427 -1.654632 1.43051e-006 -0.973366 -2.205997 1.43051e-006 -1.46081 -2.760442 1.66893e-006 -1.981263 -3.315693 1.19209e-006 -2.518414 -3.871143 1.19209e-006 -3.063189 -4.426648 1.19209e-006 -3.611098 80 10 0 0 0 -0.192974 -1.19209e-006 -0.0345913 -0.581841 -9.53674e-007 -0.19338 -1.078536 2.38419e-007 -0.500065 -1.614484 -4.76837e-007 -0.909325 -2.163095 4.76837e-007 -1.377306 -2.716475 4.76837e-007 -1.883767 -3.271338 4.76837e-007 -2.411636 -3.82663 4.76837e-007 -2.950093 -4.382048 4.76837e-007 -3.493205 81 10 0 0 0 -0.195442 1.66893e-006 -0.0355126 -0.588448 2.38419e-006 -0.198398 -1.088009 2.86102e-006 -0.510891 -1.625336 2.86102e-006 -0.925296 -2.174748 2.38419e-006 -1.398463 -2.728485 2.38419e-006 -1.909087 -3.283483 2.6226e-006 -2.439747 -3.838835 3.09944e-006 -2.980214 -4.394283 2.86102e-006 -3.524948 82 10 0 0 0 -0.198261 -1.66893e-006 -0.0365806 -0.595977 -2.38419e-007 -0.204239 -1.098792 7.15256e-007 -0.523534 -1.637991 7.15256e-007 -0.945274 -2.188309 4.76837e-007 -1.424718 -2.742376 9.53674e-007 -1.939657 -3.297503 7.15256e-007 -2.47336 -3.852907 9.53674e-007 -3.015908 -4.408384 4.76837e-007 -3.562214 83 10 0 0 0 -0.199343 0 -0.0369949 -0.59946 2.38419e-007 -0.207131 -1.104491 2.38419e-007 -0.531202 -1.644632 7.15256e-007 -0.956795 -2.195314 1.19209e-006 -1.43893 -2.749537 1.19209e-006 -1.955996 -3.304718 1.19209e-006 -2.491151 -3.860147 9.53674e-007 -3.03485 -4.415641 9.53674e-007 -3.582253 85 10 0 0 0 -0.144276 0.144276 -0.0388248 -0.434172 0.434172 -0.219463 -0.796474 0.796474 -0.560271 -1.181193 1.181194 -1.003478 -1.571819 1.571821 -1.500162 -1.964141 1.964141 -2.02731 -2.356856 2.356858 -2.569054 -2.74966 2.74966 -3.117154 -3.14248 3.142482 -3.667804 86 10 0 0 0 -0.144271 0.144267 -0.0388213 -0.434039 0.434037 -0.219264 -0.796219 0.796218 -0.559663 -1.18088 1.180879 -1.002456 -1.571479 1.571478 -1.498791 -1.963791 1.963789 -2.025696 -2.356505 2.356503 -2.567281 -2.749307 2.749304 -3.115273 -3.142125 3.142124 -3.665837 92 10 0 0 0 -0.205193 2.38419e-007 -0.039282 -0.625491 1.19209e-006 -0.231531 -1.146987 9.53674e-007 -0.595556 -1.695007 9.53674e-007 -1.059929 -2.24882 7.15256e-007 -1.571516 -2.80402 4.76837e-007 -2.107143 -3.359494 4.76837e-007 -2.653261 -3.915026 7.15256e-007 -3.203568 -4.47057 0 -3.755466 96 10 0 0 0 -0.137078 0.137077 -0.0349187 -0.426121 0.426124 -0.214243 -0.791956 0.791959 -0.567398 -1.178817 1.178819 -1.026412 -1.570299 1.570301 -1.535898 -1.96287 1.962871 -2.070964 -2.355649 2.355651 -2.61709 -2.748472 2.748473 -3.167626 -3.141301 3.141299 -3.71982 97 10 0 0 0 -0.151943 0.151944 -0.0432382 -0.457613 0.457615 -0.249833 -0.829596 0.829597 -0.62678 -1.217767 1.217767 -1.096972 -1.60944 1.609442 -1.609813 -2.002003 2.002003 -2.14453 -2.394763 2.394763 -2.689249 -2.787576 2.787574 -3.238452 -3.180399 3.180397 -3.789553 98 10 0 0 0 -0.155116 0.15512 -0.0451437 -0.463312 0.463315 -0.256207 -0.835952 0.835955 -0.63593 -1.224203 1.224207 -1.106839 -1.615847 1.615852 -1.619142 -2.008392 2.008394 -2.153137 -2.401144 2.40115 -2.697336 -2.79395 2.793954 -3.246079 -3.186773 3.186775 -3.796793 99 10 0 0 0 -0.154989 0.154989 -0.0450665 -0.46304 0.463037 -0.25587 -0.835658 0.835656 -0.635501 -1.223919 1.223916 -1.106506 -1.615573 1.615568 -1.618949 -2.00812 2.008117 -2.153078 -2.400874 2.40087 -2.697376 -2.793683 2.793679 -3.246163 -3.186505 3.1865 -3.796895 100 10 0 0 0 -0.142111 0.142111 -0.0376256 -0.436687 0.436683 -0.22563 -0.804667 0.804664 -0.586715 -1.192048 1.192045 -1.049972 -1.583637 1.583632 -1.561293 -1.976221 1.976218 -2.096951 -2.369004 2.369001 -2.643162 -2.761823 2.76182 -3.193599 -3.154649 3.154645 -3.745632 101 10 0 0 0 -0.141862 0.141855 -0.0374874 -0.437539 0.437534 -0.227268 -0.806321 0.806316 -0.5914 -1.193921 1.193916 -1.05655 -1.585547 1.585543 -1.568496 -1.978128 1.978123 -2.104056 -2.370906 2.370899 -2.649947 -2.763728 2.763721 -3.200047 -3.156555 3.156549 -3.751822 102 10 0 0 0 -0.213465 9.53674e-007 -0.0426466 -0.642811 9.53674e-007 -0.245635 -1.166528 1.19209e-006 -0.615817 -1.714377 9.53674e-007 -1.079156 -2.267865 4.76837e-007 -1.586824 -2.822927 4.76837e-007 -2.118899 -3.378352 0 -2.662569 -3.933867 -7.15256e-007 -3.211378 -4.489406 -9.53674e-007 -3.762288 103 10 0 0 0 -0.211973 2.38419e-007 -0.0420285 -0.641518 4.76837e-007 -0.245257 -1.166401 -2.38419e-007 -0.61877 -1.714955 -4.76837e-007 -1.086408 -2.268731 -9.53674e-007 -1.597532 -2.823861 -9.53674e-007 -2.131351 -3.379308 -1.19209e-006 -2.675885 -3.934827 -1.66893e-006 -3.225107 -4.490367 -1.90735e-006 -3.776334 104 10 0 0 0 -0.208253 4.76837e-007 -0.0405092 -0.63456 9.53674e-007 -0.239827 -1.158618 4.76837e-007 -0.610975 -1.707096 2.38419e-007 -1.078132 -2.260899 0 -1.589595 -2.816051 -2.38419e-007 -2.123997 -3.371505 -4.76837e-007 -2.66888 -3.927029 -9.53674e-007 -3.218376 -4.482567 -1.66893e-006 -3.769785 105 10 0 0 0 -0.199578 -9.53674e-007 -0.0370872 -0.617367 -9.53674e-007 -0.226451 -1.13857 -9.53674e-007 -0.589681 -1.686438 0 -1.053126 -2.240129 -9.53674e-007 -1.563213 -2.795271 -9.53674e-007 -2.097321 -3.350727 -1.43051e-006 -2.642431 -3.906256 -1.43051e-006 -3.192188 -4.461798 -2.38419e-006 -3.743838 106 10 0 0 0 -0.18692 -4.76837e-007 -0.0323896 -0.590669 -1.43051e-006 -0.206331 -1.106198 -1.90735e-006 -0.554834 -1.652508 -1.43051e-006 -1.009454 -2.205807 -1.43051e-006 -1.515008 -2.760885 -1.90735e-006 -2.047538 -3.316328 -2.38419e-006 -2.592123 -3.871851 -2.86102e-006 -3.141811 -4.427395 -3.33786e-006 -3.693521 107 10 0 0 0 -0.208393 0 -0.0405647 -0.634184 0 -0.239268 -1.157468 0 -0.608235 -1.705514 0 -1.072746 -2.259129 -2.38419e-007 -1.581913 -2.814235 -9.53674e-007 -2.115154 -3.369678 -1.19209e-006 -2.659551 -3.925197 -1.66893e-006 -3.208859 -4.480737 -1.90735e-006 -3.760172 108 10 0 0 0 -0.102951 0.102949 -0.0194151 -0.34831 0.348309 -0.141104 -0.693398 0.693401 -0.431203 -1.07417 1.074167 -0.850126 -1.464054 1.464054 -1.337683 -1.856264 1.856264 -1.861838 -2.248963 2.248964 -2.402567 -2.641763 2.641762 -2.950229 -3.034578 3.034578 -3.500654 109 10 0 0 0 -0.135077 0.13508 -0.0338751 -0.422676 0.422676 -0.21099 -0.787392 0.787392 -0.560131 -1.17354 1.17354 -1.013575 -1.5647 1.564698 -1.517878 -1.957168 1.957167 -2.049331 -2.349919 2.349916 -2.593251 -2.742734 2.742731 -3.142447 -3.135557 3.135553 -3.693734 110 10 0 0 0 -0.120772 0.120775 -0.0269053 -0.392328 0.392333 -0.181018 -0.751012 0.751018 -0.510003 -1.135823 1.135826 -0.953833 -1.526724 1.526728 -1.454305 -1.919164 1.919168 -1.98483 -2.311917 2.311918 -2.528786 -2.70473 2.704731 -3.078289 -3.097558 3.097559 -3.629981 111 10 0 0 0 -0.122062 0.122066 -0.0275002 -0.395276 0.395278 -0.183868 -0.75465 0.754652 -0.515052 -1.139614 1.139617 -0.959951 -1.530562 1.530562 -1.46103 -1.923007 1.923007 -1.991846 -2.315759 2.315763 -2.535852 -2.708574 2.708575 -3.085372 -3.101403 3.101403 -3.637054 112 10 0 0 0 -0.111975 0.111977 -0.023047 -0.370962 0.370962 -0.160877 -0.723399 0.7234 -0.471035 -1.106406 1.106407 -0.903089 -1.496868 1.496868 -1.397631 -1.889202 1.8892 -1.925057 -2.281924 2.281925 -2.467369 -2.674733 2.674732 -3.015893 -3.067556 3.067556 -3.566872 113 10 0 0 0 -0.127181 0.127179 -0.0299232 -0.405998 0.405995 -0.194115 -0.76755 0.767545 -0.532402 -1.153061 1.153055 -0.981179 -1.544137 1.544133 -1.484154 -1.936596 1.936591 -2.015375 -2.329351 2.329347 -2.559372 -2.722161 2.722156 -3.108733 -3.114989 3.114984 -3.660236 114 10 0 0 0 -0.175438 9.53674e-007 -0.0284266 -0.563601 0 -0.186524 -1.071062 0 -0.515971 -1.614506 -4.76837e-007 -0.956153 -2.166996 -9.53674e-007 -1.453426 -2.72192 -9.53674e-007 -1.982489 -3.277348 -1.43051e-006 -2.526165 -3.832872 -1.43051e-006 -3.075842 -4.388412 -1.90735e-006 -3.627871 115 10 0 0 0 -0.188516 -1.43051e-006 -0.0329599 -0.592069 -9.53674e-007 -0.206698 -1.106223 -9.53674e-007 -0.551812 -1.651832 -9.53674e-007 -1.002711 -2.2049 -1.90735e-006 -1.505765 -2.759906 -2.38419e-006 -2.036656 -3.315333 -2.86102e-006 -2.580234 -3.870854 -3.33786e-006 -3.129273 -4.426393 -3.33786e-006 -3.680814 116 10 0 0 0 -0.180573 2.38419e-006 -0.0301664 -0.574396 0 -0.193871 -1.084583 4.76837e-007 -0.529534 -1.629108 0 -0.974937 -2.18186 -9.53674e-007 -1.474752 -2.73678 -4.76837e-007 -2.00371 -3.2922 -1.43051e-006 -2.546915 -3.847723 -1.43051e-006 -3.0963 -4.403267 -1.90735e-006 -3.648213 117 10 0 0 0 -0.159117 -2.86102e-006 -0.0232742 -0.525732 -2.86102e-006 -0.161416 -1.023167 -2.86102e-006 -0.469577 -1.563839 -2.86102e-006 -0.897407 -2.115555 -1.90735e-006 -1.387786 -2.670237 -2.86102e-006 -1.912207 -3.22558 -4.76837e-006 -2.452343 -3.781075 -3.8147e-006 -2.999722 -4.336613 -5.72205e-006 -3.550386 118 10 0 0 0 -0.187611 -1.43051e-006 -0.0326355 -0.589668 -1.43051e-006 -0.204796 -1.103233 -1.90735e-006 -0.548474 -1.648708 -1.90735e-006 -0.998666 -2.201736 -1.90735e-006 -1.501295 -2.756723 -2.86102e-006 -2.03175 -3.312149 -2.86102e-006 -2.575178 -3.86767 -3.33786e-006 -3.124517 -4.423214 -3.8147e-006 -3.676366 119 10 0 0 0 -0.172981 -4.76837e-007 -0.0276158 -0.557417 4.76837e-007 -0.182108 -1.063164 -4.76837e-007 -0.507748 -1.606359 -4.76837e-007 -0.946761 -2.158743 -9.53674e-007 -1.443058 -2.713568 -2.38419e-006 -1.970089 -3.268963 -2.38419e-006 -2.512285 -3.82448 -2.38419e-006 -3.061148 -4.380024 -3.33786e-006 -3.612768 120 10 0 0 0 -0.117188 0.117188 -0.0252941 -0.359642 0.359643 -0.143728 -0.684298 0.684299 -0.386495 -1.047535 1.047539 -0.730484 -1.42687 1.426871 -1.14164 -1.813956 1.813958 -1.602468 -2.204194 2.204195 -2.094209 -2.595655 2.595657 -2.603299 -2.987547 2.987548 -3.12036 121 10 0 0 0 -0.127008 0.127007 -0.0298371 -0.385287 0.385289 -0.166795 -0.723085 0.723087 -0.438751 -1.094514 1.094517 -0.813415 -1.478638 1.478641 -1.252608 -1.868252 1.868254 -1.737137 -2.259848 2.25985 -2.248558 -2.652102 2.652105 -2.773888 -3.044565 3.044568 -3.305174 122 10 0 0 0 -0.124566 0.124568 -0.0286708 -0.378742 0.378741 -0.16063 -0.713095 0.713095 -0.424538 -1.082426 1.082428 -0.790809 -1.465396 1.465397 -1.222623 -1.854461 1.854463 -1.701377 -2.2458 2.245802 -2.20846 -2.637916 2.637918 -2.730386 -3.030281 3.030285 -3.258701 123 10 0 0 0 -0.127362 0.127363 -0.0300081 -0.386224 0.386225 -0.167684 -0.724492 0.724494 -0.440766 -1.096201 1.096204 -0.816582 -1.480474 1.480477 -1.256773 -1.870157 1.87016 -1.742072 -2.261784 2.261788 -2.25407 -2.654056 2.654061 -2.779859 -3.046532 3.046535 -3.311551 124 10 0 0 0 -0.129076 0.129074 -0.0308444 -0.390995 0.390993 -0.172348 -0.731861 0.73186 -0.451745 -1.10513 1.105131 -0.834153 -1.490246 1.490246 -1.280102 -1.880316 1.880319 -1.769849 -2.272116 2.272117 -2.285058 -2.664474 2.664476 -2.813214 -3.057015 3.057017 -3.347176 125 10 0 0 0 -0.12509 0.125087 -0.0289183 -0.380846 0.380846 -0.162788 -0.716889 0.716886 -0.430602 -1.087353 1.08735 -0.801347 -1.470942 1.470939 -1.237057 -1.860244 1.860242 -1.718271 -2.251653 2.251652 -2.226522 -2.643816 2.643815 -2.749513 -3.036228 3.036227 -3.279293 126 10 0 0 0 -0.186345 -1.43051e-006 -0.0321833 -0.563925 -4.76837e-007 -0.180216 -1.051413 -9.53674e-007 -0.469317 -1.582098 -4.76837e-007 -0.860511 -2.128188 -9.53674e-007 -1.313951 -2.680512 -4.76837e-007 -1.809688 -3.234921 -9.53674e-007 -2.329581 -3.78999 -1.43051e-006 -2.861866 -4.345263 -1.43051e-006 -3.399716 127 10 0 0 0 -0.172453 9.53674e-007 -0.0274425 -0.525988 2.86102e-006 -0.154448 -0.993101 2.86102e-006 -0.409255 -1.511393 1.90735e-006 -0.764778 -2.050523 1.90735e-006 -1.186237 -2.599403 2.86102e-006 -1.655918 -3.152126 2.86102e-006 -2.15544 -3.706263 9.53674e-007 -2.671309 -4.260876 2.86102e-006 -3.194516 128 10 0 0 0 -0.183025 4.76837e-007 -0.0310128 -0.554856 0 -0.173794 -1.037544 9.53674e-007 -0.454297 -1.565493 1.90735e-006 -0.836901 -2.110125 1.90735e-006 -1.282833 -2.661774 1.90735e-006 -1.772619 -3.215873 1.90735e-006 -2.28799 -3.77078 1.90735e-006 -2.816739 -4.325948 2.38419e-006 -3.351499 129 10 0 0 0 -0.186067 -9.53674e-007 -0.0320855 -0.563321 -9.53674e-007 -0.179815 -1.05109 0 -0.469431 -1.58215 -4.76837e-007 -0.861835 -2.128274 -1.43051e-006 -1.315458 -2.680578 -9.53674e-007 -1.810999 -3.234986 -4.76837e-007 -2.330876 -3.790061 -9.53674e-007 -2.863341 -4.345342 -1.43051e-006 -3.401429 130 10 0 0 0 -0.18691 -4.76837e-007 -0.0323831 -0.565527 1.43051e-006 -0.18138 -1.054029 1.43051e-006 -0.472345 -1.585254 9.53674e-007 -0.865291 -2.131602 4.76837e-007 -1.32012 -2.684045 1.43051e-006 -1.816946 -3.238511 9.53674e-007 -2.337738 -3.793605 9.53674e-007 -2.870678 -4.348899 9.53674e-007 -3.409145 131 10 0 0 0 -0.175674 4.76837e-007 -0.0285058 -0.534788 1.43051e-006 -0.160175 -1.006731 9.53674e-007 -0.422621 -1.528133 2.38419e-006 -0.786388 -2.069092 2.38419e-006 -1.215435 -2.618927 2.86102e-006 -1.691456 -3.17214 2.86102e-006 -2.196064 -3.726562 3.33786e-006 -2.716144 -4.281384 3.33786e-006 -3.243171 132 10 0 0 0 -0.141294 0.141295 -0.0371798 -0.424424 0.424425 -0.207617 -0.781527 0.781526 -0.531661 -1.1635 1.1635 -0.957475 -1.55286 1.552861 -1.439274 -1.9447 1.944701 -1.9553 -2.33724 2.337239 -2.489174 -2.729963 2.729962 -3.031426 -3.12274 3.12274 -3.577165 133 10 0 0 0 -0.14136 0.141362 -0.0372158 -0.424181 0.424183 -0.207198 -0.781208 0.781209 -0.531007 -1.163249 1.163253 -0.957213 -1.552702 1.552705 -1.439997 -1.9446 1.944603 -1.957149 -2.337166 2.337169 -2.492071 -2.729908 2.729909 -3.03527 -3.122692 3.122693 -3.581922 134 10 0 0 0 -0.133333 0.133334 -0.0329785 -0.402906 0.402907 -0.184424 -0.749811 0.749812 -0.479296 -1.126405 1.126407 -0.876751 -1.513186 1.513189 -1.335132 -1.903976 1.903979 -1.834033 -2.296087 2.296091 -2.355876 -2.688613 2.688615 -2.889276 -3.081263 3.081265 -3.427675 135 10 0 0 0 -0.13507 0.13507 -0.0338712 -0.407696 0.407697 -0.189437 -0.757131 0.757129 -0.49115 -1.135146 1.135145 -0.895542 -1.5226 1.522598 -1.35944 -1.913702 1.913701 -1.862852 -2.305971 2.30597 -2.388537 -2.698586 2.698585 -2.925382 -3.091302 3.091301 -3.467152 136 10 0 0 0 -0.134646 0.134644 -0.0336521 -0.405959 0.405959 -0.187437 -0.754257 0.754256 -0.486053 -1.131688 1.131688 -0.88755 -1.518936 1.518937 -1.349726 -1.909947 1.909948 -1.851805 -2.302168 2.302167 -2.376169 -2.694749 2.694748 -2.911639 -3.087434 3.087433 -3.451924 137 10 0 0 0 -0.136667 0.136667 -0.0347037 -0.411837 0.411839 -0.193768 -0.762826 0.762826 -0.499815 -1.141672 1.141674 -0.908434 -1.529654 1.529657 -1.376921 -1.92099 1.920991 -1.883921 -2.313348 2.313349 -2.41203 -2.705998 2.706 -2.950548 -3.098738 3.098739 -3.493597 138 10 0 0 0 -0.20776 4.76837e-007 -0.0403097 -0.616421 4.76837e-007 -0.219516 -1.124536 4.76837e-007 -0.550434 -1.665196 4.76837e-007 -0.978212 -2.215647 9.53674e-007 -1.458633 -2.769587 4.76837e-007 -1.971857 -3.324593 4.76837e-007 -2.502697 -3.879912 0 -3.042037 -4.435323 0 -3.584879 139 10 0 0 0 -0.208603 -4.76837e-007 -0.0406501 -0.622121 -7.15256e-007 -0.225199 -1.135232 -2.38419e-007 -0.56778 -1.67888 -2.38419e-007 -1.008932 -2.230858 -2.38419e-007 -1.501546 -2.785439 -4.76837e-007 -2.024219 -3.34071 -7.15256e-007 -2.561967 -3.896158 -9.53674e-007 -3.106596 -4.451655 -1.19209e-006 -3.654008 140 10 0 0 0 -0.203901 -1.43051e-006 -0.0387698 -0.607594 -9.53674e-007 -0.212652 -1.112985 -4.76837e-007 -0.537509 -1.652583 -1.43051e-006 -0.960868 -2.202651 -1.43051e-006 -1.438526 -2.756474 -1.43051e-006 -1.95023 -3.311444 -1.43051e-006 -2.480279 -3.866751 -2.38419e-006 -3.0192 -4.422154 -2.38419e-006 -3.561797 141 10 0 0 0 -0.189192 1.43051e-006 -0.0332055 -0.569623 0 -0.183899 -1.058293 9.53674e-007 -0.475173 -1.589149 -4.76837e-007 -0.866917 -2.134954 0 -1.318845 -2.686963 4.76837e-007 -1.81172 -3.241156 -4.76837e-007 -2.328364 -3.796067 -1.43051e-006 -2.857202 -4.351208 -9.53674e-007 -3.391244 142 10 0 0 0 -0.194626 1.43051e-006 -0.0352061 -0.583001 2.38419e-006 -0.193513 -1.077122 2.38419e-006 -0.495124 -1.610734 1.90735e-006 -0.896082 -2.157844 2.38419e-006 -1.355138 -2.710395 2.38419e-006 -1.852995 -3.264812 2.38419e-006 -2.373027 -3.819836 1.90735e-006 -2.904291 -4.375047 1.43051e-006 -3.440295 143 10 0 0 0 -0.210449 -2.38419e-007 -0.0414027 -0.62341 -2.38419e-007 -0.225331 -1.134382 -4.76837e-007 -0.562826 -1.676396 -4.76837e-007 -0.996465 -2.227448 0 -1.481416 -2.78162 -2.38419e-007 -1.997784 -3.336717 -4.76837e-007 -2.530763 -3.89208 -9.53674e-007 -3.071661 -4.447516 -1.43051e-006 -3.615801 144 10 0 0 0 -0.149864 0.149868 -0.0420154 -0.446257 0.446259 -0.232953 -0.811646 0.81165 -0.584493 -1.197174 1.197178 -1.033382 -1.587918 1.587922 -1.531663 -1.980209 1.980211 -2.057912 -2.37289 2.372892 -2.597823 -2.765672 2.765674 -3.143864 -3.158476 3.158478 -3.692359 146 10 0 0 0 -0.149507 0.149505 -0.0418065 -0.445464 0.445465 -0.232039 -0.810651 0.810652 -0.582856 -1.196119 1.19612 -1.031308 -1.586848 1.586849 -1.529401 -1.979136 1.979137 -2.055622 -2.371821 2.371822 -2.595562 -2.764602 2.764605 -3.141646 -3.157408 3.157407 -3.69019 147 10 0 0 0 -0.152225 0.152225 -0.0434055 -0.451149 0.451148 -0.238501 -0.817502 0.817501 -0.593517 -1.203214 1.203213 -1.043746 -1.593944 1.593943 -1.541835 -1.986204 1.986201 -2.067301 -2.378869 2.378866 -2.606414 -2.77164 2.771637 -3.151748 -3.164439 3.164436 -3.699579 150 10 0 0 0 -0.218778 5.36442e-007 -0.0448909 -0.646424 2.38419e-007 -0.245818 -1.165997 2.38419e-007 -0.60468 -1.711448 0 -1.054766 -2.263906 -4.76837e-007 -1.551746 -2.818602 -7.15256e-007 -2.07643 -3.373909 -1.13249e-006 -2.615354 -3.929369 -1.43051e-006 -3.160871 -4.48488 -1.96695e-006 -3.709165 151 10 0 0 0 -0.224512 3.57628e-007 -0.047386 -0.662984 2.98023e-007 -0.26179 -1.18931 1.78814e-007 -0.639515 -1.737543 -5.96046e-008 -1.105174 -2.290988 -4.17233e-007 -1.612349 -2.845973 -5.96046e-007 -2.142727 -3.401363 -9.53674e-007 -2.684678 -3.956858 -1.3113e-006 -3.232076 -4.512379 -1.72853e-006 -3.781821 152 10 0 0 0 -0.212018 -4.76837e-007 -0.0420476 -0.629738 -4.76837e-007 -0.231335 -1.143648 -7.15256e-007 -0.575848 -1.686996 -9.53674e-007 -1.015587 -2.238682 -1.43051e-006 -1.505685 -2.793129 -1.66893e-006 -2.026108 -3.348345 -1.90735e-006 -2.562274 -3.903776 -2.6226e-006 -3.105865 -4.459259 -2.86102e-006 -3.652613 153 10 0 0 0 -0.2206 0 -0.0456745 -0.651143 -1.78814e-007 -0.250124 -1.172412 -3.27826e-007 -0.613526 -1.718568 -5.06639e-007 -1.067323 -2.271254 -9.83477e-007 -1.566457 -2.826019 -1.2219e-006 -2.092445 -3.38135 -1.57952e-006 -2.632138 -3.936827 -1.99676e-006 -3.178185 -4.49234 -2.38419e-006 -3.726903 156 10 0 0 0 -0.146766 0.146768 -0.0402286 -0.445646 0.44565 -0.235254 -0.813894 0.813898 -0.597346 -1.20077 1.200774 -1.056476 -1.592048 1.59205 -1.562532 -1.984519 1.984521 -2.094196 -2.377267 2.377269 -2.637849 -2.770076 2.770078 -3.186697 -3.162898 3.162898 -3.737665 157 10 0 0 0 -0.155602 0.155592 -0.0454367 -0.459679 0.459677 -0.249265 -0.828141 0.828135 -0.612167 -1.214294 1.214289 -1.06569 -1.605105 1.6051 -1.564891 -1.99739 1.997384 -2.09094 -2.39007 2.390062 -2.630729 -2.782847 2.78284 -3.176845 -3.175657 3.17565 -3.725617 158 10 0 0 0 -0.152694 0.152694 -0.0436842 -0.453817 0.453819 -0.242466 -0.821039 0.821043 -0.600695 -1.206945 1.206945 -1.052317 -1.597733 1.597735 -1.551231 -1.990029 1.990031 -2.077651 -2.382711 2.382713 -2.61777 -2.775496 2.775496 -3.16413 -3.168306 3.168304 -3.713099 159 10 0 0 0 -0.152094 0.152092 -0.0433279 -0.456094 0.456088 -0.247009 -0.826172 0.826166 -0.61622 -1.213359 1.213353 -1.077889 -1.604655 1.60465 -1.584309 -1.997114 1.997108 -2.115481 -2.389853 2.389846 -2.65865 -2.782659 2.782651 -3.207112 -3.175483 3.175477 -3.757745 160 10 0 0 0 -0.155129 0.155128 -0.0451516 -0.459784 0.459785 -0.249978 -0.82897 0.82897 -0.615673 -1.215511 1.215509 -1.072166 -1.606505 1.606503 -1.573935 -1.998856 1.998852 -2.101773 -2.391553 2.391551 -2.642756 -2.784344 2.78434 -3.189703 -3.177155 3.177154 -3.739134 161 10 0 0 0 -0.148878 0.148882 -0.0414428 -0.447896 0.447902 -0.236697 -0.815509 0.815513 -0.596379 -1.201941 1.201943 -1.052014 -1.593023 1.593027 -1.555106 -1.985429 1.985429 -2.08462 -2.378151 2.378151 -2.626843 -2.77095 2.770952 -3.174699 -3.163774 3.163772 -3.724877 162 10 0 0 0 -0.20804 5.21541e-008 -0.0404233 -0.625362 -1.93715e-007 -0.229255 -1.141401 -5.36442e-007 -0.579029 -1.686218 -7.00355e-007 -1.025881 -2.238621 -1.14739e-006 -1.522336 -2.793341 -1.50502e-006 -2.047441 -3.348661 -2.04146e-006 -2.586826 -3.904131 -2.30968e-006 -3.132717 -4.459648 -2.83122e-006 -3.681609 163 10 0 0 0 -0.206627 -2.38419e-007 -0.039855 -0.62591 2.38419e-007 -0.230936 -1.144543 -3.57628e-007 -0.587338 -1.690767 -3.57628e-007 -1.041485 -2.24379 -1.07288e-006 -1.544044 -2.798729 -1.07288e-006 -2.07344 -3.354124 -1.78814e-006 -2.615676 -3.909628 -2.38419e-006 -3.163599 -4.465162 -2.86102e-006 -3.714172 164 10 0 0 0 -0.208035 0 -0.0404205 -0.623854 0 -0.227554 -1.138374 -3.57628e-007 -0.573563 -1.682392 -7.15256e-007 -1.016488 -2.23444 -1.07288e-006 -1.509711 -2.789032 -1.54972e-006 -2.032558 -3.344305 -1.78814e-006 -2.570446 -3.899757 -2.26498e-006 -3.115305 -4.455259 -2.74181e-006 -3.663124 165 10 0 0 0 -0.210138 2.08616e-007 -0.0412757 -0.628621 8.9407e-008 -0.231436 -1.144685 -2.08616e-007 -0.581262 -1.689317 -4.76837e-007 -1.027204 -2.241601 -7.7486e-007 -1.522539 -2.796271 -1.3113e-006 -2.046718 -3.351567 -1.66893e-006 -2.585449 -3.907029 -2.14577e-006 -3.1309 -4.462543 -2.563e-006 -3.679183 166 10 0 0 0 -0.203991 2.38419e-007 -0.0388063 -0.621656 4.76837e-007 -0.228032 -1.140249 2.38419e-007 -0.584334 -1.686657 -4.76837e-007 -1.039482 -2.239788 -7.15256e-007 -1.543203 -2.794769 -1.19209e-006 -2.073504 -3.35018 -1.43051e-006 -2.616326 -3.905691 -1.90735e-006 -3.164933 -4.461231 -2.14577e-006 -3.716088 167 10 0 0 0 -0.200726 0 -0.0375289 -0.616604 -4.76837e-007 -0.224726 -1.135152 -7.15256e-007 -0.580912 -1.681729 -1.19209e-006 -1.036984 -2.234965 -1.90735e-006 -1.541837 -2.78999 -1.66893e-006 -2.073089 -3.345416 -2.6226e-006 -2.616651 -3.900929 -2.6226e-006 -3.165515 -4.456464 -3.09944e-006 -3.716532 168 10 0 0 0 -0.142478 0.142482 -0.037828 -0.43665 0.436658 -0.225195 -0.803205 0.803206 -0.580942 -1.189613 1.189617 -1.036411 -1.580766 1.58077 -1.540627 -1.973207 1.973207 -2.071246 -2.365957 2.365959 -2.614837 -2.758769 2.75877 -3.164138 -3.151598 3.1516 -3.715875 169 10 0 0 0 -0.139653 0.139649 -0.0362878 -0.431275 0.431272 -0.219614 -0.797054 0.797049 -0.572542 -1.183361 1.183355 -1.027231 -1.574533 1.574528 -1.531705 -1.966997 1.96699 -2.062952 -2.359743 2.359735 -2.606549 -2.752554 2.752544 -3.155564 -3.145382 3.145372 -3.707049 170 10 0 0 0 -0.144164 0.14416 -0.0387599 -0.439999 0.439995 -0.228794 -0.807024 0.80702 -0.586298 -1.193467 1.193465 -1.042051 -1.584622 1.58462 -1.546187 -1.977085 1.977079 -2.077595 -2.369839 2.369835 -2.621818 -2.762658 2.762651 -3.171551 -3.155491 3.155485 -3.723505 171 10 0 0 0 -0.138016 0.138014 -0.0354125 -0.428233 0.428226 -0.216544 -0.793664 0.79366 -0.568244 -1.179917 1.179915 -1.022513 -1.571069 1.571065 -1.526679 -1.963524 1.963517 -2.057767 -2.35628 2.356273 -2.601913 -2.749095 2.749087 -3.151689 -3.141927 3.141919 -3.703776 172 10 0 0 0 -0.137526 0.13752 -0.0351573 -0.427083 0.427077 -0.215272 -0.792242 0.792238 -0.565991 -1.178434 1.178431 -1.019781 -1.569582 1.56958 -1.523903 -1.962032 1.962027 -2.05484 -2.35478 2.354774 -2.598669 -2.747593 2.747589 -3.148194 -3.140419 3.140413 -3.700136 173 10 0 0 0 -0.135185 0.135183 -0.033931 -0.422819 0.422821 -0.211098 -0.78741 0.787409 -0.559793 -1.173423 1.173426 -1.012254 -1.564528 1.564528 -1.515695 -1.957004 1.957005 -2.047431 -2.34977 2.349768 -2.592217 -2.742583 2.74258 -3.14242 -3.135418 3.135416 -3.694755 174 10 0 0 0 -0.205902 -2.38419e-007 -0.0395648 -0.633947 -9.53674e-007 -0.240974 -1.159788 -9.53674e-007 -0.617275 -1.709049 -1.66893e-006 -1.089434 -2.263168 -1.90735e-006 -1.605058 -2.818424 -1.90735e-006 -2.142429 -3.373913 -2.14577e-006 -2.689245 -3.929447 -2.38419e-006 -3.239888 -4.484991 -2.6226e-006 -3.791981 175 10 0 0 0 -0.190228 4.76837e-007 -0.0335817 -0.59664 0 -0.210361 -1.112415 0 -0.559472 -1.658958 -4.76837e-007 -1.01536 -2.212456 -4.76837e-007 -1.523157 -2.767611 -9.53674e-007 -2.057591 -3.323082 -9.53674e-007 -2.603658 -3.878616 -9.53674e-007 -3.154387 -4.434162 -1.90735e-006 -3.706888 176 10 0 0 0 -0.206882 9.53674e-007 -0.0399592 -0.636019 4.76837e-007 -0.242689 -1.162247 2.38419e-007 -0.620127 -1.711596 0 -1.092856 -2.265732 -2.38419e-007 -1.608695 -2.82099 -4.76837e-007 -2.146103 -3.376474 -7.15256e-007 -2.69289 -3.932009 -9.53674e-007 -3.24349 -4.487555 -1.43051e-006 -3.795537 177 10 0 0 0 -0.201447 9.53674e-007 -0.0378084 -0.619057 4.76837e-007 -0.22697 -1.138334 -9.53674e-007 -0.585058 -1.685375 -9.53674e-007 -1.04373 -2.238903 -9.53674e-007 -1.551856 -2.79405 -1.66893e-006 -2.086108 -3.349524 -1.66893e-006 -2.631999 -3.905058 -2.38419e-006 -3.182637 -4.460604 -2.86102e-006 -3.735109 178 10 0 0 0 -0.201769 -1.43051e-006 -0.0379335 -0.620491 -1.43051e-006 -0.228367 -1.141777 -1.90735e-006 -0.591813 -1.689764 -1.90735e-006 -1.055966 -2.243584 -2.6226e-006 -1.567669 -2.798794 -3.09944e-006 -2.103678 -3.354279 -3.33786e-006 -2.65031 -3.909819 -3.8147e-006 -3.201207 -4.465364 -4.29153e-006 -3.753695 179 10 0 0 0 -0.204082 -9.53674e-007 -0.0388423 -0.624159 -9.53674e-007 -0.230837 -1.144533 -1.66893e-006 -0.591837 -1.691984 -1.66893e-006 -1.05285 -2.245628 -2.38419e-006 -1.562415 -2.8008 -2.38419e-006 -2.097281 -3.356274 -2.86102e-006 -2.643391 -3.911804 -3.33786e-006 -3.194078 -4.46735 -3.57628e-006 -3.746523 180 10 0 0 0 -0.115013 0.115013 -0.024343 -0.352851 0.352854 -0.137738 -0.672367 0.672371 -0.370093 -1.031542 1.031545 -0.700636 -1.408146 1.408151 -1.098147 -1.793694 1.793695 -1.547137 -2.183105 2.183109 -2.029516 -2.574062 2.574065 -2.530766 -2.965557 2.965558 -3.040411 181 10 0 0 0 -0.115518 0.11552 -0.0245638 -0.354701 0.354703 -0.139406 -0.67548 0.675485 -0.374273 -1.035244 1.035247 -0.706709 -1.412159 1.412162 -1.1057 -1.797881 1.797884 -1.555977 -2.187437 2.18744 -2.039889 -2.578495 2.578496 -2.542621 -2.970074 2.970078 -3.053774 182 10 0 0 0 -0.108952 0.108946 -0.0217922 -0.336044 0.336038 -0.124028 -0.645174 0.645171 -0.336776 -0.996692 0.99669 -0.64431 -1.368141 1.368138 -1.019054 -1.750581 1.750578 -1.447646 -2.138144 2.138141 -1.912483 -2.527855 2.527853 -2.398091 -2.918329 2.918328 -2.892817 183 10 0 0 0 -0.117409 0.117406 -0.0253909 -0.359455 0.359453 -0.143375 -0.68291 0.682907 -0.383658 -1.044868 1.044866 -0.72331 -1.423244 1.423243 -1.129524 -1.809786 1.809786 -1.586027 -2.199779 2.199779 -2.07484 -2.591096 2.591094 -2.581587 -2.98287 2.98287 -3.096319 184 10 0 0 0 -0.112061 0.112057 -0.0230827 -0.345009 0.345007 -0.131303 -0.659761 0.659755 -0.354427 -1.015205 1.015198 -0.67342 -1.389286 1.389278 -1.059366 -1.773342 1.773337 -1.498144 -2.161928 2.161923 -1.972162 -2.552349 2.552344 -2.466192 -2.943417 2.94341 -2.969074 185 10 0 0 0 -0.113689 0.113687 -0.023771 -0.35007 0.350071 -0.135606 -0.66879 0.668787 -0.366388 -1.027496 1.027493 -0.695451 -1.403666 1.403662 -1.090879 -1.78886 1.788859 -1.537389 -2.178082 2.178081 -2.017757 -2.568909 2.568908 -2.51721 -2.960323 2.960322 -3.025532 186 10 0 0 0 -0.167622 -6.67572e-006 -0.0258927 -0.51103 -1.90735e-006 -0.144738 -0.967621 -3.8147e-006 -0.383807 -1.477533 -3.8147e-006 -0.71884 -2.010592 -1.90735e-006 -1.1179 -2.555697 -1.90735e-006 -1.566195 -3.106139 -3.8147e-006 -2.046547 -3.65874 -3.8147e-006 -2.544869 -4.212193 -4.76837e-006 -3.052161 187 10 0 0 0 -0.181348 0 -0.0304303 -0.549134 -9.53674e-007 -0.169603 -1.027932 -4.76837e-007 -0.443388 -1.553092 -9.53674e-007 -0.817693 -2.09584 -4.76837e-007 -1.254645 -2.646405 -9.53674e-007 -1.735915 -3.199917 0 -2.243842 -3.754458 -9.53674e-007 -2.765834 -4.309333 -4.76837e-007 -3.293897 188 10 0 0 0 -0.172144 5.72205e-006 -0.0273426 -0.523751 5.72205e-006 -0.152764 -0.988198 3.8147e-006 -0.403474 -1.503844 4.76837e-006 -0.752262 -2.040764 4.76837e-006 -1.16513 -2.588176 5.72205e-006 -1.625902 -3.13998 5.72205e-006 -2.117023 -3.693475 5.72205e-006 -2.624772 -4.24754 4.76837e-006 -3.139678 189 10 0 0 0 -0.148134 0 -0.0201108 -0.458176 -1.90735e-006 -0.114673 -0.8838 9.53674e-007 -0.313174 -1.371576 2.86102e-006 -0.602805 -1.889887 1.90735e-006 -0.958366 -2.426061 1.90735e-006 -1.36846 -2.971086 1.90735e-006 -1.81637 -3.52001 2.86102e-006 -2.286345 -4.07031 9.53674e-007 -2.765667 190 10 0 0 0 -0.176382 9.53674e-007 -0.0287426 -0.534981 2.86102e-006 -0.159975 -1.005346 1.90735e-006 -0.419899 -1.524788 9.53674e-007 -0.778422 -2.063982 1.90735e-006 -1.200146 -2.612612 9.53674e-007 -1.668251 -3.165072 9.53674e-007 -2.165252 -3.718967 9.53674e-007 -2.677893 -4.273344 0 -3.197249 191 10 0 0 0 -0.179117 -1.90735e-006 -0.0296664 -0.542902 -9.53674e-007 -0.165337 -1.018164 -1.90735e-006 -0.433196 -1.541017 -9.53674e-007 -0.800956 -2.082385 0 -1.231761 -2.632228 -1.90735e-006 -1.707848 -3.185359 -9.53674e-007 -2.211574 -3.739676 -1.90735e-006 -2.730037 -4.294379 -1.90735e-006 -3.254854 192 10 0 0 0 -0.125325 0.12533 -0.0290298 -0.379463 0.379469 -0.160949 -0.712146 0.712149 -0.421056 -1.079247 1.079249 -0.778825 -1.460255 1.460257 -1.199083 -1.847992 1.847994 -1.665401 -2.238506 2.238505 -2.160572 -2.630068 2.630068 -2.671441 -3.021999 3.021999 -3.189318 193 10 0 0 0 -0.131517 0.131517 -0.0320578 -0.396671 0.396673 -0.177704 -0.739091 0.739093 -0.460978 -1.112606 1.112608 -0.844442 -1.497526 1.497528 -1.289026 -1.887363 1.887363 -1.776048 -2.278996 2.278997 -2.288183 -2.671249 2.671249 -2.813393 -3.063696 3.063696 -3.344197 194 10 0 0 0 -0.122229 0.12222 -0.0275769 -0.371221 0.371214 -0.153427 -0.699464 0.699458 -0.403767 -1.063628 1.063624 -0.750981 -1.442824 1.442821 -1.161392 -1.829553 1.829546 -1.619348 -2.219501 2.219495 -2.107663 -2.610713 2.610708 -2.612755 -3.002423 3.002417 -3.126184 195 10 0 0 0 -0.131584 0.131584 -0.0320919 -0.396729 0.396725 -0.17772 -0.73904 0.739039 -0.460718 -1.112441 1.112438 -0.843683 -1.497272 1.49727 -1.287663 -1.887054 1.887053 -1.77409 -2.278659 2.278658 -2.285689 -2.670892 2.670889 -2.810413 -3.063324 3.063321 -3.340744 196 10 0 0 0 -0.136612 0.136612 -0.0346749 -0.408913 0.408914 -0.189796 -0.756077 0.756077 -0.485354 -1.131919 1.131918 -0.879275 -1.51791 1.517909 -1.331561 -1.908176 1.908177 -1.823693 -2.29999 2.29999 -2.339099 -2.692323 2.692322 -2.866501 -3.08482 3.084819 -3.398974 197 10 0 0 0 -0.132133 0.13214 -0.0323696 -0.397034 0.397039 -0.177685 -0.738236 0.738242 -0.457918 -1.110475 1.110482 -0.835928 -1.494414 1.49442 -1.273912 -1.883648 1.883653 -1.754418 -2.274937 2.274942 -2.260671 -2.666967 2.666971 -2.78059 -3.059259 3.059264 -3.30707 198 10 0 0 0 -0.177118 3.8147e-006 -0.02899 -0.536301 2.86102e-006 -0.160713 -1.006329 1.90735e-006 -0.420103 -1.524966 2.86102e-006 -0.776511 -2.063353 1.90735e-006 -1.195023 -2.611403 1.90735e-006 -1.659558 -3.163555 1.90735e-006 -2.153721 -3.717346 9.53674e-007 -2.665015 -4.271734 9.53674e-007 -3.184555 199 10 0 0 0 -0.17498 9.53674e-007 -0.028275 -0.530647 9.53674e-007 -0.157047 -0.997741 9.53674e-007 -0.411828 -1.514508 9.53674e-007 -0.763439 -2.051773 9.53674e-007 -1.177616 -2.599213 0 -1.638563 -3.151011 0 -2.129616 -3.704594 0 -2.638338 -4.258827 0 -3.155548 200 10 0 0 0 -0.206257 -2.38419e-006 -0.0397069 -0.609705 -9.53674e-007 -0.213331 -1.112615 -9.53674e-007 -0.532828 -1.649994 -1.43051e-006 -0.947439 -2.198627 -9.53674e-007 -1.415563 -2.751715 -1.90735e-006 -1.918813 -3.306327 -2.38419e-006 -2.442014 -3.861457 -2.38419e-006 -2.975842 -4.416751 -2.38419e-006 -3.514455 201 10 0 0 0 -0.179726 1.90735e-006 -0.0298731 -0.543009 2.86102e-006 -0.165108 -1.01641 3.8147e-006 -0.429915 -1.537193 3.8147e-006 -0.792011 -2.076873 2.86102e-006 -1.215705 -2.625616 3.8147e-006 -1.684531 -3.178077 3.8147e-006 -2.181532 -3.732023 1.90735e-006 -2.694804 -4.286505 2.86102e-006 -3.215796 202 10 0 0 0 -0.18652 9.53674e-007 -0.0322471 -0.561522 9.53674e-007 -0.177906 -1.044456 1.90735e-006 -0.458839 -1.571011 4.76837e-007 -0.837241 -2.114122 4.76837e-007 -1.275867 -2.664841 -4.76837e-007 -1.758267 -3.218458 -4.76837e-007 -2.267477 -3.773085 -9.53674e-007 -2.790897 -4.328045 -9.53674e-007 -3.320692 203 10 0 0 0 -0.191805 -3.33786e-006 -0.0341604 -0.574347 -1.90735e-006 -0.186845 -1.062714 -2.86102e-006 -0.477561 -1.592318 -2.38419e-006 -0.865295 -2.13694 -2.38419e-006 -1.311179 -2.688222 -2.38419e-006 -1.797948 -3.24207 -3.8147e-006 -2.309986 -3.796821 -3.8147e-006 -2.835629 -4.351869 -4.29153e-006 -3.367429 206 10 0 0 0 -0.141983 0.141982 -0.0375534 -0.422522 0.422517 -0.204213 -0.775138 0.775135 -0.514896 -1.153605 1.1536 -0.921558 -1.54083 1.540824 -1.38351 -1.931685 1.93168 -1.88334 -2.323792 2.323786 -2.405099 -2.716309 2.7163 -2.938125 -3.108944 3.108938 -3.476127 208 10 0 0 0 -0.14396 0.143954 -0.0386467 -0.427294 0.42729 -0.209383 -0.781771 0.781763 -0.525491 -1.161181 1.161178 -0.93705 -1.548841 1.548838 -1.402706 -1.939878 1.939872 -1.905096 -2.332057 2.332051 -2.428577 -2.724609 2.724602 -2.962906 -3.117277 3.117268 -3.502012 209 10 0 0 0 -0.140958 0.140957 -0.0369961 -0.4203 0.4203 -0.201942 -0.772221 0.772219 -0.510616 -1.150352 1.150348 -0.915592 -1.537511 1.537508 -1.376992 -1.928348 1.928344 -1.876626 -2.320458 2.320454 -2.398351 -2.71297 2.712965 -2.931328 -3.105605 3.105601 -3.469278 211 10 0 0 0 -0.185441 -4.76837e-007 -0.031862 -0.560705 -4.76837e-007 -0.17776 -1.045647 -1.43051e-006 -0.462251 -1.574488 -2.38419e-006 -0.847607 -2.119236 -1.90735e-006 -1.294125 -2.670782 -3.33786e-006 -1.783044 -3.224785 -2.86102e-006 -2.297071 -3.779621 -3.8147e-006 -2.824286 -4.334726 -4.29153e-006 -3.357459 213 10 0 0 0 -0.188349 -4.76837e-007 -0.0329022 -0.568995 -4.76837e-007 -0.183795 -1.058441 -9.53674e-007 -0.47651 -1.589762 0 -0.869774 -2.135815 -1.43051e-006 -1.323027 -2.68796 -4.76837e-007 -1.817091 -3.242229 -1.90735e-006 -2.334891 -3.797203 -1.90735e-006 -2.864997 -4.352402 -2.86102e-006 -3.400635 216 10 0 0 0 -0.145797 0.145797 -0.039679 -0.439584 0.43958 -0.226422 -0.803894 0.803892 -0.574146 -1.18892 1.18892 -1.019457 -1.579442 1.579438 -1.514763 -1.971655 1.971653 -2.039017 -2.364322 2.364318 -2.57813 -2.7571 2.757094 -3.124373 -3.149917 3.149913 -3.673687 217 10 0 0 0 -0.143523 0.143523 -0.0384057 -0.431518 0.431516 -0.216123 -0.791432 0.79143 -0.549042 -1.174334 1.174332 -0.98043 -1.563931 1.563927 -1.464791 -1.955809 1.955807 -1.981539 -2.348343 2.348339 -2.515338 -2.741064 2.741058 -3.057253 -3.133837 3.133832 -3.602877 218 10 0 0 0 -0.145411 0.145405 -0.0394578 -0.43835 0.438346 -0.224858 -0.802052 0.802046 -0.570451 -1.186785 1.186781 -1.013754 -1.577184 1.577176 -1.507465 -1.969349 1.969341 -2.030599 -2.362 2.361991 -2.569112 -2.75478 2.75477 -3.114961 -3.14759 3.147581 -3.66398 219 10 0 0 0 -0.145986 0.145986 -0.039786 -0.438421 0.438421 -0.22439 -0.801372 0.801368 -0.567394 -1.185699 1.185694 -1.007928 -1.575913 1.575912 -1.499426 -1.968021 1.968014 -2.02112 -2.360644 2.360641 -2.55824 -2.7534 2.753392 -3.102537 -3.146198 3.146193 -3.650122 220 10 0 0 0 -0.146747 0.146739 -0.0402144 -0.440951 0.440948 -0.227629 -0.805201 0.805195 -0.57513 -1.190139 1.190135 -1.019864 -1.580622 1.580616 -1.514659 -1.972822 1.972816 -2.038624 -2.365476 2.36547 -2.577252 -2.75825 2.758243 -3.122621 -3.151053 3.151044 -3.671112 221 10 0 0 0 -0.144567 0.144566 -0.038987 -0.434085 0.434081 -0.219038 -0.79493 0.794927 -0.555003 -1.178251 1.178247 -0.989006 -1.56803 1.568027 -1.475363 -1.959973 1.95997 -1.993491 -2.352533 2.352528 -2.528225 -2.745266 2.74526 -3.070838 -3.138042 3.138037 -3.616914 222 10 0 0 0 -0.206291 3.57628e-007 -0.0397198 -0.622293 0 -0.227055 -1.137787 -4.17233e-007 -0.575462 -1.682348 -7.7486e-007 -1.021034 -2.234669 -1.2517e-006 -1.516719 -2.789408 -1.49012e-006 -2.042176 -3.344751 -1.96695e-006 -2.582439 -3.900242 -2.44379e-006 -3.129529 -4.455772 -2.86102e-006 -3.679504 223 10 0 0 0 -0.208767 8.34465e-007 -0.0407178 -0.627718 9.53674e-007 -0.231415 -1.144676 1.19209e-007 -0.583511 -1.689867 1.19209e-007 -1.032252 -2.242428 -3.57628e-007 -1.530216 -2.797209 -8.34465e-007 -2.056388 -3.352549 -9.53674e-007 -2.596466 -3.908028 -1.3113e-006 -3.142768 -4.463547 -1.66893e-006 -3.691568 224 10 0 0 0 -0.203838 -4.76837e-007 -0.0387461 -0.614403 0 -0.220031 -1.125843 -9.53674e-007 -0.558633 -1.668482 -1.43051e-006 -0.99509 -2.220078 -1.66893e-006 -1.484425 -2.774555 -1.90735e-006 -2.005414 -3.329817 -2.86102e-006 -2.542782 -3.885273 -2.86102e-006 -3.087758 -4.440779 -3.33786e-006 -3.635951 225 10 0 0 0 -0.20018 -2.38419e-007 -0.0373175 -0.60445 -1.43051e-006 -0.211811 -1.11187 -1.19209e-006 -0.541159 -1.652678 -1.90735e-006 -0.969553 -2.203329 -1.90735e-006 -1.451459 -2.757412 -2.38419e-006 -1.966574 -3.312521 -2.86102e-006 -2.49984 -3.86791 -3.33786e-006 -3.041947 -4.423388 -3.8147e-006 -3.587969 226 10 0 0 0 -0.206448 2.38419e-007 -0.0397826 -0.620522 7.15256e-007 -0.224953 -1.134552 -2.38419e-007 -0.569756 -1.678577 -4.76837e-007 -1.012709 -2.230709 -1.19209e-006 -1.506677 -2.785355 -1.43051e-006 -2.030425 -3.340654 -1.66893e-006 -2.56912 -3.89612 -1.90735e-006 -3.114616 -4.451628 -2.38419e-006 -3.662881 227 10 0 0 0 -0.206707 1.08033e-007 -0.0398871 -0.62402 -5.96046e-008 -0.228711 -1.140533 -2.38419e-007 -0.579683 -1.685585 -8.19564e-007 -1.027707 -2.238111 -1.17719e-006 -1.525307 -2.792919 -1.47521e-006 -2.052084 -3.348286 -1.92225e-006 -2.593222 -3.903788 -2.39909e-006 -3.140906 -4.459322 -2.84612e-006 -3.691339 228 10 0 0 0 -0.152345 0.152345 -0.0434769 -0.459873 0.459878 -0.253365 -0.83271 0.832713 -0.633904 -1.22119 1.22119 -1.106895 -1.612994 1.612996 -1.622227 -2.00561 2.005609 -2.159097 -2.398389 2.398391 -2.705387 -2.791213 2.791215 -3.255571 -3.184041 3.184039 -3.807256 229 10 0 0 0 -0.148759 0.148761 -0.041374 -0.447469 0.447471 -0.236116 -0.815106 0.815108 -0.595891 -1.201709 1.201707 -1.052869 -1.592915 1.592915 -1.557871 -1.985382 1.985378 -2.089198 -2.378131 2.378127 -2.632907 -2.770936 2.770936 -3.181873 -3.163774 3.163766 -3.732821 230 10 0 0 0 -0.150224 0.150225 -0.0422287 -0.45569 0.45569 -0.248462 -0.827845 0.827845 -0.62614 -1.216175 1.216171 -1.097799 -1.607963 1.60796 -1.612717 -2.000575 2.000572 -2.149648 -2.393365 2.393358 -2.696228 -2.78618 2.786173 -3.246667 -3.179008 3.179003 -3.798587 231 10 0 0 0 -0.151941 0.151945 -0.0432391 -0.458811 0.458812 -0.25194 -0.831318 0.831317 -0.631101 -1.219698 1.219699 -1.103166 -1.611478 1.611478 -1.618009 -2.004089 2.004092 -2.154689 -2.396876 2.396876 -2.700943 -2.789695 2.789695 -3.251128 -3.182523 3.182519 -3.802827 232 10 0 0 0 -0.151412 0.151415 -0.0429265 -0.458428 0.458429 -0.251891 -0.831234 0.831236 -0.632311 -1.219744 1.219744 -1.105611 -1.611566 1.611567 -1.621296 -2.004188 2.004189 -2.158461 -2.396976 2.396976 -2.704978 -2.789792 2.789789 -3.255324 -3.182624 3.182621 -3.807145 233 10 0 0 0 -0.149344 0.149343 -0.0417106 -0.453906 0.453905 -0.246371 -0.825753 0.82575 -0.622762 -1.214021 1.214018 -1.093787 -1.605792 1.605789 -1.608471 -1.998407 1.998402 -2.145378 -2.391194 2.391191 -2.692042 -2.784017 2.78401 -3.242569 -3.176849 3.17684 -3.794574 234 10 0 0 0 -0.227316 -3.57628e-007 -0.0486344 -0.674931 -4.76837e-007 -0.275127 -1.206266 -8.34465e-007 -0.668432 -1.756346 -1.43051e-006 -1.146174 -2.310425 -1.54972e-006 -1.6613 -2.865601 -1.90735e-006 -2.196308 -3.421045 -2.26498e-006 -2.740824 -3.976563 -2.74181e-006 -3.2897 -4.532095 -2.86102e-006 -3.840459 235 10 0 0 0 -0.227486 1.78814e-007 -0.0487095 -0.673471 -1.19209e-007 -0.272994 -1.204002 -6.85453e-007 -0.663682 -1.753822 -1.01328e-006 -1.139602 -2.307804 -1.40071e-006 -1.65336 -2.862951 -1.72853e-006 -2.187592 -3.418388 -2.02656e-006 -2.731678 -3.9739 -2.41399e-006 -3.280146 -4.529428 -2.68221e-006 -3.830358 236 10 0 0 0 -0.221012 -4.76837e-007 -0.0458536 -0.663237 -4.76837e-007 -0.265135 -1.193399 -1.19209e-006 -0.654642 -1.743483 -1.90735e-006 -1.132404 -2.297699 -2.14577e-006 -1.649342 -2.852934 -2.14577e-006 -2.186119 -3.408403 -2.6226e-006 -2.731951 -3.963928 -2.6226e-006 -3.281676 -4.519468 -3.09944e-006 -3.832957 237 10 0 0 0 -0.216509 2.38419e-007 -0.0439254 -0.655087 -2.38419e-007 -0.258465 -1.184305 -7.15256e-007 -0.644996 -1.734251 -7.15256e-007 -1.12179 -2.288469 -9.53674e-007 -1.638817 -2.843725 -1.66893e-006 -2.176083 -3.399202 -2.14577e-006 -2.722362 -3.954733 -2.38419e-006 -3.27244 -4.510273 -2.6226e-006 -3.824025 238 10 0 0 0 -0.223145 -1.19209e-007 -0.046783 -0.666355 -4.76837e-007 -0.267364 -1.196531 -7.15256e-007 -0.656927 -1.746548 -1.19209e-006 -1.134223 -2.300718 -1.54972e-006 -1.650592 -2.855936 -1.90735e-006 -2.186779 -3.411398 -2.26498e-006 -2.732205 -3.966921 -2.5034e-006 -3.281645 -4.522457 -2.86102e-006 -3.832654 239 10 0 0 0 -0.217649 2.08616e-007 -0.0444083 -0.65057 -2.98023e-008 -0.251795 -1.175322 -3.27826e-007 -0.624924 -1.723516 -8.04663e-007 -1.090335 -2.277157 -1.16229e-006 -1.599808 -2.832253 -1.63913e-006 -2.132754 -3.387684 -2.02656e-006 -2.67651 -3.943193 -2.38419e-006 -3.224991 -4.498724 -2.68221e-006 -3.775331 240 10 0 0 0 -0.111485 0.111487 -0.0228403 -0.342364 0.342368 -0.128918 -0.654401 0.654405 -0.346969 -1.007436 1.00744 -0.658859 -1.379472 1.379477 -1.036035 -1.762036 1.762038 -1.465372 -2.14955 2.149557 -1.92981 -2.539168 2.53917 -2.414335 -2.929519 2.929522 -2.90751 241 10 0 0 0 -0.111145 0.111153 -0.0227067 -0.341415 0.34142 -0.128154 -0.652824 0.652829 -0.345046 -1.005363 1.005367 -0.6555 -1.377039 1.377041 -1.031169 -1.759367 1.759369 -1.459081 -2.14673 2.146734 -1.922238 -2.536236 2.53624 -2.405604 -2.926504 2.926506 -2.89769 242 10 0 0 0 -0.102274 0.102274 -0.0191586 -0.316561 0.31656 -0.10909 -0.611937 0.611937 -0.298385 -0.951975 0.951973 -0.575747 -1.314604 1.314599 -0.917654 -1.690981 1.690978 -1.314093 -2.074467 2.074464 -1.749142 -2.461197 2.461195 -2.207118 -2.849218 2.849216 -2.675968 243 10 0 0 0 -0.108501 0.108509 -0.021611 -0.334005 0.334016 -0.122266 -0.64065 0.640656 -0.330568 -0.989464 0.989474 -0.630596 -1.35845 1.358459 -0.995526 -1.739042 1.739049 -1.413432 -2.125299 2.125308 -1.867758 -2.514147 2.514155 -2.344331 -2.904049 2.904059 -2.832147 244 10 0 0 0 -0.119603 0.119602 -0.0263765 -0.364226 0.364233 -0.147235 -0.688626 0.688634 -0.389473 -1.050193 1.050197 -0.727807 -1.42773 1.427737 -1.129833 -1.813512 1.813519 -1.580581 -2.202929 2.202934 -2.062932 -2.593841 2.593845 -2.563557 -2.985335 2.985342 -3.073284 245 10 0 0 0 -0.104948 0.104951 -0.0201929 -0.324091 0.324093 -0.114666 -0.62442 0.624421 -0.312106 -0.968391 0.968391 -0.599314 -1.333916 1.333918 -0.951339 -1.712248 1.712251 -1.357323 -2.097038 2.097039 -1.801017 -2.484732 2.48473 -2.266933 -2.873642 2.873646 -2.744157 246 10 0 0 0 -0.138029 -2.86102e-006 -0.0174229 -0.429049 -3.8147e-006 -0.0997472 -0.834167 -1.90735e-006 -0.27514 -1.304672 -9.53674e-007 -0.535282 -1.809631 -9.53674e-007 -0.859197 -2.33652 -9.53674e-007 -1.238595 -2.875738 -9.53674e-007 -1.660412 -3.421026 -2.86102e-006 -2.109647 -3.968788 -1.90735e-006 -2.572454 247 10 0 0 0 -0.145 2.86102e-006 -0.0192568 -0.450251 3.8147e-006 -0.110631 -0.871712 3.8147e-006 -0.304231 -1.357829 4.76837e-006 -0.590838 -1.876266 3.8147e-006 -0.946737 -2.413518 3.8147e-006 -1.360855 -2.959863 3.8147e-006 -1.815675 -3.510129 3.8147e-006 -2.294736 -4.06182 2.86102e-006 -2.784901 248 10 0 0 0 -0.139994 9.53674e-007 -0.0179279 -0.434636 3.8147e-006 -0.102495 -0.843595 3.8147e-006 -0.282025 -1.317287 1.90735e-006 -0.547312 -1.824692 9.53674e-007 -0.876627 -2.353283 9.53674e-007 -1.261204 -2.893525 0 -1.687217 -3.439495 0 -2.140032 -3.987835 -9.53674e-007 -2.606347 249 10 0 0 0 -0.145356 0 -0.0193512 -0.451286 9.53674e-007 -0.111171 -0.873564 9.53674e-007 -0.305722 -1.360352 9.53674e-007 -0.593548 -1.879263 -2.86102e-006 -0.950682 -2.416813 -1.90735e-006 -1.365944 -2.963335 -2.86102e-006 -1.821724 -3.513714 -9.53674e-007 -2.3016 -4.065495 -2.86102e-006 -2.792499 250 10 0 0 0 -0.148643 -3.8147e-006 -0.0202534 -0.459078 -3.8147e-006 -0.115079 -0.884358 -9.53674e-007 -0.313171 -1.37121 -2.86102e-006 -0.601115 -1.88843 -3.8147e-006 -0.953876 -2.423689 -1.90735e-006 -1.360641 -2.968246 -4.76837e-006 -1.806192 -3.517189 -4.76837e-006 -2.276296 -4.067934 -5.72205e-006 -2.758898 251 10 0 0 0 -0.158689 -3.8147e-006 -0.0231431 -0.48701 -2.86102e-006 -0.130538 -0.92977 -5.72205e-006 -0.350524 -1.429839 -3.8147e-006 -0.664064 -1.956359 -5.72205e-006 -1.042364 -2.497567 -5.72205e-006 -1.472476 -3.045975 -4.76837e-006 -1.939201 -3.597494 -5.72205e-006 -2.427931 -4.150221 -7.62939e-006 -2.927477 252 10 0 0 0 -0.117332 0.117326 -0.0253551 -0.359133 0.359128 -0.143072 -0.681551 0.681548 -0.381233 -1.042313 1.042308 -0.716923 -1.419966 1.419961 -1.119514 -1.806198 1.806192 -1.573608 -2.196058 2.196052 -2.060889 -2.587334 2.587328 -2.566952 -2.979118 2.97911 -3.081819 253 10 0 0 0 -0.115675 0.115664 -0.0246296 -0.354832 0.354827 -0.13945 -0.675091 0.675081 -0.373272 -1.034744 1.034734 -0.705359 -1.411816 1.411803 -1.105075 -1.797712 1.797705 -1.556712 -2.18742 2.18741 -2.042243 -2.578606 2.578595 -2.546923 -2.970322 2.970311 -3.060522 254 10 0 0 0 -0.112829 0.112819 -0.0234044 -0.346731 0.346729 -0.132619 -0.661884 0.661881 -0.35651 -1.01738 1.017373 -0.675645 -1.391632 1.391626 -1.062341 -1.77599 1.775984 -1.503104 -2.164839 2.164834 -1.979719 -2.555513 2.555507 -2.477009 -2.946859 2.946851 -2.984198 255 10 0 0 0 -0.112152 0.11215 -0.0231211 -0.344772 0.34477 -0.130994 -0.65864 0.658636 -0.352461 -1.013029 1.013022 -0.668305 -1.386452 1.386443 -1.051368 -1.770306 1.770295 -1.48879 -2.158876 2.158866 -1.962722 -2.549374 2.549363 -2.457696 -2.940584 2.940573 -2.962812 256 10 0 0 0 -0.114902 0.114905 -0.0242927 -0.351961 0.35196 -0.136847 -0.669722 0.669721 -0.365757 -1.026737 1.026734 -0.689526 -1.401375 1.401373 -1.077949 -1.785767 1.785763 -1.518903 -2.174614 2.174611 -1.995484 -2.565292 2.565286 -2.492805 -2.956627 2.956623 -2.999861 257 10 0 0 0 -0.123577 0.123571 -0.0282023 -0.375793 0.375793 -0.157833 -0.70714 0.707135 -0.414952 -1.073563 1.073565 -0.770261 -1.454552 1.454548 -1.190387 -1.842525 1.842521 -1.658792 -2.233282 2.233279 -2.157276 -2.625071 2.625066 -2.672154 -3.017208 3.017203 -3.194684 258 10 0 0 0 -0.160503 9.53674e-007 -0.0236897 -0.496763 2.86102e-006 -0.13701 -0.950502 3.8147e-006 -0.371999 -1.460363 2.86102e-006 -0.706908 -1.994404 2.86102e-006 -1.109352 -2.540592 2.86102e-006 -1.563313 -3.091909 2.86102e-006 -2.050387 -3.645185 1.90735e-006 -2.555692 -4.199125 9.53674e-007 -3.068894 259 10 0 0 0 -0.171188 0 -0.0270321 -0.52566 1.90735e-006 -0.154812 -0.995031 9.53674e-007 -0.41316 -1.514839 1.90735e-006 -0.772647 -2.054703 0 -1.197101 -2.603909 0 -1.668904 -3.156796 -9.53674e-007 -2.170066 -3.71101 -9.53674e-007 -2.687074 -4.265667 -9.53674e-007 -3.21103 260 10 0 0 0 -0.160149 9.53674e-007 -0.0235829 -0.493546 1.90735e-006 -0.13474 -0.944063 9.53674e-007 -0.365219 -1.451717 0 -0.695096 -1.984453 1.90735e-006 -1.093059 -2.529968 1.90735e-006 -1.543461 -3.080959 0 -2.027938 -3.634092 0 -2.531661 -4.187977 -1.90735e-006 -3.044135 261 10 0 0 0 -0.158493 9.53674e-007 -0.0230877 -0.49094 -9.53674e-007 -0.133537 -0.941116 -9.53674e-007 -0.363541 -1.448578 -2.86102e-006 -0.692987 -1.981163 -1.90735e-006 -1.090444 -2.526571 -9.53674e-007 -1.540303 -3.077492 -2.86102e-006 -2.024239 -3.630552 -2.86102e-006 -2.527179 -4.18434 -2.86102e-006 -3.038472 262 10 0 0 0 -0.148602 9.53674e-007 -0.0202425 -0.46177 4.76837e-006 -0.116919 -0.893395 2.86102e-006 -0.322703 -1.387744 2.86102e-006 -0.624764 -1.911963 9.53674e-007 -0.99637 -2.45267 2.86102e-006 -1.42434 -3.001012 1.90735e-006 -1.890666 -3.552514 9.53674e-007 -2.379226 -4.10513 9.53674e-007 -2.877732 263 10 0 0 0 -0.155869 -1.90735e-006 -0.0223129 -0.481076 9.53674e-007 -0.127443 -0.923365 -9.53674e-007 -0.34681 -1.425167 0 -0.663965 -1.954235 -9.53674e-007 -1.050018 -2.497734 -1.90735e-006 -1.490461 -3.047644 -9.53674e-007 -1.96701 -3.600134 -1.90735e-006 -2.46429 -4.153543 -1.90735e-006 -2.97104 266 10 0 0 0 -0.12277 0.122767 -0.0278251 -0.376728 0.376727 -0.159526 -0.711525 0.711528 -0.424461 -1.080919 1.08092 -0.790955 -1.463728 1.463725 -1.221769 -1.852613 1.852614 -1.69874 -2.243818 2.243817 -2.203676 -2.635838 2.635835 -2.723341 -3.028118 3.028114 -3.249264 268 10 0 0 0 -0.129086 0.129088 -0.0308499 -0.391994 0.391999 -0.173611 -0.733465 0.733469 -0.454513 -1.106663 1.106662 -0.836588 -1.491489 1.491491 -1.280536 -1.881325 1.881324 -1.767536 -2.272982 2.272982 -2.280143 -2.66527 2.665272 -2.80626 -3.05776 3.05776 -3.338485 269 10 0 0 0 -0.127924 0.127923 -0.0302811 -0.389887 0.38989 -0.171841 -0.73081 0.730811 -0.451374 -1.103689 1.10369 -0.832103 -1.488356 1.488356 -1.274922 -1.878119 1.878117 -1.761098 -2.269745 2.269744 -2.273107 -2.662012 2.662012 -2.79877 -3.054487 3.054486 -3.330461 270 10 0 0 0 -0.183216 0 -0.0310788 -0.558786 9.53674e-007 -0.177258 -1.044567 9.53674e-007 -0.463259 -1.573898 -4.76837e-007 -0.850147 -2.118959 -9.53674e-007 -1.298232 -2.670637 -1.43051e-006 -1.788241 -3.224697 -1.43051e-006 -2.303097 -3.779562 -2.38419e-006 -2.830922 -4.334682 -1.90735e-006 -3.364493 276 10 0 0 0 -0.144257 0.144252 -0.0388149 -0.434555 0.434545 -0.220066 -0.796247 0.79624 -0.558834 -1.180027 1.180021 -0.995787 -1.57003 1.570024 -1.484722 -1.96207 1.962061 -2.004761 -2.35466 2.354651 -2.540833 -2.747402 2.747394 -3.084385 -3.140192 3.140183 -3.631026 277 10 0 0 0 -0.147789 0.147794 -0.0408145 -0.443569 0.443573 -0.23076 -0.808506 0.808513 -0.580696 -1.193653 1.193656 -1.026845 -1.584196 1.584199 -1.522393 -1.976408 1.976411 -2.046648 -2.369064 2.369066 -2.585298 -2.761837 2.761836 -3.130508 -3.154636 3.154635 -3.678377 278 10 0 0 0 -0.142975 0.142978 -0.0381025 -0.430693 0.430694 -0.215398 -0.790485 0.790484 -0.547918 -1.173473 1.17347 -0.979857 -1.56319 1.563189 -1.465512 -1.955143 1.95514 -1.983787 -2.347715 2.347713 -2.519027 -2.740458 2.740454 -3.062248 -3.133247 3.133245 -3.608821 279 10 0 0 0 -0.141975 0.14197 -0.0375487 -0.427835 0.427833 -0.212042 -0.786587 0.786586 -0.541252 -1.169052 1.169048 -0.969979 -1.558512 1.558504 -1.45281 -1.95035 1.950345 -1.968813 -2.342878 2.342871 -2.502318 -2.735596 2.735587 -3.044157 -3.128363 3.128354 -3.5895 280 10 0 0 0 -0.139515 0.139515 -0.0362155 -0.421974 0.421978 -0.205676 -0.778345 0.77835 -0.527466 -1.159452 1.159453 -0.948255 -1.548273 1.548276 -1.424565 -1.93989 1.939892 -1.936358 -2.332344 2.332342 -2.467287 -2.725029 2.725029 -3.007682 -3.117786 3.117785 -3.552236 281 10 0 0 0 -0.131462 0.131464 -0.0320315 -0.400427 0.400425 -0.182668 -0.746597 0.746593 -0.475592 -1.122347 1.122347 -0.869099 -1.508545 1.508543 -1.322933 -1.899015 1.89901 -1.817532 -2.290968 2.290964 -2.335866 -2.683414 2.683409 -2.866518 -3.076006 3.076 -3.402633 282 10 0 0 0 -0.190571 -4.76837e-007 -0.0337102 -0.578768 -9.53674e-007 -0.191842 -1.074614 -4.76837e-007 -0.496843 -1.60948 -1.43051e-006 -0.902206 -2.157316 -9.53674e-007 -1.365465 -2.710217 -1.43051e-006 -1.866791 -3.264811 -2.38419e-006 -2.389641 -3.819939 -3.8147e-006 -2.923356 -4.375227 -3.8147e-006 -3.461745 283 10 0 0 0 -0.209385 -4.76837e-007 -0.0409694 -0.623219 4.76837e-007 -0.22587 -1.134121 4.76837e-007 -0.563203 -1.675568 0 -0.994343 -2.226156 -1.43051e-006 -1.475771 -2.780079 -1.43051e-006 -1.988703 -3.335052 -1.43051e-006 -2.518778 -3.890345 -1.43051e-006 -3.057236 -4.445736 -1.43051e-006 -3.599152 284 10 0 0 0 -0.188408 9.53674e-007 -0.032923 -0.572302 1.43051e-006 -0.186896 -1.065058 0 -0.485867 -1.598362 -4.76837e-007 -0.88578 -2.145477 -9.53674e-007 -1.344865 -2.698092 -1.90735e-006 -1.843311 -3.252563 -9.53674e-007 -2.364237 -3.807636 -1.90735e-006 -2.896617 -4.362892 -2.38419e-006 -3.433976 285 10 0 0 0 -0.218973 -2.38419e-007 -0.044975 -0.647049 -2.38419e-007 -0.246423 -1.166386 -2.38419e-007 -0.604667 -1.711544 -9.53674e-007 -1.053242 -2.263727 -1.66893e-006 -1.547678 -2.818256 -1.90735e-006 -2.069489 -3.37347 -2.38419e-006 -2.605578 -3.928883 -2.86102e-006 -3.148424 -4.484348 -3.33786e-006 -3.694012 286 10 0 0 0 -0.212181 1.43051e-006 -0.0421134 -0.633997 1.19209e-006 -0.236128 -1.151548 7.15256e-007 -0.589733 -1.696606 9.53674e-007 -1.037805 -2.248976 0 -1.533971 -2.80364 0 -2.058103 -3.358925 -4.76837e-007 -2.596348 -3.914375 -9.53674e-007 -3.141087 -4.469868 -1.43051e-006 -3.688452 287 10 0 0 0 -0.218718 -1.19209e-006 -0.0448654 -0.64694 -1.43051e-006 -0.246489 -1.166595 -1.90735e-006 -0.605567 -1.711962 -2.38419e-006 -1.055212 -2.264256 -2.6226e-006 -1.550641 -2.818832 -2.86102e-006 -2.073219 -3.374067 -3.33786e-006 -2.609876 -3.929489 -3.8147e-006 -3.153164 -4.484962 -3.8147e-006 -3.699137 288 10 0 0 0 -0.163929 0.163928 -0.0506806 -0.480736 0.480736 -0.277742 -0.855049 0.855045 -0.664692 -1.24309 1.243086 -1.133723 -1.634495 1.634491 -1.641823 -2.026926 2.026918 -2.172169 -2.419638 2.41963 -2.713643 -2.81242 2.812412 -3.260218 -3.205235 3.205227 -3.809035 289 10 0 0 0 -0.157526 0.157534 -0.0466268 -0.46693 0.466937 -0.259866 -0.838903 0.838912 -0.636791 -1.226624 1.226631 -1.10292 -1.618019 1.618026 -1.610943 -2.010473 2.010484 -2.141945 -2.403198 2.403205 -2.684171 -2.795992 2.795995 -3.231448 -3.188807 3.188813 -3.780728 290 10 0 0 0 -0.163494 0.163495 -0.0503998 -0.480358 0.480358 -0.277572 -0.855194 0.855194 -0.666884 -1.243473 1.243473 -1.138047 -1.63497 1.634966 -1.647733 -2.027437 2.027433 -2.17914 -2.420156 2.420153 -2.721319 -2.812948 2.812944 -3.268403 -3.205759 3.205755 -3.817599 291 10 0 0 0 -0.155045 0.155051 -0.0451018 -0.455671 0.455683 -0.243047 -0.820759 0.820769 -0.593514 -1.205027 1.205034 -1.033663 -1.594864 1.594875 -1.520737 -1.986719 1.986727 -2.036928 -2.379198 2.379206 -2.568871 -2.771875 2.771883 -3.1085 -3.164614 3.164621 -3.651636 292 10 0 0 0 -0.155649 0.155642 -0.0454646 -0.457693 0.457681 -0.245803 -0.823807 0.823796 -0.599936 -1.208582 1.208574 -1.043542 -1.598663 1.598655 -1.533449 -1.99062 1.990612 -2.051735 -2.383144 2.383133 -2.585196 -2.775845 2.775833 -3.125988 -3.168594 3.168583 -3.669883 293 10 0 0 0 -0.165531 0.165523 -0.0517235 -0.483429 0.483425 -0.280911 -0.857424 0.857416 -0.666466 -1.245283 1.245272 -1.13384 -1.636597 1.636589 -1.640538 -2.028996 2.028988 -2.169852 -2.421692 2.421684 -2.710592 -2.81447 2.814466 -3.256768 -3.207283 3.207272 -3.805366 294 10 0 0 0 -0.221581 -1.19209e-006 -0.0461003 -0.64941 -9.53674e-007 -0.247249 -1.167046 -9.53674e-007 -0.601075 -1.711004 -1.66893e-006 -1.043715 -2.262533 -1.66893e-006 -1.532514 -2.816771 -2.38419e-006 -2.049824 -3.371857 -2.6226e-006 -2.582499 -3.927198 -2.86102e-006 -3.122696 -4.482628 -3.33786e-006 -3.666404 295 10 0 0 0 -0.224571 -1.19209e-006 -0.0474123 -0.655504 -7.15256e-007 -0.252342 -1.17445 -1.19209e-006 -0.609563 -1.718832 -1.66893e-006 -1.054273 -2.270502 -2.14577e-006 -1.544212 -2.824778 -2.6226e-006 -2.062088 -3.37989 -2.6226e-006 -2.595476 -3.935263 -3.09944e-006 -3.136708 -4.490711 -3.33786e-006 -3.681442 296 10 0 0 0 -0.229076 -3.57628e-007 -0.0494281 -0.673954 -2.98023e-007 -0.27223 -1.202101 -7.7486e-007 -0.655434 -1.750599 -1.07288e-006 -1.122719 -2.304058 -1.19209e-006 -1.63003 -2.859053 -1.54972e-006 -2.160565 -3.414448 -1.84774e-006 -2.702805 -3.969944 -2.20537e-006 -3.250443 -4.525471 -2.32458e-006 -3.800337 297 10 0 0 0 -0.229761 -1.49012e-007 -0.0497376 -0.675222 -3.8743e-007 -0.273319 -1.203547 -6.85453e-007 -0.657074 -1.752087 -9.83477e-007 -1.124617 -2.305544 -1.40071e-006 -1.631914 -2.860531 -1.63913e-006 -2.162353 -3.415926 -1.99676e-006 -2.704471 -3.971422 -2.20537e-006 -3.252021 -4.526947 -2.47359e-006 -3.801846 298 10 0 0 0 -0.229036 -1.19209e-007 -0.0494071 -0.673172 -4.47035e-007 -0.271217 -1.200762 -8.34465e-007 -0.652729 -1.749012 -1.13249e-006 -1.118491 -2.302374 -1.49012e-006 -1.624727 -2.857338 -1.66893e-006 -2.154589 -3.412724 -1.96695e-006 -2.696432 -3.96822 -2.26498e-006 -3.243795 -4.523743 -2.5034e-006 -3.793471 299 10 0 0 0 -0.225208 -7.15256e-007 -0.047693 -0.656811 -4.76837e-007 -0.253448 -1.176136 -4.76837e-007 -0.611661 -1.720705 -1.43051e-006 -1.057256 -2.272455 -1.90735e-006 -1.547917 -2.826767 -2.38419e-006 -2.066323 -3.381886 -2.38419e-006 -2.599786 -3.937254 -2.6226e-006 -3.140999 -4.492702 -2.86102e-006 -3.685713' )
		self.hairShape.boundingBox()
		self.editGuidesShape.boundingBox()

	def test_CommandPerformance( self ):
		self.assertPerformance( 'MayaTest.Self.SetupPerformance()', 'MayaTest.Self.CommandPerformance()' )

	def SetupBrushPerformance( self, addMfs = False ):
		self.SetupPerformance()

		changeWidth = pm.PyNode( pm.mel.OxAddStrandOperator( self.hairShape, TestUtilities.ChangeWidthNodeName ) )
		changeWidth.width.set( 0.04 )

		if addMfs:
			meshFromStrands = pm.PyNode( pm.mel.OxAddStrandOperator( self.hairShape, TestUtilities.MeshFromStrandsNodeName ) )

		pm.select( self.editGuidesShape )

		try:
			pm.mel.eval( "OxCombBrushTool oxCombBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxCombBrushTool1" )

		viewportName = 'modelPanel4'
		self.viewportWidth = pm.control(viewportName, query=True, width=True)
		self.viewportHeight = pm.control(viewportName, query=True, height=True)

		# Set move brush
		pm.mel.eval( "OxCombBrushTool -e -s " + str( self.viewportHeight / 2 ) + " oxCombBrushTool1" )

	def BrushPerformance( self ):
		for i in range(0, 5):
			pm.mel.OxDragBrush( 10 + i * ( self.viewportWidth - 20 ) / 5, self.viewportHeight / 2.0, 10 + ( i + 1 ) * ( self.viewportWidth - 20 ) / 5, self.viewportHeight / 2.0, 4 )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_BrushPerformance( self ):
		self.assertPerformance( 'MayaTest.Self.SetupBrushPerformance()', 'MayaTest.Self.BrushPerformance()' )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_BrushPerformanceMesh( self ):
		self.assertPerformance( 'MayaTest.Self.SetupBrushPerformance( True )', 'MayaTest.Self.BrushPerformance()' )

	def test_CopyGuideShapeOnPlanting( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )

		# Spread the guides apart using a surface comb node
		guidesFromMeshNode = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		surfaceCombNode = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromMeshNode, TestUtilities.SurfaceCombNodeName ) )
		surfaceCombNode.algorithm.set( 0 )

		# Select one of the guides and plant a new guide
		pm.select( editGuidesShape + ".ep[2]" );
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0.08669 0.2385;" )

		# The vector formed by planted guide in object space should be the same as its source guide
		vertices = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		sourceGuideVector = dt.Vector( vertices[2 * 2 + 1] ) - dt.Vector( vertices[2 * 2] )
		newGuideVector = dt.Vector( vertices[4 * 2 + 1] ) - dt.Vector( vertices[4 * 2] )
		self.assertAlmostEqual( sourceGuideVector[0], newGuideVector[0], places = 4 )
		self.assertAlmostEqual( sourceGuideVector[1], newGuideVector[1], places = 4 )
		self.assertAlmostEqual( sourceGuideVector[2], newGuideVector[2], places = 4 )

	def test_CopiedGuideShapePreservesRootData( self ):
		# Create some guides, add a root channel, and set some values to it
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )
		pm.select( editGuidesShape )
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -crc 1 Groups;" )
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -rcd 2 4 0 0 0 0 1 2 1 2;" )

		strandIds = pm.mel.OxGetStrandIds( editGuidesShape )

		expectedValuesByStrandIds = TestUtilities.MakeDictionary( [0, 1, 2, 3], [1, 2, 1, 2] )
		actualValuesByStrandId = TestUtilities.MakeDictionary( strandIds, pm.mel.OxGetRootValuesForChannel( editGuidesShape, 1 ) )

		TestUtilities.AssertEqualByKeys( self, expectedValuesByStrandIds, actualValuesByStrandId )

		# Select the last guide and plant a new guide, which should copy its settings
		pm.select( editGuidesShape + ".ep[3]" );
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0.08669 0.2385;" )

		# Check that the newly planted guide copied the root channel data from last root
		self.assertEqual( actualValuesByStrandId[strandIds[3]], pm.mel.OxGetRootValuesForChannel( editGuidesShape, 1 )[4] )

	def test_CopiedGuidePreservesStrandGroup( self ):
		# Create some guides, add a root channel, and set some values to it
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )
		pm.select( editGuidesShape )

		# Assign group 5 to last guide
		editGuidesShape.useStrandGroups.set( True )
		pm.select( editGuidesShape + ".ep[3]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "5" )

		# Select the last guide and plant a new guide, which should copy its settings
		pm.select( editGuidesShape + ".ep[3]" );
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0.08669 0.2385;" )

		# The group of the last guide should be 5
		pm.mel.OxSelectStrandGroup( editGuidesShape, 5 )
		self.assertEqual( [0, 0, 0, 1, 1], pm.mel.OxGetSelectedStrands( editGuidesShape ) )

	# There should be no problems planting two guides in the same spot and later editing them
	def test_PlantTwoGuidesInSameSpot( self ):
		# Plant two guides in the exact same location
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )
		pm.select( editGuidesShape )
		plantCommand = "OxPlantGuideDirectly " + editGuidesShape + " 0 0.08669 0.2385;"
		for i in range( 0, 2 ):
			pm.mel.eval( plantCommand )
		pm.mel.OxEditGuides( cc = True )

		# Select and change tip position of just the last guide
		pm.select( editGuidesShape + ".ep[4]" );
		pm.move( 0.5, 0, 0.5, relative = True )
		pm.mel.OxEditGuides( cc = True )
		guide6tipPosition = pm.mel.OxGetStrandPoint( editGuidesShape, 4, 1 )

		# Change first planeted guide's tip position
		pm.select( editGuidesShape + ".ep[5]" );
		pm.move( -0.5, 0, -0.5, relative = True )
		pm.mel.OxEditGuides( cc = True )

		# The position of last guide's tip should remain as it was
		self.assertEqual( guide6tipPosition, pm.mel.OxGetStrandPoint( editGuidesShape, 4, 1 ) )

	def test_PlantNewHair( self ):
		# Create some guides, add a root channel, and set some values to it
		hairShape = TestUtilities.AddHairToNewPlane()
		editGuidesShape = pm.PyNode( pm.mel.OxAddStrandOperator( hairShape, TestUtilities.EditGuidesShapeName ) )
		pm.select( editGuidesShape )

		boundingBoxBeforePlanting = TestUtilities.GetBoundingBox( TestUtilities.GetVerticesInObjectCoordinates( hairShape ) )

		# Select the first guide and plant a new guide, which should copy its settings
		pm.select( editGuidesShape + ".ep[0]" );
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0.08669 0.2385;" )

		# Check that the newly planted hair is correctly positioned
		boundingBoxAfterPlanting = TestUtilities.GetBoundingBox( TestUtilities.GetVerticesInObjectCoordinates( hairShape ) )

		TestUtilities.CheckPointsAllNearEqual( self, boundingBoxBeforePlanting, boundingBoxAfterPlanting )

		# Repeat one more time
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0.08669 0.2385;" )

		# Check that the newly planted hair is correctly positioned
		boundingBoxAfterPlantingTwice = TestUtilities.GetBoundingBox( TestUtilities.GetVerticesInObjectCoordinates( hairShape ) )

		TestUtilities.CheckPointsAllNearEqual( self, boundingBoxBeforePlanting, boundingBoxAfterPlantingTwice )

	# TODO: I cannot get this test to work. Need to make a special OxCommitEditGuides command which is not undoable.
	#def test_UndoRootMoveDoesNotRotateStrand( self ):
	#	# Create EditGuides and move the tips of all 4 strands
	#	editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )

	#	pm.select( editGuidesShape + ".ep[0:3]" );
	#	pm.move( 0.5, 0, 0.5, relative = True )
	#	pm.mel.OxEditGuides( cc = True )

	#	# Move the first root such that it ends up on a different face
	#	pm.select( editGuidesShape + ".f[0]" )
	#	pm.move( 0.82, 0, -0.66, relative = True )
	#	pm.mel.OxEditGuides( cc = True )

	#	verticesBeforeUndo = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

	#	# Undo the move
	#	pm.undo()
	#	pm.undo()
	#	pm.mel.OxEditGuides( cc = True )

	#	verticesAfterUndo = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

	#	# The object-space direction vector of the first guide should not change
	#	TestUtilities.CheckPointsNearEqual( self, dt.Vector( verticesBeforeUndo[1] ) - dt.Vector( verticesBeforeUndo[0] ), dt.Vector( verticesAfterUndo[1] ) - dt.Vector( verticesAfterUndo[0] ) )

	def test_ReloadMayaAsciiWithMovedRootAndShapeChanges( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )
		pm.select( editGuidesShape )

		# Select and move all roots
		pm.select( editGuidesShape + ".f[0:3]" )
		pm.move( 0.5, 0, 0.5, relative = True )
		pm.mel.OxEditGuides( cc = True )

		# Move guide's tip as well
		pm.select( editGuidesShape + ".ep[0:2]" );
		pm.move( 0.5, 0, 0.5, relative = True )
		pm.mel.OxEditGuides( cc = True )

		verticesBeforeSave = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# Save and reload the scene into an Ascii file
		fileName = 'temp.ma'
		filePath = pm.saveAs( fileName )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		verticesAfterLoad = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeSave, verticesAfterLoad )

	def test_ReferenceSceneWithPlantedGuides( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( rootGenerationMethod = 6, pointsPerStrandCount = 2, guideLength = 1, lengthRandomness = 0 )
		pm.select( editGuidesShape )

		## Move all three guides using control points
		#pm.select( editGuidesShape + ".ep[0]" );
		#pm.move( 0.5, 0, 0.5, relative = True )
		#pm.mel.OxEditGuides( cc = True )

		# Plant a couple of guides
		pm.select( editGuidesShape + ".ep[0]" );
		pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 0 0.250322878 0.209610820;" )
		#pm.mel.eval( "OxPlantGuideDirectly " + editGuidesShape + " 1 0.7 0.3 0;" )

		# Move all three guides using control points
		pm.select( editGuidesShape + ".ep[0:1]" );
		pm.move( 0.5, 0, 0.5, relative = True )
		pm.mel.OxEditGuides( cc = True )

		## Select and move all roots
		#pm.select( editGuidesShape + ".f[0:2]" )
		#pm.move( 0, 0, 0.5, relative = True )
		#pm.mel.OxEditGuides( cc = True )

		# Assign different strand groups to guides
		editGuidesShape.useStrandGroups.set( True )
		pm.select( editGuidesShape + ".ep[0]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "1" )
		pm.select( editGuidesShape + ".ep[1]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "2" )

		vertexPositionsBeforeReferencing = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		# Save referenced file
		referencedFilePath = pm.saveAs( 'referencedFile.mb' )

		#print( referencedFilePath + "\n" )
		# Create a new scene and reference the file we just saved
		pm.newFile()
		pm.createReference( referencedFilePath )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]

		# Change the strand groups of the guides and save the scene
		pm.select( editGuidesShape + ".ep[0]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "3" )
		pm.select( editGuidesShape + ".ep[1]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "4" )

		# Reload the scene, the positions of strands should remain the same
		referecingFilePath = pm.saveAs( "temp.mb" )
		pm.newFile()
		pm.openFile( referecingFilePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		vertexPositionsAfterReloading = TestUtilities.GetVerticesInObjectCoordinates( editGuidesShape )

		TestUtilities.CheckPointsAllNearEqual( self, vertexPositionsBeforeReferencing, vertexPositionsAfterReloading )

	# When using multi-threaded strand generation the roots can be generated in a different order each time. This test makes sure that strand groups
	# assigned to these roots are always consistent with the strand ids.
	def test_MultiThreadedStrandGenerationAndStrandGroupAssignment( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( pointsPerStrandCount = 2, rootGenerationMethod = 1, guideCount = 30 )

		# Create and assign some strand groups
		pm.mel.eval( TestUtilities.EditGuidesCommandName  + " -crc 1 Groups;" )

		strandCount = pm.mel.OxGetStrandCount( editGuidesShape )
		initialStrandIds = pm.mel.OxGetStrandIds( editGuidesShape )

		# Construct and execute a command which will set unique root indices for each strand
		editRootDataCommand = TestUtilities.EditGuidesCommandName + " -sd " + str( strandCount );
		for i in range( 0, strandCount ):
			# Changing strand id at i, for 1 channel, second one
			editRootDataCommand += " " + str( initialStrandIds[i] ) + " 1 1 " + str( i )

		editRootDataCommand += ";"
		pm.mel.eval( editRootDataCommand )

		#print( editRootDataCommand )
		initialStrandGroups = pm.mel.OxGetRootValuesForChannel( editGuidesShape, 1 )

		# Save and reload the scene
		fileName = 'temp.ma'
		filePath = pm.saveAs( fileName )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		reloadedStrandGroups = pm.mel.OxGetRootValuesForChannel( editGuidesShape, 1 )
		reloadedStrandIds = pm.mel.OxGetStrandIds( editGuidesShape )

		errorMessage = "Initial ids: " + str( initialStrandIds ) + "\nInitial groups: " + str( initialStrandGroups ) + "\nReloaded ids: " + str( reloadedStrandIds ) + "\nReloadedgroups:" + str( reloadedStrandGroups )
		#print( errorMessage )

		# Make sure that the groups still match the strands
		for strandIndex in range( 0, len( initialStrandGroups ) ):
			initialStrandGroup = initialStrandGroups[strandIndex]
			initialStrandId = initialStrandIds[strandIndex]
			for reloadedStrandIndex in range( 0, len( reloadedStrandGroups ) ):
				reloadedStrandGroup = reloadedStrandGroups[reloadedStrandIndex]
				reloadedStrandId = reloadedStrandIds[reloadedStrandIndex]

				if initialStrandId == reloadedStrandId:
					# For the strand with the same id the strand groups must also match
					self.assertEqual( initialStrandGroup, reloadedStrandGroup, errorMessage )
					break

		# Repeat test saving as a binary file
		fileName = 'temp.mb'
		filePath = pm.saveAs( fileName )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		reloadedStrandGroups = pm.mel.OxGetRootValuesForChannel( editGuidesShape, 1 )
		reloadedStrandIds = pm.mel.OxGetStrandIds( editGuidesShape )

		# Make sure that the groups still match the strands
		for strandIndex in range( 0, len( initialStrandGroups ) ):
			initialStrandGroup = initialStrandGroups[strandIndex]
			initialStrandId = initialStrandIds[strandIndex]
			for reloadedStrandIndex in range( 0, len( reloadedStrandGroups ) ):
				reloadedStrandGroup = reloadedStrandGroups[reloadedStrandIndex]
				reloadedStrandId = reloadedStrandIds[reloadedStrandIndex]

				if initialStrandId == reloadedStrandId:
					# For the strand with the same id the strand groups must also match
					self.assertEqual( initialStrandGroup, reloadedStrandGroup )
					break

	@unittest.skip( "No idea how to fix this yet. Crashes in Maya 2018.2+" ) # If( MayaTest.InBatchMode, "Requires GUI" )
	def test_SwitchToMoveInComponentAxis( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()

		pm.select( "EditGuides1.ep[0:*]", replace = True )
		from pymel.core.context import manipMoveContext
		componentMoveContext = manipMoveContext( mode = 10 )
		pm.setToolTo( componentMoveContext )

	def test_SelectInverse( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		pm.select( editGuidesShape )

		# Select first two roots
		pm.select( editGuidesShape + ".ep[0:1]" )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [1, 1, 0, 0] )

		# Invert selection and test
		# InvertSelection only works in GUI mode
		# pr.InvertSelection();
		pm.select( editGuidesShape + ".ep[*]", toggle = True )

		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 0, 1, 1] )

	def test_SelectFrozenRoots( self ):
		hairShape = TestUtilities.AddHairToNewPlane()
		editGuidesShape = pm.PyNode( pm.mel.OxAddStrandOperator( hairShape, TestUtilities.EditGuidesShapeName ) )
		pm.select( editGuidesShape + ".f[0:1]" )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [1, 1, 0, 0] )

		# Freeze the selected roots.
		pm.mel.OxSetSelectedGuidesVisibility( freeze = True )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 0, 0, 0] )

		# Try to select the frozen roots.
		pm.select( editGuidesShape + ".f[0:1]" )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 0, 0, 0] )

		# Select unfrozen roots and invert the selection.
		pm.select( editGuidesShape + ".f[2:3]" )
		pm.select( editGuidesShape + ".f[*]", tgl = True )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 0, 0, 0] )

	def test_SelectFrozenGuides( self ):
		hairShape = TestUtilities.AddHairToNewPlane()
		editGuidesShape = pm.PyNode( pm.mel.OxAddStrandOperator( hairShape, TestUtilities.EditGuidesShapeName ) )
		pm.selectType( allComponents = False, editPoint = True )
		pm.select( editGuidesShape + ".ep[0:1]" )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [1, 1, 0, 0] )

		# Freeze the selected guides.
		pm.mel.OxSetSelectedGuidesVisibility( freeze = True )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 0, 0, 0] )

		# Select all strands. Only the last two strands should be selected.
		pm.select( editGuidesShape + ".f[*]" )
		selectedStrands = pm.mel.OxGetSelectedStrands( editGuidesShape )
		self.assertEqual( selectedStrands, [0, 0, 1, 1] )

	def test_DeleteGuidesPreservesStrandGroups( self ):
		# Create guides and assign the last guide a unique strand group
		editGuidesShape1 = TestUtilities.AddEditGuidesToNewPlane()
		editGuidesShape1.useStrandGroups.set( True )
		pm.select( editGuidesShape1 + ".ep[3]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape1, "1" )

		# Add another edit guides and delete a guide
		editGuidesShape2 = pm.PyNode( pm.mel.OxAddStrandOperator( editGuidesShape1, TestUtilities.EditGuidesShapeName ) )
		editGuidesShape2.changeTracking.set( 1 )
		pm.select( editGuidesShape2 + ".ep[2]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape2 )

		# Ensure that last guide still has the correct strand group
		self.assertEqual( [0, 0, 1], pm.mel.OxGetStrandGroups( editGuidesShape2 ) )

	def test_DeleteHair( self ):
		hairShape = TestUtilities.AddHairToNewPlane()
		editGuidesShape = pm.PyNode( pm.mel.OxAddStrandOperator( hairShape, TestUtilities.EditGuidesShapeName ) )
		editGuidesShape.changeTracking.set( 1 )
		pm.select( editGuidesShape )

		# Select and delete the last two hair strands
		pm.select( editGuidesShape + ".ep[2:3]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape )

		# We should have 2 hairs now
		self.assertEqual( 2, pm.mel.OxGetStrandCount( hairShape ) )

	def test_DeleteAllStrands( self ):
		hairShape = TestUtilities.AddHairToNewPlane()
		pm.PyNode( pm.mel.OxAddStrandOperator( hairShape, TestUtilities.GroundStrandsNodeName ) )
		editGuidesShape = pm.PyNode( pm.mel.OxAddStrandOperator( hairShape, TestUtilities.EditGuidesShapeName ) )
		editGuidesShape.changeTracking.set( 1 )
		pm.select( editGuidesShape )

		# Select and delete all the hair strands.
		pm.select( editGuidesShape + ".f[0:3]" )
		pm.mel.OxDeleteSelectedGuides( editGuidesShape )

		# We should have no hairs now, and no crashes should occur when selecting.
		self.assertEqual( 0, pm.mel.OxGetStrandCount( hairShape ) )
		pm.select( editGuidesShape )

	def test_RetargetChangesPreservesStrandGroups( self ):
		editGuidesShape1 = TestUtilities.AddEditGuidesToNewPlane( rootGenerationMethod = 6 )
		editGuidesShape1.useStrandGroups.set( True )
		pm.select( editGuidesShape1 + ".ep[0]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape1, "1" )

		# Need to evaluate once
		self.assertEqual( [1], pm.mel.OxGetStrandGroups( editGuidesShape1 ) )

		plane = pm.ls( type = "polyPlane" )[0]
		plane.subdivisionsWidth.set( 2 )

		#guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		pm.mel.OxRetargetChanges( editGuidesShape1 )
		self.assertEqual( [1, 1], pm.mel.OxGetStrandGroups( editGuidesShape1 ) )

	def test_RetargetChangesAfterSceneReload( self ):
		editGuidesShape1 = TestUtilities.AddEditGuidesToNewPlane( rootGenerationMethod = 6 )
		editGuidesShape1.useStrandGroups.set( True )
		pm.select( editGuidesShape1 + ".ep[0]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape1, "1" )

		self.assertEqual( [1], pm.mel.OxGetStrandGroups( editGuidesShape1 ) )

		# Save and reload the scene
		filePath = pm.saveAs( 'temp.mb' )
		pm.openFile( filePath, force = True )

		editGuidesShape1 = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		plane = pm.ls( type = "polyPlane" )[0]

		# Evaluate once
		self.assertEqual( [1], pm.mel.OxGetStrandGroups( editGuidesShape1 ) )
		plane.subdivisionsWidth.set( 2 )

		pm.mel.OxRetargetChanges( editGuidesShape1 )
		self.assertEqual( [1, 1], pm.mel.OxGetStrandGroups( editGuidesShape1 ) )

	def test_ControlPointSoftSelection( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()
		tipPositionsBeforeMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )
		pm.setAttr( editGuidesShape + ".useSoftSelection", True )
		pm.setAttr( editGuidesShape + ".softSelectionDistance", 10.0 )
		pm.select( editGuidesShape + ".ep[3]" )

		# Move the selected guide control point
		pm.move( 0, 0, 5, relative = True )
		pm.mel.OxEditGuides( cc = True )

		tipPositionsAfterMove = TestUtilities.GetTipPositions( editGuidesShape, useObjectCoordinates = True )

		# All tips should have moved due to soft selection
		for i in range( len( tipPositionsBeforeMove ) ):
			self.assertGreaterEqual( dt.Vector( tipPositionsBeforeMove[i] ).distanceTo( tipPositionsAfterMove[i] ), 3,
			'Move: i: %(i)d, before: %(before)s, after: %(after)s' % { 'i': i, 'before': tipPositionsBeforeMove[i], 'after' : tipPositionsAfterMove[i] } )

	@staticmethod
	def MoveAllControlPoints( editGuidesShape, vector ):
		pm.select( editGuidesShape + ".ep[0:" + str( pm.mel.OxGetStrandCount( editGuidesShape ) - 1 ) + "]" )
		pm.move( vector[0], vector[1], vector[2], relative = True )
		pm.mel.OxEditGuides( cc = True )

	# Tests that planting guides with precise surface coordinates and shape edits (which hash said surface coordinates) will load the scene correctly when saved as ascii
	def test_ReloadAsciiSceneWithPlantedGuidesAndShapeChanges( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane()

		# Plant a new guide. Note that we are using precise coordinates which can be rounded when being hashed to generate the guide id. So they need to be stored precisely in ASCII file.
		TestUtilities.CreateStrand( 0, [0.0213666093, 0.774024785], 10, 5 )
		self.MoveAllControlPoints( editGuidesShape, [5, 0, 10] )

		verticesBeforeReload = pm.mel.OxGetVertices( editGuidesShape, os = True )

		# Save and reload the scene
		filePath = pm.saveAs( self.addTempFile( 'test_ReloadAsciiSceneWithPlantedGuidesAndShapeChanges.ma' ) )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]
		verticesAfterReload = pm.mel.OxGetVertices( editGuidesShape, os = True )

		TestUtilities.CheckPointsAllNearEqual( self, verticesBeforeReload, verticesAfterReload )

	@unittest.skipIf( MayaTest.ApiVersion < 20190200, "The test .mb file was saved in Maya 2019 and won't load in earlier versions" )
	def test_LoadV2SceneWithEditGuidesCreationAndEditing( self ):
		testScenePath = self.findTestFile( 'Ox2-Maya2019-CreateBrushEG-Windows.mb' )
		pm.openFile( testScenePath, force = True )
		hair = pm.ls( type = TestUtilities.HairShapeName )[0]

		expectedVertices = [-13.071115, 0, 12.514814, -14.35962, 13.249234, 5.601243, -21.404285, 24.655617, -1.126844, 17.881746, 0, 15.534288, 18.971281, 14.517077, 11.919377, 21.937824, 26.568584, 3.495513, 8.561945, 0, -9.118245, 10.226149, 14.226541, -13.571969, 14.501925, 23.869396, -24.236551, -9.360306, 0, -9.727772, -11.02355, 13.667352, -15.680714, -17.919374, 25.295593, -22.179287]
		actualVertices = pm.mel.OxGetVertices( hair, os = True )

		TestUtilities.CheckPointsAllNearEqual( self, expectedVertices, actualVertices )

	@unittest.skipIf( MayaTest.ApiVersion < 20190200, "The test .mb file was saved in Maya 2019 and won't load in earlier versions" )
	def test_LoadV2SceneWithEditGuidesCreationEditingMovingAndDeleting( self ):
		testScenePath = self.findTestFile( 'Ox2-Maya2019-CreateBrushDeleteEG-Windows.mb' )
		pm.openFile( testScenePath, force = True )
		hair = pm.ls( type = TestUtilities.HairShapeName )[0]

		expectedVertices = [-18.606062, 0, 19.457792, -19.894567, 13.249234, 12.544221, -26.939232, 24.655617, 5.816133, 19.453434, 0, 4.422605, 20.542969, 14.517077, 0.807693, 23.509512, 26.568584, -7.616171, -5.311346, 0, -19.191042, -3.647142, 14.226541, -23.644766, 0.628636, 23.869396, -34.309349]
		actualVertices = pm.mel.OxGetVertices( hair, os = True )

		TestUtilities.CheckPointsAllNearEqual( self, expectedVertices, actualVertices )

	def test_AssignStrandGroupToPlantedGuides( self ):
		# Add EG to a plane and make sure we have 0 guides to start with
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( rootGenerationMethod = 2, guideCount = 0 )
		TestUtilities.CreateStrand( 0, [0.0213666093, 0.774024785], 1, 2 )
		TestUtilities.CreateStrand( 0, [0.0213666093, 0], 1, 2 )
		TestUtilities.CreateStrand( 0, [0.1, 0.1], 1, 2 )
		TestUtilities.CreateStrand( 0, [0.5, 0.5], 1, 2 )

		editGuidesShape.useStrandGroups.set( True )
		pm.select( editGuidesShape )
		pm.select( editGuidesShape + ".f[0]" )
		pm.mel.OxAssignStrandGroup( editGuidesShape, "1" )

		self.assertEqual( 4, pm.mel.OxGetStrandCount( editGuidesShape ) )

	def test_InterpolationOptionsPersistence( self ):
		editGuidesShape = TestUtilities.AddEditGuidesToNewPlane( rootGenerationMethod = 2, guideCount = 0 )

		# Check for default settings
		self.assertEqual( 0, pm.getAttr( editGuidesShape + ".interpolation" ) )
		self.assertEqual( 0, pm.getAttr( editGuidesShape + ".inputRootInterpolation" ) )

		# Set non-default values
		pm.setAttr( editGuidesShape + ".interpolation", 1 )
		pm.setAttr( editGuidesShape + ".inputRootInterpolation", 1 )

		# Save and reload
		filePath = pm.saveAs( self.addTempFile( 'test_InterpolationOptionsPersistence.mb' ) )
		pm.openFile( filePath, force = True )

		editGuidesShape = pm.ls( type = TestUtilities.EditGuidesShapeName )[0]

		# Make sure non-default values are persisted
		self.assertEqual( 1, pm.getAttr( editGuidesShape + ".interpolation" ) )
		self.assertEqual( 1, pm.getAttr( editGuidesShape + ".inputRootInterpolation" ) )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_CombBrushWithVertexColors( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		self.SetupBrushTest( editGuidesShape )

		editGuidesShape.applyChannel.set( 2001 )

		viewportName = 'modelPanel4'
		viewportWidth = pm.control(viewportName, query=True, width=True)
		viewportHeight = pm.control(viewportName, query=True, height=True)

		# Set move brush
		try:
			pm.mel.eval( "OxCombBrushTool oxCombBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxCombBrushTool1" )
		pm.mel.eval( "OxCombBrushTool -e -s " + str( viewportHeight / 2 ) + " oxCombBrushTool1" )

		# Get the positions of the four tip vertices before moving them with a brush
		strandTipsBeforeBrushing = [pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 1, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 2, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 3, 1 )]

		# Begin brushing in the center of the view in upper corner and drag brush horizontally to the right
		pm.mel.OxDragBrush( viewportWidth / 2.0, viewportHeight / 2.0, viewportWidth * ( 2.0 / 3.0 ), viewportHeight / 2.0, 3 )
		strandTipsAfterBrushing = [pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 1, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 2, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 3, 1 )]

		# Make sure that the strand tips have indeed moved due to brushing
		epsilon = 0.01
		for i in range(0, 3):
			distanceDelta = dt.Vector( strandTipsBeforeBrushing[i] ).distanceTo( dt.Vector( strandTipsAfterBrushing[i] ) )
			self.assertGreater( distanceDelta, epsilon )

		# Now repeat the test but make sure the tips do not move if the vertices are black
		blackFaces = meshShape.f[:]
		coloredVertices = TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces)

		strandTipsBeforeBrushing = [pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 1, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 2, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 3, 1 )]
		pm.mel.OxDragBrush( viewportWidth / 2.0, viewportHeight / 2.0, viewportWidth * ( 2.0 / 3.0 ), viewportHeight / 2.0, 3 )
		strandTipsAfterBrushing = [pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 1, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 2, 1 ), pm.mel.OxGetStrandPointInObjectCoordinates( editGuidesShape, 3, 1 )]

		# Make sure that the strand tips have indeed moved due to brushing
		epsilon = 0.01
		for i in range(0, 3):
			distanceDelta = dt.Vector( strandTipsBeforeBrushing[i] ).distanceTo( dt.Vector( strandTipsAfterBrushing[i] ) )
			self.assertLess( distanceDelta, epsilon )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_CreationBrushWithVertexColors( self ):
		plane = pm.polyPlane( sx = 5, sy = 5, width = 10, height = 10 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		self.SetupBrushTest( editGuidesShape )

		editGuidesShape.applyChannel.set( 2001 )

		viewportName = 'modelPanel4'
		viewportWidth = pm.control(viewportName, query=True, width=True)
		viewportHeight = pm.control(viewportName, query=True, height=True)

		# Set move brush
		try:
			pm.mel.eval( "OxCreateBrushTool oxCreateBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxCreateBrushTool1" )
		pm.mel.eval( "OxCreateBrushTool -e -s " + str( viewportHeight / 2.0 ) + " oxCreateBrushTool1" )

		# Get the number of guides before creation 
		guideCountBeforeCreation = pm.mel.OxGetStrandCount( editGuidesShape )

		# Begin brushing in the center of the view in upper corner and drag brush horizontally to the right
		pm.mel.OxDragBrush( viewportWidth / 4.0, viewportHeight * 0.2, viewportWidth * ( 7.0 / 8.0 ), viewportHeight * 0.2, 30 )
		guideCountAfterCreation = pm.mel.OxGetStrandCount( editGuidesShape )

		# Make sure that the guides were deleted
		self.assertGreater( guideCountAfterCreation, guideCountBeforeCreation )

		# Now repeat the test but make sure the guides are not created if the faces are black
		blackFaces = meshShape.f[:]
		coloredVertices = TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces)

		guideCountBeforeCreation = pm.mel.OxGetStrandCount( editGuidesShape )
		pm.mel.OxDragBrush( viewportWidth / 4.0, viewportHeight * 0.2, viewportWidth * ( 7.0 / 8.0 ), viewportHeight * 0.2, 30 )
		guideCountAfterCreation = pm.mel.OxGetStrandCount( editGuidesShape )

		self.assertEqual( guideCountAfterCreation, guideCountBeforeCreation )

	@unittest.skipIf( MayaTest.InBatchMode, "Requires GUI" )
	def test_DeletionBrushWithVertexColors( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )
		pm.select( plane )
		pm.polyTriangulate( plane )
		pm.polyColorPerVertex( colorRGB = [1.0, 1.0, 1.0] )
		meshShape = plane[0]

		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		self.SetupBrushTest( editGuidesShape )

		editGuidesShape.applyChannel.set( 2001 )

		viewportName = 'modelPanel4'
		viewportWidth = pm.control(viewportName, query=True, width=True)
		viewportHeight = pm.control(viewportName, query=True, height=True)

		# Set delete brush
		try:
			pm.mel.eval( "OxDeleteBrushTool oxDeleteBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxDeleteBrushTool1" )
		pm.mel.eval( "OxDeleteBrushTool -e -s " + str( viewportHeight / 2 ) + " oxDeleteBrushTool1" )

		# Get the number of guides before deleting
		guideCountBeforeDeleting = pm.mel.OxGetStrandCount( editGuidesShape )

		# Begin brushing in the center of the view in upper corner and drag brush horizontally to the right
		pm.mel.OxDragBrush( viewportWidth / 2.0, viewportHeight / 2.0, viewportWidth * ( 2.0 / 3.0 ), viewportHeight / 2.0, 3 )
		guideCountAfterDeleting = pm.mel.OxGetStrandCount( editGuidesShape )

		# Make sure that the guides were deleted
		self.assertGreater( guideCountBeforeDeleting, guideCountAfterDeleting )

		# Now repeat the test but make sure the guides are not deleted if the faces are black
		pm.delete( editGuidesShape )
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( meshShape, pointsPerStrandCount = 10, guideLength = 5.0 )
		self.SetupBrushTest( editGuidesShape )
		editGuidesShape.applyChannel.set( 2001 )

		# Set delete brush
		try:
			pm.mel.eval( "OxDeleteBrushTool oxDeleteBrushTool1" )
		except:
			pass
		pm.setToolTo( "oxDeleteBrushTool1" )
		pm.mel.eval( "OxDeleteBrushTool -e -s " + str( viewportHeight / 2 ) + " oxDeleteBrushTool1" )

		blackFaces = meshShape.f[:]
		coloredVertices = TestUtilities.ColorFaceVerticesBlack( meshShape, blackFaces)

		pm.mel.OxDragBrush( viewportWidth / 2.0, viewportHeight / 2.0, viewportWidth * ( 2.0 / 3.0 ), viewportHeight / 2.0, 3 )
		guideCountAfterDeleting = pm.mel.OxGetStrandCount( editGuidesShape )

		self.assertGreater( guideCountBeforeDeleting, guideCountAfterDeleting )
